/** 
 * Defines a data source object. A data source manages the connection to a source
 * system and the flow of data to and from that source system into one or more 
 * datasets.
 *
 * @param p_name name of the data source
 * @param p_type type classification of the data source
 * @param p_isLive flag indicating if the data source is live or static data
 * @param p_source source object reference
 * @param p_connection connection object reference from the source
 * @param p_onReload callback function to execute on reload of the data source from 
 *          the source system
 */
function DataSource(p_name, p_type, p_isLive, p_source, p_connection, p_onReload) {
	// Set properties
	var name		= p_name;
	var type		= p_type;
	var isLive		= p_isLive;
	var source		= p_source;
	var connection	= p_connection;
	
	// Set callback functions
	var onReload	= p_onReload;

	// Schema for the data source
	var schema		= new DatasetSchema();
	
	/**
	 * Returns the name.
	 */
	this.getName = function() {
		return name;
	};
	
	/**
	 * Returns the type.
	 */
	this.getType = function() {
		return type;
	};
	
	/**
	 * Returns the isLive flag for the data source.
	 */
	this.isLive = function() {
		return isLive;
	};
	
	/**
	 * Returns the source object.
	 */
	this.getSource = function() {
		return source;
	};
	
	/**
	 * Returns the connection object.
	 */
	this.getConnection = function() {
		return connection;
	};
	
	/**
	 * Returns the schema object.
	 */
	this.getSchema = function() {
		return schema;
	};
	
	/**
	 * Sets the schema for the data source using the set of fields 
	 * @param fields list of fields
	 */
	this.setSchema = function(fields) {
		for(var idx = 0; idx < fields.length; idx++)
			schema.addField(fields[idx].name, fields[idx].type);
	};
	
	/** 
	 * Reloads the data from the source
	 * @param onSuccess callback on reload success
	 * @param onError callback on reload error
	 */
	this.reload = function(onSuccess, onError) {
		if(onReload != undefined) 
			onReload(onSuccess, onError);
	};
};


/** 
 * Defines a data sink object. A data sink renders data from a dataset into a UI
 * component.
 *
 * @param p_name name of the data sink
 * @param p_type type classification of the data sink
 * @param p_callbacks object containing callback functions for all event types
 */
function DataSink(p_name, p_type, p_callbacks) {
	var name		= p_name;
	var type		= p_type;
	var callbacks	= p_callbacks;
	
	/**
	 * Changes the schema on a data sink.
	 * @param schema the schema
	 */
	this.onSchemaChange = function(schema) {
		if(callbacks.onSchemaChange != undefined)
			callbacks.onSchemaChange(schema);
	};
	
	/**
	 * Starts a snapshot on the sink.
	 */
	this.onSnapshotStart = function() {
		if(callbacks.onSnapshotStart != undefined)
			callbacks.onSnapshotStart();
	};
	
	/**
	 * Ends a snapshot on the sink.
	 */
	this.onSnapshotEnd = function() {
		if(callbacks.onSnapshotEnd != undefined)
			callbacks.onSnapshotEnd();
	};
	
	/**
	 * Insert a new object into the sink.
	 * @param object object being inserted
	 */
	this.onInsert = function(object) {
		if(callbacks.onInsert != undefined)
			callbacks.onInsert(object);
		if(callbacks.onUpsert != undefined)
			callbacks.onUpsert(object);
	};
	
	/**
	 * Update an object on the sink.
	 * @param object object being updated
	 */
	this.onUpdate = function(object) {
		if(callbacks.onUpdate != undefined)
			callbacks.onUpdate(object);
		if(callbacks.onUpsert != undefined)
			callbacks.onUpsert(object);
	};

	/**
	 * Delete an object on the sink.
	 * @param object object being deleted
	 */
	this.onDelete = function(object) {
		if(callbacks.onDelete != undefined)
			callbacks.onDelete(object);
	};
	
	/**
	 * Clears all data in the sink.
	 */
	this.onClear = function() {
		if(callbacks.onClear != undefined)
			callbacks.onClear();
	};
	
};


/** 
 * Defines a dataset schema. Each dataset has a defined schema consisting
 * of one to many fields.  
 */
function DatasetSchema() {
	var fieldMap = {};
	
	this.addField = function(name, type) {
		fieldMap[name] = {
			name:	name,
			type:	type
		};
	}
	
	this.getFieldLength = function() {
		return Object.keys(fieldMap).length;
	};
	
	this.getFields = function() {
		var fields = [];
		for(var idx = 0; idx < this.getFieldLength(); idx++)
			fields.push(this.getFieldByIndex(idx));
		return fields;
	};
	
	this.getField = function(name) {
		return fieldMap[name];
	};
	
	this.getFieldByIndex = function(index) {
		var fieldName = Object.keys(fieldMap)[index];
		return fieldMap[fieldName];
	};
};


/** 
 * Defines a dataset object. A dataset contains data from one to many data sources
 * and manages the flow of this data into one to many data sinks.
 *
 * @param p_scope scope object containing the dataset
 * @param p_id unique identifier for the dataset
 * @param p_setKey array of fields in the objects that defines a unique identifier
 * @param p_onCreate callback function to execute on create of a new dataset object
 * @param p_onMerge callback function to execute on merge of fields into an existing dataset object
 * @param p_onDelete callback function to execute on delete of a dataset object
 * @param p_onClear callback function to execute on clear of a dataset
 */
function Dataset(p_scope, p_id, p_setKey, p_onCreate, p_onMerge, p_onDelete, p_onClear) {
	// Set properties
	var $scope		= p_scope;
	var id			= p_id;
	var setKey		= p_setKey;
	var live		= false;
	
	// Set callback functions
	var onCreate 	= p_onCreate;
	var onMerge 	= p_onMerge;
	var onDelete 	= p_onDelete;
	var onClear 	= p_onClear;
	
	// Debug flag
	var debug = 0;

	// Setup the data collection
	var collection	= {};
	
	// Schema for the dataset
	var schema		= new DatasetSchema();
	
	// Initialize logging
	var $init 		= angular.injector(['ng']);
	var $log 		= $init.get('$log');

	// Setup self variable
	var self		= this;
	
	// Create a list of sinks
	var sinks		= [];
	
	// Create a list of sources
	var sources		= [];
	
	/**
	 * Returns the collection.
	 */
	this.getCollection = function() {
		return collection;
	};
	
	/**
	 * Returns the id for the dataset.
	 */
	this.getId = function() {
		return id;
	};

	/**
	 * Returns true if any datasource in the dataset is live.
	 */
	this.isLive = function() {
		return live;
	};
	
	/**
	 * Returns an array of field names that define the key for the dataset.
	 */
	this.getDatasetKey = function() {
		return setKey;
	};
	
	/**
	 * Returns the object at the specified index from the dataset.
	 */
	var getByIndex = function(index) {
		var key = Object.keys(collection)[index];
		return getByKey(key);
	};

	/**
	 * Generates a key string for the given object.
	 */
	var getKey = function(object) {
		if(setKey == undefined) return undefined;
		var key = object[setKey[0]];
		if(key == undefined) return undefined;
		for(var idx = 1; idx < setKey.length; idx++) {
			var thisKey = object[setKey[idx]];
			if(thisKey == undefined) return undefined;
			key = key + '|' + thisKey;
		}
		return key;
	};
	
	/**
	 * Retrieve the existing object for the given object fields using
	 * the specified key.
	 */
	var getByKey = function(key) {
		if(key == undefined) return undefined;
		return collection[key];
	};
	
	/**
	 * Retrieve the existing object for the given object fields using
	 * the external primary key.
	 */
	this.getByObject = function(object) {
		var key = getKey(object);
		if(key == undefined) return undefined;
		return collection[key];
	};
	
	/**
	 * Puts the object in the dataset.
	 */
	var put = function(object) {
		var key = getKey(object);
		collection[key] = object;
	};
	
	/**
	 * Removes the object from the dataset.
	 */
	var remove = function(object) {
		var key = getKey(object);
		delete collection[key];		
	};
	
	/**
	 * Returns the length of the dataset.
	 */
	this.length = function() {
		return Object.keys(collection).length;
	};
	
	/**
	 * Removes all objects from the dataset.
	 */
	var removeAll = function() {
		var keys = [];
		for(var idx = 0; idx < self.length(); idx++)
			keys.push(getByIndex(idx));
		for(var idx = 0; idx < keys.length; idx++)
			remove(keys[idx]);
	};
	
	/**
	 * Performs a data operation on the dataset with the specified object.
	 */
	this.operation = function(operation, fieldMap, fields) {
		// Choose what to do based on operation
		if(operation == 'onSnapshotStart') {
			for(var idx = 0; idx < sinks.length; idx++)
				sinks[idx].onSnapshotStart();
		}
		else if(operation == 'onSnapshotEnd') {
			for(var idx = 0; idx < sinks.length; idx++)
				sinks[idx].onSnapshotEnd();
		}
		else if(operation == 'onInsert' || operation == 'onUpdate') {
			var key = getKey(fieldMap);
			var existing = getByKey(key);
			if(existing == undefined)
				createNewDSObject(key, fieldMap);
			else
				mergeDSObject(key, fieldMap, existing);
		}
		else if(operation == 'onDelete') {
			var existing = this.getByObject(fieldMap);
			if(existing != undefined) {
				deleteDSObject(existing, fields);
			}
		}
		else if(operation == 'onClear') {
			removeAll();
			if(onClear != undefined) onClear();			
			for(var idx = 0; idx < sinks.length; idx++)
				sinks[idx].onClear();
		}
		else {
			$log.warn(operation + " unknown operation");
		}
		
		if(debug >= 1) $log.debug(collection);
	};

	/**
	 * Creates a new dataset object from the field map.
	 */
	var createNewDSObject = function(key, fieldMap) {
		// Create the object and set the _key
		var object = {
			_key:	key,
		};
		
		// Set the fields in the fieldMap on the object
		for(var idx = 0; idx < Object.keys(fieldMap).length; idx++) {
			var fieldName = Object.keys(fieldMap)[idx];
			object[fieldName] = fieldMap[fieldName];
		}
		
		// Put the object in the dataset
		put(object);
		
		// Run callbacks
		if(onCreate != undefined) onCreate(object);
		for(var idx = 0; idx < sinks.length; idx++)
			sinks[idx].onInsert(object);
	};
	
	/**
	 * Merges the fieldMap into the existing dataset object.
	 */
	var mergeDSObject = function(key, fieldMap, existing) {
		// Set the fields in the fieldMap on the object
		for(var idx = 0; idx < Object.keys(fieldMap).length; idx++) {
			var fieldName = Object.keys(fieldMap)[idx];
			existing[fieldName] = fieldMap[fieldName];
		}

		// Run callbacks
		if(onMerge != undefined) onMerge(existing);
		for(var idx = 0; idx < sinks.length; idx++)
			sinks[idx].onUpdate(existing);
	};
	
	/**
	 * Deletes the existing dataset object.
	 */
	var deleteDSObject = function(existing, fields) {
		for(var idx = 0; idx < fields.length; idx++) {
			var field = fields[idx];
			if(self.getDatasetKey().indexOf(field.name) == -1) {
				delete existing[field.name];
			}
		}

		// Determine empty object field count
		var emptyFieldCount = self.getDatasetKey().length + 1;
		if(existing.$$hashKey != undefined) emptyFieldCount = emptyFieldCount + 1;
		
		// Determine if there are any more fields remaining
		if(emptyFieldCount == Object.keys(existing).length) {
			// Remove the existing object and run callbacks
			remove(existing);
			if(onDelete != undefined) onDelete(existing);
			for(var idx = 0; idx < sinks.length; idx++)
				sinks[idx].onDelete(existing);
		}
		else {
			// Keep existing object and run callbacks
			if(onMerge != undefined) onMerge(existing);
			for(var idx = 0; idx < sinks.length; idx++) {
				//$log.debug('blah');
				//$log.debug(existing);
				sinks[idx].onUpdate(existing);
			}
		}
	};

	/**
	 * Sets the fields on the schema.
	 */
	this.setSchema = function(fields) {
		for(var idx = 0; idx < fields.length; idx++)
			schema.addField(fields[idx].name, fields[idx].type);
		for(var idx = 0; idx < sinks.length; idx++)
			sinks[idx].onSchemaChange(schema);
		if(debug >= 1) $log.debug(schema);
	};
	
	/** 
	 * Returns the schema.
	 */
	this.getSchema = function() {
		return schema;
	};
	
	/**
	 * Register a new data sink.
	 */
	this.registerSink = function(sink) {
		sinks.push(sink);
		if(this.length() > 0) {
			sink.onSchemaChange(schema);
			sink.onSnapshotStart();
			for(var idx = 0; idx < this.length(); idx++) {
				var object = getByIndex(idx);
				sink.onInsert(object);
			}
			sink.onSnapshotEnd();
		}
	}

	/**
	 * Register a new data source.
	 */
	this.registerSource = function(source) {
		live = live || source.isLive();
		sources.push(source);
	 }

	/**
	 * Unregister an existing data source.
	 */
	this.unregisterSource = function(source) {
		var idx = sources.indexOf(source);
		if(idx > -1) sources.splice(idx, 1);

		// Rebuild the schema
		schema = new DatasetSchema();
		for(var sourceIdx = 0; sourceIdx < sources.length; sourceIdx++) {
			var source = sources[sourceIdx];
			var sourceSchema = source.schema;
			for(var fieldIdx = 0; fieldIdx < sourceSchema.getFieldLength(); fieldIdx++) {
				var f = sourceSchema.getFieldByIndex(fieldIdx);
				schema.addField(f.name, f.type);
			}
		}		
		
		// Run callbacks
		for(var idx = 0; idx < sinks.length; idx++)
			sinks[idx].onSchemaChange(schema);		
		if(debug >= 1) $log.debug(schema);
		
		// Reload all sources
		this.reload();
	}

	/**
	 * Returns an array of datasources.
	 */
	this.getSources = function() {
		 return sources;
	}
	 
	/**
	 * Returns an array of datasinks.
	 */
	this.getSinks = function() {
		 return sinks;
	 }

	/**
	 * Clears the dataset and reloads all sources
	 */
	 this.reload = function() {
		removeAll();
		if(onClear != undefined) onClear();			
		for(var idx = 0; idx < sinks.length; idx++)
			sinks[idx].onClear();		
		for(var idx = 0; idx < this.getSources().length; idx++) {
			var source = self.getSources()[idx];
			source.reload();
		}
	 };
}


/**
 * Defines an arbitrary collection object.
 *
 * @param p_keys an array listing the fields that define a unique key
 */
function Collection(p_keys) {
	var keys = p_keys;
	var collection = {};
	
	/**
	 * Returns the collection lookup key for the given object.
	 */
	this.getKeyForObject = function(object) {
		if(keys.length == 0)
			return null;
		var key = object[keys[0]];
		for(var idx = 1; idx < keys.length; idx++)
			key = key + '|' + object[keys[idx]];
		return key;
	};
	
	/**
	 * Puts the object in the collection.
	 */
	this.put = function(object) {
		var key = this.getKeyForObject(object);
		collection[key] = object;
	};
	
	/**
	 * Returns the object for the given key.
	 */
	this.get = function(key) {
		return collection[key];
	};

	/** 
	 * Returns the object at the given index.
	 */
	this.getByIndex = function(index) {
		var key = Object.keys(collection)[index];		
		return this.get(key);
	}
	
	/**
	 * Returns the length of the collection.
	 */
	this.length = function() {
		return Object.keys(collection).length;
	}

	/**
	 * Removes the object with the key.
	 */
	this.remove = function(key) {
		delete collection[key];
	}
	
	/**
	 * Removes all objects.
	 */
	this.removeAll = function() {
		collection = {};
	};
	
	/**
	 * Converts the collection into an array.
	 */
	this.toArray = function() {
		var array = [];
		for(var idx = 0; idx < this.length(); idx++)
			array.push(this.getByIndex(idx));
		return array;
	}
};

