function LDMGlobalConnection() {
	// Debug flag
	var debug = 0;

	// Connection state constants
	var ConnectionState = {
		UNCONNECTED: 	'UNCONNECTED',
		CONNECTING: 	'CONNECTING',
		CONNECTED: 		'CONNECTED',
		ERROR:			'ERROR'
	};

	// Set the configuration parameters
	var url 		= '/lv/client';

	// Initialize objects
	var conn		= null;
	var dataSources	= new Collection(['controller']);

	// Initialize logging
	var $init 		= angular.injector(['ng']);
	var $log 		= $init.get('$log');

	// Initialize class variables
	this.state		= ConnectionState.UNCONNECTED;

	// Initialize a local variable to hold current reference
	var self 		= this;
	
	/** 
	 * Connect to LiveView and prepare the queries.
	 */
	this.connect = function() {
		if(debug >= 1) $log.debug('[Global] ' + 'starting connection to LDM');
		this.state = ConnectionState.CONNECTING;
		LiveView.connect({
				url: url,
				onClose: function(connection, actor, error) {
					if(debug >= 1) $log.debug('[Global] ' + 'connection closed... reconnecting');
					self.connect();
				}
			}
		).then(
			function(connection){
				//once connected, keep the reference to the connection and run queries
				conn = connection;
				self.state = ConnectionState.CONNECTED;
				if(debug >= 1) $log.debug('[Global] ' + 'connected');
				self.processDataSources();
			},
			function(error){
				$log.error('[Global] ' + 'Connect: ' + error.message);
				$log.error(error);
				self.state = ConnectionState.ERROR;
			}
		);
	};

	/**
	 * Processes all registered data sources 
	 */
	this.processDataSources = function() {
		for(var idx = 0; idx < dataSources.length(); idx++) {
			dataSources.getByIndex(idx).processConfiguredQueries(conn);
		}
	};
	
	/** 
	 * Closes this connection. 
	 */
	this.close = function() {
		if(debug >= 1) $log.debug('[Global] ' + 'closing connection to LDM');
		LiveView.closeAllConnections(
		).then(
			function(connection){
				self.state = ConnectionState.UNCONNECTED;
				if(debug >= 1) $log.debug('[Global] ' + 'disconnected');
			},
			function(error){
				$log.error('[Global] ' + 'Close: ' + error.message);
				$log.error(error);
				self.state = ConnectionState.UNCONNECTED;
			}
		);
	};
	
	/**
	 * Adds a data source to the connection 
	 */
	 this.addDataSource = function(dataSource) {
		// Add to the collection
		dataSources.put(dataSource);
		
		// If the connection has already been made, process the queries for 
		//   the data source
		if(conn != null)
			dataSource.processConfiguredQueries(conn);
		
		// Otherwise the data source will have queries processed once the
		//   connection is obtained
	 };
};

// Create a LV Global Connection object
var ldmGlobalConn = new LDMGlobalConnection();
ldmGlobalConn.connect();


function LDMDataSource(config) {
	// Debug flag
	var debug = 0;
	
	// Set the configuration fields
	var $scope 		= config.scope;
	var configurl 	= config.configurl;
	var onConnect   = config.onConnect;
	this.controller = $scope.controllerName;
	
	// Set the logging fields
	var $init 		= angular.injector(['ng']);
	var $log 		= $init.get('$log');
	
	// Initialize objects
	var conn		= null
	var queries 	= new Collection(['id']);

	// Initialize a local variable to hold current reference
	var self 		= this;

	/** 
	 * Add the data source to the connection. If the connection 
	 * is pending then the queries will be processed once completed.
	 * Otherwise the queries will be processed immediately.
	 */
	this.init = function() {
		ldmGlobalConn.addDataSource(this);
	};
	
	/** 
	 * Sets the LV connection then processes all configured queries. Each query
	 * is registered which will run it as subscription or snapshot, as appropriate. 
	 * Called from LDMGlobalConnection. 
	 */
	this.processConfiguredQueries = function(connection) {
		// Set the connection
		conn = connection;
		
		// If there is an onConnect function, run that
		if(onConnect != null) {
			onConnect();
		}
		
		// If there is a configurl, load queries
		if(configurl != null) {
			$.ajax({
				type: 'GET',
				url: configurl,
				cache: false,
				dataType: 'json',
				success: function(data, textStatus, xhr) {
					// Add each loaded query to the queries collection
					for(var idx = 0; idx < data.queries.length; idx++) {
						if(data.queries[idx].disabled != true)
							self.register(data.queries[idx]);
					}
				},
				error: function(xhr, textStatus, errorThrown) {
					$log.error(textStatus);
					$log.error(xhr);
				}
			});	
		}
	};
	
	/** 
	 * Register a new query on the system. 
	 */
	this.register = function(query, onSuccess, onError) {
		// If grouping is enabled, set a flag
		if(query.grouping != undefined) {
			query.isGrouped = true;
		}
			
		// Link to the datasets
		for(var idx = 0; idx < query.dataset.length; idx++) {
			// Retrieve the dataset
			var datasetName = query.dataset[idx];
			var dataset = $scope[datasetName];
			
			// If dataset is not found, create it
			if(dataset == undefined) {
				dataset = new Dataset($scope, datasetName, query.isGrouped == true ? query.grouping.groupBy : query.primaryKey);
				$scope[datasetName] = dataset;
				$scope.$apply();
			}
			
			// Set the live flag
			var live = !(query.snapshot == true);

			// Setup onReload function for the data source to call when the query is reloaded
			var onReload = function(onSuccess, onError) {
				if(query.snapshot == true)
					self.snapshot(query, onSuccess, onError);
				else {
					self.unsubscribe(query, function(result) {
						self.subscribe(query, onSuccess, onError);
					});
				}
			};
			
			// Create the datasource, schema will be set later
			query.datasource = new DataSource(query.name, 'LDM', live, query, this, onReload);
			
			// Map of tupleId to primary key for caching
			query.collection  = {};

			// Register the data source with the dataset
			dataset.registerSource(query.datasource);
		}
		
		// Snapshot or subscribe depending on the query type
		if(query.snapshot == true)
			this.snapshot(query, onSuccess, onError);
		else
			this.subscribe(query, onSuccess, onError);
	};
	
	/** 
	 * Subscribes to a query.
	 */
	this.subscribe = function(query, onSuccess, onError) {
		// Subscribe to the query on the server
		conn.subscribe(
			new LiveView.Query(query.template, query.parameters), {
				onSnapshotStart: function(event){					
					query.schema = event.schema;
					query.status = event.subscription.status.string;	

					// Get the fields array from the query schema
					var fields = getSchemaFields(query);
					
					// Set the schema on the datasource
					query.datasource.setSchema(fields);
					
					// Set the schema on the dataset
					for(var idx = 0; idx < query.dataset.length; idx++)
						$scope[query.dataset[idx]].setSchema(fields);
					
					// Call snapshot start
					for(var idx = 0; idx < query.dataset.length; idx++)
						$scope[query.dataset[idx]].operation('onSnapshotStart');
					if(debug >= 1) $log.debug('[' + self.controller + '] ' + 'query:' + query.id + ':subscribe:onSnapshotStart');					
				},
				onInsert: function(event){
					query.collection[event.tuple.id] = self.getKeyFields(query, event.tuple.fieldMap);
					var operation = 'onInsert';
					var fieldMap = event.tuple.fieldMap;
					if(query.isGrouped == true) {
						var g = LDMGroup.onObject(fieldMap, query);
						operation = g.operation;
						fieldMap = g.fieldMap;
					}
					for(var idx = 0; idx < query.dataset.length; idx++) {
						$scope[query.dataset[idx]].operation(operation, fieldMap);
					}						
					
					if(debug >= 2) $log.debug('[' + self.controller + '] ' + 'query:' + query.id + ':subscribe:onInsert tid=' + event.tuple.id);
				},
				onUpdate: function(event){
					var keyFields = query.collection[event.tuple.id];
					if(keyFields == undefined) return;
					for(var idx = 0; idx < Object.keys(keyFields).length; idx++) {
						var k = Object.keys(keyFields)[idx];
						event.tuple.fieldMap[k] = keyFields[k];
					}
					var operation = 'onUpdate';
					var fieldMap = event.tuple.fieldMap;
					if(query.isGrouped == true) {
						var g = LDMGroup.onObject(fieldMap, query);
						operation = g.operation;
						fieldMap = g.fieldMap;
					}
					for(var idx = 0; idx < query.dataset.length; idx++) {
						$scope[query.dataset[idx]].operation(operation, fieldMap);
					}						
					
					if(debug >= 2) $log.debug('[' + self.controller + '] ' + 'query:' + query.id + ':subscribe:onUpdate tid=' + event.tuple.id);
				},
				onDelete: function(event){
					var keyFields = query.collection[event.tuple.id];
					if(keyFields == undefined) return;
					var operation = 'onDelete';
					var fieldMap = {};
					for(var idx = 0; idx < Object.keys(keyFields).length; idx++) {
						var k = Object.keys(keyFields)[idx];
						fieldMap[k] = keyFields[k];
					}
					if(query.isGrouped == true) {
						var g = LDMGroup.onDelete(fieldMap, query);
						operation = g.operation;
						fieldMap = g.fieldMap;
					}
					for(var idx = 0; idx < query.dataset.length; idx++) {
						$scope[query.dataset[idx]].operation(operation, fieldMap, getSchemaFields(query));
					}
					
					if(debug >= 2) $log.debug('[' + self.controller + '] ' + 'query:' + query.id + ':subscribe:onDelete tid=' + event.tuple.id);
				},
				onSnapshotEnd: function(event){
					query.snapshot = undefined;
					for(var idx = 0; idx < query.dataset.length; idx++)
						$scope[query.dataset[idx]].operation('onSnapshotEnd');

					if(debug >= 1) $log.debug('[' + self.controller + '] ' + 'query:' + query.id + ':subscribe:onSnapshotEnd');
				},
				onError: function(error){
					query.error = error.message;
					$log.error('[' + self.controller + '] ' + 'There was an error while running query ');
					$log.error(event);
					$log.error(error);
				}
			}
		).then(
			function(result){
				queries.put(query);
				if(debug >= 1) $log.debug('[' + self.controller + '] ' + 'query:' + query.id + ':subscribe:success');
				query.error = undefined;
				query.subscription = result.subscription;
				if(onSuccess != undefined) onSuccess(query);
				$scope.$apply();
			},
			function(error){
				$log.error('[' + self.controller + '] ' + 'Query: ' + error.message);
				query.error = error.message;
				if(onError != undefined) onError(error);
			}
		);					
	};
	
	/** 
	 * Unsubscribes a query.
	 */
	this.unsubscribe = function(query, onSuccess, onError) {
		conn.unsubscribe(
			query.subscription
		).then(
			function(result){
				delete query.schema;
				if(onSuccess != undefined) onSuccess(result);
			},
			function(result){
				$log.error('[' + controller + '] ' + 'Query unsubscribe: ' + result.error);
				delete query.schema;
				if(onError != undefined) onError(result);
			}
		);					
	};
	
	/** 
	 * Executes the query as a snapshot 
	 */
	this.snapshot = function(query, onSuccess, onError) {
		// Execute the query on the server
		conn.execute(
			new LiveView.Query(query.template, query.parameters), {
				onSuccess: function(event){
					query.schema = event.schema;
					query.status = 'COMPLETE';

					// Get the fields array from the query schema
					var fields = getSchemaFields(query);
					
					// Set the schema on the datasource
					query.datasource.setSchema(fields);
					
					// Set the schema on the dataset
					for(var idx = 0; idx < query.dataset.length; idx++)
						$scope[query.dataset[idx]].setSchema(fields);

					// Call snapshot start
					for(var idx = 0; idx < query.dataset.length; idx++)
						$scope[query.dataset[idx]].operation('onSnapshotStart');
					
					// Loop over the event data and send event for each
					for(var idx = 0; idx < event.data.length; idx++) {
						var tuple = event.data[idx];
						var operation = 'onInsert';
						var fieldMap = tuple.fieldMap;
						query.collection[tuple.id] = self.getKeyFields(query, fieldMap);
						if(query.isGrouped == true) {
							var g = LDMGroup.onObject(fieldMap, query);
							operation = g.operation;
							fieldMap = g.fieldMap;
						}
						for(var dsIndex = 0; dsIndex < query.dataset.length; dsIndex++) {
							$scope[query.dataset[dsIndex]].operation(operation, fieldMap);
						}						
					}					

					for(var idx = 0; idx < query.dataset.length; idx++)
						$scope[query.dataset[idx]].operation('onSnapshotEnd');

					if(debug >= 1) $log.debug('[' + controller + '] ' + 'query:' + query.id + ':snapshot:onSuccess');					
				},
				onProgress: function(event){
					if(debug >= 2) $log.debug('[' + controller + '] ' + 'query:' + query.id + ':snapshot:onProgress');					
				},
				onError: function(error){
					query.error = error.message;
					$log.error('[' + controller + '] ' + 'There was an error while executing query ' + event.subscription.id);
					$log.error('[' + controller + '] ' + 'ERROR:' + error);
				}
			}
		).then(
			function(result){
				queries.put(query);
				query.error = undefined;
				if(onSuccess != undefined) onSuccess(query);
				$scope.$apply();
			},
			function(error){
				$log.error('[' + controller + '] ' + 'Query: ' + error.message);
				query.error = error.message;
				if(onError != undefined) onError(error);
			}
		);				
	};
	
	/**
	 * Extracts the primary key fields from the field map into a new
	 * object.
	 */
	this.getKeyFields = function(query, fieldMap) {
		var obj = {};
		for(var idx = 0; idx < query.primaryKey.length; idx++) {
			var pk = query.primaryKey[idx];
			obj[pk] = fieldMap[pk];
		}
		return obj;
	};
	
	/**
	 * Extracts the fields from the query schema and creates an array.
	 */
	var	getSchemaFields = function(query) {
		var fields = [];
		for(var idx = 0; idx < query.schema.fields.length; idx++) {
			fields[idx] = {
				name:	query.schema.fields[idx].name,
				type:	query.schema.fields[idx].type,
			}
		}		
		return fields;
	}
	
	/** 
	 * Sends a tuple 
	 */
	this.sendTuple = function(key, serverUri, inputStream, input, outputStream, onSuccess, onError) {		
		var inputTuple = new LiveView.Tuple(null, input);
		if(outputStream != undefined)
			inputTuple.fieldMap.__SendTupleResponseKey = null;
			
		conn.sendTuple(inputTuple, inputStream, outputStream, {
			serverUri: serverUri
		}).then(
			function(result){
				var output = undefined;				
				if(outputStream != undefined) {
					output = result.tuple.fieldMap;
					delete output.__SendTupleResponseKey;
				}
				if(onSuccess != undefined) onSuccess(result);
			},
			function(error){
				$log.error('[' + controller + '] ' + 'SendTuple: ' + error);
				if(onError != undefined) onError(error);
			}
		);
	};
	
	/** Applies parameters to a previously registered query.
 	 *  If queryId is undefined then the parameters are applied to all queries.
	 */
	this.applyParameters = function(parameters, queryId) {

		if(queryId == undefined) {
			for(var idx = 0; idx < queries.length(); idx++) {
				var query = queries.getByIndex(idx);
				applyParametersToQuery(query, parameters, this);
			}
		}
		else {
			var query = queries.get(queryId);
			applyParametersToQuery(query, parameters, this);
		}
	};
	
	/** 
	 * Applies parameters to the specified query 
	 */
	var applyParametersToQuery = function(query, parameters, self) {
		if(query == null || query.subscription == null)
			return;

		// Loop over the keys and apply to the query
		for(var idx = 0; idx < Object.keys(parameters).length; idx++) {
			var paramName = Object.keys(parameters)[idx];
			query.parameters[paramName] = parameters[paramName];
		}

		// Clear the dataset
		for(var idx = 0; idx < query.dataset.length; idx++)
			$scope[query.dataset[idx]].operation('onClear');
		
		// Clear the local key cache
		query.collection = {};
		
		// Apply to the query subscription
		/*query.subscription.applyParameters(query.parameters).then(
			function(success) {
			},
			function(error) {
			}
		);*/
		
		// THIS IS A WORKAROUND
		// Unsubscribe the existing query
		self.unsubscribe(query, 
			function(result) {
				self.subscribe(query);
			},
			function(result) {
				$log.error(result);
			});
		
	};	

	/**
	 * Returns the queries collection.
	 */
	this.getQueries = function() {
		return queries;
	};
	
	/** 
	 * Retrieve a list of registered queries in the data source.
	 * Not implemented in LDM JS API.
	 */
	this.listQueries = function(onSuccess, onError) {
		var err = "listQueries not implemented";
		$log.error(err);
		if(onError != undefined) onError(err);
	};
	
	/** 
	 * Retrieve a list of tables in the system. If the call is successful,
	 * the the success callback will be called with the result.
	 */
	this.listTables = function(includeSystemTables, onSuccess, onError) {
		conn.getTables({
			includeSystemTables: includeSystemTables
		}).then(
			function(result){		
				if(onSuccess != undefined) onSuccess(result);
			},
			function(error){
				$log.error('[' + controller + '] ' + 'List Tables: ' + error);
				if(onError != undefined) onError(error);
			}
		);
	};
		
	/** 
	 * Validates a query but does not store either the query or the results.
	 * On completion of the call, the success callback will always be called 
	 * whether or not the query is valid. The callback will be passed the result.
	 */
	this.validate = function(query, onSuccess, onError) {
		conn.execute(
			new LiveView.Query(query.template, query.parameters), {
				onSuccess: function(event){
				},
				onProgress: function(event){
				},
				onError: function(error){
				}
			}
		).then(
			function(result){
				var response = {
					valid:		true,
					message:	"Success"
				};
				if(onSuccess != undefined) onSuccess(response);
			},
			function(error){
				var response = {
					valid:		error.code == 0,
					message:	error.message
				};
				if(onSuccess != undefined) onSuccess(response);
			}
		);				
	};
	
	/**
	 * Unregisters an existing query.
	 */
	this.unregister = function(query) {
		this.remove(query);
		
		for(var idx = 0; idx < query.dataset.length; idx++) {
			// Retrieve the dataset
			var datasetName = query.dataset[idx];
			var dataset = $scope[datasetName];
			dataset.unregisterSource(query.datasource);
		}
	};
	
	/** 
	 * Removes an existing query from the list of registered queries. If the query 
	 * is a subscription it will be unsubscribed first.
	 */
	this.remove = function(query) {
		if(query.snapshot != true)
			this.unsubscribe(query);
		
		var id = query.id;
		queries.remove(query.id);	
	};

	/** 
	 * Closes this connection 
	 */
	this.close = function() {
		// Remove all queries
		for(var idx = queries.length() - 1; idx >= 0; idx--) {
			var query = queries.getByIndex(idx);
			this.remove(query);
		}
	};	

	/**
	 * Grouping functions.
	 */
	var LDMGroup = {
		onObject: function(fieldMap, query) {
			// Get the group collection, this will create if not found
			var groupCollection = this.getGroupCollection(query);
			
			// Get the specified group object from the collection
			var groupObject = groupCollection.get(groupCollection.getKeyForObject(fieldMap));
			var operation = 'onUpdate';
			
			// Get the grouping definition
			var grouping = query.grouping;
						
			// If not found, create the group object and change the operation to insert
			if(groupObject == undefined) {
				groupObject = this.createGroupObject(fieldMap, grouping, query);
				groupCollection.put(groupObject);
				operation = 'onInsert';
			}
			
			// Copy the fieldMap group properties to the group object
			for(var idx = 0; idx < grouping.groupProperties.length; idx++)
				groupObject[grouping.groupProperties[idx]] = fieldMap[grouping.groupProperties[idx]];
			
			// Create the new element object for the group object and copy all the fieldMap properties 
			//   to this object
			var elementObject = this.createElementObject(fieldMap, query);
			
			// Put the element into the group
			groupObject[grouping.groupList].push(elementObject);
			groupObject._keyMap[elementObject._key] = elementObject;

			// Return the operation and fieldMap
			return {operation: operation, fieldMap: groupObject};
		},
		
		onDelete: function(fieldMap, query) {
			// Get the group collection, this will create if not found
			var groupCollection = this.getGroupCollection(query);

			// Get the specified group object from the collection
			var groupObject = groupCollection.get(groupCollection.getKeyForObject(fieldMap));
			var operation = 'onUpdate';
			
			// Get the grouping definition
			var grouping = query.grouping;
						
			// Get the element object for the fieldMap
			var key = this.getKeyForElementObject(fieldMap, query);
			var elementObject = groupObject._keyMap[key];

			var arrIndex = groupObject[grouping.groupList].indexOf(elementObject);
			groupObject[grouping.groupList].splice(arrIndex, 1);
			delete groupObject._keyMap[key];

			// If there are no more elements, return delete
			if(groupObject[grouping.groupList].length == 0)
				operation = 'onDelete';

			// Return the operation and fieldMap
			return {operation: operation, fieldMap: groupObject};
		},
		
		// Create a new group object
		createGroupObject: function(fieldMap, grouping, query) {
			var groupObject = {
				_orderBy:	grouping.orderBy,
				_keyMap:	{}
			};
			for(var idx = 0; idx < grouping.groupBy.length; idx++)
				groupObject[grouping.groupBy[idx]] = fieldMap[grouping.groupBy[idx]];
			groupObject[grouping.groupList] = [];
			return groupObject;
		},
		
		// Create a new element object
		createElementObject: function(fieldMap, query) {
			var self = this;
			var elementObject = {
				_key: self.getKeyForElementObject(fieldMap, query)
			};
			for(var idx = 0; idx < Object.keys(fieldMap).length; idx++) {
				var key = Object.keys(fieldMap)[idx];
				elementObject[key] = fieldMap[key];
			}
			return elementObject;
		},
		
		// Get the key for an element object
		getKeyForElementObject: function(fieldMap, query) {
			if(query.primaryKey.length == 0)
				return null;
			var key = fieldMap[query.primaryKey[0]];
			for(var idx = 1; idx < query.primaryKey.length; idx++)
				key = key + '|' + fieldMap[query.primaryKey[idx]];
			return key;
		},
		
		// Retrieve the group collection from the query
		getGroupCollection: function(query) {
			var groupCollection = query.groups;
			if(groupCollection == undefined) {
				groupCollection = new Collection(query.grouping.groupBy);
				query.groups = groupCollection;
			}
			return groupCollection;
		}
	};
	
};
