app.directive('liveplotly', function($log, $filter, Overlay, Modal, EventListener, ContextMenu, Clipboard) {
	return {
		templateUrl: './templates/live-plotly.html',
		replace: false,
		transclude: true,
		restrict: 'E',
		scope: {
			dataset: 	'=',
		},
		controller: function($scope) {
			this.addSeries = function(series) {
				if($scope.series == undefined)
					$scope.series = [];
				$scope.series.push(series);
			};
			this.getChart = function() {
				return $scope.chart;
			};
		},
		link: function(scope, element, attrs, ctrl, transclude) {

			// Add flexbox classes to the directive element
			$(element).addClass('flex-vertical-container fill-remainder');

			// Dependency check
			scope.dependencies = function() {
				try {
					var p = Plotly;
					return true;
				}
				catch(err) {
					return false;
				}
				return false;
			};
			
			// DEFINE OBJECTS
			{
				// LiveChart object
				{
					/**
					 * Create a new chart.
					 * @param category attribute on the data object that provides the domain value for a data point
					 * @param inverted boolean value indicating if the axes are inverted. By default the domain is presented
					 *    on the x-axis and the range on the y-axis. If this flag is set to true then the domain will be 
					 *    prensented on the y-axis and the range on the x-axis.
					 * @param options series options
					 */
					function LiveChart(graphDiv, category, inverted, options, layout, series, is3D) {
						var self 		= this;

						// These properties are used for microbatching the updates
						this.microbatch = {
							rtime:		new Date(1, 1, 2000, 12,00,00),
							timeout:	false,
							delta:		500
						}				

						// Set chart properties
						this.graphDiv	= graphDiv;
						this.category 	= category;
						this.inverted	= inverted;
						this.options	= options;
						this.series 	= series;
						this.paused		= false;
						
						// Setup plotly properties
						this.data 		= [];						
						this.layout 	= layout;
						this.maxAxis	= 0;
						this.minAxis	= 0;
						
						// Setup color scale
						this.colorscale = ['#1f77b4','#aec7e8','#ff7f0e','#ffbb78','#2ca02c','#98df8a',
							'#d62728','#ff9896','#9467bd','#c5b0d5','#8c564b','#c49c94','#e377c2','#f7b6d2',
							'#7f7f7f','#c7c7c7','#bcbd22','#dbdb8d','#17becf','#9edae5'];
						
						if(is3D == true) {
							this.layout.scene.xaxis = this.layout.xaxis;
							this.layout.scene.yaxis = this.layout.yaxis;
							this.layout.scene.zaxis = this.layout.zaxis;
						}
						
						// If chart is inverted, swap the layout for x and y axes
						if(this.inverted == true && this.layout.xaxis != undefined) {
							var xtemp = this.layout.xaxis;
							this.layout.xaxis = this.layout.yaxis;
							this.layout.yaxis = xtemp;
						}

						this.configuration = {
							displayModeBar: false	
						};

						// Draw the chart
						this.draw();
						
						// Call chart ready to populate the series
						this.chartReady();
					};
					
					LiveChart.prototype.draw = function() {
						// Create a new empty Plotly graph
						Plotly.newPlot(this.graphDiv, this.data, this.layout, this.configuration);							
					};
					
					/**
					 * Function that prepares the chart by applying configured series data
					 * on to the plot area.
					 */
					LiveChart.prototype.chartReady = function() {	
						if(this.series == undefined) return;
						
						// Iterate over the series and push the trace into the data for the chart
						for(var idx = 0; idx < this.series.length; idx++) {
							if(this.inverted == true) this.series[idx].invert();
							this.data.push(this.series[idx].trace);
						}

						// Update the chart
						this.update();
					};
					
					/**
					 * Clears the entire contents of the chart.
					 */
					LiveChart.prototype.clear = function() {
						// Iterate over the series and remove all points
						for(var idx = 0; idx < this.series.length; idx++) {
							this.series[idx].removeAllPoints();
						}
					}
					
					/**
					 * Resizes the chart to fit the parent element dimensions.
					 */
					LiveChart.prototype.resize = function() {
						// Get the new size for the chart
						var update = {
							width: element.parent().width(), 
							height: element.parent().height()
						}

						// Relayout then redraw the chart						
						Plotly.relayout(this.graphDiv, update);	
						Plotly.redraw(this.graphDiv);
					};
					
					/**
					 * Updates and redraws the chart.
					 */
					LiveChart.prototype.update = function() {
						var self = this;

						// Microbatch handler to prevent too many updates
						if(self.microbatch.timeout === false) {
							self.microbatch.timeout = true;
							setTimeout(eventEnd, self.microbatch.delta);
						}					
						function eventEnd() {
							var d = new Date() - self.microbatch.rtime;
							if (new Date() - self.microbatch.rtime < self.microbatch.delta)
								setTimeout(eventEnd, self.microbatch.delta);
							else {
								self.microbatch.rtime = new Date();
								self.microbatch.timeout = false;
								run();
							}               
						};
						
						// Run the update
						function run() {
							// Get the width and height to determine visibility
							var w = element.parent().width();
							var h = element.parent().height();
							
							// If the chart isn't visible, don't redraw to prevent errors
							//if(w == 0 || h == 0) return; 
							
							// Redraw the chart
							if(self.paused == false)
								Plotly.redraw(self.graphDiv);
						};
					};
					
					LiveChart.prototype.pause = function() {
						this.paused = true;
					};
					
					LiveChart.prototype.resume = function() {
						this.paused = false;
						this.update();
					};
						
				}				
			}
			
			// Transclude own scope
			transclude(scope, function (clone, scope) {
				element.append(clone);
			});
						
			// Set the element ID to scope if it exists
			if(attrs.id != undefined) scope.elementId = attrs.id;
			
			// Set the name to scope if it exists
			if(attrs.name != undefined) scope.elementName = attrs.name;
			
			// Check dependencies
			if(scope.dependencies() == false) return;
			
			// Get the graph div
			scope.graphDiv = $(element).find('.chart-container')[0];
			
			// Read the options into scope
			scope.options = {
				configIcon: 			attrs.configIcon == undefined ? 
											false : 
											eval(attrs.configIcon),
										// Optional. Show or hide config icon. May be true or false.
										
				datasetIcon: 			attrs.datasetIcon == undefined ? 
											false : 
											eval(attrs.datasetIcon),
										// Optional. Show or hide dataset icon field. May be true or false.
										
				datasetModal:			attrs.datasetModal == "R" || attrs.datasetModal == "RW" ?
											attrs.datasetModal :
											undefined,
										// Optional. Show the dataset modal if set. May be one of:
										//   R -- read-only modal
										//   RW -- read-write modal
										
				equalAxes:				attrs.equalAxes == undefined ? false : eval(attrs.equalAxes),
										// Optional. Constrain all axes to equal ranges.
										
			};

			// Read the Plotly chart options into scope
			scope.chartOptions = {
				barmode:				attrs.barmode,
				
				font: {
					family: 			attrs.fontFamily,
										// Optional. Sets the global font family.
										
					size:				attrs.fontSize == undefined || attrs.fontSize.length == 0 ? undefined : eval(attrs.fontSize),
										// Optional. Sets the global font size. 
										
					color:				attrs.fontColor,
										// Optional. Sets the global font color.
				},
				
				title:					attrs.title,
										// Optional. Chart title to display.
										
				titlefont: {
					family: 			attrs.titlefontFamily,
										// Optional. Sets the title font family.

					size:				attrs.titlefontSize == undefined || attrs.titlefontSize.length == 0 ? undefined : eval(attrs.titlefontSize),
										// Optional. Sets the title font size.

					color:				attrs.titlefontColor,
										// Optional. Sets the title font color.
				},
				
				autosize:				true, 
										// Optional. Determines whether or not a layout width or height that has been left undefined by the user is 
										//   initialized on each relayout. Note that, regardless of this attribute, an undefined layout width or 
										//   height is always initialized on the first call to plot.
				
				margin: {
					l:					attrs.marginLeft == undefined ? undefined : eval(attrs.marginLeft),
										// Optional. Sets the left margin (in px).

					r:					attrs.marginRight == undefined ? undefined : eval(attrs.marginRight),
										// Optional. Sets the right margin (in px).
					
					t:					attrs.marginTop == undefined ? undefined : eval(attrs.marginTop),
										// Optional. Sets the top margin (in px).
					
					b:					attrs.marginBottom == undefined ? undefined : eval(attrs.marginBottom),
										// Optional. Sets the bottom margin (in px).
					
					pad:				attrs.marginPad  == undefined ? undefined : eval(attrs.marginPad),
										// Optional. Sets the amount of padding (in px) between the plotting area and the axis lines.
					
					autoexpand:			attrs.marginAutoexpand == undefined ? undefined : eval(attrs.marginAutoexpand),
										// Optional.
				},
				
				paper_bgcolor:			attrs.paperBgcolor,
										// Optional. Sets the color of paper where the graph is drawn.
										
				plot_bgcolor:			attrs.plotBgcolor,
										// Optional. Sets the color of plotting area in-between x and y axes.
										
				showlegend:				attrs.showlegend == undefined ? undefined : eval(attrs.showlegend),
										// Optional. Determines whether or not a legend is drawn.

				scene: {
					camera: {
						eye: {
							x:			attrs.sceneCameraEyeX == undefined ? undefined : eval(attrs.sceneCameraEyeX),
							y:			attrs.sceneCameraEyeY == undefined ? undefined : eval(attrs.sceneCameraEyeY),
							z:			attrs.sceneCameraEyeZ == undefined ? undefined : eval(attrs.sceneCameraEyeZ),
						},
					},
				},
				
				xaxis: {
					color:				attrs.xaxisColor,
										// Optional. Sets default for all colors associated with this axis all at once: line, font, 
										//   tick, and grid colors. Grid color is lightened by blending this with the plot background 
										//   Individual pieces can override this.
										
					title:				attrs.xaxisTitle,
										// Optional. Sets the title of this axis.
										
					titlefont: {
						family: 		attrs.xaxisTitlefontFamily,
										// Optional. Sets the title font family.

						size:			attrs.xaxisTitlefontSize == undefined || attrs.xaxisTitlefontSize.length == 0 ? undefined : eval(attrs.xaxisTitlefontSize),
										// Optional. Sets the title font size.

						color:			attrs.xaxisTitlefontColor,
										// Optional. Sets the title font color.
					},
					
					type:				attrs.xaxisType,
										// Optional. Sets the axis type. 
										//   enumerated: "-" | "linear" | "log" | "date" | "category"
										
					autorange:			attrs.xaxisAutorange,
										// Optional. Determines whether or not the range of this axis is computed in relation to the 
										//   input data. See `rangemode` for more info. If `range` is provided, then `autorange` is set to "false".
					
					rangemode:			attrs.xaxisRangemode,
										// Optional. If "normal", the range is computed in relation to the extrema of the input data. If "tozero"`, 
										//   the range extends to 0, regardless of the input data If "nonnegative", the range is non-negative, 
										//   regardless of the input data.
										//   enumerated: "normal" | "tozero" | "nonnegative"
										
					range:				attrs.xaxisRange == undefined ? undefined : eval(attrs.xaxisRange),
										// Optional. Sets the range of this axis. If the axis `type` is "log", then you must take the log of your 
										//   desired range (e.g. to set the range from 1 to 100, set the range from 0 to 2). If the axis `type` is 
										//   "date", it should be date strings, like date data, though Date objects and unix milliseconds will be 
										//   accepted and converted to strings. If the axis `type` is "category", it should be numbers, using the 
										//   scale where each category is assigned a serial number from zero in the order it appears. 

					fixedrange:			attrs.xaxisFixedrange == undefined ? undefined : eval(attrs.xaxisFixedrange),
										// Optional. Determines whether or not this axis is zoom-able. If true, then zoom is disabled.
										
					tickmode:			attrs.xaxisTickmode,
										// Optional. Sets the tick mode for this axis. If "auto", the number of ticks is set via `nticks`. If 
										//   "linear", the placement of the ticks is determined by a starting position `tick0` and a tick step 
										//   `dtick` ("linear" is the default value if `tick0` and `dtick` are provided). If "array", the 
										//   placement of the ticks is set via `tickvals` and the tick text is `ticktext`. ("array" is the default 
										//   value if `tickvals` is provided).
										//   enumerated: "auto" | "linear" | "array"
										
					nticks:				attrs.xaxisNticks == undefined ? undefined : eval(attrs.xaxisNticks),
										// Optional. Specifies the maximum number of ticks for the particular axis. The actual number of ticks will 
										//   be chosen automatically to be less than or equal to `nticks`. Has an effect only if `tickmode` is set to "auto".
					
					tick0:				attrs.xaxisTick0,
										// Optional. Sets the placement of the first tick on this axis. Use with `dtick`. If the axis `type` is "log", 
										//   then you must take the log of your starting tick (e.g. to set the starting tick to 100, set the `tick0` 
										//   to 2) except when `dtick`="L<f>" (see `dtick` for more info). If the axis `type` is "date", it should be 
										//   a date string, like date data. If the axis `type` is "category", it should be a number, using the scale 
										//   where each category is assigned a serial number from zero in the order it appears.
										
					dtick:				attrs.xaxisDtick,
										// Optional. Sets the step in-between ticks on this axis.
										
					tickvals:			attrs.xaxisTickvals == undefined ? undefined : eval(attrs.xaxisTickvals),
										// Optional. Sets the values at which ticks on this axis appear. Only has an effect if `tickmode` is set to 
										//   "array". Used with `ticktext`.
										
					ticktext:			attrs.xaxisTicktext == undefined ? undefined : eval(attrs.xaxisTicktext),
										// Optional. Sets the text displayed at the ticks position via `tickvals`. Only has an effect if `tickmode` 
										//   is set to "array". Used with `tickvals`.
										
					ticks:				attrs.xaxisTicks,
										// Optional. Determines whether ticks are drawn or not. If "", this axis' ticks are not drawn. If "outside" 
										//   ("inside"), this axis' are drawn outside (inside) the axis lines.

					ticklen:			attrs.xaxisTicklen,
										// Optional. Sets the tick length (in px).
					
					tickwidth:			attrs.xaxisTickwidth,
										// Optional. Sets the tick width (in px).
										
					tickcolor:			attrs.xaxisTickcolor,
										// Optional. Sets the tick color.
					
					showticklabels:		attrs.xaxisShowticklabels == undefined ? undefined : eval(attrs.xaxisShowticklabels),
										// Optional. Determines whether or not the tick labels are drawn.

					tickfont: {
						family: 		attrs.xaxisTickfontFamily,
										// Optional. Sets the tick font family.

						size:			attrs.xaxisTickfontSize == undefined || attrs.xaxisTickfontSize.length == 0 ? undefined : eval(attrs.xaxisTickfontSize),
										// Optional. Sets the tick font size.

						color:			attrs.xaxisTickfontColor,
										// Optional. Sets the tick font color.
					},
					
					tickangle:			attrs.xaxisTickangle,
										// Optional. Sets the angle of the tick labels with respect to the horizontal. For example, a `tickangle` 
										//   of -90 draws the tick labels vertically.
										
					tickprefix:			attrs.xaxisTickprefix,
										// Optional. Sets a tick label prefix.
										
					showtickprefix:		attrs.xaxisShowtickprefix,
										// Optional. If "all", all tick labels are displayed with a prefix. If "first", only the first tick is 
										//   displayed with a prefix. If "last", only the last tick is displayed with a suffix. If "none", tick 
										//   prefixes are hidden.
										//   enumerated: "all" | "first" | "last" | "none"

					ticksuffix:			attrs.xaxisTicksuffix,
										// Optional. Sets a tick label suffix.
										
					showticksuffix:		attrs.xaxisShowticksuffix,
										// Optional. If "all", all tick labels are displayed with a suffix. If "first", only the first tick is 
										//   displayed with a suffix. If "last", only the last tick is displayed with a suffix. If "none", tick 
										//   suffixes are hidden.
										//   enumerated: "all" | "first" | "last" | "none"
										
					showline:			attrs.xaxisShowline == undefined || attrs.xaxisShowline.length == 0 ? false : eval(attrs.xaxisShowline),
										// Optional. Determines whether or not a line bounding this axis is drawn.
										
					linecolor:			attrs.xaxisLinecolor,
										// Optional. Sets the axis line color.
										
					linewidth:			attrs.xaxisLinewidth,
										// Optional. Sets the width (in px) of the axis line.
										
					showgrid: 			attrs.xaxisShowgrid == undefined ? undefined : eval(attrs.xaxisShowgrid),
										// Optional. Determines whether or not grid lines are drawn. If "true", the grid lines are drawn at every tick mark.
										
					gridcolor:			attrs.xaxisGridcolor,
										// Optional. Sets the color of the grid lines.
										
					gridwidth:			attrs.xaxisGridwidth,
										// Optional. Sets the width (in px) of the grid lines.
					
					zeroline:			attrs.xaxisZeroline == undefined || attrs.xaxisZeroline.length == 0 ? false : eval(attrs.xaxisZeroline),
										// Optional. Determines whether or not a line is drawn at along the 0 value of this axis. If "true", the 
										//   zero line is drawn on top of the grid lines.
										
					zerolinecolor:		attrs.xaxisZerolinecolor,
										// Optional. Sets the line color of the zero line.
										
					zerolinewidth:		attrs.xaxisZerolinewidth,
										// Optional. Sets the width (in px) of the zero line.
					
					side:				attrs.xaxisSide,
										// Optional. Determines whether a x (y) axis is positioned at the "bottom" ("left") or "top" ("right") 
										//   of the plotting area.
				},
			
				yaxis: {
					color:				attrs.yaxisColor,
										// Optional. Sets default for all colors associated with this axis all at once: line, font, 
										//   tick, and grid colors. Grid color is lightened by blending this with the plot background 
										//   Individual pieces can override this.
										
					title:				attrs.yaxisTitle,
										// Optional. Sets the title of this axis.
										
					titlefont: {
						family: 		attrs.yaxisTitlefontFamily,
										// Optional. Sets the title font family.

						size:			attrs.yaxisTitlefontSize == undefined || attrs.yaxisTitlefontSize.length == 0 ? undefined : eval(attrs.yaxisTitlefontSize),
										// Optional. Sets the title font size.

						color:			attrs.yaxisTitlefontColor,
										// Optional. Sets the title font color.
					},
					
					type:				attrs.yaxisType,
										// Optional. Sets the axis type. 
										//   enumerated: "-" | "linear" | "log" | "date" | "category"
										
					autorange:			attrs.yaxisAutorange,
										// Optional. Determines whether or not the range of this axis is computed in relation to the 
										//   input data. See `rangemode` for more info. If `range` is provided, then `autorange` is set to "false".
					
					rangemode:			attrs.yaxisRangemode,
										// Optional. If "normal", the range is computed in relation to the extrema of the input data. If "tozero"`, 
										//   the range extends to 0, regardless of the input data If "nonnegative", the range is non-negative, 
										//   regardless of the input data.
										//   enumerated: "normal" | "tozero" | "nonnegative"
										
					range:				attrs.yaxisRange == undefined ? undefined : eval(attrs.yaxisRange),
										// Optional. Sets the range of this axis. If the axis `type` is "log", then you must take the log of your 
										//   desired range (e.g. to set the range from 1 to 100, set the range from 0 to 2). If the axis `type` is 
										//   "date", it should be date strings, like date data, though Date objects and unix milliseconds will be 
										//   accepted and converted to strings. If the axis `type` is "category", it should be numbers, using the 
										//   scale where each category is assigned a serial number from zero in the order it appears. 

					fixedrange:			attrs.yaxisFixedrange == undefined ? undefined : eval(attrs.yaxisFixedrange),
										// Optional. Determines whether or not this axis is zoom-able. If true, then zoom is disabled.
										
					tickmode:			attrs.yaxisTickmode,
										// Optional. Sets the tick mode for this axis. If "auto", the number of ticks is set via `nticks`. If 
										//   "linear", the placement of the ticks is determined by a starting position `tick0` and a tick step 
										//   `dtick` ("linear" is the default value if `tick0` and `dtick` are provided). If "array", the 
										//   placement of the ticks is set via `tickvals` and the tick text is `ticktext`. ("array" is the default 
										//   value if `tickvals` is provided).
										//   enumerated: "auto" | "linear" | "array"
										
					nticks:				attrs.yaxisNticks == undefined ? undefined : eval(attrs.yaxisNticks),
										// Optional. Specifies the maximum number of ticks for the particular axis. The actual number of ticks will 
										//   be chosen automatically to be less than or equal to `nticks`. Has an effect only if `tickmode` is set to "auto".
					
					tick0:				attrs.yaxisTick0,
										// Optional. Sets the placement of the first tick on this axis. Use with `dtick`. If the axis `type` is "log", 
										//   then you must take the log of your starting tick (e.g. to set the starting tick to 100, set the `tick0` 
										//   to 2) except when `dtick`="L<f>" (see `dtick` for more info). If the axis `type` is "date", it should be 
										//   a date string, like date data. If the axis `type` is "category", it should be a number, using the scale 
										//   where each category is assigned a serial number from zero in the order it appears.
										
					dtick:				attrs.yaxisDtick,
										// Optional. Sets the step in-between ticks on this axis.
										
					tickvals:			attrs.yaxisTickvals == undefined ? undefined : eval(attrs.yaxisTickvals),
										// Optional. Sets the values at which ticks on this axis appear. Only has an effect if `tickmode` is set to 
										//   "array". Used with `ticktext`.
										
					ticktext:			attrs.yaxisTicktext == undefined ? undefined : eval(attrs.yaxisTicktext),
										// Optional. Sets the text displayed at the ticks position via `tickvals`. Only has an effect if `tickmode` 
										//   is set to "array". Used with `tickvals`.
										
					ticks:				attrs.yaxisTicks,
										// Optional. Determines whether ticks are drawn or not. If "", this axis' ticks are not drawn. If "outside" 
										//   ("inside"), this axis' are drawn outside (inside) the axis lines.

					ticklen:			attrs.yaxisTicklen,
										// Optional. Sets the tick length (in px).
					
					tickwidth:			attrs.yaxisTickwidth,
										// Optional. Sets the tick width (in px).
										
					tickcolor:			attrs.yaxisTickcolor,
										// Optional. Sets the tick color.
					
					showticklabels:		attrs.yaxisShowticklabels == undefined ? undefined : eval(attrs.yaxisShowticklabels),
										// Optional. Determines whether or not the tick labels are drawn.

					tickfont: {
						family: 		attrs.yaxisTickfontFamily,
										// Optional. Sets the tick font family.

						size:			attrs.yaxisTickfontSize == undefined || attrs.yaxisTickfontSize.length == 0 ? undefined : eval(attrs.yaxisTickfontSize),
										// Optional. Sets the tick font size.

						color:			attrs.yaxisTickfontColor,
										// Optional. Sets the tick font color.
					},
					
					tickangle:			attrs.yaxisTickangle,
										// Optional. Sets the angle of the tick labels with respect to the horizontal. For example, a `tickangle` 
										//   of -90 draws the tick labels vertically.
										
					tickprefix:			attrs.yaxisTickprefix,
										// Optional. Sets a tick label prefix.
										
					showtickprefix:		attrs.yaxisShowtickprefix,
										// Optional. If "all", all tick labels are displayed with a prefix. If "first", only the first tick is 
										//   displayed with a prefix. If "last", only the last tick is displayed with a suffix. If "none", tick 
										//   prefixes are hidden.
										//   enumerated: "all" | "first" | "last" | "none"

					ticksuffix:			attrs.yaxisTicksuffix,
										// Optional. Sets a tick label suffix.
										
					showticksuffix:		attrs.yaxisShowticksuffix,
										// Optional. If "all", all tick labels are displayed with a suffix. If "first", only the first tick is 
										//   displayed with a suffix. If "last", only the last tick is displayed with a suffix. If "none", tick 
										//   suffixes are hidden.
										//   enumerated: "all" | "first" | "last" | "none"
										
					showline:			attrs.yaxisShowline == undefined || attrs.yaxisShowline.length == 0 ? false : eval(attrs.yaxisShowline),
										// Optional. Determines whether or not a line bounding this axis is drawn.
										
					linecolor:			attrs.yaxisLinecolor,
										// Optional. Sets the axis line color.
										
					linewidth:			attrs.yaxisLinewidth,
										// Optional. Sets the width (in px) of the axis line.
										
					showgrid: 			attrs.yaxisShowgrid == undefined ? undefined : eval(attrs.yaxisShowgrid),
										// Optional. Determines whether or not grid lines are drawn. If "true", the grid lines are drawn at every tick mark.
										
					gridcolor:			attrs.yaxisGridcolor,
										// Optional. Sets the color of the grid lines.
										
					gridwidth:			attrs.yaxisGridwidth,
										// Optional. Sets the width (in px) of the grid lines.
					
					zeroline:			attrs.yaxisZeroline == undefined || attrs.yaxisZeroline.length == 0 ? false : eval(attrs.yaxisZeroline),
										// Optional. Determines whether or not a line is drawn at along the 0 value of this axis. If "true", the 
										//   zero line is drawn on top of the grid lines.
										
					zerolinecolor:		attrs.yaxisZerolinecolor,
										// Optional. Sets the line color of the zero line.
										
					zerolinewidth:		attrs.yaxisZerolinewidth,
										// Optional. Sets the width (in px) of the zero line.
					
					side:				attrs.yaxisSide,
										// Optional. Determines whether a x (y) axis is positioned at the "bottom" ("left") or "top" ("right") 
										//   of the plotting area.
										
					overlaying:			attrs.yaxisOverlaying,
										// Optional. If set a same-letter axis id, this axis is overlaid on top of the corresponding same-letter axis. 
										//   If "false", this axis does not overlay any same-letter axes.																				
				},

				zaxis: {
					color:				attrs.zaxisColor,
										// Optional. Sets default for all colors associated with this axis all at once: line, font, 
										//   tick, and grid colors. Grid color is lightened by blending this with the plot background 
										//   Individual pieces can override this.
										
					title:				attrs.zaxisTitle,
										// Optional. Sets the title of this axis.
										
					titlefont: {
						family: 		attrs.zaxisTitlefontFamily,
										// Optional. Sets the title font family.

						size:			attrs.zaxisTitlefontSize == undefined || attrs.zaxisTitlefontSize.length == 0 ? undefined : eval(attrs.zaxisTitlefontSize),
										// Optional. Sets the title font size.

						color:			attrs.zaxisTitlefontColor,
										// Optional. Sets the title font color.
					},
					
					type:				attrs.zaxisType,
										// Optional. Sets the axis type. 
										//   enumerated: "-" | "linear" | "log" | "date" | "category"
										
					autorange:			attrs.zaxisAutorange,
										// Optional. Determines whether or not the range of this axis is computed in relation to the 
										//   input data. See `rangemode` for more info. If `range` is provided, then `autorange` is set to "false".
					
					rangemode:			attrs.zaxisRangemode,
										// Optional. If "normal", the range is computed in relation to the extrema of the input data. If "tozero"`, 
										//   the range extends to 0, regardless of the input data If "nonnegative", the range is non-negative, 
										//   regardless of the input data.
										//   enumerated: "normal" | "tozero" | "nonnegative"
										
					range:				attrs.zaxisRange == undefined ? undefined : eval(attrs.zaxisRange),
										// Optional. Sets the range of this axis. If the axis `type` is "log", then you must take the log of your 
										//   desired range (e.g. to set the range from 1 to 100, set the range from 0 to 2). If the axis `type` is 
										//   "date", it should be date strings, like date data, though Date objects and unix milliseconds will be 
										//   accepted and converted to strings. If the axis `type` is "category", it should be numbers, using the 
										//   scale where each category is assigned a serial number from zero in the order it appears. 

					fixedrange:			attrs.zaxisFixedrange == undefined ? undefined : eval(attrs.zaxisFixedrange),
										// Optional. Determines whether or not this axis is zoom-able. If true, then zoom is disabled.
										
					tickmode:			attrs.zaxisTickmode,
										// Optional. Sets the tick mode for this axis. If "auto", the number of ticks is set via `nticks`. If 
										//   "linear", the placement of the ticks is determined by a starting position `tick0` and a tick step 
										//   `dtick` ("linear" is the default value if `tick0` and `dtick` are provided). If "array", the 
										//   placement of the ticks is set via `tickvals` and the tick text is `ticktext`. ("array" is the default 
										//   value if `tickvals` is provided).
										//   enumerated: "auto" | "linear" | "array"
										
					nticks:				attrs.zaxisNticks == undefined ? undefined : eval(attrs.zaxisNticks),
										// Optional. Specifies the maximum number of ticks for the particular axis. The actual number of ticks will 
										//   be chosen automatically to be less than or equal to `nticks`. Has an effect only if `tickmode` is set to "auto".
					
					tick0:				attrs.zaxisTick0,
										// Optional. Sets the placement of the first tick on this axis. Use with `dtick`. If the axis `type` is "log", 
										//   then you must take the log of your starting tick (e.g. to set the starting tick to 100, set the `tick0` 
										//   to 2) except when `dtick`="L<f>" (see `dtick` for more info). If the axis `type` is "date", it should be 
										//   a date string, like date data. If the axis `type` is "category", it should be a number, using the scale 
										//   where each category is assigned a serial number from zero in the order it appears.
										
					dtick:				attrs.zaxisDtick,
										// Optional. Sets the step in-between ticks on this axis.
										
					tickvals:			attrs.zaxisTickvals == undefined ? undefined : eval(attrs.zaxisTickvals),
										// Optional. Sets the values at which ticks on this axis appear. Only has an effect if `tickmode` is set to 
										//   "array". Used with `ticktext`.
										
					ticktext:			attrs.zaxisTicktext == undefined ? undefined : eval(attrs.zaxisTicktext),
										// Optional. Sets the text displayed at the ticks position via `tickvals`. Only has an effect if `tickmode` 
										//   is set to "array". Used with `tickvals`.
										
					ticks:				attrs.zaxisTicks,
										// Optional. Determines whether ticks are drawn or not. If "", this axis' ticks are not drawn. If "outside" 
										//   ("inside"), this axis' are drawn outside (inside) the axis lines.

					ticklen:			attrs.zaxisTicklen,
										// Optional. Sets the tick length (in px).
					
					tickwidth:			attrs.zaxisTickwidth,
										// Optional. Sets the tick width (in px).
										
					tickcolor:			attrs.zaxisTickcolor,
										// Optional. Sets the tick color.
					
					showticklabels:		attrs.zaxisShowticklabels == undefined ? undefined : eval(attrs.zaxisShowticklabels),
										// Optional. Determines whether or not the tick labels are drawn.

					tickfont: {
						family: 		attrs.zaxisTickfontFamily,
										// Optional. Sets the tick font family.

						size:			attrs.zaxisTickfontSize == undefined || attrs.zaxisTickfontSize.length == 0 ? undefined : eval(attrs.zaxisTickfontSize),
										// Optional. Sets the tick font size.

						color:			attrs.zaxisTickfontColor,
										// Optional. Sets the tick font color.
					},
					
					tickangle:			attrs.zaxisTickangle,
										// Optional. Sets the angle of the tick labels with respect to the horizontal. For example, a `tickangle` 
										//   of -90 draws the tick labels vertically.
										
					tickprefix:			attrs.zaxisTickprefix,
										// Optional. Sets a tick label prefix.
										
					showtickprefix:		attrs.zaxisShowtickprefix,
										// Optional. If "all", all tick labels are displayed with a prefix. If "first", only the first tick is 
										//   displayed with a prefix. If "last", only the last tick is displayed with a suffix. If "none", tick 
										//   prefixes are hidden.
										//   enumerated: "all" | "first" | "last" | "none"

					ticksuffix:			attrs.zaxisTicksuffix,
										// Optional. Sets a tick label suffix.
										
					showticksuffix:		attrs.zaxisShowticksuffix,
										// Optional. If "all", all tick labels are displayed with a suffix. If "first", only the first tick is 
										//   displayed with a suffix. If "last", only the last tick is displayed with a suffix. If "none", tick 
										//   suffixes are hidden.
										//   enumerated: "all" | "first" | "last" | "none"
										
					showline:			attrs.zaxisShowline == undefined || attrs.zaxisShowline.length == 0 ? false : eval(attrs.zaxisShowline),
										// Optional. Determines whether or not a line bounding this axis is drawn.
										
					linecolor:			attrs.zaxisLinecolor,
										// Optional. Sets the axis line color.
										
					linewidth:			attrs.zaxisLinewidth,
										// Optional. Sets the width (in px) of the axis line.
										
					showgrid: 			attrs.zaxisShowgrid == undefined ? undefined : eval(attrs.zaxisShowgrid),
										// Optional. Determines whether or not grid lines are drawn. If "true", the grid lines are drawn at every tick mark.
										
					gridcolor:			attrs.zaxisGridcolor,
										// Optional. Sets the color of the grid lines.
										
					gridwidth:			attrs.zaxisGridwidth,
										// Optional. Sets the width (in px) of the grid lines.
					
					zeroline:			attrs.zaxisZeroline == undefined || attrs.zaxisZeroline.length == 0 ? false : eval(attrs.zaxisZeroline),
										// Optional. Determines whether or not a line is drawn at along the 0 value of this axis. If "true", the 
										//   zero line is drawn on top of the grid lines.
										
					zerolinecolor:		attrs.zaxisZerolinecolor,
										// Optional. Sets the line color of the zero line.
										
					zerolinewidth:		attrs.zaxisZerolinewidth,
										// Optional. Sets the width (in px) of the zero line.
					
					side:				attrs.zaxisSide,
										// Optional. Determines whether a x (y) axis is positioned at the "bottom" ("left") or "top" ("right") 
										//   of the plotting area.
										
					overlaying:			attrs.zaxisOverlaying,
										// Optional. If set a same-letter axis id, this axis is overlaid on top of the corresponding same-letter axis. 
										//   If "false", this axis does not overlay any same-letter axes.																				
				},
				
			}
			
			// Enable second y-axis
			if(attrs.yaxis2Overlaying != undefined) {
				scope.chartOptions.yaxis2 = {
						color:				attrs.yaxis2Color,
											// Optional. Sets default for all colors associated with this axis all at once: line, font, 
											//   tick, and grid colors. Grid color is lightened by blending this with the plot background 
											//   Individual pieces can override this.
											
						title:				attrs.yaxis2Title,
											// Optional. Sets the title of this axis.
											
						titlefont: {
							family: 		attrs.yaxis2TitlefontFamily,
											// Optional. Sets the title font family.

							size:			attrs.yaxis2TitlefontSize == undefined || attrs.yaxis2TitlefontSize.length == 0 ? undefined : eval(attrs.yaxis2TitlefontSize),
											// Optional. Sets the title font size.

							color:			attrs.yaxis2TitlefontColor,
											// Optional. Sets the title font color.
						},
						
						type:				attrs.yaxis2Type,
											// Optional. Sets the axis type. 
											//   enumerated: "-" | "linear" | "log" | "date" | "category"
											
						autorange:			attrs.yaxis2Autorange,
											// Optional. Determines whether or not the range of this axis is computed in relation to the 
											//   input data. See `rangemode` for more info. If `range` is provided, then `autorange` is set to "false".
						
						rangemode:			attrs.yaxis2Rangemode,
											// Optional. If "normal", the range is computed in relation to the extrema of the input data. If "tozero"`, 
											//   the range extends to 0, regardless of the input data If "nonnegative", the range is non-negative, 
											//   regardless of the input data.
											//   enumerated: "normal" | "tozero" | "nonnegative"
											
						range:				attrs.yaxis2Range == undefined ? undefined : eval(attrs.yaxis2Range),
											// Optional. Sets the range of this axis. If the axis `type` is "log", then you must take the log of your 
											//   desired range (e.g. to set the range from 1 to 100, set the range from 0 to 2). If the axis `type` is 
											//   "date", it should be date strings, like date data, though Date objects and unix milliseconds will be 
											//   accepted and converted to strings. If the axis `type` is "category", it should be numbers, using the 
											//   scale where each category is assigned a serial number from zero in the order it appears. 

						fixedrange:			attrs.yaxis2Fixedrange == undefined ? undefined : eval(attrs.yaxis2Fixedrange),
											// Optional. Determines whether or not this axis is zoom-able. If true, then zoom is disabled.
											
						tickmode:			attrs.yaxis2Tickmode,
											// Optional. Sets the tick mode for this axis. If "auto", the number of ticks is set via `nticks`. If 
											//   "linear", the placement of the ticks is determined by a starting position `tick0` and a tick step 
											//   `dtick` ("linear" is the default value if `tick0` and `dtick` are provided). If "array", the 
											//   placement of the ticks is set via `tickvals` and the tick text is `ticktext`. ("array" is the default 
											//   value if `tickvals` is provided).
											//   enumerated: "auto" | "linear" | "array"
											
						nticks:				attrs.yaxis2Nticks == undefined ? undefined : eval(attrs.yaxis2Nticks),
											// Optional. Specifies the maximum number of ticks for the particular axis. The actual number of ticks will 
											//   be chosen automatically to be less than or equal to `nticks`. Has an effect only if `tickmode` is set to "auto".
						
						tick0:				attrs.yaxis2Tick0,
											// Optional. Sets the placement of the first tick on this axis. Use with `dtick`. If the axis `type` is "log", 
											//   then you must take the log of your starting tick (e.g. to set the starting tick to 100, set the `tick0` 
											//   to 2) except when `dtick`="L<f>" (see `dtick` for more info). If the axis `type` is "date", it should be 
											//   a date string, like date data. If the axis `type` is "category", it should be a number, using the scale 
											//   where each category is assigned a serial number from zero in the order it appears.
											
						dtick:				attrs.yaxis2Dtick,
											// Optional. Sets the step in-between ticks on this axis.
											
						tickvals:			attrs.yaxis2Tickvals == undefined ? undefined : eval(attrs.yaxis2Tickvals),
											// Optional. Sets the values at which ticks on this axis appear. Only has an effect if `tickmode` is set to 
											//   "array". Used with `ticktext`.
											
						ticktext:			attrs.yaxis2Ticktext == undefined ? undefined : eval(attrs.yaxis2Ticktext),
											// Optional. Sets the text displayed at the ticks position via `tickvals`. Only has an effect if `tickmode` 
											//   is set to "array". Used with `tickvals`.
											
						ticks:				attrs.yaxis2Ticks,
											// Optional. Determines whether ticks are drawn or not. If "", this axis' ticks are not drawn. If "outside" 
											//   ("inside"), this axis' are drawn outside (inside) the axis lines.

						ticklen:			attrs.yaxis2Ticklen,
											// Optional. Sets the tick length (in px).
						
						tickwidth:			attrs.yaxis2Tickwidth,
											// Optional. Sets the tick width (in px).
											
						tickcolor:			attrs.yaxis2Tickcolor,
											// Optional. Sets the tick color.
						
						showticklabels:		attrs.yaxis2Showticklabels == undefined ? undefined : eval(attrs.yaxis2Showticklabels),
											// Optional. Determines whether or not the tick labels are drawn.

						tickfont: {
							family: 		attrs.yaxis2TickfontFamily,
											// Optional. Sets the tick font family.

							size:			attrs.yaxis2TickfontSize == undefined || attrs.yaxis2TickfontSize.length == 0 ? undefined : eval(attrs.yaxis2TickfontSize),
											// Optional. Sets the tick font size.

							color:			attrs.yaxis2TickfontColor,
											// Optional. Sets the tick font color.
						},
						
						tickangle:			attrs.yaxis2Tickangle,
											// Optional. Sets the angle of the tick labels with respect to the horizontal. For example, a `tickangle` 
											//   of -90 draws the tick labels vertically.
											
						tickprefix:			attrs.yaxis2Tickprefix,
											// Optional. Sets a tick label prefix.
											
						showtickprefix:		attrs.yaxis2Showtickprefix,
											// Optional. If "all", all tick labels are displayed with a prefix. If "first", only the first tick is 
											//   displayed with a prefix. If "last", only the last tick is displayed with a suffix. If "none", tick 
											//   prefixes are hidden.
											//   enumerated: "all" | "first" | "last" | "none"

						ticksuffix:			attrs.yaxis2Ticksuffix,
											// Optional. Sets a tick label suffix.
											
						showticksuffix:		attrs.yaxis2Showticksuffix,
											// Optional. If "all", all tick labels are displayed with a suffix. If "first", only the first tick is 
											//   displayed with a suffix. If "last", only the last tick is displayed with a suffix. If "none", tick 
											//   suffixes are hidden.
											//   enumerated: "all" | "first" | "last" | "none"
											
						showline:			attrs.yaxis2Showline == undefined || attrs.yaxis2Showline.length == 0 ? false : eval(attrs.yaxis2Showline),
											// Optional. Determines whether or not a line bounding this axis is drawn.
											
						linecolor:			attrs.yaxis2Linecolor,
											// Optional. Sets the axis line color.
											
						linewidth:			attrs.yaxis2Linewidth,
											// Optional. Sets the width (in px) of the axis line.
											
						showgrid: 			attrs.yaxis2Showgrid == undefined ? undefined : eval(attrs.yaxis2Showgrid),
											// Optional. Determines whether or not grid lines are drawn. If "true", the grid lines are drawn at every tick mark.
											
						gridcolor:			attrs.yaxis2Gridcolor,
											// Optional. Sets the color of the grid lines.
											
						gridwidth:			attrs.yaxis2Gridwidth,
											// Optional. Sets the width (in px) of the grid lines.
						
						zeroline:			attrs.yaxis2Zeroline == undefined || attrs.yaxis2Zeroline.length == 0 ? false : eval(attrs.yaxis2Zeroline),
											// Optional. Determines whether or not a line is drawn at along the 0 value of this axis. If "true", the 
											//   zero line is drawn on top of the grid lines.
											
						zerolinecolor:		attrs.yaxis2Zerolinecolor,
											// Optional. Sets the line color of the zero line.
											
						zerolinewidth:		attrs.yaxis2Zerolinewidth,
											// Optional. Sets the width (in px) of the zero line.
						
						side:				attrs.yaxis2Side,
											// Optional. Determines whether a x (y) axis is positioned at the "bottom" ("left") or "top" ("right") 
											//   of the plotting area.
											
						overlaying:			attrs.yaxis2Overlaying,
											// Optional. If set a same-letter axis id, this axis is overlaid on top of the corresponding same-letter axis. 
											//   If "false", this axis does not overlay any same-letter axes.										
					};
			}
			
			// Read the context menu options from attributes
			scope.chartContextMenu = [];
			$(element).find('chart-context-menu').each(function(index, chartContextMenuElement) {
				var options = {
					"divider":				$(chartContextMenuElement).attr('divider'),
					
					"label":				$(chartContextMenuElement).attr('label'),
					
					"displayCondition":		$(chartContextMenuElement).attr('display-condition'),
					
					"eventId":				$(chartContextMenuElement).attr('event-id'),
				};
				scope.chartContextMenu.push(options);
			});
						
			// Declare the overlay counter
			scope.overlay = 0;

			// Watch for the dataset creation
			scope.$watch('dataset', function() {
				if(scope.dataset != undefined) {
					scope.registerSink();
				}
			});
			
			// Watch the overlay count and show and hide overlay div
			scope.$watch('overlay', function() {
				if(scope.overlay > 0)
					Overlay.show(element);
				else
					Overlay.hide(element);
			});
			
			// Process an object event
			scope.processObject = function(object, operation) {
				if(scope.chart.series == undefined) return;
				
				for(var idx = 0; idx < scope.chart.series.length; idx++) {
					var series = scope.chart.series[idx];
					scope.processSeriesObject(series, object, operation);
				}
			};
			
			// Process an object event for a series
			scope.processSeriesObject = function(series, object, operation) {
				// Generate the pointId and retrieve the existing point from the series
				var pointId = 'point.' + series.data.range1 + "." + object._key;				
				var point = series.getPoint(pointId);

				// If operation is delete and the point exists, then remove it
				if(operation == 'delete' && point != undefined) {
					point.remove();
				}
				// Else if operation is insert or update
				else if(operation == 'insert' || operation == 'update') {
					// Create a datum object
					var datum = {
						id: 		pointId,
									// Unique identifier

						domain:		object[scope.chart.category],
									// Domain
									
						range1:		object[series.data.range1],
									// Range1
									
						range2:		object[series.data.range2],
									// Range2
									
						radius:		series.data.radius == undefined ? undefined : object[series.data.radius],
									// Radius, if specified

						open:		series.data.open == undefined ? undefined : object[series.data.open],
									// OHLC open if specified
									
						high:		series.data.high == undefined ? undefined : object[series.data.high],
									// OHLC high if specified
									
						low:		series.data.low == undefined ? undefined : object[series.data.low],
									// OHLC low if specified
									
						close:		series.data.close == undefined ? undefined : object[series.data.close],
									// OHLC close if specified
								
						heat: 		series.data.heat == undefined ? undefined : object[series.data.heat],
									// Heatmap heat if specified
						
						object:		object,
									// Data object
									
						events: 	{
										click: function(e) {
											$('.context-menu').hide();
										},
										contextmenu: function(e) {
											scope.chartMenu(object[series.data.range1], object, series, e);
										}
									},
									// Event handler
					};

					// If the point is undefined, add it to the series
					if(point == undefined)
						series.addPoint(datum);
					// Otherwise update the point
					else
						point.update(datum);
				}
			};
			
			// Create and register a sink
			scope.registerSink = function() {
				var is3D = false;
				for(var idx = 0; idx < scope.series.length; idx++) {
					if(scope.series[idx].seriesOptions.type == 'scatter3d') {
						is3D = true;
						break;
					}
				}
				
				// Create the chart and call chart ready
				var category = attrs.category;
				var inverted = attrs.inverted == undefined ? false : eval(attrs.inverted);

				scope.chart = new LiveChart(scope.graphDiv, category, inverted, scope.options, scope.chartOptions, scope.series, is3D);

				var sink = {				
					onSchemaChange:	function(schema) {
					},
					onSnapshotStart: function() {
						scope.snapshot = true;
						scope.overlay = scope.overlay + 1;
						scope.$applyAsync();
					},
					onSnapshotEnd: function() {
						scope.snapshot = undefined;
						scope.overlay = scope.overlay - 1;
						scope.$applyAsync();
					},
					onInsert: function(object) {
						scope.processObject(object, 'insert');
						scope.chart.update();
					},
					onUpdate: function(object) {
						scope.processObject(object, 'update');
						scope.chart.update();
					},
					onDelete: function(object) {
						scope.processObject(object, 'delete');
						scope.chart.update();
					},
					onClear: function() {
						scope.chart.clear();
						scope.chart.update();
					}
				};
				
				// Register the datasink with the dataset
				var datasink = new DataSink(scope.elementName, 'LivePlotly', sink);
				scope.dataset.registerSink(datasink);
				
			}
		
			// Bind listener on window resize to reflow chart
			EventListener.windowResize(function() {
				if(scope.chart == null) return;
				setTimeout(function() {
					scope.chart.resize();
				},0);
			});
			
			// Bind listener on component display for tabs, applications, and visibility-toggle divs
			EventListener.componentDisplay(element, function() {
				if(scope.chart == null) return;
				setTimeout(function() {
					scope.chart.resize();
				},100);
			});
		
			// Display the dataset modal
			scope.displayDatasetModal = function() {
				if(scope.options.datasetModal == undefined) return;
				
				// Setup modal options
				var options = {
					template: '<dataset-modal></dataset-modal>',
					inputContext: {
						dataset: scope.dataset,
						display: scope.options.datasetModal == 'R' || scope.options.datasetModal == 'RW',
						editor:  scope.options.datasetModal == 'RW',
					}
				};

				// Show the modal with the options
				var modal = Modal.show(options);
			};

			// Display chart config modal
			scope.chartConfig = function() {
				// TODO chart config
				$log.debug('not implemented yet');
			};
			
			// CHART MENU ACTIONS
			// Display the chart context menu
			scope.chartMenu = function(point, object, series, e) {
				// Add custom chart context menu items
				var customMenuItems = [];
				for(var idx = 0; idx < scope.chartContextMenu.length; idx++) {
					var chartContextMenuItem = scope.chartContextMenu[idx];
					var customMenuItem = {
						self:		this,
						config:		chartContextMenuItem,
						divider:	chartContextMenuItem.divider != undefined ? eval(chartContextMenuItem.divider) : false,
						eventId:	chartContextMenuItem.eventId,
						label:		chartContextMenuItem.label,
						display:	function() {
							var include = true;
							if(this.config.displayCondition != undefined) {								
								include = false;
								try {
									include = eval(this.config.displayCondition);
								}
								catch(err) {
									//$log.debug(err);
								}
							}
							return include;
						},
						action:		function(e) {
							scope.customAction(point, object, series, e);
						}
					};					
					customMenuItems.push(customMenuItem);
				}

				// Add conditional menu items
				var conditionalMenuItems = [];	
				conditionalMenuItems.push({
					menuItemClass:	'copy-point',
					display:		point != undefined
				});
				conditionalMenuItems.push({
					menuItemClass:	'copy-object',
					display:		object != undefined
				});
				
				// Create the menu
				var menu = ContextMenu.show(e, element, ".chart-context-menu", customMenuItems, conditionalMenuItems);			
				
				// Unbind existing click handlers
				$(menu).find('.copy-point').unbind('click');
				$(menu).find('.copy-object').unbind('click');

				// Bind new click handlers
				$(menu).find('.copy-point').click(function(e) {
					scope.copyPoint(point, e);
				});
				
				$(menu).find('.copy-object').click(function(e) {
					scope.copyObject(object, e);
				});
				
				return false;
			};
			
			// Determine if copy is supported
			scope.copySupported = document.queryCommandSupported('copy');
			
			// Copy point to clipboard
			scope.copyPoint = function(point, e) {
				var menu = $(element).find(".chart-context-menu");
				var copyText = point;
				Clipboard.copy(copyText, menu);				
			};

			// Copy object to clipboard
			scope.copyObject = function(object, e) {
				var mode = 'csv'
				var copyText = ''				
				if(mode == 'csv') {
					for(var idx = 0; idx < Object.keys(object).length; idx++) {
						if(idx > 0) copyText = copyText + ',';
						var k = Object.keys(object)[idx];
						copyText = copyText + object[k];
					}
				}
				else if(mode == 'json') {
					copyText = JSON.stringify(object, null, 2);
				}

				var menu = $(element).find(".chart-context-menu");				
				Clipboard.copy(copyText, menu);				
			};

			// Custom action
			scope.customAction = function(point, object, series, e) {
				var operation = $(e.target).attr('event-id');
				var topic = 'custom:' + scope.elementId;
				scope.$emit(topic, operation, point, object, series);
			};		

			// Event handlers to pause and resume for wheel and drag events for 3d charts
			$(element).bind('mousewheel', function(e) {
				scope.chart.pause();
				setTimeout(function() {
					scope.chart.resume();
				},500);
			});

			$(element).on('mousedown', function (evtDown) {
				$(element).on('mouseup', function handler(evt) {
					if(scope.chart.paused == true)
						scope.chart.resume();
					$(element).off('mouseup');
					$(element).off('mousemove');
				});
				$(element).on('mousemove', function handler(evt) {
					var delta = Math.max(Math.abs(evtDown.pageX - evt.pageX), Math.abs(evtDown.pageY- evt.pageY));
					if(delta > 5 && scope.chart.paused == false) {
						scope.chart.pause();
					}
				});
			});
		}		
	};	
});

app.directive('liveplotlySeries', function($log, $filter, Overlay, Modal, EventListener, ContextMenu, Clipboard) {
	return {
		replace: false,
		transclude: true,
		restrict: 'E',
		scope: {},
		require: '^liveplotly',
		controller: function($scope) {
		},
		link: function(scope, element, attrs, ctrl, transclude) {
			// DEFINE OBJECTS
			{
				// LiveChartSeries object
				{
					/**
					 * Create a new series for the chart.
					 * @param data fields
					 * @param options the general options
					 * @param seriesOptions the series options
					 */
					function LiveChartSeries(data, options, seriesOptions) {
						var self			= this;

						// Set series properties
						this.data			= data;
						this.options		= options;
						this.seriesOptions	= seriesOptions;
						this.domain			= [];
						this.range1			= [];
						this.range2			= [];
						this.heat			= [];
						
						// Collections to hold the data points
						this.pointData 		= new Collection(['id']);
						this.pointArr		= [];
						
						// Setup plotly properties
						this.trace 			= this.seriesOptions;
						
						// Setup x-y for scatter and bar chart types
						this.trace.x		= this.domain;
						this.trace.y		= this.range1;
						
						// Setup x-y-z for 3D chart types
						this.trace.z		= this.range2;
						
						// Setup labels and values for pie chart type
						this.trace.labels	= this.domain;
						this.trace.values	= this.range1;
						
						// If the radius is defined then set the marker size for bubble charts
						if(this.data.radius != undefined) {
							this.trace.marker.size	= [];
						}

						// If OHLC is defined the set the trace arrays for those data fields
						if(this.data.open != undefined && this.data.high != undefined && this.data.low != undefined && this.data.close != undefined) {
							this.trace.open		= [];
							this.trace.high 	= [];
							this.trace.low		= [];
							this.trace.close	= [];
						}
							
						// If heat is defined then set the z axis for heatmap
						if(this.data.heat != undefined) {
							this.trace.z	= this.heat;
						}
						
						// If point labels are enabled, override the text settings with a new array
						if(this.options.pointLabels == true || this.options.textExpression != undefined) {
							this.trace.text = [];
						}
					}
					
					/**
					 * Adds a new point to the series.
					 * @param datum the data to add
					 */
					LiveChartSeries.prototype.addPoint = function(datum) {
						if(this.seriesOptions.filter != undefined) {
							var object = datum.object;
							var b = eval(this.seriesOptions.filter);
							if(b == false) return;
						}
						
						// Create a new point
						var point = new LiveChartPoint(datum, this);

						// Put the point in the collection and array
						this.pointData.put(point);
						this.pointArr.push(point);

						// THIS IS A HACK
						if(this.inverted == true) {
							this.trace.x = this.range1;
							this.trace.y = this.domain;
						}
						else {
							this.trace.x = this.domain;
							this.trace.y = this.range1;
						}
						this.trace.z = this.range2;
					};
					
					/**
					 * Removes a point from the series.
					 * @param point the point to remove
					 */
					LiveChartSeries.prototype.removePoint = function(point) {
						// Get the index for the point
						var index = this.pointArr.indexOf(point);
						
						// Remove the point data from the domain and range
						this.domain.splice(index, 1);
						this.range1.splice(index, 1);
						this.range2.splice(index, 1);

						// Special case for pie charts when a purge happens
						//   Plotly doesn't like zero length arrays in pie charts, 
						//   so if all data has been removed, call draw to regen the chart
						if(this.seriesOptions.type == 'pie' && this.domain.length == 0) {
							var c = ctrl.getChart().draw();
						}
						
						// Remove the point data from the radius, if specified
						if(this.data.radius != undefined)
							this.trace.marker.size.splice(index, 1);
						
						// Remove the point data from OHLC, if specified
						if(this.data.open != undefined && this.data.high != undefined && this.data.low != undefined && this.data.close != undefined) {
							this.trace.open.splice(index, 1);
							this.trace.high.splice(index, 1);
							this.trace.low.splice(index, 1);
							this.trace.close.splice(index, 1);
						}
						
						// Remove the point data from the heat, if specified
						if(this.data.heat != undefined) {
							this.trace.heat.splice(index, i);
						}
						
						// Remove the point from the point label, if specified
						if(this.options.pointLabels == true || this.options.textExpression != undefined)
							this.trace.text.splice(index, 1);
						
						// Remove the point from the collections
						this.pointArr.splice(index, 1);
						this.pointData.remove(point.id);						
					};
					
					/**
					 * Removes all points from the series.
					 */
					LiveChartSeries.prototype.removeAllPoints = function() {
						// Remove the point data from the domain and range
						this.domain = [];
						this.range1 = [];
						this.range2 = [];
						
						// THIS IS A HACK
						this.trace.x = [];
						this.trace.y = [];
						this.trace.z = [];
						
						// Remove the point data from the radius, if specified
						if(this.data.radius != undefined)
							this.trace.marker.size = [];
						
						// Remove the point data from OHLC, if specified
						if(this.data.open != undefined && this.data.high != undefined && this.data.low != undefined && this.data.close != undefined) {
							this.trace.open 	= [];
							this.trace.high		= [];
							this.trace.low		= [];
							this.trace.close	= [];
						}
						
						// Remove the heat data from the heat array, if specified
						if(this.data.heat != undefined) {
							this.heat = [];
						}
						
						// Remove the point data from the point label, if specified
						if(this.options.pointLabels == true || this.options.textExpression != undefined)
							this.trace.text = [];
						
						// Remove the points from the collections
						this.pointArr = [];
						this.pointData.removeAll();
					};
					
					/**
					 * Returns the point data for a given id.
					 * @param id the id for the point to retrieve
					 */
					LiveChartSeries.prototype.getPoint = function(id) {
						return this.pointData.get(id);
					};

					/**
					 * Inverts the domain and range for the series.
					 */
					LiveChartSeries.prototype.invert = function() {
						this.inverted 		= true;
						this.trace.x		= this.range1;
						this.trace.y		= this.domain;
					};
				}
			
				// LiveChartPoint object
				{
					/**
					 * Create a new point for the series.
					 * @param datum data object 
					 * @param series the series to add the point to
					 */
					function LiveChartPoint(datum, series) {
						// Set the properties
						this.id 	= datum.id;
						this.datum 	= datum;
						this.series = series;

						// Push the domain and range into the series
						this.series.domain.push(datum.domain);
						this.series.range1.push(datum.range1);
						this.series.range2.push(datum.range2);
						
						// Push the radius into the series, if configured
						if(this.series.data.radius != undefined)
							this.series.trace.marker.size.push(datum.radius);						

						// Push the OHLC into the series, if configured
						if(this.series.data.open != undefined && this.series.data.high != undefined && this.series.data.low != undefined && this.series.data.close != undefined) {
							this.series.trace.open.push(datum.open);
							this.series.trace.high.push(datum.high);
							this.series.trace.low.push(datum.low);
							this.series.trace.close.push(datum.close);
						}

						// Push the heat into the series, if configured
						if(this.series.data.heat != undefined) {
							this.series.heat.push(datum.heat);
						}

						// Push the range into point label into the series, if configured
						if(this.series.options.textExpression != undefined) {
							var object = datum.object;							
							var t = eval(this.series.options.textExpression);
							this.series.trace.text.push(t);
						}
						else if(this.series.options.pointLabels == true)
							this.series.trace.text.push(datum.range1);

						// TODO						
						var c = ctrl.getChart();
						if(c.options.equalAxes == true) {
							if(isNaN(datum.domain) == false) {
								c.maxAxis = Math.max(c.maxAxis, datum.domain);
								c.minAxis = Math.min(c.minAxis, datum.domain);
							}
							if(isNaN(datum.range1) == false) {
								c.maxAxis = Math.max(c.maxAxis, datum.range1);
								c.minAxis = Math.min(c.minAxis, datum.range1);
							}
							if(isNaN(datum.range2) == false) {
								c.maxAxis = Math.max(c.maxAxis, datum.range2);
								c.minAxis = Math.min(c.minAxis, datum.range2);
							}
							c.layout.xaxis.range = [c.minAxis, c.maxAxis];
							c.layout.yaxis.range = [c.minAxis, c.maxAxis];
							c.layout.zaxis.range = [c.minAxis, c.maxAxis];
						}
					};
					
					/**
					 * Removes this point from it's series.
					 */
					LiveChartPoint.prototype.remove = function() {
						this.series.removePoint(this);
					};
					
					/**
					 * Update the point with new data.
					 * @param datum data object 
					 */
					LiveChartPoint.prototype.update = function(datum) {
						// Set the properties
						this.datum = datum;
						
						// Get the index for the current point in the series
						var index = this.series.pointArr.indexOf(this);
						
						// Update the domain and range in the series
						this.series.domain[index] = datum.domain;
						this.series.range1[index] = datum.range1;
						this.series.range2[index] = datum.range2;

						
						// Update the radius in the series, if configured
						if(this.series.data.radius != undefined) {
							this.series.trace.marker.size[index] = datum.radius;
						}

						// Update the OHLC into the series, if configured
						if(this.series.data.open != undefined && this.series.data.high != undefined && this.series.data.low != undefined && this.series.data.close != undefined) {
							this.series.trace.open[index] = datum.open;
							this.series.trace.high[index] = datum.high;
							this.series.trace.low[index] = datum.low;
							this.series.trace.close[index] = datum.close;
						}
						
						// Update the heat in the series, if configured
						if(this.series.data.heat != undefined) {
							this.series.heat[index] = datum.heat;
						}

						// Update the range into point label into the series, if configured
						if(this.series.options.textExpression != undefined) {
							var object = datum.object;
							var t = eval(this.series.options.textExpression);
							this.series.trace.text[index] = t;
						}
						else if(this.series.options.pointLabels == true)
							this.series.trace.text[index] = datum.range1;
						
						// TODO
						var c = ctrl.getChart();
						if(c.options.equalAxes == true) {
							if(isNaN(datum.domain) == false) {
								c.maxAxis = Math.max(c.maxAxis, datum.domain);
								c.minAxis = Math.min(c.minAxis, datum.domain);
							}
							if(isNaN(datum.range1) == false) {
								c.maxAxis = Math.max(c.maxAxis, datum.range1);
								c.minAxis = Math.min(c.minAxis, datum.range1);
							}
							if(isNaN(datum.range2) == false) {
								c.maxAxis = Math.max(c.maxAxis, datum.range2);
								c.minAxis = Math.min(c.minAxis, datum.range2);
							}
							c.layout.xaxis.range = [c.minAxis, c.maxAxis];
							c.layout.yaxis.range = [c.minAxis, c.maxAxis];
							c.layout.zaxis.range = [c.minAxis, c.maxAxis];
						}
					}
					
				}
			}

			// Transclude own scope
			transclude(scope, function (clone, scope) {
				element.append(clone);
			});

			// Set the element ID to scope if it exists
			if(attrs.id != undefined) scope.elementId = attrs.id;
			
			// Set the name to scope if it exists
			if(attrs.name != undefined) scope.elementName = attrs.name;
						
			// Read the options from attributes
			scope.options = {
				pointLabels:			attrs.pointLabels == undefined ? undefined : eval(attrs.pointLabels),
										// Optional. Display point labels.
										
				textExpression:			attrs.textExpression,
										// Optional. Expression to generate the text value.
			};
			
			// Read the series options from the attributes
			scope.seriesOptions = {
				filter:					attrs.filter,
				
				type: 					attrs.type,
										// Mandatory. Plotly chart type.
										
				visible: 				attrs.visible == "true" || attrs.visible == "false" ? eval(attrs.visible) : attrs.visible,
										// Optional. Display or hide series.
										//   true | false | "legendonly"
									
				showlegend:				attrs.showlegend == undefined ? undefined : eval(attrs.showlegend),
										// Optional. Show or hide the legend.
										
				legendgroup:			attrs.legendgroup,
										// Optional. Sets the legend group.
										
				opacity:				attrs.opacity,
										// Optional. Sets the opacity of the trace.
										
				name:					attrs.name,
										// Optional. Sets the trace name.
										
				hoverinfo:				attrs.hoverinfo,
										// Optional. Determines the trace info to show on hover.
										//   Any combination of "x", "y", "z", "text", "name" joined with a "+" OR "all" or "none" or "skip"
										
				text:					attrs.text == undefined ? undefined : attrs.text,
										// Optional. Sets text elements associated with each (x,y) pair. If a single string, the same 
										//   string appears over all the data points. If an array of string, the items are mapped in 
										//   order to the this trace's (x,y) coordinates.
										
				mode:					attrs.mode,
										// Optional. Determines the drawing mode for the scatter trace.
										//   Any combination of "lines", "markers", "text" joined with a "+" OR "none". 
										
				hoveron:				attrs.hoveron,
										// Optional. Hover effects.
										//   Any combination of "points", "fills" joined with a "+" 
										
				line: {
					color:				attrs.lineColor,
										// Optional. Sets the line color.
										
					width:				attrs.lineWidth,
										// Optional. Sets the line width in px.
										
					shape:				attrs.lineShape,
										// Optional. Determines the line shape. With "spline" the lines are drawn using spline interpolation. 
										//   The other available values correspond to step-wise line shapes.
										//   "linear" | "spline" | "hv" | "vh" | "hvh" | "vhv"
									
					smoothing:			attrs.lineSmoothing,
										// Optional. Has an effect only if `shape` is set to "spline" Sets the amount of smoothing. "0" corresponds to no smoothing .
										
					dash:				attrs.lineDash,
										// Optional. Sets the style of the lines. Set to a dash string type or a dash length in px.
										
					simplify:			attrs.lineSimplify,
										// Optional. Simplifies lines by removing nearly-collinear points. When transitioning lines, it may be desirable to disable 
										//   this so that the number of points along the resulting SVG path is unaffected..
				},
										
				connectgaps:			attrs.connectgaps,
										// Optional. Determines whether or not gaps (i.e. {nan} or missing values) in the provided data arrays are connected.
										
				fill:					attrs.fill,
										// Optional. Sets the area to fill with a solid color. Use with `fillcolor` if not "none". 
										//   "none" | "tozeroy" | "tozerox" | "tonexty" | "tonextx" | "toself" | "tonext"
										
				fillcolor:				attrs.fillColor,
										// Optional. Sets the fill color.
										
				marker: {
					colors:				attrs.pieMarkerColors == undefined ? undefined : eval(attrs.pieMarkerColors),
										// Optional. Pie. Sets the color of each sector of this pie chart. If not specified, the default trace color set 
										//   is used to pick the sector colors.
										
					symbol:				attrs.markerSymbol,
										// Optional. Sets the marker symbol type.
										//   "0" | "circle" | "100" | "circle-open" | "200" | "circle-dot" | "300" | "circle-open-dot" | "1" | "square" | "101" | 
										//   "square-open" | "201" | "square-dot" | "301" | "square-open-dot" | "2" | "diamond" | "102" | "diamond-open" | "202" | 
										//   "diamond-dot" | "302" | "diamond-open-dot" | "3" | "cross" | "103" | "cross-open" | "203" | "cross-dot" | "303" | 
										//   "cross-open-dot" | "4" | "x" | "104" | "x-open" | "204" | "x-dot" | "304" | "x-open-dot" | "5" | "triangle-up" | "105" | 
										//   "triangle-up-open" | "205" | "triangle-up-dot" | "305" | "triangle-up-open-dot" | "6" | "triangle-down" | "106" | 
										//   "triangle-down-open" | "206" | "triangle-down-dot" | "306" | "triangle-down-open-dot" | "7" | "triangle-left" | "107" | 
										//   "triangle-left-open" | "207" | "triangle-left-dot" | "307" | "triangle-left-open-dot" | "8" | "triangle-right" | "108" | 
										//   "triangle-right-open" | "208" | "triangle-right-dot" | "308" | "triangle-right-open-dot" | "9" | "triangle-ne" | "109" | 
										//   "triangle-ne-open" | "209" | "triangle-ne-dot" | "309" | "triangle-ne-open-dot" | "10" | "triangle-se" | "110" | 
										//   "triangle-se-open" | "210" | "triangle-se-dot" | "310" | "triangle-se-open-dot" | "11" | "triangle-sw" | "111" | 
										//   "triangle-sw-open" | "211" | "triangle-sw-dot" | "311" | "triangle-sw-open-dot" | "12" | "triangle-nw" | "112" | 
										//   "triangle-nw-open" | "212" | "triangle-nw-dot" | "312" | "triangle-nw-open-dot" | "13" | "pentagon" | "113" | "pentagon-open" | 
										//   "213" | "pentagon-dot" | "313" | "pentagon-open-dot" | "14" | "hexagon" | "114" | "hexagon-open" | "214" | "hexagon-dot" | 
										//   "314" | "hexagon-open-dot" | "15" | "hexagon2" | "115" | "hexagon2-open" | "215" | "hexagon2-dot" | "315" | "hexagon2-open-dot" | 
										//   "16" | "octagon" | "116" | "octagon-open" | "216" | "octagon-dot" | "316" | "octagon-open-dot" | "17" | "star" | "117" | "star-open" | 
										//   "217" | "star-dot" | "317" | "star-open-dot" | "18" | "hexagram" | "118" | "hexagram-open" | "218" | "hexagram-dot" | "318" | 
										//   "hexagram-open-dot" | "19" | "star-triangle-up" | "119" | "star-triangle-up-open" | "219" | "star-triangle-up-dot" | "319" | 
										//   "star-triangle-up-open-dot" | "20" | "star-triangle-down" | "120" | "star-triangle-down-open" | "220" | "star-triangle-down-dot" | 
										//   "320" | "star-triangle-down-open-dot" | "21" | "star-square" | "121" | "star-square-open" | "221" | "star-square-dot" | "321" | 
										//   "star-square-open-dot" | "22" | "star-diamond" | "122" | "star-diamond-open" | "222" | "star-diamond-dot" | "322" | 
										//   "star-diamond-open-dot" | "23" | "diamond-tall" | "123" | "diamond-tall-open" | "223" | "diamond-tall-dot" | "323" | 
										//   "diamond-tall-open-dot" | "24" | "diamond-wide" | "124" | "diamond-wide-open" | "224" | "diamond-wide-dot" | "324" | 
										//   "diamond-wide-open-dot" | "25" | "hourglass" | "125" | "hourglass-open" | "26" | "bowtie" | "126" | "bowtie-open" | "27" | 
										//   "circle-cross" | "127" | "circle-cross-open" | "28" | "circle-x" | "128" | "circle-x-open" | "29" | "square-cross" | "129" | 
										//   "square-cross-open" | "30" | "square-x" | "130" | "square-x-open" | "31" | "diamond-cross" | "131" | "diamond-cross-open" | "32" | 
										//   "diamond-x" | "132" | "diamond-x-open" | "33" | "cross-thin" | "133" | "cross-thin-open" | "34" | "x-thin" | "134" | "x-thin-open" | 
										//   "35" | "asterisk" | "135" | "asterisk-open" | "36" | "hash" | "136" | "hash-open" | "236" | "hash-dot" | "336" | "hash-open-dot" | 
										//   "37" | "y-up" | "137" | "y-up-open" | "38" | "y-down" | "138" | "y-down-open" | "39" | "y-left" | "139" | "y-left-open" | "40" | 
										//   "y-right" | "140" | "y-right-open" | "41" | "line-ew" | "141" | "line-ew-open" | "42" | "line-ns" | "142" | "line-ns-open" | 
										//   "43" | "line-ne" | "143" | "line-ne-open" | "44" | "line-nw" | "144" | "line-nw-open"
										
					opacity:			attrs.markerOpacity,
										// Optional. Sets the marker opacity.
										
					size:				attrs.markerSize,
										// Optional. Sets the marker size.
										
					maxdisplayed:		attrs.markerMaxdisplayed,
										// Optional. Sets the marker maximum displayed.
										
					sizeref:			attrs.markerSizeref,
										// Optional. Has an effect only if `marker.size` is set to a numerical array. 
										//   Sets the scale factor used to determine the rendered size of marker points. 
										//   Use with `sizemin` and `sizemode`.
										
					sizemin:			attrs.markerSizemin,
										// Optional. Has an effect only if `marker.size` is set to a numerical array. 
										//   Sets the minimum size (in px) of the rendered marker points.
										
					sizemode:			attrs.markerSizemode,
										// Optional. Has an effect only if `marker.size` is set to a numerical array. 
										//   Sets the rule for which the data in `size` is converted to pixels.
										
					showscale:			attrs.markerShowscale,
										// Optional. Has an effect only if `marker.color` is set to a numerical array. 
										//   Determines whether or not a colorbar is displayed.
										
					color:				attrs.markerColor,
										// Optional. Sets the marker color.
										
					colorscale:			attrs.markerColorscale,
										// Optional. Sets the marker colorscale.
										
					cauto:				attrs.markerCauto,
										// Optional. Has an effect only if `marker.color` is set to a numerical array and `cmin`, `cmax` are set by the user. 
										
					cmax:				attrs.markerCmax,
										// Optional. Has an effect only if `marker.color` is set to a numerical array. Sets the upper bound of the color domain. 
										//   Value should be associated to the `marker.color` array index, and if set, `marker.cmin` must be set as well.
										
					cmin:				attrs.markerCmin,
										// Optional. Has an effect only if `marker.color` is set to a numerical array. Sets the lower bound of the color domain. 
										//   Value should be associated to the `marker.color` array index, and if set, `marker.cmax` must be set as well.
										
					autocolorscale:		attrs.markerAutocolorscale,
										// Optional. Has an effect only if `marker.color` is set to a numerical array. Determines whether the colorscale is a 
										//   default palette (`autocolorscale: true`) or the palette determined by `marker.colorscale`. In case `colorscale` 
										//   is unspecified or `autocolorscale` is true, the default palette will be chosen according to whether numbers in the 
										//   `color` array are all positive, all negative or mixed.
										
					reversescale:		attrs.markerReversescale,
										// Optional. Has an effect only if `marker.color` is set to a numerical array. Reverses the color mapping if true (`cmin` 
										//   will correspond to the last color in the array and `cmax` will correspond to the first color).
				},
										
				textposition:			attrs.textposition,
										// Optional. Sets the positions of the `text` elements with respects to the (x,y) coordinates.
										//   "top left" | "top center" | "top right" | "middle left" | "middle center" | "middle right" | 
										//   "bottom left" | "bottom center" | "bottom right" 
										
				textfont: {
					family:				attrs.textfontFamily,
					size:				attrs.textfontSize,
					color:				attrs.textfontColor,
				},

				insidetextfont: {
					family:				attrs.insidetextfontFamily,
					size:				attrs.insidetextfontSize,
					color:				attrs.insidetextfontColor,
				},

				outsidetextfont: {
					family:				attrs.outsidetextfontFamily,
					size:				attrs.outsidetextfontSize,
					color:				attrs.outsidetextfontColor,
				},
				
				orientation:			attrs.barOrientation, 
										// Optional. Bar. Sets the orientation of the bars. With "v" ("h"), the value of the each bar spans 
										//   along the vertical (horizontal).
										
				base:					attrs.barBase,
										// Optional. Bar. Sets where the bar base is drawn (in position axis units). In "stack" or "relative" 
										//   barmode, traces that set "base" will be excluded and drawn in "overlay" mode instead.
										
				offset:					attrs.barOffset,
										// Optional. Bar. Shifts the position where the bar is drawn (in position axis units). In "group" barmode, 
										//   traces that set "offset" will be excluded and drawn in "overlay" mode instead.
										
				width:					attrs.barWidth, 
										// Optional. Bar. Sets the bar width (in position axis units).
										
				labels:					attrs.pieLabels == undefined ? undefined : eval(attrs.pieLabels),
										// Optional. Pie. Sets the sector labels.
										
				textinfo:				attrs.pieTextinfo,
										// Optional. Pie. Any combination of "label", "text", "value", "percent" joined with a "+" OR "none". 
					
				hole:					attrs.pieHole,
										// Optional. Pie. Sets the fraction of the radius to cut out of the pie. Use this to make a donut chart.
										
				sort:					attrs.pieSort,
										// Optional. Pie. Determines whether or not the sectors of reordered from largest to smallest.
										
				direction:				attrs.pieDirection,
										// Optional. Pie. Specifies the direction at which succeeding sectors follow one another.
										
				rotation:				attrs.pieRotation,
										// Optional. Pie. Instead of the first slice starting at 12 o'clock, rotate to some other angle.
										
				pull:					attrs.piePull == undefined ? undefined : eval(attrs.piePull),
										// Optional. Pie.Sets the fraction of larger radius to pull the sectors out from the center. This can be a 
										//   constant to pull all slices apart from each other equally or an array to highlight one or more slices.

				xaxis:					attrs.xaxis,
										// Optional. Sets a reference between this trace's x coordinates and a 2D cartesian x axis. If "x" (the default value), 
										//   the x coordinates refer to `layout.xaxis`. If "x2", the x coordinates refer to `layout.xaxis2`, and so on.
										
				yaxis:					attrs.yaxis,
										// Optional. Sets a reference between this trace's y coordinates and a 2D cartesian y axis. If "y" (the default value), 
										//   the y coordinates refer to `layout.yaxis`. If "y2", the y coordinates refer to `layout.xaxis2`, and so on.
			};

			// Create the series
			var data = {
				range1:		attrs.data != undefined ? attrs.data : attrs.data1,
				range2:		attrs.data2,
				radius:		attrs.radius,
				open:		attrs.open,
				high:		attrs.high,
				low:		attrs.low,
				close:		attrs.close,
				heat:		attrs.heat,
			}
			
			scope.series = new LiveChartSeries(data, scope.options, scope.seriesOptions);
			ctrl.addSeries(scope.series);
		}
	};	
});
