/** 
 * Factory object for creating dataset instances from file or configuration.
 */
DatasetFactory = {
	/** 
	 * Initializes the factory. Datasets are created from the uri or configuration 
	 * passed through in the options object.
	 * @param {Object} options - options object containing dataset creation options
	 * @return {Promise} 
	 */
	init: function(options) {
		var self = this;
		
		var promise = new Promise(function(resolve, reject) {
			// Validate the options
			if(options == undefined)
				reject(Error('DatasetFactory options are undefined'));
			if(options.container == undefined)
				reject(Error('DatasetFactory container is undefined'));		

			// Setup the container objects
			container 			= options.container;
			container.datasets 	= [];
			container.warnings	= [];
			
			// Options can be either uri or configuration
			// If uri, load configuration from the file which will return it's own promise
			// If both uri and configuration are specified then uri will take precedence
			if(options.uri != null) {
				self.loadFile(options.uri).then(
					function(configuration) {
						// Process the configuration loaded from the file
						self.processConfiguration(configuration, resolve, reject);
					},
					function(err) {
						// Error happened loading the file, reject with the passed error
						reject(Error('URI ' + err.message));
					}
				);
			}
			// If configuration object then process that configuration
			else if(options.configuration != null) {
				self.processConfiguration(options.configuration, resolve, reject);
			}
			// Otherwise reject with an error for unknown configuration method
			else {
				reject(Error('DatasetFactory uri and configuration are both undefined'));		
			}
		});	
		
		return promise;
	},
	
	/** 
	 * Processes the configuration object. This can be passed directly or loaded
	 * from URI.
	 * @param {Object} configuration - the configuration object
	 */
	processConfiguration: function(configuration, resolve, reject) {
		var self = this;
		
		// Setup an object to hold dataset id to detect duplicates
		var duplicates = {};
		
		// Count the configured datasets. If there are none, resolve the promise
		var count = configuration.dataset == undefined ? 0 : configuration.dataset.length;
		if(count == 0) resolve();
		
		// Function to complete processing
		//   Once the count reaches zero, resolve the promise
		var complete = function() {
			count = count - 1;
			if(count == 0) resolve();
		};
		
		// Loop over each dataset in the configuration
		for(var idx = 0; idx < configuration.dataset.length; idx++) {
			// Get the configuration and validate
			var thisConfiguration = configuration.dataset[idx];
			var valid = self.validate(thisConfiguration, idx, duplicates, container.warnings);
			
			// If valid, create and initialize the dataset which will return it's own promise
			if(valid == true) {
				// Add the ID to the duplicate container
				duplicates[thisConfiguration.id] = thisConfiguration.id;
				
				// Create and initialize the dataset			
				new Dataset(container, thisConfiguration).init().then(
					function(dataset) {
						// Add the dataset to the container, then call complete
						container[dataset.id] = dataset;
						container.datasets.push(dataset);
						complete();
					},
					function(err) {
						// Log the error, then call complete
						console.error(err);
						complete();
					}
				);
			}
			// Otherwise skip this dataset and just call complete
			else {
				complete();
			}
		}
	},
	
	/** 
	 * Loads dataset configuration from a file.
	 * @param {String} uri - the uri for the file
	 * @return {Promise}
	 */
	loadFile: function(uri) {
		var self = this;
		var promise = new Promise(function(resolve, reject) {
			// Load the file asynchronously
			$.ajax({
				type: 'GET',
				url: uri,
				dataType: 'json',
				cache: false,
				success: function(data, textStatus, xhr) {
					// Resolve with the loaded data
					resolve(data);
				},
				error: function(xhr, textStatus, errorThrown) {
					// Reject with an error
					reject(Error(xhr.statusText));
				}
			});	
		});
		
		return promise;
	},
	
	/**
	 * Validates the configuration for a single dataset config.
	 * @param {Object} configuration - the configuration for the individual dataset
	 * @param {Integer} index - the index of the dataset in the overall configuration
	 * @param {Object} duplicates - object containing dataset keys to detect duplicates
	 * @param {String[]} warnings - list of warnings
	 * @param {Boolean} true if the configuration is valid, otherwise false
	 */
	validate: function(configuration, index, duplicates, warnings) {
		var valid = true;
		
		// Check the dataset has an id
		if(configuration.id == undefined) {
			warnings.push('Dataset ' + index + ' has no id');
			valid = false;
		}
		
		// Check the dataset id is not duplicated
		if(duplicates[configuration.id] != undefined) {
			warnings.push('Dataset ' + configuration.id + ' already exists');
			valid = false;
		}
		
		// Check the dataset has a key defined and it is an array
		if(configuration.key == undefined || Array.isArray(configuration.key) == false || configuration.key.length == 0) {
			warnings.push('Dataset ' + configuration.id + ' has no key or invalid key');
			valid = false;
		}
		
		// Check that if the dataset has a grouped config
		if(configuration.grouped != undefined) {
			// Check the grouped has a property 
			if(configuration.grouped.property == undefined || configuration.grouped.property.length == 0) {
				warnings.push('Dataset ' + configuration.id + ' grouped has no property');
				valid = false;
			}
			
			// Check the grouped has a key and it is an array
			if(configuration.grouped.key == undefined || Array.isArray(configuration.grouped.key) == false || configuration.grouped.key.length == 0) {
				warnings.push('Dataset ' + configuration.id + ' grouped has no key or invalid key');
				valid = false;
			}

			// Check the grouped has fields and it is an array
			if(configuration.grouped.fields == undefined || Array.isArray(configuration.grouped.fields) == false || configuration.grouped.fields.length == 0) {
				warnings.push('Dataset ' + configuration.id + ' grouped has no fields or invalid fields');
				valid = false;
			}
		}
		return valid;
	},
	
};