/**
 * Represents a dataset.
 * @constructor
 * @param {Object} container - container DOM object this dataset will be added to
 * @param {Object} configuration - configuration object defining the options
 *   for this dataset
 */
function Dataset(container, configuration) {
	// Parent container object containing this dataset
	this.container		= container;
	
	// Dataset properties from configuration
	this.id				= configuration.id;
	this.name			= configuration.name;
	this.description	= configuration.description;
	this.key			= configuration.key;
	this.grouped		= configuration.grouped;
	this.links			= configuration.links == undefined ? [] : configuration.links;
	
	// Full configuration object
	this.configuration	= configuration;
	
	// Schema for the dataset
	this.schema			= null;		
	
	// Vars to hold the objects being hovered and marked
	this.hovered		= null;		
	this.marked			= null;

	// List of sinks registered on this dataset
	this.sinks 			= [];
	
	// List of sources registered on this dataset
	this.sources 		= [];
	
	// List of data objects and a map of key to object for this dataset
	this.data			= [];
	this.dataMap		= {};

	// Map of grouped key to object for this dataset
	this.groupedDataMap	= {};
	
	// List of warnings for this dataset
	this.warnings		= [];
	
	// Live flat
	this.live			= false;
}

/**
 * Initializes the dataset by reading the configuration and then creating
 * and initializing all contained datasources.
 * @return {Promise} 
 */
Dataset.prototype.init = function() {	
	var self = this;
	var promise = new Promise(function(resolve, reject) {
		// Setup an object for datasource id to detect duplicates
		var duplicates = {};
		
		// Count the configured datasources. If there are none, resolve the promise
		var count = self.configuration.source == undefined ? 0 : self.configuration.source.length;
		if(count == 0) resolve(self);
		
		// Function to complete processing by decrementing the count
		//   Once the count reaches zero, rebuild the dataset schema and
		//   resolve the promise
		var complete = function() {
			count = count - 1;
			if(count == 0) {
				self.rebuildSchema();			
				resolve(self);
			}
		};
		
		// Loop over each source in the configuration
		for(var idx = 0; idx < self.configuration.source.length; idx++) {			
			// Get the configuration and validate
			var thisConfiguration = self.configuration.source[idx];
			var valid = DatasourceFactory.validate(thisConfiguration, idx, duplicates, self.warnings);

			// If valid, create and initialize the source, which will return it's own promise
			if(valid == true) {
				// Add the ID to the duplicate container
				duplicates[thisConfiguration.id] = thisConfiguration.id;
				
				// Create and initialize the datasource
				DatasourceFactory.createDatasource(thisConfiguration, self).init().then(
					function(source) {
						// Add to the source collection, then call complete
						if(source.live == true) 
							self.live = true;
						self.sources.push(source);
						complete();
					},
					function(err) {
						// Log the error, then call complete
						console.error(err);
						complete();
					}
				);
			}
			// Otherwise skip this source and just call complete
			else {
				complete();
			}
		}
	});
	
	return promise;
};

/**
 * Trigger an event on the dataset. Depending on the trigger type, all contained sinks
 * and/or sources will be notified with the appropriate information.
 * @param {String} eventType - string describing the event to trigger
 * @param {Object} eventSource - object event source for the event, can be either a 
 *   datasink or a datasource depending on the event type
 * @param {Object} object - data object associated with the event, may also be 
 *   null for certain event types
 */
Dataset.prototype.trigger = function(eventType, eventSource, object) {
	// Choose what to do based on trigger type

	// Snapshot start, notify all sinks of the event
	// eventSource - datasource that called trigger
	// object - null or undefined
	if(eventType == 'snapshotStart') {
		for(var idx = 0; idx < this.sinks.length; idx++) {
			this.sinks[idx].trigger(eventType);
		}
	}
	// Snapshot end, notify all sinks of the event
	// eventSource - datasource that called trigger
	// object - null or undefined
	else if(eventType == 'snapshotEnd') {
		for(var idx = 0; idx < this.sinks.length; idx++) {
			this.sinks[idx].trigger(eventType);
		}
	}
	// Insert, Update, Upsert, get the object key for the object and any existing
	//   instance matching the object key. If not found, create a new dataset object,
	//   otherwise merge the fields into the existing dataset object. 
	//   Note that insert/update/upsert are all handled the same here. An 'insert' 
	//   eventType might result in a merge if there are multiple datasources.
	// eventSource - datasource that called trigger
	// object - field map defining all fields on the insert or changed fields on the update
	else if(eventType == 'insert' || eventType == 'update' || eventType == 'upsert') {
		var objectKey = this.getObjectKey(object);
		var existing = this.dataMap[objectKey];

		if(existing == undefined)
			this.createNewDSObject(eventType, objectKey, object, eventSource);
		else
			this.mergeDSObject(eventType, objectKey, object, existing, eventSource);
	}
	// Delete, get the object key for the object and any existing instance matching the
	//   object key. If not found, do nothing, otherwise perform a dataset delete, 
	//   which may result in an internal update or a delete of the object depending
	//   on circumstances. See deleteDSObject for details.
	// eventSource - datasource that called trigger
	// object - field map defining at a minimum only key fields for the deleted object
	else if(eventType == 'delete') {
		var objectKey = this.getObjectKey(object);
		var existing = this.dataMap[objectKey];
		if(existing != undefined) {
			this.deleteDSObject(eventType, objectKey, object, existing, eventSource);
		}
	}
	// Clear, purge all data from the dataset, then notify all sinks to clear all data
	// eventSource - 
	// object - null or undefined
	else if(eventType == 'clear') {
		this.data = [];
		this.dataMap = {};
		for(var idx = 0; idx < this.sinks.length; idx++) {
			this.sinks[idx].trigger(eventType);
		}
	}
	// Hover, a data object is being hovered over in a UI element. If the dataset has
	//   any links defined, loop over the links and find any with eventType of hover.
	//   For each of these links, call apply link take any linking action on the
	//   target dataset.
	//   Also loop over all configured data sinks and notify all sinks of the event
	//     except for the sink that caused the trigger
	// eventSource - datasink that produced the hover event
	// object - object being hovered over, or null when the hover has stopped
	else if(eventType == 'hover') {
		if(eventSource != null && eventSource.dataset == this) {
			for(var idx = 0; idx < this.links.length; idx++) {
				var link = this.links[idx];
				if(link.trigger == 'hover') {
					this.applyLink(link, object);
				}
			}
		}

		for(var idx = 0; idx < this.sinks.length; idx++) {
			if(this.sinks[idx] != eventSource)
				this.sinks[idx].trigger(eventType, object);
		}
		
		this.hovered = object;
		if(this.container.onUpdate != undefined) 
			this.container.onUpdate();
	}
	// Marked, a data object is being marked in a UI element. If the dataset has
	//   any links defined, loop over the links and find any with eventType of marked.
	//   For each of these links, call apply link take any linking action on the
	//   target dataset.
	//   Also loop over all configured data sinks and notify all sinks of the event
	//     except for the sink that caused the trigger
	// eventSource - datasink that produced the marking event
	// object - object being marked, or null when the marking has stopped
	else if(eventType == 'marked') {
		if(eventSource != null && eventSource.dataset == this) {
			for(var idx = 0; idx < this.links.length; idx++) {
				var link = this.links[idx];
				if(link.trigger == 'marked') {
					this.applyLink(link, object);
				}
			}
		}

		for(var idx = 0; idx < this.sinks.length; idx++) {
			if(this.sinks[idx] != eventSource)
				this.sinks[idx].trigger(eventType, object);
		}
		
		this.marked = object;
		if(this.container.onUpdate != undefined) 
			this.container.onUpdate();
	}
	// Filter, a data set is being filtered in a UI element. If the dataset has
	//   any links defined, loop over the links and find any with eventType of filter.
	//   For each of these links, call apply link take any linking action on the
	//   target dataset.
	//   Also loop over all configured data sinks and notify all sinks of the event
	//     except for the sink that caused the trigger
	// eventSource - datasink that produced the marking event
	// filter - filter definition object with predicate that defines when a row
	//   is included in the result
	else if(eventType == 'filter') {
		for(var idx = 0; idx < this.sinks.length; idx++) {
			if(this.sinks[idx] != eventSource)
				this.sinks[idx].trigger(eventType, object);
		}
	}
	// Otherwise log a warning about an unknown event type
	else {
		console.warn('Dataset ' + this.id + ' unknown eventType ' + eventType);
	}
};

/** 
 * Registers a new sink on the dataset. This is called whenever a sink is added,
 * either dynamically or during application initialization.
 * @param {Object} sink - datasink to register on the dataset
 */
Dataset.prototype.registerSink = function(sink) {
	// Add the sink to the array
	this.sinks.push(sink);
	
	// Trigger a schema change, then a snapshot start on the sink
	sink.trigger('schemaChange', this.schema);
	sink.trigger('snapshotStart');
	
	// If there is any data in the dataset, iterate over and trigger
	//   an insert for each object on the sink
	if(this.data.length > 0) {
		for(var idx = 0; idx < this.data.length; idx++) {
			object = this.data[idx];
			sink.trigger('insert', object);
		}
	}
	
	// Trigger a snapshot end on the sink
	sink.trigger('snapshotEnd');
};

/** 
 * Unregisters an existing sink on the dataset. This will remove the sink
 * and it will stop receiving events from this dataset.
 * @param {Object} sink - datasink to unregister from the dataset
 */
Dataset.prototype.unregisterSink = function(sink) {
	// Get the index of the sink and remove it from the array, if it exists
	var idx = this.sinks.indexOf(sink);
	if(idx > -1) this.sinks.splice(idx, 1);
};

/**
 * Registers a new source on the dataset. This is called only when a source
 * is added dynamically and NOT during intialization. The sources will
 * automatically be added during initialization based on initial configuration.
 * @param {Object} source - the datasource to register on the dataset
 */
Dataset.prototype.registerSource = function(source) {
	// Add the source to the array, rebuild the schema, and reload all
	//   of the datasources. This will cause a full rebuild of the 
	//   dataset and all associated datasinks.
	source.init().then(
		function(source) {
			// Add to the source collection, then call complete
			if(source.live == true) 
				self.live = true;
			self.sources.push(source);
			self.rebuildSchema();	
			self.reload();
		},
		function(err) {
			// Log the error, then call complete
			console.error(err);
		}
	);
};

/**
 * Unregisters an existing sink on the dataset. This will remove the sink
 * and it will stop producing events for this dataset.
 * @param {Object} source - the datasource to unregister from the dataset
 */
Dataset.prototype.unregisterSource = function(source) {
	var idx = this.sources.indexOf(source);
	if(idx > -1) this.sources.splice(idx, 1);
	this.rebuildSchema();	
	this.reload();
};

/**
 * Rebuild the dataset schema from all the configured datasource schemas.
 * This will automatically be called on initialization once all datasources
 * have been added to the dataset. It will also automatically be called when
 * a new datasource is registered with the registerSource() function.
 */
Dataset.prototype.rebuildSchema = function() {
	// Create a new schema
	this.schema = new DatasetSchema();
	
	// Loop over each of the sources
	for(var sourceIdx = 0; sourceIdx < this.sources.length; sourceIdx++) {
		var source = this.sources[sourceIdx];
		
		// Get the source schema and throw an error if it is undefined
		var sourceSchema = source.schema;
		if(sourceSchema == undefined) 
			throw Error('Datasource ' + source.id + ' has no schema');
		
		// Loop over the source fields and add to the dataset schema
		//   Duplicate field names will overwrite existing configured fields
		//   therefore it's important that field configurations are consistent
		//   in particular for key fields which will always be duplicated
		for(var fieldIdx = 0; fieldIdx < sourceSchema.fields.length; fieldIdx++) {
			var f = sourceSchema.fields[fieldIdx];
			this.schema.addField(f.name, f.type);
		}
	}		
	
	// Trigger a schema change on all sinks
	for(var idx = 0; idx < this.sinks.length; idx++) {
		this.sinks[idx].trigger('schemaChange', this.schema);
	}
};

/**
 * Clear the existing set of data objects and reload all from each of the datasources.
 * This will automatically be called when a new datasource is registered with the 
 * registerSource() function.
 */
Dataset.prototype.reload = function() {
	// Clear all data from the local data
	this.data = [];
	this.dataMap = {};
	
	// Trigger a clear on all sinks, this will purge all existing rendered data
	for(var idx = 0; idx < this.sinks.length; idx++) {
		this.sinks[idx].trigger('clear');
	}

	// Run reload on all sources. This will cause a snapshotStart, insert set, snapshotEnd
	//   sequence to repeat which will rebuild the full set of data in the dataset
	//   and notify all sinks of these events.  
	for(var idx = 0; idx < this.sources.length; idx++) {
		this.sources[idx].reload();
	}	
};

/**
 * Builds the object key for the given object. The dataset must have a key configured
 * otherwise an error will be thrown during dataset creation.
 * @param {Object} object - the data object
 * @return {String} the object key
 */
Dataset.prototype.getObjectKey = function(object) {
	// The dataset will be validated at create time to ensure
	//   there is actually a key defined
	
	// Get the field value for the first key element and set this
	//   as the initial object key
	var objectKey = object[this.key[0]];
	if(objectKey == undefined) {
		console.warn('Dataset ' + this.id + ' object key not found for ' + this.key[0]);
		objectKey = "";
	}
	
	// For each remaining key elements, get the field value and concatenate
	//   to the object key with pipes between elements
	for(var idx = 1; idx < this.key.length; idx++) {
		var thisKey = object[this.key[idx]];
		if(thisKey == undefined) {
			console.warn('Dataset ' + this.id + ' object key not found for ' + this.key[idx]);
			thisKey = "";
		}
		objectKey = objectKey + '|' + thisKey;
	}
	
	return objectKey;
};

/**
 * Builds the grouped object key for the given object. The dataset group must have a key configured
 * otherwise an error will be thrown during dataset creation.
 * @param {Object} object - the data object
 * @return {String} the group object key
 */
Dataset.prototype.getGroupedObjectKey = function(object) {
	// The dataset will be validated at create time to ensure
	//   there is actually a group key defined
	
	// Get the field value for the first key element and set this
	//   as the initial object key
	var objectKey = object[this.grouped.key[0]];
	if(objectKey == undefined) {
		console.warn('Dataset ' + this.id + ' grouped object key not found for ' + this.grouped.key[0]);
		objectKey = "";
	}
	
	// For each remaining key elements, get the field value and concatenate
	//   to the group object key with pipes between elements
	for(var idx = 1; idx < this.grouped.key.length; idx++) {
		var thisKey = object[this.grouped.key[idx]];
		if(thisKey == undefined) {
			console.warn('Dataset ' + this.id + ' grouped object key not found for ' + this.grouped.key[idx]);
			thisKey = "";
		}
		objectKey = objectKey + '|' + thisKey;
	}
	
	return objectKey;
};

/** 
 * Apply a configured link against a target dataset in response to a UI action.
 * @param {Object} link - the link configuration to apply
 * @param {Object} object - the data object that caused the link, or null 
 */
Dataset.prototype.applyLink = function(link, object) {
	// Decision based on the action attribute of the link

	// Navigation
	if(link.action == 'navigation') {
		// Check to see if there is a navigate function on the container navigator object
		if(this.container.navigator == undefined) return;
		if(this.container.navigator.navigate == undefined) return;
		this.container.navigator.navigate(link.navigate);
	}
	// All others are dataset related
	else {
		// Get the linked dataset from the target property
		var linkedDataset = this.container[link.target];
		if(linkedDataset == undefined) return;
		
		// Parameters link action
		if(link.action == 'parameters') {
			// Create the target parameters
			var targetParameters = {};
			
			// If object is null, reset the parameters
			if(object == null)
				linkedDataset.resetParameters();
			// Otherwise evaluate the parameters
			else {
				for(var parameterIdx = 0; parameterIdx < Object.keys(link.parameters).length; parameterIdx++) {
					var parameterName = Object.keys(link.parameters)[parameterIdx];
					var parameterExpression = link.parameters[parameterName];
					var thisObject = object;
					if(Array.isArray(thisObject) == true) 
						thisObject = object[0];
					var source = thisObject;
					var parameterValue = eval(parameterExpression);
					targetParameters[parameterName] = parameterValue;
				}

				// Call apply on the linked dataset
				linkedDataset.applyParameters(targetParameters);
			}
		}
		// Other link actions, hovered and marked
		else {
			// Loop over the objects in the target dataset and apply conditions
			var targetObjects = [];
			if(object == null) {
				linkedDataset.trigger(link.action, this, null);
			}
			else {
				for(var datasetIdx = 0; datasetIdx < linkedDataset.data.length; datasetIdx++) {
					var target = linkedDataset.data[datasetIdx];
					if(Array.isArray(object)) {
						for(var objectIdx = 0; objectIdx < object.length; objectIdx++) {
							var source = object[objectIdx];
							var result = eval(link.condition);
							if(result == true)
								targetObjects.push(target);
						}
					}
					else {
						var source = object;
						var result = eval(link.condition);
						if(result == true)
							targetObjects.push(target);
					}
				}

				if(targetObjects.length == 0)
					targetObjects = null;
				else if(targetObjects.length == 1)
					targetObjects = targetObjects[0];
				linkedDataset.trigger(link.action, this, targetObjects);
			}
		}
	}
};

/**
 * Creates a new data object in the dataset from a field map. 
 * @param {String} eventType - string describing the event that triggered
 * @param {String} objectKey - internal unique identifier object key for the dataset object
 * @param {Object} fieldMap - field map defining all fields on the new object
 * @param {Object} source - datasource that produced the event causing this create
 */
Dataset.prototype.createNewDSObject = function(eventType, objectKey, fieldMap, source) {
	// Create the object and set the _key to the object key and set the _source to the
	//   id of the original source. Source will be used as part of the delete step.
	var object = {
		_key:		objectKey,
		_source:	[source.id],
	};
		
	// If grouped, add a group field
	var groupedObject = null;
	if(this.grouped != undefined) {
		object[this.grouped.property] = [];
	}

	// Set all the fields in the field map on the object
	for(var idx = 0; idx < Object.keys(fieldMap).length; idx++) {
		var fieldName = Object.keys(fieldMap)[idx];
		if(this.grouped != undefined && this.grouped.fields.indexOf(fieldName) > -1) {
			if(groupedObject == null) {
				groupedObject = {
					_key:		this.getGroupedObjectKey(fieldMap),
					_source:	[source.id]
				}				
			}
			groupedObject[fieldName] = fieldMap[fieldName];
		}
		else {
			object[fieldName] = fieldMap[fieldName];
		}
	}

	// Put the object in the data array and in the data map keyed on the object key
	this.data.push(object);
	this.dataMap[object._key] = object;
		
	// If there is a grouped object, add it to the appropriate arrays and maps
	if(groupedObject != null) {
		object[this.grouped.property].push(groupedObject);
		this.groupedDataMap[groupedObject._key] = groupedObject;
	}

	// Run callbacks on all sinks and notify of the inserted object
	for(var idx = 0; idx < this.sinks.length; idx++)
		this.sinks[idx].trigger('insert', object);	
};

/**
 * Merges a field map into an existing dataset object.
 * @param {String} eventType - string describing the event that triggered
 * @param {String} objectKey - internal unique identifier object key for the dataset object
 * @param {Object} fieldMap - field map defining all changed fields on the object
 * @param {Object} existing - existing dataset object
 * @param {Object} source - datasource that produced the event causing this create
 */
Dataset.prototype.mergeDSObject = function(eventType, objectKey, fieldMap, existing, source) {
	// Check the if the source is set on the object, if not set it now
	if(existing._source.indexOf(source.id) == -1)
		existing._source.push(source.id);
	
	// If grouped, retrieve the existing grouped field
	var groupedObject = null;
	var newGroupedObject = false;
	if(this.grouped != undefined) {
		groupedObject = this.groupedDataMap[this.getGroupedObjectKey(fieldMap)];
	}
	
	// Set all the fields in the field map on the object
	for(var idx = 0; idx < Object.keys(fieldMap).length; idx++) {
		var fieldName = Object.keys(fieldMap)[idx];
		if(this.grouped != undefined && this.grouped.fields.indexOf(fieldName) > -1) {
			if(groupedObject == null) {
				groupedObject = {
					_key:		this.getGroupedObjectKey(fieldMap),
					_source:	[source.id]
				}	
				newGroupedObject = true;
			}
			else {
				if(groupedObject._source.indexOf(source.id) == -1)
					groupedObject._source.push(source.id);
			}
			groupedObject[fieldName] = fieldMap[fieldName];
		}
		else {
			existing[fieldName] = fieldMap[fieldName];
		}
	}

	// If there is a new grouped object, add it to the appropriate arrays and maps
	if(groupedObject != null && newGroupedObject == true) {
		existing[this.grouped.property].push(groupedObject);
		this.groupedDataMap[groupedObject._key] = groupedObject;
	}

	// Run callbacks on all sinks and notify of the merged object
	for(var idx = 0; idx < this.sinks.length; idx++)
		this.sinks[idx].trigger('update', existing);		
};

/**
 * Deletes or updates an existing dataset object.
 * @param {String} eventType - string describing the event that triggered
 * @param {String} objectKey - internal unique identifier object key for the dataset object
 * @param {Object} fieldMap - field map defining all changed fields on the object
 * @param {Object} existing - existing dataset object
 * @param {Object} source - datasource that produced the event causing this create
 */
Dataset.prototype.deleteDSObject = function(eventType, objectKey, fieldMap, existing, source) {
	// If grouped, retrieve the existing grouped field
	var groupedObject = null;
	if(this.grouped != undefined) {
		groupedObject = this.groupedDataMap[this.getGroupedObjectKey(fieldMap)];
	}

	// If there is a grouped object, remove it now
	if(groupedObject != null) {
		delete this.groupedDataMap[groupedObject._key];
		existing[this.grouped.property].splice(existing[this.grouped.property].indexOf(groupedObject), 1);
	}
	
	// If not grouped, or all group elements are removed, then 
	//   remove all the other fields from the object based on the source 
	//   schema, unless the field is part of the key in which case leave it
	if(this.grouped == undefined || existing[this.grouped.property].length == 0) {
		for(var idx = 0; idx < source.schema.fields.length; idx++) {
			var field = source.schema.fields[idx];
			if(this.key.indexOf(field.name) == -1) {
				delete existing[field.name];
			}
			
			// Remove the source from the list of contributing sources
			var sourceIdx = existing._source.indexOf(source.id);
			if(sourceIdx > -1) existing._source.splice(sourceIdx, 1);
		}
		if(this.grouped != undefined) {
			delete existing[this.grouped.property];
			groupedObject = null;
		}
	}

	// If the contributing source array is length zero, then delete the whole object, 
	//   otherwise just update the sinks
	if(existing._source.length == 0) {
		delete this.dataMap[existing._key];
		var sourceIdx = this.data.indexOf(existing);
		if(sourceIdx > -1) this.data.splice(sourceIdx, 1);
		for(var idx = 0; idx < this.sinks.length; idx++)
			this.sinks[idx].trigger('delete', existing);		
	}
	else {
		for(var idx = 0; idx < this.sinks.length; idx++)
			this.sinks[idx].trigger('update', existing);		
	}
};

/**
 * Clears all data in the dataset and notifies all data sinks to clear data. Then applies
 * passed parameters on all data sources in the dataset. 
 * data which will be rebuilt from the data sources once the parameters are applied.
 * @param {Object} parameters - object defining the parameters to set
 */
Dataset.prototype.applyParameters = function(parameters) {
	// Clear all data from the local data
	this.data = [];
	this.dataMap = {};
	this.parameters = parameters;
	
	// Run sink.trigger clear data, hover, and marking
	for(var idx = 0; idx < this.sinks.length; idx++) {
		this.sinks[idx].trigger('clear');
		this.sinks[idx].trigger('hover', null);
		this.sinks[idx].trigger('marked', null);
	}
	
	this.hovered = null;
	this.marked = null;	
	if(this.container.onUpdate != undefined) 
		this.container.onUpdate();
	
	// Run source.applyParameters on each of the sources
	for(var idx = 0; idx < this.sources.length; idx++) {
		this.sources[idx].applyParameters(parameters);
	}	
};

/**
 * Clears all data in the dataset and notifies all data sinks to clear data. Then applies
 * default parameters on all data sources in the dataset.
 */
Dataset.prototype.resetParameters = function() {
	// Clear all data from the local data
	this.data = [];
	this.dataMap = {};
	this.parameters = {};
	
	// Run sink.trigger clear data, hover, and marking
	for(var idx = 0; idx < this.sinks.length; idx++) {
		this.sinks[idx].trigger('clear');
		this.sinks[idx].trigger('hover', null);
		this.sinks[idx].trigger('marked', null);
	}

	this.hovered = null;
	this.marked = null;	
	if(this.container.onUpdate != undefined) 
		this.container.onUpdate();

	// Run source.resetParameters on each of the sources
	for(var idx = 0; idx < this.sources.length; idx++) {
		this.sources[idx].resetParameters();
	}	
};