/**
 * Represents a grid data sink.
 * @constructor
 * @param {Object} dataset - dataset that this sink will be attached to
 * @param {Object} configuration - configuration object defining the sink
 */
function GridDatasink(id, configuration, dataset, containerDiv, overlay) {
	this.id = id;
	this.dataset = dataset;
	this.grid = new LiveGrid.Grid(id, configuration, containerDiv, overlay, this);
}

/**
 * Event handler for dataset events. 
 * @param {String} eventType - string describing the event that occurred
 * @param {Object} object - data object for the event, or null
 */
GridDatasink.prototype.trigger = function(eventType, object) {
	this.grid.trigger(eventType, object);
};


// Define a module to hold object defintions for the grid objects
var LiveGrid = (function() {
	var module = {};

	/**
	 * Creates a new Grid object.
	 * @constructor
	 * @param {String} id - unique identifier
	 * @param {Object} configuration - the configuration for the grid
	 * @param {Object} containerDiv - this node object that grid will be appended
	 * @param (Function} overlay - function to call on start and end of the overlay
	 * @param {Object} sink - datasink that will feed this grid
	 */
	function Grid(id, configuration, containerDiv, overlay, sink) {
		this.id				= id;
		this.configuration	= configuration;
		this.containerDiv 	= containerDiv;
		this.overlay		= overlay;
		this.sink			= sink;
		
		// Pause and snapshot flags
		this.paused			= false;
		this.snapshot		= false;
		
		// Schema for the dataset
		this.schema			= null;		
		
		// Vars to hold the objects being hovered and marked
		this.hovered		= [];		
		this.marked			= [];
		
		// Filter definition
		this.filterDef		= {
			search: {
				expression:	undefined,
				fields:		this.configuration.options.search.fields
			},
			filter:	{
				filters:	[],
				fields:		this.configuration.options.filter.fields
			}
		};
		
		// Flag for filters being visible
		this.filtersVisible	= false;
		
		// Ordering definition
		this.orderDef		= [];
		
		// Pagination defintion
		this.pagination		= null;
		
		// Color maps
		this.colorMaps		= {};

		// Columns
		this.columnMap		= {};
		this.columnArr		= [];
		
		// Rows
		this.rowMap			= {};
		this.rowArr			= [];
		
		// Ordered, filtered, and final result rows
		this.orderedArr		= [];
		this.filterArr		= [];
		this.resultArr		= [];
		
		// Check dependencies. This will throw Error if dependencies are not satisfied.
		this.dependencyCheck();
		
		// Microbatch is used to batch up updates to prevent too many chart redraws
		this.microbatch = {
			rtime:		new Date(1, 1, 2000, 12,00,00),
			timeout:	false,
			delta:		500 // msec between updates
		}

		// Color maps
		if(this.configuration.options.colormaps != undefined) {
			for(var idx = 0; idx < Object.keys(this.configuration.options.colormaps).length; idx++) {
				var thisColorMapName = Object.keys(this.configuration.options.colormaps)[idx];
				var thisColors = this.configuration.options.colormaps[thisColorMapName];
				this.colorMaps[thisColorMapName] = new DataFlow.ColorMap(thisColorMapName, thisColors);
			}
		}
		
		// Pagination
		if(this.configuration.options.pagination.enabled == true) {
			this.pagination = {
				size:	configuration.options.pagination.size,
				begin: 	0,
				end:	configuration.options.pagination.size
			};
		}
		
		// On document ready, register UI handlers
		var self = this;
		setTimeout(function() {
			self.registerEventHandlers();
		}), 0;
		
		// Draw the initial empty grid
		this.draw();
	};

	/**
	 * Performs a dependency check for rendering the grid.
	 */
	Grid.prototype.dependencyCheck = function() {
		// No dependencies for this component
	};
		
	/**
	 * Registers event handlers on UI components.
	 */
	Grid.prototype.registerEventHandlers = function() {
		var self = this;
		
		// Event handler to hide any open context menu on click
		$(document).on('click', function() {
			$(document).find('.live-context-menu').remove();
		});
		
		// Event handler for horizontal scrolling to bind the position of header row with
		//   body of the grid
		this.lastLeftPos = $(this.containerDiv).find('.grid-rows').scrollLeft();
		$(this.containerDiv).find('.grid-rows').scroll(function () {
			var currLeftPos = $(this).scrollLeft();
			if(self.lastLeftPos != currLeftPos) {
				$(self.containerDiv).find('.header-table').offset({ left: -1 * this.scrollLeft + 11});
				self.lastLeftPos = currLeftPos;
			}
			$(document).find('.live-context-menu').remove();
		});
		
		// Event handler to show/hide the header filter row
		$(this.containerDiv).find('.grid-filtered').on('click', function() {
			self.toggleFilters();
		});
		
		// Event handler to display the configuration modal
		$(this.containerDiv).find('.grid-config').on('click', function(e) {
			self.displayConfiguration();
		});

		// Event handler to display the dataset modal
		$(this.containerDiv).find('.grid-icon').on('click', function(e) {
			self.displayDataset();
		});
		
		// Event handler to apply search
		$(this.containerDiv).find('.grid-search').find('input').on('input', DataFlow.Utilities.Debounce.debounce(function(e) {
			var val = $(e.target).val();
			self.search(val);
		}, this.configuration.options.search.delay));

		// Event handler to clear search
		$(this.containerDiv).find('.grid-search').find('.grid-search-clear').on('click', function(e) {
			$(self.containerDiv).find('.grid-search').find('input').val('');
			self.search('');
		});

		// Event handler to clear hover when mouse moves out of boundary for the grid
		if(this.configuration.options.hover.enabled != false) {
			$(this.containerDiv).on('mouseout', function() {
				self.hover(null, true);
			});
		}	
		
		// Event handler for adjusting pagination on scroll
		if(this.configuration.options.pagination.enabled == true) {
			var dataTableDiv = $(this.containerDiv).find('.data-table');
			$(this.containerDiv).find('.grid-rows').scroll(function(e) {
				var scrollPercentage = $(this).scrollTop() / (dataTableDiv.height() - $(this).height());
				if(scrollPercentage >= 0.75 && self.resultArr.length < self.rowArr.length) {
					self.pagination.end = self.pagination.end + self.pagination.size;
					self.update(true);
				}
			});			
		}

		// Event handler to draw selection area for multiple marking
		if(this.configuration.options.marking.enabled != false && this.configuration.options.marking.mode == 'multiple') {
			$(self.containerDiv).on('dragstart', function(e) { return false; });
			$(self.containerDiv).find('.data-table').on('mousedown', function(e) {				
				// Function to update the size and position of the selection area
				var positionSelectionArea = function(startY, endY, dataTableDiv, gridRowsDiv, selectionAreaDiv, startScrollTop) {
					if(endY + (gridRowsDiv.scrollTop() - startScrollTop) - startY > 0) {
						//console.log("DN " + startY + " -> " + endY + " --- " + gridRowsDiv.scrollTop() + " " + gridRowsDiv.height() + " " + dataTableDiv.height());
						if(endY >= dataTableDiv.height())
							endY = dataTableDiv.height();
						selectionAreaDiv.css({height: endY + (gridRowsDiv.scrollTop() - startScrollTop) - startY});
					}
					else { 
						//console.log("UP " + startY + " -> " + endY + " --- " + gridRowsDiv.scrollTop() + " " + gridRowsDiv.height() + " " + dataTableDiv.height());
						selectionAreaDiv.css({top: endY + (gridRowsDiv.scrollTop() - startScrollTop), height: startY - (endY + (gridRowsDiv.scrollTop() - startScrollTop))});
					}
				};
				
				// Create a selection area div
				var appended = false;
				var selectionAreaDiv = $(document.createElement('div')).attr('class', 'selection-area');
				
				// Data table, containing all the rows
				var dataTableDiv = $(self.containerDiv).find('.data-table');
				var offset = dataTableDiv.offset();
				
				// Grid rows, the scrolling container
				var gridRowsDiv = $(self.containerDiv).find('.grid-rows');				
				var startScrollTop = gridRowsDiv.scrollTop();

				// Get initial start and end position
				var startY = e.pageY - offset.top;	
				var endY = null;
				var ctrlKey = e.ctrlKey;
				
				// Set intial position for the selection area
				selectionAreaDiv.css({'left': 0, 'top': startY});
				
				// Mouse exits the scroll area, create an auto-scroll
				var interval = null;
				gridRowsDiv.on('mouseleave', function(e) {
					// Compute the direction
					var direction = e.pageY - offset.top < startY ? 'moveup' : 'movedown';
					
					// If already auto-scrolling don't start another one
					if(interval == null) {
						var thisStartScroll = gridRowsDiv.scrollTop(); 
						interval = setInterval(function() {
							var scrollDelta = gridRowsDiv.scrollTop() - thisStartScroll;
							if(direction == 'movedown' && endY + scrollDelta < dataTableDiv.height() - 1) {
								gridRowsDiv.scrollTop(gridRowsDiv.scrollTop() + 10);
							}
							else if(direction == 'moveup') {
								gridRowsDiv.scrollTop(gridRowsDiv.scrollTop() - 10);
							}
							else {
								clearInterval(interval);
								interval = null;
							}
							
							// Update the selection area
							positionSelectionArea(startY, endY, dataTableDiv, gridRowsDiv, selectionAreaDiv, startScrollTop);
						},50);
					}
				});

				// Mouse enters the scroll area, cancel any auto-scroll
				gridRowsDiv.on('mouseenter', function(e) {
					if(interval != null) {
						clearInterval(interval);
						interval = null;
					}
				});
				
				// Move the mouse in the scroll area
				gridRowsDiv.on('mousemove', function(e) {
					endY = e.pageY - offset.top;
					if(appended == false && Math.abs(startY - endY) > 10) {
						dataTableDiv.append(selectionAreaDiv);				
						appended = true;
					}
					
					// Update the selection area
					positionSelectionArea(startY, endY, dataTableDiv, gridRowsDiv, selectionAreaDiv, startScrollTop);
				});
				
				// Complete the select
				$(document).on('mouseup', function() {
					// If the selection area exists, execute marking
					if(selectionAreaDiv.height() > 0) {

						// Create array of objects to mark
						var objects = [];

						// If CTRL key is down, then pre-populate the array with existing
						if(ctrlKey == true) {
							for(var idx = 0; idx < self.marked.length; idx++) 
								objects.push(self.marked[idx]);
						}
						
						// Loop over all rows and mark those inside the selection area
						var selectionStart = selectionAreaDiv.offset().top - offset.top + gridRowsDiv.scrollTop();
						var selectionEnd = selectionStart + selectionAreaDiv.height();
						for(var idx = 0; idx < Object.keys(self.rowMap).length; idx++) {
							var id = Object.keys(self.rowMap)[idx];
							var row = self.rowMap[id];
							if(row.node == null) continue;
							var rowStart = row.node.offset().top - offset.top + gridRowsDiv.scrollTop();
							var rowEnd = rowStart + row.node.height();
							if(rowEnd >= selectionStart && rowStart <= selectionEnd && objects.indexOf(row.object) == -1) {
								objects.push(row.object);
							}
						}
						self.mark(objects, true);
					}
					
					// Remove the selection area
					selectionAreaDiv.remove();
					
					// If there is an auto-scroll, then clear it
					if(interval != null) {
						clearInterval(interval);
						interval = null;
					}					
					
					// Turn off other listeners
					$(document).off('mouseup');
					gridRowsDiv.off('mousemove');
					gridRowsDiv.off('mouseenter');
					gridRowsDiv.off('mouseleave');
				});
			});
		}			
		
	};
	
	/**
	 * Event handler for all objects delegated from the datasink.
	 * @param {String} eventType - the event type
	 * @param {Object} object - optional object
	 */
	Grid.prototype.trigger = function(eventType, object) {
		var self = this;

		if(eventType == 'schemaChange') {
			this.schema = object;			
			this.buildColumnModel();
		}
		else if(eventType == 'snapshotStart') {
			this.snapshot = true;
			this.overlay(true);
		}
		else if(eventType == 'snapshotEnd') {
			setTimeout(function() {
				self.snapshot = false;
				self.overlay(false);
			},500);
		}
		else if(eventType == 'clear') {
			this.clear();
		}
		else if(eventType == 'insert') {
			this.addRow(object);
		}
		else if(eventType == 'update') {
			this.updateRow(object);
		}
		else if(eventType == 'upsert') {
		}
		else if(eventType == 'delete') {
			this.removeRow(object);
		}
		else if(eventType == 'hover') {
			if(this.configuration.options.hover.accept != false)
				this.hover(object);
		}
		else if(eventType == 'marked') {
			if(this.configuration.options.marking.accept != false)
				this.mark(object);
		}					
		else if(eventType == 'filter') {
			if(this.configuration.options.filter.accept != false)
				this.filter(object);
		}					
	};

	/**
	 * Builds the column model based on the column configuration.
	 */
	Grid.prototype.buildColumnModel = function() {
		// Clear the header row
		var headerRowDiv = $(this.containerDiv).find('.header-row');
		headerRowDiv.empty();

		// Clear the header filter row
		var headerFilterRowDiv = $(this.containerDiv).find('.header-filter-row');
		headerFilterRowDiv.empty();

		// Reset the column list and then build new column objects
		this.columnMap = {};
		this.columnArr = [];
		for(var idx = 0; idx < this.configuration.columns.length; idx++) {
			var column = new Column(this.configuration.columns[idx], this, headerRowDiv, headerFilterRowDiv);
			this.columnMap[column.field] = column;
			this.columnArr.push(column);
			column.render();
		}
	};
		
	/**
	 * Add a new row bound to the specified object on the grid.
	 * @param {Object} object - the object to add the bound row
	 */
	Grid.prototype.addRow = function(object) {
		var row = new Row(object, this.columnArr, this);
		this.rowMap[row._key] = row;
		this.rowArr.push(row);
		this.update();
	};

	/**
	 * Updates the row bound to the specified object on the grid.
	 * @param {Object} object - the object to update the bound row
	 */
	Grid.prototype.updateRow = function(object) {
		var row = this.rowMap[object._key];
		if(row == undefined) return;		
		row.object = object;
		row.update();
		this.update();
	};

	/**
	 * Removes the row bound to the specified object from the grid.
	 * @param {Object} object - the object to remove the bound row
	 */
	Grid.prototype.removeRow = function(object) {
		var row = this.rowMap[object._key];
		delete this.rowMap[object._key];
		this.rowArr.splice(this.rowArr.indexOf(row), 1);
		row.node = null;
		this.update();
	};

	/**
	 * Clears all objects and all rows from the grid.
	 */
	Grid.prototype.clear = function() {	
		this.rowMap			= {};
		this.rowArr			= [];
		this.update();
	};

	/**
	 * Draws the UI components for the grid.
	 */
	Grid.prototype.draw = function() {
		var self = this;
		var context = {};
		
		// Set classes on the container div
		if(this.configuration.options.classes != undefined) {
			$(this.containerDiv).addClass(DataFlow.Utilities.Evaluate.classes(this.configuration.options.classes, context));
		}
		
		// Set styles on the container div, always must be an object
		if(this.configuration.options.styles != undefined) {
			var styles = DataFlow.Utilities.Evaluate.styles(this.configuration.options.styles, context);
			for(style in styles) $(this.containerDiv).css(style, styles[style]);
		}	
	};

	/**
	 * Updates the contents of the grid. Normally this will happen in a microbatch
	 * to prevent too many updates from occurring. However it is possible to trigger
	 * an immediate update by setting the flag.
	 * @param {Boolean} immediate - trigger an immediate update
	 */
	Grid.prototype.update = function(immediate) {
		var self = this;

		if(immediate == true) {
			run();
			return;
		}
		
		// Microbatch handler to prevent too many updates
		if(self.microbatch.timeout === false) {
			self.microbatch.timeout = true;
			setTimeout(eventEnd, self.microbatch.delta);
		}					
		function eventEnd() {
			var d = new Date() - self.microbatch.rtime;
			if (new Date() - self.microbatch.rtime < self.microbatch.delta)
				setTimeout(eventEnd, self.microbatch.delta);
			else {
				self.microbatch.rtime = new Date();
				self.microbatch.timeout = false;
				run();
			}               
		};
		
		// Run the update
		function run() {
			if(self.paused == true) return;
			
			// Detach all child data rows, and reattach later in the correct order
			var dataTableDiv = $(self.containerDiv).find('.data-table');
			dataTableDiv.children('.data-row').detach();

			// Apply ordering if set
			self.orderedArr = self.rowArr;
			if(self.configuration.options.ordering.enabled != false) {
				self.orderedArr = DataFlow.Utilities.Ordering.order(self.orderedArr, self.orderDef, 'object');
			}

			// Apply search and filters if set
			self.filterArr = self.orderedArr;
			if(self.configuration.options.search.enabled != false || self.configuration.options.filter.enabled != false) {
				self.filterArr = DataFlow.Utilities.Filtering.filter(self.filterArr, self.filterDef, 'object');
			}
		
			// Apply pagination if enabled
			self.resultArr = self.filterArr;
			if(self.pagination != null) {
				self.resultArr = DataFlow.Utilities.Filtering.paginate(self.resultArr, self.pagination);
			}
			
			// Show or hide the no data message depending on the number of rows in the result
			var noDataDiv = $(self.containerDiv).find('.no-data');
			if(self.resultArr.length > 0)
				noDataDiv.hide();
			else
				noDataDiv.show();
				
			// Loop over all the result rows
			for(var idx = 0; idx < self.resultArr.length; idx++) {
				// If the node isn't rendered for this row, render it now
				if(self.resultArr[idx].node == null) 
					self.resultArr[idx].render();
				dataTableDiv.append(self.resultArr[idx].node);
			}

			// Update the info panel
			self.updateInfo();
		};
	};

	/**
	 * Pauses updates on a grid.
	 */
	Grid.prototype.pause = function() {
		this.paused = true;
	};
	
	/**
	 * Resumes a paused grid.
	 */
	Grid.prototype.resume = function() {
		if(this.paused == false) return;
		this.paused = false;
		this.update();
	};

	/**
	 * Resizes a grid.
	 */
	Grid.prototype.resize = function() {
		// No implementation for this component
	};
	
	/**
	 * Applies a search value against the grid. This will not be passed through
	 * to the dataset and is local only to the current grid.
	 * @param {String} value - the search string
	 */
	Grid.prototype.search = function(value) {
		if(value == undefined || value.length == 0) {
			delete this.filterDef.search.expression;
		}
		else {
			this.filterDef.search.expression = value;
		}
		
		this.update();
	};
	
	/**
	 * Hovers a single row bound to the specified object.
	 * Will also optionally notify the dataset of a hover event, if enabled.
	 * @param {Object} object - object that the associated row will be hovered
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Grid.prototype.hover = function(object, notify) {
		if(this.configuration.options.hover.enabled != false) {
			// Unhover all existing rows
			for(var idx = this.hovered.length - 1; idx >= 0; idx--) {
				var thisObject = this.hovered[idx];
				var row = this.rowMap[thisObject._key];
				if(row != undefined) row.hover(false);
				this.hovered.splice(this.hovered.indexOf(thisObject), 1);
			}	
					
			// If the object is not null, process the hovering
			if(object != null) {
				if(Array.isArray(object)) {
					for(var idx = 0; idx < object.length; idx++) {
						var thisObject = object[idx];
						var row = this.rowMap[thisObject._key];
						if(row != undefined) row.hover(true);
						this.hovered.push(thisObject);
					}
				}
				else {
					var row = this.rowMap[object._key];
					if(row != undefined) row.hover(true);
					this.hovered.push(object);
				}
			}
					
			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.hover.passthrough == true) {
				this.sink.dataset.trigger('hover', this.sink, object);
			}			
		}
	};

	/**
	 * Marks a single row or a set of rows bound to the specified object or array of objects.
	 * Will also optionally notify the dataset of a marked event, if enabled.
	 * @param {Object|Array} object - if a single data object then the associated row will be marked
	 *                                if an array of objects then the associated rows will be marked
	 *                                  for single marking rows that receive a list of objects then only
	 *                                  the first row will be marked
	 *                                if null then all marked rows will be unmarked
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Grid.prototype.mark = function(object, notify) {
		if(this.configuration.options.marking.enabled != false) {
			var mode = this.configuration.options.marking.mode;

			// Unmark all existing rows
			for(var idx = this.marked.length - 1; idx >= 0; idx--) {
				var thisObject = this.marked[idx];
				var row = this.rowMap[thisObject._key];
				if(row != undefined) row.mark(false);
				this.marked.splice(this.marked.indexOf(thisObject), 1);
			}	
			
			// If the object is not null, process the marking
			if(object != null) {
				if(Array.isArray(object)) {
					if(this.configuration.options.marking.mode == 'multiple') {
						for(var idx = 0; idx < object.length; idx++) {
							var thisObject = object[idx];
							var row = this.rowMap[thisObject._key];
							row.mark(true);
							this.marked.push(thisObject);
						}
					}
					// If we receive a list of objects but can only mark one, mark only the first one
					else if(this.configuration.options.marking.mode == 'single' && object.length > 0) {
						var thisObject = object[0];
						var row = this.rowMap[thisObject._key];
						row.mark(true);
						this.marked.push(thisObject);
					}
				}
				else {
					var row = this.rowMap[object._key];
					row.mark(true);
					this.marked.push(object);
				}
			}
			
			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.marking.passthrough == true) {
				this.sink.dataset.trigger('marked', this.sink, object);
			}			

			this.updateInfo();
		}
	};

	/**
	 * Applies a filter and updates the grid. Will also optionally notify the
	 * dataset of a filter event, if enabled.
	 * @param {Array} filters - array of filter objects
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Grid.prototype.filter = function(filters, notify) {
		if(this.configuration.options.filter.enabled != false) {			
			// This clones the array so we don't get duplicates
			this.filterDef.filter.filters = filters.slice(0);
			this.filterDef.filter.fields = this.configuration.options.filter.fields;
			
			// Clear all existing filters
			for(var idx = 0; idx < this.columnArr.length; idx++) {
				this.columnArr[idx].applyFilter(null, false);
			}
			
			// Apply the new filters on each column
			for(var idx = 0; idx < filters.length; idx++) {
				var column = this.columnMap[filters[idx].field];
				column.applyFilter(filters[idx].expression, false);
			}

			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.filter.passthrough == true) {
				this.sink.dataset.trigger('filter', this.sink, this.filterDef.filter.filters);
			}
			
			this.update();
		}
	};
	
	/**
	 * Updates the contents of the information bar, if present.
	 */
	Grid.prototype.updateInfo = function() {
		$(this.containerDiv).find('.grid-showing').text(this.filterArr.length + " of " + this.rowArr.length + ' rows ' + this.marked.length + ' marked');			
	}

	/**
	 * Clears all ordering and updates the grid.
	 */
	Grid.prototype.clearOrdering = function() {
		if(this.orderDef.length > 0) {
			for(var idx = this.orderDef.length - 1; idx >= 0; idx--) {
				var prev = this.orderDef[idx];
				prev.applyOrder(null, false);
				this.orderDef.splice(idx, 1);
			}
		}
		this.update();
	}
	
	/** 
	 * Clears all field-level filters and updates the grid. Does NOT
	 * clear any Search value.
	 */
	Grid.prototype.clearFilters = function() {
		for(var idx = 0; idx < this.columnArr.length; idx++) {
			this.columnArr[idx].applyFilter(null, false);
		}
		this.filter([], true);
	};
	
	/**
	 * Toggles the visibility of the field-level filters.
	 */
	Grid.prototype.toggleFilters = function() {
		if(this.filtersVisible == true)
			$(this.containerDiv).find('.grid-header').removeClass('display-filters');
		else
			$(this.containerDiv).find('.grid-header').addClass('display-filters');
		this.filtersVisible = !this.filtersVisible;
	};
	
	/**
	 * Display the grid configuration modal.
	 */
	Grid.prototype.displayConfiguration = function() {
		if(this.configuration.options.configurable != true) return;
		console.log('display config');
		// TODO display configuration modal
	};

	/**
	 * Displays the dataset configuration modal.
	 */
	Grid.prototype.displayDataset = function() {
		if(this.configuration.options.dataset.modal == undefined) return;
		console.log('display dataset');
		// TODO display dataset modal
	};
	
	/** 
	 * Returns a color for the given value from the specified color map.
	 * @param {Object} value - the value to lookup, can be anything but typically a String
	 * @param (Object} name - the name of the color map
	 * @return {String} the color
	 */
	Grid.prototype.colormap = function(value, name) {
		var colorMap = this.colorMaps[name];
		if(colorValMap == undefined) return;
		return colorMap.getColor(value);
	};
	


	
	/**
	 * Creates a new Column object which represents a column in the grid.
	 * @constructor
	 * @param {Object} configuration - the configuration for the column
	 * @param {Object} grid - the grid containing this column
	 * @param {Object} containerDiv - this node object that header columns will be appended
	 * @param (Object} filterContainerDiv - the node object that column filters will be appended
	 */
	function Column(configuration, grid, containerDiv, filterContainerDiv) {
		this.field				= configuration.field;
		this.configuration 		= configuration;
		this.grid				= grid;
		this.containerDiv		= containerDiv;
		this.filterContainerDiv	= filterContainerDiv;

		// Set the title and datatype based on configuration
		this.title				= this.configuration.title == undefined ? this.configuration.field : this.configuration.title;
		this.type				= this.grid.schema.fieldMap[this.field] == undefined ? "string" : this.grid.schema.fieldMap[this.field].type;
		
		// The UI node object which is the DIV in the HTML for the column header
		this.node				= null;

		// The UI node object which is the DIV in the HTML for the column title
		this.titleNode			= null;
		
		// The UI node object which is the DIV in the HTML for the column filter
		this.filterNode			= null;
		
		// The UI node object which is the INPUT in the HTML for the column filter input box
		this.filterInputNode	= null;

		// Content to render for each row cell for this column. Ultimately this will end up being one of the following:
		//   1. undefined - default option, only the formatted data field will be appended to the cell 
		//   2. string - content field in column configuration in the form of a JS string template
		//   3. string - contentUrl loaded from the server in the form of a JS string template, used for complex, 
		//        multi-line type contents		
		this.content			= null;
		
		// Set the content value based on configuration, prioritize contentUrl over content
		var self = this;
		if(this.configuration.contentUrl != undefined) {
			$.ajax({
				type: 'GET',
				url: this.configuration.contentUrl,
				dataType: 'html',
				cache: false,
				aync: false,
				success: function(data, textStatus, xhr) {
					self.content = data;
				},
				error: function(xhr, textStatus, errorThrown) {
					console.error("Grid content url not found or invalid - " + self.configuration.contentUrl);
				}
			});	
		}
		else if(this.configuration.content != undefined) {
			this.content = this.configuration.content;
		}
		
		// Color map for this column
		this.colormap 			= new DataFlow.ColorMap(null, this.configuration.colormap);
		
		// Filter definition applied on the column
		this.filterDef			= null;
		
		// Flag to indicate if the column can be filtered
		this.filterable			= this.grid.configuration.options.filter.fields == undefined || 
									this.grid.configuration.options.filter.fields.length == 0 || 
									this.grid.configuration.options.filter.fields.indexOf(this.field) > -1;
									
		// Order direction, asc, desc, or null
		this.orderDir			= null;
		
		// Flag to indicate if the column can be ordered
		this.orderable			= this.grid.configuration.options.ordering.fields == undefined || 
									this.grid.configuration.options.ordering.fields.length == 0 ||
									this.grid.configuration.options.ordering.fields.indexOf(this.field) > -1;
		
	}
	
	/**
	 * Renders the UI components for a column header. This is a one-time render,
	 * if already rendered it will not be re-rendered.
	 */
	Column.prototype.render = function() {
		var self = this;
		
		// If node doesn't exist, then create it now
		if(this.node == null) {
			var context = {
				column:		this,
				grid:		this.grid
			}

			// Create the node and append to the header row
			this.node = $(document.createElement('div')).attr('class', 'header');
			this.containerDiv.append(this.node);
			
			// Create a filter node
			var filterIconNode = $(document.createElement('div')).attr('class', 'header-filter');
			this.node.append(filterIconNode);
			
			// Create a title node
			this.titleNode = $(document.createElement('div')).attr('class', 'header-title');
			this.node.append(this.titleNode);
						
			// Append default classes
			if(this.grid.configuration.options.columnHeaders.classes != undefined) {
				this.node.addClass(DataFlow.Utilities.Evaluate.classes(this.grid.configuration.options.columnHeaders.classes, context));
			}
			
			// Append default styles
			if(this.grid.configuration.options.columnHeaders.styles != undefined) {
				var styles = DataFlow.Utilities.Evaluate.styles(this.grid.configuration.options.columnHeaders.styles, context);
				for(style in styles) this.node.css(style, styles[style]);
			}

			// If filtering is enabled, append filter row
			if(this.grid.configuration.options.filter.enabled != false) {
				// Create the filter node and append to the header row
				this.filterNode = $(document.createElement('div')).attr('class', 'header-filter');
				this.filterContainerDiv.append(this.filterNode);
				
				// Append an input box to this filter node if there are no columns specified, or this column
				//   is one of the specified columns
				if(this.filterable == true) {
					var filterContainerNode = $(document.createElement('div'));
					this.filterNode.append(filterContainerNode);
					
					// Create the filter input node and append to the filter node
					this.filterInputNode = $(document.createElement('input')).attr('type', 'text');
					filterContainerNode.append(this.filterInputNode);
					
					// Create the filter input node clear and append
					var filterInputClearNode = $(document.createElement('div')).attr('class', 'header-filter-clear');
					filterContainerNode.append(filterInputClearNode);
					
					// Create event handler for change in input box value
					this.filterInputNode.on('input', DataFlow.Utilities.Debounce.debounce(function(e) {
						var val = $(e.target).val();
						self.applyFilter(val, true);
					}, this.grid.configuration.options.filter.delay));
					
					// Create event handler for clearing input box value
					filterInputClearNode.on('click', function(e) {
						self.filterInputNode.val('');
						var val = $(e.target).val();
						self.applyFilter(null, true);
					});
				}
			}
		
			// If ordering is enabled, create listeners
			if(this.grid.configuration.options.ordering.enabled == true) {
				// Apply initial ordering to this column if no columns are specified, or this 
				//   column is one of the specified columns
				if(this.orderable == true) {					
					// Add class
					this.node.addClass('orderable');
					
					// If there is initial ordering for this column, apply that sort
					if(this.grid.configuration.options.ordering.initial != undefined) {
						var initOrder = this.grid.configuration.options.ordering.initial[this.field];
						if(initOrder != undefined) {
							this.applyOrder(initOrder, false); // no update because that will happen below
						}
					}
					
					// Create a listener for click of the column header for sorting
					this.node.on('click', function(e) {
						if(self.orderDir == null)
							self.applyOrder('asc', true)
						else if(self.orderDir == 'asc')
							self.applyOrder('desc', true)
						else if(self.orderDir == 'desc')
							self.applyOrder(null, true)
					});
				}
			}
		
			// Add listener for the context menu
			this.node.on('contextmenu', function(e) {
				var context = {
					column:		self,
					grid:		self.grid
				}
				var container = self.grid.containerDiv;
				if(self.grid.configuration.options.menuContainer != undefined)
					container = $(self.grid.containerDiv).closest('.' + self.grid.configuration.options.menuContainer)[0];
				DataFlow.Utilities.ContextMenu.display(container, e, self.grid.configuration.headerMenu, context);
				return false;
			});
		}
		this.update();
	};
	
	/**
	 * Updates the column header.
	 */
	Column.prototype.update = function() {
		// Column header content will be the title if configured, otherwise
		//   set to the field name
		var content = this.title;
		this.titleNode.html(content);
		
		// Set the visibility
		if(this.configuration.visible == false) 
			this.node.css('display', 'none');
		else
			this.node.css('display', '');
		
		// Set the width
		if(this.configuration.width != undefined && this.node.width() != this.configuration.width) {
			this.node.css({'width': this.configuration.width});
		}
	};
	
	/**
	 * Applies a column ordering on the grid.
	 * @param {String} dir - the direction, 'asc', 'desc', or null to remove ordering
	 * @param {Boolean} update - flag indicating if the grid should be updated
	 */
	Column.prototype.applyOrder = function(dir, update) {
		// If multiple ordering is disabled, remove previous ordering
		if(this.grid.configuration.options.ordering.multiple == false) {
			if(this.grid.orderDef.length > 0) {
				var prev = this.grid.orderDef[0];
				this.grid.orderDef.splice(0, 1);
				prev.applyOrder(null, false); // no update because that will be happen below
			}
		}
		
		// Remove previous ordering class
		this.node.removeClass('order-asc order-desc');
		
		// Set the orderDir to the specified direction
		//   May be one of asc, desc, or null for no ordering
		this.orderDir = dir;
		
		// Add appropriate classes to the header
		if(dir == 'asc')
			this.node.addClass('order-asc');
		else if(dir == 'desc')
			this.node.addClass('order-desc');
		else if(dir == null) {
			this.grid.orderDef.splice(this.grid.orderDef.indexOf(this), 1);
		}

		if(dir != null && this.grid.orderDef.indexOf(this) == -1)
			this.grid.orderDef.push(this);
		
		if(update == true)
			this.grid.update();
	};
	
	/** 
	 * Applies a column-level filter on the grid.
	 * @param {String} val - the filter string to apply
	 */
	Column.prototype.applyFilter = function(val, notify) {
		// Create the filter definition
		if(val == null) {
			this.filterDef = null;
			if(this.filterInputNode != null)
				this.filterInputNode.val('');
		}
		else {
			this.filterDef = {
				field:		this.field,
				expression:	val
			}
			if(this.filterInputNode != null)
				this.filterInputNode.val(val);
		}
		
		// Call grid filter
		if(notify == true) {
			// Create an array of filters and populate with the current filters from the grid
			//   except for any filter for the current column
			var filters = [];
			for(var idx = 0; idx < this.grid.filterDef.filter.filters.length; idx++) {
				var thisFilter = this.grid.filterDef.filter.filters[idx];
				if(thisFilter.field != this.field)
					filters.push(thisFilter);
			}
			
			// Add this filter definition if it isn't null		
			if(this.filterDef != null)
				filters.push(this.filterDef);
		
			this.grid.filter(filters, true);
		}
		
	};
	
	/** 
	 * Returns a color for the given value from the color map for this column.
	 * @param {Object} value - the value to lookup, can be anything but typically a String
	 * @return {String} the color
	 */
	Column.prototype.colormap = function(value) {
		var color = this.colorValMap[value];
		if(color == undefined) {
			color = this.colorscale[this.colorIdx++];
			if(this.colorIdx == this.colorscale.length) {
				this.colorIdx = 0;
			}
			this.colorValMap[value] = color; 
		}
		return color;
	};
	
	
	
	/**
	 * Creates a new Row object which represents a row in the grid.
	 * @constructor
	 * @param {Object} object - the data object for the row
	 * @param {Object} columns - this list of columns for the row
	 * @param {Object} grid - the grid containing this row
	 */
	function Row(object, columns, grid) {
		this._key			= object._key;
		this.object 		= object;
		this.columns 		= columns;
		this.grid			= grid;
		this.configuration	= grid.configuration;
		
		// The UI node object which is the DIV in the HTML
		this.node			= null;
		
		// List of cell objects inside this row
		this.cells			= [];
		
		// Flags for hovered, marked, and filtered
		this.hovered		= false;
		this.marked			= false;
		this.filtered		= false;
		
		// Create a cell for each column in the configuration
		for(var idx = 0; idx < this.columns.length; idx++) {
			var thisColumn = this.columns[idx];
			this.cells.push(new Cell(this, thisColumn));
		}
	}
	
	/**
	 * Renders the UI components for a row. This is a one-time render,
	 * if already rendered it will not be re-rendered.
	 */
	Row.prototype.render = function() {
		var self = this;
		
		var context = {
			row:		this,
			grid:		this.grid
		}

		// Create the node element
		this.node = $(document.createElement('div')).attr('class', 'data-row');
		
		// Append default classes
		if(this.grid.configuration.options.rows.classes != undefined) {
			this.node.addClass(DataFlow.Utilities.Evaluate.classes(this.grid.configuration.options.rows.classes, context));
		}
		
		// If stripe classes are enabled, add class
		if(this.grid.configuration.options.rows.stripeClasses == true) {
			this.node.addClass('striped');
		}
		
		// Append default styles
		if(this.grid.configuration.options.rows.styles != undefined) {
			var styles = DataFlow.Utilities.Evaluate.styles(this.grid.configuration.options.rows.styles, context);
			for(style in styles) this.node.css(style, styles[style]);
		}
		
		// If marking is enabled, add a markable class and add the marking classes
		if(this.configuration.options.marking.enabled != false) {
			this.node.addClass('markable');
			this.node.addClass(this.configuration.options.marking.classes);
		}
	
		// If hover is enabled, add the hovering classes
		if(this.configuration.options.hover.enabled != false) {
			this.node.addClass(this.configuration.options.hover.classes);
		}

		// If row is already marked, add a marked class
		if(this.marked == true) {
			this.node.addClass('marked');
		}
		
		// Render each of the cells
		for(var idx = 0; idx < this.cells.length; idx++) {
			this.cells[idx].render();
		}
		
		// Create listener for mouseover to trigger hover
		this.node.on('mouseover', function() {
			self.grid.hover(self.object, true);
		});
		
		// Create listener for click to trigger marking
		this.node.on('click', function(e) {
			if(self.configuration.options.marking.enabled != false) {
				var ctrlKey = e.ctrlKey;					
				var mode = self.configuration.options.marking.mode;
				var isMarked = self.marked == true;
				
				// If unmarking is disabled and this row is marked, then just call mark again
				//   and take no further action
				if(self.configuration.options.marking.unmark == false && isMarked == true) {
					self.grid.mark(self.object, true);
					return;
				}
				
				// Single mode
				if(mode == 'single') {
					if(isMarked == true)
						self.grid.mark(null, true);
					else
						self.grid.mark(self.object, true);
				}
				// Multiple mode
				else if(mode == 'multiple') {
					if(ctrlKey == true) {
						var objects = [];
						for(var idx = 0; idx < self.grid.marked.length; idx++)
							objects.push(self.grid.marked[idx]);
						if(isMarked == false)
							objects.push(self.object);
						else if(isMarked == true)
							objects.splice(objects.indexOf(self.object), 1);
						self.grid.mark(objects, true);
					}
					else {
						if(isMarked == true)
							self.grid.mark(null, true);
						else
							self.grid.mark(self.object, true);
					}
				}
			}
		});			
		
		// Update the row
		this.update();
	};

	/** 
	 * Updates the contents of the row with the given object data.
	 * This will trigger cell.update() on each of the cells in the row.
	 */
	Row.prototype.update = function() {
		if(this.node == null) return;
		for(var idx = 0; idx < this.cells.length; idx++)
			this.cells[idx].update(this.object);
	};
	
	/**
	 * Copies the data for all cells to the clipboard in CSV format. 
	 * This will not include any formatting, just the raw data from the object.
	 */
	Row.prototype.copyToClipboard = function() {
		var text = "";
		for(key in this.object) text = text + "," + this.object[key];
		text = text.substring(1, text.length);
		DataFlow.Utilities.Clipboard.copy(text, this.node);
	};
	
	/**
	 * Sets the row hovered state based on the flag and updates the row classes.
	 * @param {Boolean} flag - true if the row is hovered, otherwise false
	 */
	Row.prototype.hover = function(flag) {
		if(flag == true && this.hovered == false) {
			this.hovered = true;
			if(this.node == null) return;		
			this.node.addClass('hover');
		}
		else if(flag == false && this.hovered == true) {
			this.hovered = false;
			if(this.node == null) return;		
			this.node.removeClass('hover');
		}
	};
	
	/**
	 * Sets the row marked state based on the flag and updates the row classes.
	 * @param {Boolean} flag - true if the row is marked, otherwise false
	 */
	Row.prototype.mark = function(flag) {
		if(flag == true && this.marked == false) {
			this.marked = true;
			if(this.node == null) return;
			this.node.addClass('marked');			
		}
		else if(flag == false && this.marked == true) {
			this.marked = false;
			if(this.node == null) return;
			this.node.removeClass('marked');			
		}
	}
	
	

	
	/**
	 * Creates a new Cell object which represents a cell in the grid.
	 * @constructor
	 * @param {Object} row - the row this cell belongs to
	 * @param {Object} column - this column this cell belongs to
	 */
	function Cell(row, column) {
		this.row			= row;
		this.column			= column;

		// The UI node object which is the DIV in the HTML
		this.node			= null;

		// Data object for the row, all fields
		this.object			= null;
		
		// Field data from the data object for this column
		this.data			= null;
		
		// Formatted field data
		this.formatted		= null;
	};
	
	/**
	 * Renders the UI components for a cell. This is a one-time render,
	 * if already rendered it will not be re-rendered.
	 */
	Cell.prototype.render = function() {
		var self = this;
		
		// Create a context object.
		var context = {
			data:		this.data,
			formatted:	this.formatted,
			object:		this.object,
			cell:		this,
			column:		this.column,
			row:		this.row,
			grid:		this.row.grid,
		}

		// Create the element
		this.node = $(document.createElement('div')).attr('class', 'cell');
		
		// Add default classes 
		if(this.row.grid.configuration.options.cells.classes != undefined) {
			this.node.addClass(DataFlow.Utilities.Evaluate.classes(this.row.grid.configuration.options.cells.classes, context));
		}
		
		// Add custom class if there is custom content for padding
		if(this.column.content != undefined)
			this.node.addClass('custom');
		
		// Add default styles
		if(this.row.grid.configuration.options.cells.styles != undefined) {
			var styles = DataFlow.Utilities.Evaluate.styles(this.row.grid.configuration.options.cells.styles, context);
			for(style in styles) this.node.css(style, styles[style]);
		}

		// Add listener for the context menu
		this.node.on('contextmenu', function(e) {
			var context = {
				data:		self.data,
				formatted:	self.formatted,
				object:		self.object,
				cell:		self,
				column:		self.column,
				row:		self.row,
				grid:		self.row.grid,
			}
			var container = self.row.grid.containerDiv;
			if(self.row.grid.configuration.options.menuContainer != undefined)
				container = $(self.row.grid.containerDiv).closest('.' + self.row.grid.configuration.options.menuContainer)[0];
			DataFlow.Utilities.ContextMenu.display(container, e, self.row.grid.configuration.bodyMenu, context);
			return false;
		});
		
		// Append the cell to the row
		this.row.node.append(this.node);
		
		// No update here, will be triggered by row.update()
	};

	/**
	 * Updates the contents of the cell with the given object data.
	 * @param {Object} object - the data object to update
	 */
	Cell.prototype.update = function(object) {
		var self 		= this;
		this.data 		= object[this.column.field] != undefined ? object[this.column.field] : null;
		this.object 	= object;
		
		// Function to retrieve a color for the given value
		//   If name is undefined then the column colormap is used, 
		//   otherwise the grid colormap is used
		function colormap(value, name) {
			if(name == undefined)
				return self.column.colormap.getColor(value);
			else {
				var colormap = self.row.grid.colorMaps[name];
				if(colormap == undefined) return;
				return colormap.getColor(value);
			}
		}

		// Function to format the specified value, which can be timestamp or number
		function format(val, format) {
			return DataFlow.Utilities.Format.format(val, format);
		};
				
		// Apply any default formatting by data type
		this.formatted	= this.data;
		if(this.row.grid.configuration.options.format[self.column.type] != undefined) {
			var thisFormat = this.row.grid.configuration.options.format[self.column.type];
			if(self.column.type == 'timestamp')
				this.formatted = DataFlow.Utilities.Format.formatTimestamp(this.data, thisFormat);
			else if(self.column.type == 'double' || self.column.type == 'int' || self.column.type == 'long')
				this.formatted = DataFlow.Utilities.Format.formatNumber(this.data, thisFormat);
		}
		
		// Create the context
		var context = {
			data:		this.data,
			formatted:	this.formatted,
			object:		this.object,
			cell:		this,
			column:		this.column,
			row:		this.row,
			grid:		this.row.grid,
			colormap:	colormap,
			format:		format
		}
			
		// Set the node content
		var content = this.formatted;
		if(this.column.content != undefined) {
			content = function(str){
				for(key in this) eval('var ' + key + ' = this.' + key);
				return eval('`' + str + '`');
			}.call(context, this.column.content);
		}
		this.node.empty().html(content);
		if(this.node.children().length > 0)
			$(Utilities.Template.process(this.node.children()[0]));
		
		// Add or remove classes
		if(this.column.configuration.classes != undefined) {
			var classes = DataFlow.Utilities.Evaluate.classes(this.column.configuration.classes, context);
			if(typeof classes === 'object') {
				for(className in classes) {
					var value = classes[className];
					if(value == true)
						this.node.addClass(className)
					else
						this.node.removeClass(className);
				}
			}
			else {
				this.node.addClass(classes);
			}			
		}

		// Set styles
		if(this.column.configuration.styles != undefined) {
			var styles = DataFlow.Utilities.Evaluate.styles(this.column.configuration.styles, context);
			for(style in styles) this.node.css(style, styles[style]);
		}
	
		// Set the visibility
		if(this.column.configuration.visible == false) 
			this.node.css('display', 'none');
		else
			this.node.css('display', '');
		
		// Set the width
		if(this.column.configuration.width != undefined && this.node.width() != this.column.configuration.width) {
			this.node.css({'width': this.column.configuration.width});
		}
	};
	
	/**
	 * Copies the cell data to the clipboard. This will not include
	 * any formatting, just the raw data from the object.
	 */
	Cell.prototype.copyToClipboard = function() {
		DataFlow.Utilities.Clipboard.copy(this.data, this.node);
	};
	
	
	

	/** 
	 * Grid configuration helper object.
	 */
	GridConfiguration = {
		// Default grid configuration object
		defaultConfiguration: {
				columns:	[],
				options: {
					//classes:			undefined,
					configurable:		false,
					cells: {
						//classes:		undefined,
						//styles:		undefined
					},
					columnHeaders: {
						//classes:		undefined,
						//styles:		undefined
					},
					copy: {
						enabled:		true
					},
					dataset: {
						icon:			true
						//modal:			undefined
					},
					filter: {	
						accept:			true,
						fields:			undefined,
						delay:			500,
						enabled:		true,
						passthrough:	true
					},
					format: {
						timestamp:		"yyyy-MM-dd HH:mm:ss"
					},
					hover: {
						accept:			true,
						classes:		"hover-background",
						enabled:		true,
						passthrough:	true
					},
					// info:			false,
					marking: {
						accept:			true,
						classes:		"mark-background",
						enabled:		true,
						mode:			"single",
						passthrough:	true,
						unmark:			true
					},
					ordering: {
						//classes:		undefined,
						fields:			undefined,
						enabled:		true,
						//initial:		{},
						multiple:		false
					},
					pagination: {
						enabled:		true,
						size:			25
					},
					rows: {
						//classes:		undefined,
						stripes:		false
						//styles:		undefined
					},
					search: {
						fields:			undefined,
						delay:			500,
						enabled:		true,
						label:			"Search"
					},
					//styles:			undefined,
					//title:			undefined,
				},
				
				headerMenu: {
					ordering: [
						{
							label:		"Order ${column.title} Ascending",
							condition:	"column.orderable == true && (column.orderDir == null || column.orderDir == 'desc')",
							action:		"column.applyOrder('asc', true)"
						},
						{
							label:		"Order ${column.title} Descending",
							condition:	"column.orderable == true && (column.orderDir == null || column.orderDir == 'asc')",
							action:		"column.applyOrder('desc', true)"
						},
						{
							label:		"Clear ${column.title} Ordering",
							condition:	"column.orderable == true && (column.orderDir != null)",
							action:		"column.applyOrder(null, true)"
						},
						{
							label:		"Clear All Ordering",
							condition:	"grid.configuration.options.ordering.enabled != false && grid.orderDef.length > 0",
							action:		"grid.clearOrdering()"
						}
					],
					filtering: [
						{
							label:		"Clear ${column.title} Filter",
							condition:	"column.filterable == true && column.filterDef != null",
							action:		"column.applyFilter(null, true, true)"
						},
						{
							label:		"Clear All Filters",
							condition:	"grid.configuration.options.filter.enabled != false && grid.filterDef.filter.filters != undefined && grid.filterDef.filter.filters.length > 0",
							action:		"grid.clearFilters()"
						},
						{
							label:		"Show Field Filters",
							condition:	"grid.configuration.options.filter.enabled != false && grid.filtersVisible == false",
							action:		"grid.toggleFilters()"
						},
						{
							label:		"Hide Field Filters",
							condition:	"grid.configuration.options.filter.enabled != false && grid.filtersVisible == true",
							action:		"grid.toggleFilters()"
						}
					],
					panels: [
						{
							label:		"Display Configuration",
							condition:	"grid.configuration.options.configurable == true",
							action:		"grid.displayConfiguration()"
						},
						{
							label:		"Display Dataset",
							condition:	"grid.configuration.options.dataset.modal != undefined",
							action:		"grid.displayDataset()"
						}
					]
				},
				
				bodyMenu: {
					clipboard: [
						{
							label:		"Copy Cell",
							condition:	"document.queryCommandSupported('copy') == true && grid.configuration.options.copy.enabled != false",
							action:		"cell.copyToClipboard()"
						},
						{
							label:		"Copy Row",
							condition:	"document.queryCommandSupported('copy') == true && grid.configuration.options.copy.enabled != false",
							action:		"row.copyToClipboard()"
						}
					],
					filtering: [
						{
							label:		"Filter ${column.title} to value ${data}",
							condition:	"column.filterable == true",
							action:		"column.applyFilter(\"== '\" + data + \"'\", true, true)"
						},
						{
							label:		"Clear ${column.title} Filter",
							condition:	"column.filterable == true && column.filterDef != null",
							action:		"column.applyFilter(null, true, true)"
						},
						{
							label:		"Clear All Filters",
							condition:	"grid.configuration.options.filter.enabled != false && grid.filterDef.filter.filters != undefined && grid.filterDef.filter.filters.length > 0",
							action:		"grid.clearFilters()"
						}
					],
					panels: [
						{
							label:		"Display Configuration",
							condition:	"grid.configuration.options.configurable == true",
							action:		"grid.displayConfiguration()"
						},
						{
							label:		"Display Dataset",
							condition:	"grid.configuration.options.dataset.modal != undefined",
							action:		"grid.displayDataset()"
						}
					]
				}
			},
			
		// Default grid column configuration object
		defaultColumnConfiguration: {
				//field:				undefined,
				//classes:				undefined, 
				//styles:				undefined,
				//title:				undefined,
				visible:				true,
				//width:				undefined
			},
		
		// Merges the source configuration object into the target configuration object
		//   Source values will overwrite target values.
		//   Should always merge a DEFAULT configuration source into a USER configuration target
		mergeConfiguration:		function(source, target) {
									for(var idx = 0; idx < Object.keys(source).length; idx++) {
										var key = Object.keys(source)[idx];
										var val = source[key];
										if(typeof val === 'object') {
											if(target[key] == undefined) 
												target[key] = val;
											else
												this.mergeConfiguration(val, target[key]);
										}
										else {
											if(target[key] == undefined) 
												target[key] = val;
										}
									}
									return target;
								},

		// Processes the specified configuration object and merges in default configuration
		//   for both base object and column objects
		processConfiguration:	function(config) {
									var configuration = this.mergeConfiguration(this.defaultConfiguration, config);
									for(var idx = 0; idx < configuration.columns.length; idx++) {
										configuration.columns[idx] = this.mergeConfiguration($.extend({}, this.defaultColumnConfiguration), 
											configuration.columns[idx]);
									}
									return configuration;
								},
	};
	
	
	
	
	// Setup module objects
	module.Grid 				= Grid;
	module.GridConfiguration	= GridConfiguration;
	
	return module;
})();