/**
 * Represents a Highcharts chart data sink.
 * @constructor
 * @param {Object} dataset - dataset that this sink will be attached to
 * @param {Object} configuration - configuration object defining the sink
 */
function HighchartsChartDatasink(id, configuration, dataset, containerDiv, overlay) {
	this.id = id;
	this.dataset = dataset;
	this.chart = new LiveHighchartsChart.Chart(id, configuration, containerDiv, overlay, this);
}

/**
 * Event handler for dataset events. 
 * @param {String} eventType - string describing the event that occurred
 * @param {Object} object - data object for the event, or null
 */
HighchartsChartDatasink.prototype.trigger = function(eventType, object) {
	this.chart.trigger(eventType, object);
};


// Define a module to hold object defintions for the chart objects
var LiveHighchartsChart = (function() {
	var module = {};

	/**
	 * Creates a new Chart object.
	 * @constructor
	 * @param {String} id - unique identifier
	 * @param {Object} configuration - the configuration for the chart
	 * @param {Object} containerDiv - this node object that chart will be appended
	 * @param (Function} overlay - function to call on start and end of the overlay
	 * @param {Object} sink - datasink that will feed this chart
	 */
	function Chart(id, configuration, containerDiv, overlay, sink) {
		this.id				= id;
		this.configuration	= configuration;
		this.containerDiv 	= containerDiv;
		this.overlay		= overlay;
		this.sink			= sink;

		// Pause and snapshot flags
		this.paused			= false;
		this.snapshot		= false;

		// Schema for the dataset
		this.schema			= null;		
		
		// Highcharts charting object
		this.highcharts		= null;
		
		// Vars to hold the objects being hovered and marked
		this.hovered		= [];		
		this.marked			= [];

		// Filter definition
		this.filterDef		= {
			search: {
				expression:	undefined,
				fields:		this.configuration.options.search.fields
			},
			filter:	{
				filters:	[],
				fields:		this.configuration.options.filter.fields
			}
		};

		// Color maps
		this.colorMaps		= {};

		// Series array, but not Highcharts series objects, LiveHighchartsChart.Series objects
		this.seriesArr 		= [];
		this.seriesMap		= {};
		
		// Check dependencies. This will throw Error if dependencies are not satisfied.
		this.dependencyCheck();
		
		// Microbatch is used to batch up updates to prevent too many chart redraws
		this.microbatch = {
			rtime:		new Date(1, 1, 2000, 12,00,00),
			timeout:	false,
			delta:		500 // msec between updates
		}

		// Color maps
		if(this.configuration.options.colormaps != undefined) {
			for(var idx = 0; idx < Object.keys(this.configuration.options.colormaps).length; idx++) {
				var thisColorMapName = Object.keys(this.configuration.options.colormaps)[idx];
				var thisColors = this.configuration.options.colormaps[thisColorMapName];
				this.colorMaps[thisColorMapName] = new DataFlow.ColorMap(thisColorMapName, thisColors);
			}
		}
		
		// Loop over the series configuration and create a series object for each
		for(var idx = 0; idx < configuration.highcharts.series.length; idx++) {
			var thisSeries = new Series(configuration.highcharts.series[idx], this, idx);
			this.seriesArr.push(thisSeries);
			this.seriesMap[thisSeries.id] = thisSeries;
		}
		
		// Register UI handlers
		this.registerEventHandlers();

		// Draw the initial empty chart
		this.draw();
	};
	
	/**
	 * Performs a dependency check for rendering the chart.
	 */
	Chart.prototype.dependencyCheck = function() {
		var msg = "Required Highcharts libraries not loaded or not present";
		try {
			var h = Highcharts;
		}
		catch(err) {
			throw Error("Required Highcharts libraries not loaded or not present");
		}
	};
		
	/**
	 * Registers event handlers on UI components.
	 */
	Chart.prototype.registerEventHandlers = function() {
		var self = this;
		
		// Event handler to hide any open context menu on click
		$(document).on('click', function() {
			$(document).find('.live-context-menu').remove();
		});
		
		// Event handler to display the configuration modal
		$(this.containerDiv).find('.chart-config').on('click', function(e) {
			self.displayConfiguration();
		});

		// Event handler to display the dataset modal
		$(this.containerDiv).find('.chart-icon').on('click', function(e) {
			self.displayDataset();
		});
		
		// Event handler to detect mouse over and mouse out of points on the chart to set hover
		if(this.configuration.options.hover.enabled == true) {
			this.configuration.highcharts.plotOptions.series.point.events.mouseOver = function() {
				var object = self.seriesArr[0].pointMap[this._key].object;
				self.hover(object, true);
			};
			this.configuration.highcharts.plotOptions.series.point.events.mouseOut = function() {
				self.hover(null, true);
			};
		}
		
		// Event handler to detect clicks on points to set marking
		if(this.configuration.options.marking.enabled == true) {
			// Adjust default configuration to allow point select
			this.configuration.highcharts.plotOptions.series.allowPointSelect = true;
			
			// Register event handler
			this.configuration.highcharts.plotOptions.series.point.events.click = function(e) {
				var ctrlKey = e.ctrlKey;					
				var mode = self.configuration.options.marking.mode;
				var point = self.seriesArr[0].pointMap[this._key];
				var object = point.object;
				var isMarked = point.marked == true;
				
				// If unmarking is disabled and this component is marked, then just call mark again
				//   and take no further action
				if(self.configuration.options.marking.unmark == false && isMarked == true) {
					self.mark(object, true);
					return;
				}
				
				// Single mode
				if(mode == 'single') {
					if(isMarked == true)
						self.mark(null, true);
					else
						self.mark(object, true);
				}
				// Multiple mode
				else if(mode == 'multiple') {
					if(ctrlKey == true) {
						var objects = [];
						for(var idx = 0; idx < self.marked.length; idx++)
							objects.push(self.marked[idx]);
						if(isMarked == false)
							objects.push(object);
						else if(isMarked == true)
							objects.splice(objects.indexOf(self.object), 1);
						self.mark(objects, true);
					}
					else {
						if(isMarked == true)
							self.mark(null, true);
						else
							self.mark(object, true);
					}
				}
						
				return;
			};
		}
		
		// Event handler for 3D rotation, if enabled
		if(this.configuration.options.rotate3d.enabled == true) {
			$(this.containerDiv).on('mousedown.hc touchstart.hc', function (eStart) {
				var eStart = self.highcharts.pointer.normalize(eStart);

				var posX = eStart.pageX;
				var posY = eStart.pageY;
				var alpha = self.highcharts.options.chart.options3d.alpha;
				var beta = self.highcharts.options.chart.options3d.beta;
				var newAlpha = null;
				var newBeta = null;
				var sensitivity = 5; // lower is more sensitive

				$(document).on({
					'mousemove.hc touchdrag.hc': function (e) {
						// Run beta
						newBeta = beta + (posX - e.pageX) / sensitivity;
						self.highcharts.options.chart.options3d.beta = newBeta;

						// Run alpha
						newAlpha = alpha + (e.pageY - posY) / sensitivity;
						self.highcharts.options.chart.options3d.alpha = newAlpha;

						self.highcharts.redraw(false);
					},
					'mouseup touchend': function () {
						$(document).off('.hc');
					}
				});
			});
		}
	};
	
		
	/**
	 * Event handler for all objects delegated from the datasink.
	 * @param {String} eventType - the event type
	 * @param {Object} object - optional object
	 */
	Chart.prototype.trigger = function(eventType, object) {
		var self = this;
		
		if(eventType == 'schemaChange') {
			this.schema = object;
		}
		else if(eventType == 'snapshotStart') {
			this.snapshot = true;
			this.overlay(true);
		}
		else if(eventType == 'snapshotEnd') {
			setTimeout(function() {
				self.snapshot = false;
				self.overlay(false);
			},500);
		}
		else if(eventType == 'clear') {
			this.clear();
		}
		else if(eventType == 'insert') {
			this.addPoint(object);
		}
		else if(eventType == 'update') {
			this.updatePoint(object);
		}
		else if(eventType == 'upsert') {
		}
		else if(eventType == 'delete') {
			this.removePoint(object);
		}
		else if(eventType == 'hover') {
			if(this.configuration.options.hover.accept != false)
				this.hover(object);
		}
		else if(eventType == 'marked') {
			if(this.configuration.options.marking.accept != false)
				this.mark(object);
		}
		else if(eventType == 'filter') {
			if(this.configuration.options.filter.accept != false)
				this.filter(object);
		}					
		
	};
	
	/**
	 * Add a new point bound to the specified object on the chart.
	 * @param {Object} object - the object to add the bound point
	 */
	Chart.prototype.addPoint = function(object) {
		for(var idx = 0; idx < this.seriesArr.length; idx++)
			this.seriesArr[idx].addPoint(object);
		this.update();
	};

	/**
	 * Updates the point bound to the specified object on the chart.
	 * @param {Object} object - the object to update the bound point
	 */
	Chart.prototype.updatePoint = function(object) {
		for(var idx = 0; idx < this.seriesArr.length; idx++)
			this.seriesArr[idx].updatePoint(object);
		this.update();
	};
	
	/**
	 * Removes the point bound to the specified object from the chart.
	 * @param {Object} object - the object to remove the bound point
	 */
	Chart.prototype.removePoint = function(object) {
		for(var idx = 0; idx < this.seriesArr.length; idx++)
			this.seriesArr[idx].removePoint(object);
		this.update();
	};
					
	/**
	 * Clears all objects and all points from the chart.
	 */
	Chart.prototype.clear = function() {
		for(var idx = 0; idx < this.seriesArr.length; idx++)
			this.seriesArr[idx].removeAll();
		this.update();
	};
	
	/**
	 * Draws the UI components for the chart.
	 */
	Chart.prototype.draw = function() {
		this.highcharts = Highcharts.chart(this.containerDiv, this.configuration.highcharts);	;
	};
	
	/**
	 * Updates the contents of the chart. Normally this will happen in a microbatch
	 * to prevent too many updates from occurring. However it is possible to trigger
	 * an immediate update by setting the flag.
	 * @param {Boolean} immediate - trigger an immediate update
	 */
	Chart.prototype.update = function(immediate) {
		var self = this;

		if(immediate == true) {
			run();
			return;
		}

		// Microbatch handler to prevent too many updates
		if(self.microbatch.timeout === false) {
			self.microbatch.timeout = true;
			setTimeout(eventEnd, self.microbatch.delta);
		}					
		function eventEnd() {
			var d = new Date() - self.microbatch.rtime;
			if (new Date() - self.microbatch.rtime < self.microbatch.delta)
				setTimeout(eventEnd, self.microbatch.delta);
			else {
				self.microbatch.rtime = new Date();
				self.microbatch.timeout = false;
				run();
			}               
		};
		
		// Run the update
		function run() {
			if(self.paused == true) return;
			self.highcharts.redraw();
		};
	};
	
	/**
	 * Pauses updates on a chart.
	 */
	Chart.prototype.pause = function() {
		this.paused = true;
	};
	
	/**
	 * Resumes a paused chart.
	 */
	Chart.prototype.resume = function() {
		if(this.paused == false) return;
		this.paused = false;
		this.update();
	};
	
	/**
	 * Resizes a chart.
	 */
	Chart.prototype.resize = function() {
		this.highcharts.setSize($(this.containerDiv).parent().width(), 
			$(this.containerDiv).parent().height());
	};
			
	/**
	 * Applies a search value against the chart. This will not be passed through
	 * to the dataset and is local only to the current chart.
	 * @param {String} value - the search string
	 */
	Chart.prototype.search = function(value) {
		if(value == undefined || value.length == 0) {
			delete this.filterDef.search.expression;
		}
		else {
			this.filterDef.search.expression = value;
		}
		
		// Pass the filterDef through to the series
		for(var idx = 0; idx < this.seriesArr.length; idx++) {
			this.seriesArr[idx].filter(this.filterDef);
		}
	};
	
	/**
	 * Hovers a single point bound to the specified object.
	 * Will also optionally notify the dataset of a hover event, if enabled.
	 * @param {Object} object - object that the associated point will be hovered
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Chart.prototype.hover = function(object, notify) {
		if(this.configuration.options.hover.enabled != false) {
			// Unhover all existing series
			for(var idx = 0; idx < this.seriesArr.length; idx++) {
				this.seriesArr[idx].unhover();
			}
			
			// If the object is not null, process the hovering
			if(object != null) {
				if(Array.isArray(object)) {
					this.hovered = object.slice(0);
				}
				else {
					this.hovered = [object];
				}

				for(var idx = 0; idx < this.seriesArr.length; idx++) {
					this.seriesArr[idx].hover(this.hovered);
				}
			}

			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.hover.passthrough == true) {
				this.sink.dataset.trigger('hover', this.sink, object);
			}			
		}
	};
		
	/**
	 * Marks a single point or a set of points bound to the specified object or array of objects.
	 * Will also optionally notify the dataset of a marked event, if enabled.
	 * @param {Object|Array} object - if a single data object then the associated point will be marked
	 *                                if an array of objects then the associated points will be marked
	 *                                  for single marking points that receive a list of objects then only
	 *                                  the first point will be marked
	 *                                if null then all marked points will be unmarked
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Chart.prototype.mark = function(object, notify) {
		if(this.configuration.options.marking.enabled != false) {
			var mode = this.configuration.options.marking.mode;

			// Unmark all existing series
			for(var idx = 0; idx < this.seriesArr.length; idx++) {
				this.seriesArr[idx].unmark();
			}
			this.marked = [];
			
			// If the object is not null, process the marking
			if(object != null) {
				if(Array.isArray(object)) {
					if(this.configuration.options.marking.mode == 'multiple') {
						for(var idx = 0; idx < object.length; idx++) {
							var thisObject = object[idx];
							this.marked.push(thisObject);
						}
					}
					// If we receive a list of objects but can only mark one, mark only the first one
					else if(this.configuration.options.marking.mode == 'single' && object.length > 0) {
						var thisObject = object[0];
						this.marked.push(thisObject);
					}
				}
				else {
					this.marked.push(object);
				}

				// Update the series marked
				for(var idx = 0; idx < this.seriesArr.length; idx++) {
					this.seriesArr[idx].mark(this.marked);
				}
			}
			
			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.marking.passthrough == true) {
				this.sink.dataset.trigger('marked', this.sink, this.marked);
			}			
		}
	};
		
	/**
	 * Applies a filter and updates the chart. Will also optionally notify the
	 * dataset of a filter event, if enabled.
	 * @param {Array} filters - array of filter objects
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Chart.prototype.filter = function(filters, notify) {
		if(this.configuration.options.filter.enabled != false) {
			// This clones the array so we don't get duplicates
			this.filterDef.filter.filters = filters.slice(0);
			this.filterDef.filter.fields = this.configuration.options.filter.fields;

			// Pass the filterDef through to the series
			for(var idx = 0; idx < this.seriesArr.length; idx++) {
				this.seriesArr[idx].filter(this.filterDef);
			}

			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.filter.passthrough == true) {
				this.sink.dataset.trigger('filter', this.sink, this.filtered.filter);
			}
			
			this.update();
		}
	};
	
	/**
	 * Display the chart configuration modal.
	 */
	Chart.prototype.displayConfiguration = function() {
		if(this.configuration.options.configurable != true) return;
		console.log('highcharts display config');
		// TODO display configuration modal
	};

	/**
	 * Displays the dataset configuration modal.
	 */
	Chart.prototype.displayDataset = function() {
		if(this.configuration.options.dataset.modal == undefined) return;
		console.log('highcharts display dataset');
		// TODO display dataset modal
	};
						
	/** 
	 * Returns a color for the given value from the color map.
	 * @param {Object} value - the value to lookup, can be anything but typically a String
	 * @return {String} the color
	 */
	Chart.prototype.colormap = function(value) {
		return this.colorMap.getColor(value);
	};
	
	
	
	
	/**
	 * Creates a new Series object which represents a set of data on a chart. 
	 * A Series will represent different things depending on the underlying chart type.
	 * Typically a set of points or a line on a chart will be one series. A pie chart
	 * consists of a single series.
	 * @constructor
	 * @param {Object} configuration - the configuration for the series
	 * @param {Object} chart - the chart containing this series
	 */
	function Series(configuration, chart, index) {
		// Set configuration
		this.configuration 		= configuration;
		this.chart 				= chart;
		this.id					= 'series-' + index;
		
		this.highchartsSeries	= null;
		
		this.hovered			= [];
		this.marked				= [];
		this.filterDef			= {
			search: {
				expression:	undefined,
				fields:		this.chart.configuration.options.search.fields
			},
			filter:	{
				filters:	[],
				fields:		this.chart.configuration.options.filter.fields
			}
		};

		// Initialize map for point objects
		this.pointMap			= {};
		
		// Create the property index
		var pointPropertiesIndexResult = DataFlow.Utilities.PropertyMap.index(this.configuration.pointProperties, ['property', 'eval']);
		if(pointPropertiesIndexResult != null) {
			this.pointPropertiesIndex = pointPropertiesIndexResult.propertyIndex;
			this.pointProperties = pointPropertiesIndexResult.properties;
		}

		this.configuration.id 	= this.id;
		if(this.configuration.data == undefined)
			this.configuration.data = [];
	};

	/**
	 * Add a new object to the series as a new point.
	 * @param {Object} object - the object to add
	 */
	Series.prototype.addPoint = function(object) {
		if(this.pointPropertiesIndex == null) return;
		
		var filtered = this.checkFilter(object);
		if(filtered == true) return;

		var data = {
			_key:				object._key,
			id:					this.id + "-" + object._key,
			object:				object,
			selected:			false
		}
		var point = {
			data:				data,
			object:				object,
			highchartsPoint: 	null,
			marked:				false,
			added:				false
		}

		// Build a context and evaluate the point properties, append to the data object
		var context = {
			object:		object,
			series:		this,
			chart:		this.chart
		};
		DataFlow.Utilities.PropertyMap.evaluate(this.pointPropertiesIndex, context, object, data);
		
		// Add the point to the point map
		this.pointMap[object._key] = point;

		// Check the filter
		var include = DataFlow.Utilities.Filtering.filterItem(data, 
			this.filterDef.filter.filters, this.filterDef.filter.fields, 'object');
		
		if(include == true) {
			if(this.highchartsSeries == null) 
				this.highchartsSeries = this.chart.highcharts.get(this.id);
			this.highchartsSeries.addPoint(data, false);
			point.added = true;
		}
	};

	/**
	 * Update an existing object point on the series.
	 * @param {Object} object - the object to update
	 */
	Series.prototype.updatePoint = function(object) {
		if(this.pointPropertiesIndex == null) return;

		var point = this.pointMap[object._key];
		if(point == undefined) return;
		var data = point.data;
		
		// Build a context and evaluate the point properties, append to the data object
		var context = {
			object:		object,
			series:		this,
			chart:		this.chart
		};
		DataFlow.Utilities.PropertyMap.evaluate(this.pointPropertiesIndex, context, object, data);
		
		// Check the filter
		var include = DataFlow.Utilities.Filtering.filterItem(data, 
			this.filterDef.filter.filters, this.filterDef.filter.fields, 'object');
		
		var self = this;
		var doUpdate = function(thisData) {
			if(include == true) {
				if(point.added == false) {
					if(self.highchartsSeries == null) 
						self.highchartsSeries = self.chart.highcharts.get(self.id);
					self.highchartsSeries.addPoint(data, false);
					point.added = true;
				}
				else {
					point.highchartsPoint.update(thisData, false);
				}
			}
			else if(include == false && point.added == true) {
				if(point.highchartsPoint != null) {
					data.color = point.highchartsPoint.color;
					point.highchartsPoint.select(false);		
					point.highchartsPoint.remove(data);
					point.highchartsPoint = null;
				}
				point.added = false;
			}				
		};

		if(point.highchartsPoint == null)
			point.highchartsPoint = this.chart.highcharts.get(data.id);
		if(point.highchartsPoint == null) {
			var self = this;
			setTimeout(function() {
				point.highchartsPoint = self.chart.highcharts.get(data.id);		
				if(point.highchartsPoint == null && point.added == true) return;
				doUpdate(data);
			},1000); // this is highcharts nonsense because sometimes we can't find the point too soon after add
		}
		else {
			doUpdate(data);
		}
	};
	
	/**
	 * Remove an existing object point from the series.
	 * @param {Object} object - the object to remove
	 */
	Series.prototype.removePoint = function(object) {
		if(this.pointPropertiesIndex == null) return;

		var point = this.pointMap[object._key];
		if(point == undefined) return;
		var data = point.data;
		
		var doRemove = function(thisData) {
			point.highchartsPoint.remove(thisData);
			point.added = false;
		};
		
		if(point.highchartsPoint == null)
			point.highchartsPoint = this.chart.highcharts.get(data.id);
		if(point.highchartsPoint == null) {
			var self = this;
			setTimeout(function() {
				point.highchartsPoint = self.chart.highcharts.get(data.id);		
				if(point.highchartsPoint == null) return;
				doRemove(data);
			},1000); // this is highcharts nonsense because sometimes we can't find the point too soon after add
		}
		else {
			doRemove(data);
		}
		
		delete this.pointMap[object._key];		
	};
	
	/**
	 * Remove all existing object points from the series.
	 */
	Series.prototype.removeAll = function() {
		if(this.pointPropertiesIndex == null) return;

		if(this.highchartsSeries == null) 
			this.highchartsSeries = this.chart.highcharts.get(this.id);
		this.highchartsSeries.setData([]);
		this.pointMap = {};		
	};
	
	/** 
	 * Updates the trace array with the current property arrays.
	 */
	Series.prototype.update = function() {

	};

	/**
	 * Checks to see if the given object has been filtered by any configured
	 * filter condition.
	 * @param {Object} object - the object to test
	 * @return false if the object is included, true if the object is filtered
	 */
	Series.prototype.checkFilter = function(object) {
		if(this.configuration.pointProperties == null) return false;
		if(this.configuration.pointProperties.eval_filter == undefined) return false;
		var context = {
			object:		object,
			series:		this,
			chart:		this.chart
		};
		var b = DataFlow.Utilities.Evaluate.expression(this.configuration.pointProperties.eval_filter, context);
		return !b;
	};
	
	/**
	 * Unhovers the hovered point if there is one.
	 */
	Series.prototype.unhover = function() {
		for(var idx = 0; idx < this.hovered.length; idx++) {
			var thisObject = this.hovered[idx];
			var point = this.pointMap[thisObject._key];
			if(point == undefined) return;
			var data = point.data;
			if(point.highchartsPoint == null)
				point.highchartsPoint = this.chart.highcharts.get(data.id);
			point.highchartsPoint.setState();
		}
		this.hovered = [];
	};
	
	/**
	 * Hovers the point for the specified object.
	 * @param {Object} object - the object to hover the point for
	 */
	Series.prototype.hover = function(object) {
		for(var idx = 0; idx < object.length; idx++) {
			var thisObject = object[idx];
			var point = this.pointMap[thisObject._key];
			if(point == undefined) return;
			var data = point.data;
			if(point.highchartsPoint == null)
				point.highchartsPoint = this.chart.highcharts.get(data.id);
			point.highchartsPoint.setState('hover');
		}
		this.hovered = object;
	};
		
	/**
	 * Unmarks all marked points on the series.
	 */
	Series.prototype.unmark = function() {
		if(this.marked == null) return;
		for(var idx = 0; idx < this.marked.length; idx++) {
			var thisObject = this.marked[idx];
			var point = this.pointMap[thisObject._key];
			if(point == undefined) continue;
			var data = point.data;
			if(point.highchartsPoint == null)
				point.highchartsPoint = this.chart.highcharts.get(data.id);
			if(point.highchartsPoint != null)
				point.highchartsPoint.select(false);		
			point.marked = false;
			point.data.selected = false;
		}
		this.marked = [];
	};
	
	/**
	 * Marks the point for the specified object or array.
	 * @param {Object[]} object - the array of objects to mark points for
	 */
	Series.prototype.mark = function(object) {
		if(object == null) return;
		for(var idx = 0; idx < object.length; idx++) {
			var thisObject = object[idx];
			var point = this.pointMap[thisObject._key];
			if(point == undefined) continue;
			var data = point.data;
			if(point.highchartsPoint == null)
				point.highchartsPoint = this.chart.highcharts.get(data.id);
			if(point.highchartsPoint != null)
				point.highchartsPoint.select(true, true);
			point.marked = true;
			point.data.selected = true;
			this.marked.push(thisObject);;
		}
	};
	
	/**
	 * Applies a filter definition on a series.
	 * @param {Object} filterDef - filter definition
	 */
	Series.prototype.filter = function(filterDef) {
		var self = this;
		this.filterDef = filterDef;
		
		// Remove all points from the series, but don't discard them from the component
		for(var idx = 0; idx < Object.keys(this.pointMap).length; idx++) {
			var key = Object.keys(this.pointMap)[idx];
			var point = this.pointMap[key];
			var data = point.data;
			if(point.highchartsPoint == null)
				point.highchartsPoint = this.chart.highcharts.get(data.id);
			if(point.highchartsPoint != undefined) {
				data.color = point.highchartsPoint.color;
				point.highchartsPoint.select(false);		
				point.highchartsPoint.remove(data);
				point.highchartsPoint = null;
				point.added = false;
			}
		}

		// Add the points back in to the series
		for(var idx = 0; idx < Object.keys(this.pointMap).length; idx++) {
			var key = Object.keys(this.pointMap)[idx];
			var point = this.pointMap[key];
			var data = point.data;
			var include = DataFlow.Utilities.Filtering.filterItem(data, 
				this.filterDef.filter.filters, this.filterDef.filter.fields, 'object');
			if(include == true) {
				this.highchartsSeries.addPoint(data, false);
				point.added = true;
			}
		}

		setTimeout(function() {
			var marked = self.marked;
			self.unmark();
			self.mark(marked);
		}, 750);
	};
	
	
	
	
	/** 
	 * Highcharts configuration helper object.
	 */
	HighchartsConfiguration = {
		defaultConfiguration: {
				highcharts: {
					chart: {
						animation:		false
					},
					credits: {
						enabled:		false
					},
					exporting: {
						enabled: 		false
					},
					plotOptions: {
						series: {
							point: {
								events: {
								}
							}
						}
					}
				},
				options: {
					configurable:		false,
					copy: {
						enabled:		true
					},
					dataset: {
						icon:			true
						//modal:			undefined
					},
					filter: {	
						accept:			true,
						fields:			undefined,
						delay:			500,
						enabled:		true,
						passthrough:	true
					},
					format: {
						timestamp:		"yyyy-MM-dd HH:mm:ss"
					},
					hover: {
						accept:			true,
						enabled:		true,
						passthrough:	true					
					},
					marking: {
						accept:			true,
						enabled:		true,
						mode:			"single",
						passthrough:	true,
						unmark:			true
					},
					rotate3d: {
					},
					search: {
						fields:			undefined,
						delay:			500,
						enabled:		true,
						label:			"Search"
					}
				},
				
			},
		
		// Merges the source configuration object into the target configuration object
		//   Source values will overwrite target values.
		//   Should always merge a DEFAULT configuration source into a USER configuration target
		mergeConfiguration:		function(source, target) {
									for(var idx = 0; idx < Object.keys(source).length; idx++) {
										var key = Object.keys(source)[idx];
										var val = source[key];
										if(typeof val === 'object') {
											if(target[key] == undefined) 
												target[key] = val;
											else
												this.mergeConfiguration(val, target[key]);
										}
										else {
											if(target[key] == undefined) 
												target[key] = val;
										}
									}
									return target;
								},

		// Processes the specified configuration object and merges in default configuration
		//   for both base object and column objects
		processConfiguration:	function(config) {
									var configuration = this.mergeConfiguration(this.defaultConfiguration, config);
									return configuration;
								},
	};


	
	module.Chart 					= Chart;
	module.HighchartsConfiguration	= HighchartsConfiguration;
	
	return module;
})();
