/**
 * Represents a Leaflet map data sink.
 * @constructor
 * @param {Object} dataset - dataset that this sink will be attached to
 * @param {Object} configuration - configuration object defining the sink
 */
function LeafletLayerDatasink(id, configuration, dataset, map, overlay) {
	this.id = id;
	this.dataset = dataset;
	this.layer = new LiveLeafletMap.Layer(id, configuration, map, overlay, this);
}

/**
 * Event handler for dataset events. 
 * @param {String} eventType - string describing the event that occurred
 * @param {Object} object - data object for the event, or null
 */
LeafletLayerDatasink.prototype.trigger = function(eventType, object) {
	this.layer.trigger(eventType, object);
};


// Define a module to hold object defintions for the map objects
var LiveLeafletMap = (function() {
	var module = {};

	/**
	 * Creates a new Map object.
	 * @constructor
	 * @param {String} id - unique identifier
	 * @param {Object} configuration - the configuration for the map
	 * @param {Object} containerDiv - this node object that map will be appended
	 */
	function Map(id, configuration, containerDiv) {
		this.id					= id;
		this.configuration		= configuration;
		this.containerDiv 		= containerDiv;

		// Pause and snapshot flags
		this.paused				= false;
		this.snapshot			= false;

		// Leaflet node object, this is the leaflet map
		this.node				= null;

		// Color maps
		this.colorMaps			= {};

		// Layers
		this.layersArr			= [];

		// Control layers
		this.controlLayers		= {};
		
		// Check dependencies. This will throw Error if dependencies are not satisfied.
		this.dependencyCheck();
		
		// Color maps
		if(this.configuration.options.colormaps != undefined) {
			for(var idx = 0; idx < Object.keys(this.configuration.options.colormaps).length; idx++) {
				var thisColorMapName = Object.keys(this.configuration.options.colormaps)[idx];
				var thisColors = this.configuration.options.colormaps[thisColorMapName];
				this.colorMaps[thisColorMapName] = new DataFlow.ColorMap(thisColorMapName, thisColors);
			}
		}

		// On document ready, register UI handlers
		var self = this;
		setTimeout(function() {
			self.registerEventHandlers();
		}), 0;
		
		// Draw the initial empty map
		this.draw();
	};
	
	/**
	 * Performs a dependency check for rendering the map.
	 */
	Map.prototype.dependencyCheck = function() {
		var msg = "Required Leaflet libraries not loaded or not present";
		try {
			var p = L;
		}
		catch(err) {
			console.log(err);
			throw Error("Required Leaflet libraries not loaded or not present");
		}
	};
		
	/**
	 * Registers event handlers on UI components.
	 */
	Map.prototype.registerEventHandlers = function() {
		var self = this;
		
		this.node.on('mousedown', function(e) {
			if(e.originalEvent.ctrlKey == true) {
				self.node.dragging.disable();
				
				// Create a selection area div
				var appended = false;
				var selectionAreaDiv = $(document.createElement('div')).attr('class', 'selection-area');
				
				// Get initial start event
				var startEvent = e;
				var startX = e.containerPoint.x;
				var startY = e.containerPoint.y;

				// Function to update the size and position of the selection area
				var positionSelectionArea = function(endX, endY) {
					selectionAreaDiv.css({
						top: Math.min(startY, endY), 
						left: Math.min(startX, endX),
						height: Math.abs(startY - endY),
						width: Math.abs(startX - endX)
					});
				};
				
				self.node.on('mousemove', function(e) {
					var endX = e.containerPoint.x;
					var endY = e.containerPoint.y;
					if(appended == false && (Math.abs(startY - endY) > 10 || Math.abs(startX - endX) > 10)) {
						$(self.containerDiv).append(selectionAreaDiv);
						appended = true;
					}
					
					// Update the selection area
					positionSelectionArea(endX, endY);
				});
				
				self.node.on('mouseup', function(e) {
					var endEvent = e;
					
					// Instruct each layer to mark bounds
					for(var idx = 0; idx < self.layersArr.length; idx++)
						self.layersArr[idx].markBounds(startEvent.latlng, endEvent.latlng);
					
					// Remove the selection area
					selectionAreaDiv.remove();
					self.node.off('mouseup');
					self.node.off('mousemove');
					self.node.dragging.enable();
				});
			}
		});
	};
			
	/**
	 * Draws the UI components for the map.
	 */
	Map.prototype.draw = function() {
		var self = this;
		
		// Create the map
		this.node = L.map(this.containerDiv, this.configuration.leaflet.map);
		
		// Setup controls
		if(this.configuration.leaflet.control.attribution != undefined) {
			this.controlLayers.attributionControl = L.control.attribution(this.configuration.leaflet.control.attribution);
			this.controlLayers.attributionControl.addTo(this.node);
		}

		// Setup multiple marking control
		var multipleMarkingEnabled = false;
		for(var layerId in this.configuration.leaflet.layers) {
			var thisConfiguration = this.configuration.leaflet.layers[layerId];
			if(thisConfiguration.options.marking.enabled == true && thisConfiguration.options.marking.mode == 'multiple') {
				multipleMarkingEnabled = true;
				break;
			}
		}
		
		if(this.configuration.options.search.enabled == true) {
			L.Control.Search = L.Control.extend({
				onAdd: function(map) {
					var searchDiv = $(document.createElement('div')).attr('class', 'leaflet-control-search');
					
					var searchInputDiv = $(document.createElement('div')).attr('class', 'leaflet-control-search-input');
					searchDiv.append(searchInputDiv);
					searchInputDiv.append($(document.createElement('span')).attr('class', 'leaflet-control-search-layers'));
					var searchSelect = $(document.createElement('select'));
					searchInputDiv.append(searchSelect);
					var searchInput = $(document.createElement('input'));
					searchInputDiv.append(searchInput);
					var searchClear = $(document.createElement('div')).attr('class', 'leaflet-control-search-clear');
					searchInputDiv.append(searchClear);					
					var searchLink = $(document.createElement('a')).attr('class', 'leaflet-control-search-toggle').attr('href', '#');
					searchDiv.append(searchLink);
					
					var opened = false;
					searchDiv.on('dblclick', function(e) {
						return false;
					});
					searchDiv.on('mouseover', function(e) {
						self.node.dragging.disable();
					});
					searchDiv.on('mouseout', function(e) {
						self.node.dragging.enable();
					});
					searchLink.on('click', function(e) {
						if(opened == false) {
							searchSelect.empty();
							searchSelect.append($(document.createElement('option')).attr('value', null).html('All'));
							for(var idx = 0; idx < self.layersArr.length; idx++) {
								var layer = self.layersArr[idx];
								searchSelect.append($(document.createElement('option')).attr('value', self.layersArr[idx].name).html(self.layersArr[idx].name));
							}
							opened = true;
							searchInputDiv.addClass('visible');
							searchInput.focus();
						}
						else if(searchInput.val().length == 0) {
							opened = false;
							searchInputDiv.removeClass('visible');						
						}
						return false;
					});
					searchClear.on('click', function(e) {
						searchInput.val('');
						applySearch(searchSelect.val(), searchInput.val());
						return false;
					});
					
					var applySearch = function(layer, val) {
						for(var idx = 0; idx < self.layersArr.length; idx++) {
							if(val == '')
								self.layersArr[idx].search('');
							else if(layer == 'All' || layer == self.layersArr[idx].name) 
								self.layersArr[idx].search(val);
							else
								self.layersArr[idx].search('');
						}
					};
					
					searchSelect.on('change', function(e) {
						applySearch(searchSelect.val(), searchInput.val());
					});
					
					searchInput.on('input', DataFlow.Utilities.Debounce.debounce(function(e) {
						applySearch(searchSelect.val(), searchInput.val());
					}, self.configuration.options.search.delay));

					return searchDiv[0];
				},

				onRemove: function(map) {
					
				}
			});

			L.control.search = function(opts) {
				return new L.Control.Search(opts);
			}
			var thisConfiguration = this.configuration.leaflet.control.search == undefined ? {position: 'topright'} : this.configuration.leaflet.control.search;
			this.controlLayers.searchControl = L.control.search(thisConfiguration);
			this.controlLayers.searchControl.addTo(this.node);
		}
		
		if(this.configuration.leaflet.control.layers != undefined) {
			this.controlLayers.layers = L.control.layers(this.baseLayers, this.overlayLayers, this.configuration.leaflet.control.layers);
			this.controlLayers.layers.addTo(this.node);
		}
		
		if(this.configuration.leaflet.control.scale != undefined) {
			this.controlLayers.scale = L.control.scale(this.configuration.leaflet.control.scale);
			this.controlLayers.scale.addTo(this.node);
		}
		
		if(this.configuration.leaflet.control.zoom != undefined) {
			this.controlLayers.zoomControl = L.control.zoom(this.configuration.leaflet.control.zoom);
			this.controlLayers.zoomControl.addTo(this.node);
		}

		// Setup configuration control
		if(this.configuration.options.configurable == true) {
			L.Control.Configuration = L.Control.extend({
				onAdd: function(map) {
					var configurationDiv = L.DomUtil.create('div', 'leaflet-control-configuration');
					var configurationLink = L.DomUtil.create('a', 'leaflet-control-configuration-toggle', configurationDiv);
					configurationLink.href = '#';
					L.DomEvent.on(configurationLink, 'click', function(e) {
						self.displayConfiguration();
					});
					return configurationDiv;
				},

				onRemove: function(map) {
					
				}
			});

			L.control.configuration = function(opts) {
				return new L.Control.Configuration(opts);
			}
			var thisConfiguration = this.configuration.leaflet.control.configuration == undefined ? {position: 'topright'} : this.configuration.leaflet.control.configuration;
			this.controlLayers.configurationControl = L.control.configuration(thisConfiguration);
			this.controlLayers.configurationControl.addTo(this.node);
		}
		
		// Setup dataset control
		if(this.configuration.options.dataset.icon == true) {
			L.Control.Dataset = L.Control.extend({
				onAdd: function(map) {
					var live = true;
					var datasetDiv = L.DomUtil.create('div', 'leaflet-control-dataset');
					var datasetLink = L.DomUtil.create('a', 'leaflet-control-dataset-toggle ' + (live == true ? 'live' : 'notlive'), datasetDiv);
					datasetLink.href = '#';
					L.DomEvent.on(datasetLink, 'click', function(e) {
						self.displayDataset();
					});
					return datasetDiv;
				},

				onRemove: function(map) {
					
				}
			});

			L.control.dataset = function(opts) {
				return new L.Control.Dataset(opts);
			}
			var thisConfiguration = this.configuration.leaflet.control.dataset == undefined ? {position: 'topright'} : this.configuration.leaflet.control.dataset;
			this.controlLayers.datasetControl = L.control.dataset(thisConfiguration);
			this.controlLayers.datasetControl.addTo(this.node);
		}
		
		// Create and append tile layer
		var tileLayer = L.tileLayer(this.configuration.leaflet.tileLayer.provider, this.configuration.leaflet.tileLayer.options);
		tileLayer.addTo(this.node);
		if(this.controlLayers.layers != null) {
			this.controlLayers.layers.addBaseLayer(tileLayer, this.configuration.leaflet.tileLayer.name);
		}
	};
	
	/**
	 * Pauses updates on a map.
	 */
	Map.prototype.pause = function() {
		this.paused = true;
	};
	
	/**
	 * Resumes a paused map.
	 */
	Map.prototype.resume = function() {
		if(this.paused == false) return;
		this.paused = false;
		for(var idx = 0; idx < this.layersArr.length; idx++)
			this.layersArr[idx].update();
	};
	
	/**
	 * Resizes a map.
	 */
	Map.prototype.resize = function() {
		this.node.invalidateSize();		
	};
			
	/**
	 * Display the map configuration modal.
	 */
	Map.prototype.displayConfiguration = function() {
		if(this.configuration.options.configurable != true) return;
		console.log('leaflet display config');
		// TODO display configuration modal
	};

	/**
	 * Displays the dataset configuration modal.
	 */
	Map.prototype.displayDataset = function() {
		if(this.configuration.options.dataset.modal == undefined) return;
		console.log('leaflet display dataset');
		// TODO display dataset modal
	};
	
	/** 
	 * Returns a color for the given value from the specified color map.
	 * @param {Object} value - the value to lookup, can be anything but typically a String
	 * @param (Object} name - the name of the color map
	 * @return {String} the color
	 */
	Map.prototype.colormap = function(value, name) {
		var colorMap = this.colorMaps[name];
		if(colorMap == undefined) return;
		return colorMap.getColor(value);
	};
	
	
	
	
	/**
	 * Creates a new Layer object.
	 * @constructor
	 * @param {String} id - unique identifier
	 * @param {Object} configuration - the configuration for the layer
	 * @param {Object} map - the map that this layer will be added to
	 * @param (Function} overlay - function to call on start and end of the overlay
	 * @param {Object} sink - datasink that will feed this layer
	 */
	function Layer(id, configuration, map, overlay, sink) {
		this.id				= id;
		this.name			= configuration.name;
		this.configuration	= configuration;
		this.map			= map;
		this.overlay		= overlay;
		this.sink			= sink;
		
		// Pause and snapshot flags
		this.paused			= false;
		this.snapshot		= false;

		// Schema for the dataset
		this.schema			= null;		
		
		// Leaflet layer object
		this.node			= null;
		
		// Vars to hold the objects being hovered and marked
		this.hovered		= [];		
		this.marked			= [];

		// Filter definition
		this.filterDef		= {
			search: {
				expression:	undefined,
				fields:		this.configuration.options.search.fields
			},
			filter:	{
				filters:	[],
				fields:		this.configuration.options.filter.fields
			}
		};

		// Color map
		this.colormap 		= new DataFlow.ColorMap(null, this.configuration.colormap);

		// Component array and map, AbstractComponent objects
		this.componentArr	= [];
		this.componentMap	= {};
		
		// Add this layer to the map
		this.map.layersArr.push(this);
				
		// Microbatch is used to batch up updates to prevent too many map redraws
		this.updateQueue = {};
		this.microbatch = {
			rtime:		new Date(1, 1, 2000, 12,00,00),
			timeout:	false,
			delta:		500 // msec between updates
		}

		// Draw initial empty layer
		this.draw();
	};
	
	/**
	 * Event handler for all objects delegated from the datasink.
	 * @param {String} eventType - the event type
	 * @param {Object} object - optional object
	 */
	Layer.prototype.trigger = function(eventType, object) {
		var self = this;
		
		if(eventType == 'schemaChange') {
			this.schema = object;
		}
		else if(eventType == 'snapshotStart') {
			this.snapshot = true;
			this.overlay(true);
		}
		else if(eventType == 'snapshotEnd') {
			setTimeout(function() {
				self.snapshot = false;
				self.overlay(false);
			},500);
		}
		else if(eventType == 'clear') {
			this.clear();
		}
		else if(eventType == 'insert') {
			this.addObject(object);
		}
		else if(eventType == 'update') {
			this.updateObject(object);
		}
		else if(eventType == 'upsert') {
		}
		else if(eventType == 'delete') {
			this.removeObject(object);
		}
		else if(eventType == 'hover') {
			if(this.configuration.options.hover.accept != false)
				this.hover(object);
		}
		else if(eventType == 'marked') {
			if(this.configuration.options.marking.accept != false)
				this.mark(object);
		}
		else if(eventType == 'filter') {
			if(this.configuration.options.filter.accept != false)
				this.filter(object);
		}
	};
	
	/**
	 * Add a new object bound to the specified object on the layer group.
	 * @param {Object} object - the object to add the bound object
	 */
	Layer.prototype.addObject = function(object) {
		var component = this.createComponent(object);
		this.componentMap[component._key] = component;
		this.componentArr.push(component);
		this.updateQueue[component._key] = component;
		this.update();
	};

	/**
	 * Updates the component bound to the specified object on the layer group.
	 * @param {Object} object - the object to update the bound object
	 */
	Layer.prototype.updateObject = function(object) {
		var component = this.componentMap[object._key];
		if(component == undefined) return;		
		component.object = object;
		this.updateQueue[component._key] = component;
		this.update();
	};
	
	/**
	 * Removes the component bound to the specified object from the layer group.
	 * @param {Object} object - the object to remove the bound object
	 */
	Layer.prototype.removeObject = function(object) {
		var component = this.componentMap[object._key];
		component.remove();
		delete this.componentMap[object._key];
		this.componentArr.splice(this.componentArr.indexOf(component), 1);
		this.update();
	};
					
	/**
	 * Clears all objects and all components from the layer group.
	 */
	Layer.prototype.clear = function() {
		for(var idx = 0; idx < this.componentArr.length; idx++)
			this.componentArr[idx].remove();
		this.componentMap	= {};
		this.componentArr	= [];
		this.update();
	};
	
	/**
	 * Hovers a single component bound to the specified object.
	 * Will also optionally notify the dataset of a hover event, if enabled.
	 * @param {Object} object - object that the associated component will be hovered
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Layer.prototype.hover = function(object, notify) {
		if(this.configuration.options.hover.enabled != false) {

			// Unhover all existing rows
			for(var idx = this.hovered.length - 1; idx >= 0; idx--) {
				var thisObject = this.hovered[idx];
				var component = this.componentMap[thisObject._key];
				if(component != undefined) {
					component.hovered = false;
					component.update();
				};
				this.hovered.splice(this.hovered.indexOf(thisObject), 1);
			}	

			// If the object is not null, process the hovering
			if(object != null) {
				if(Array.isArray(object)) {
					for(var idx = 0; idx < object.length; idx++) {
						var thisObject = object[idx];
						var component = this.componentMap[thisObject._key];
						if(component != undefined) {
							component.hovered = true;
							component.update();
						}
						this.hovered.push(thisObject);
					}
				}
				else {
					var component = this.componentMap[object._key];
					if(component != undefined) {
						component.hovered = true;
						component.update();
					}
					this.hovered.push(object);
				}
			}
			
			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.hover.passthrough == true) {
				this.sink.dataset.trigger('hover', this.sink, object);
			}
		}
	};
		
	/**
	 * Marks a single component or a set of components bound to the specified object or array of objects.
	 * Will also optionally notify the dataset of a marked event, if enabled.
	 * @param {Object|Array} object - if a single data object then the associated component will be marked
	 *                                if an array of objects then the associated components will be marked
	 *                                  for single marking components that receive a list of objects then only
	 *                                  the first component will be marked
	 *                                if null then all marked components will be unmarked
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Layer.prototype.mark = function(object, notify) {
		if(this.configuration.options.marking.enabled != false) {
			var mode = this.configuration.options.marking.mode;

			// Unmark all existing components
			for(var idx = this.marked.length - 1; idx >= 0; idx--) {
				var thisObject = this.marked[idx];				
				var component = this.componentMap[thisObject._key];
				if(component != undefined) {
					component.marked = false;
					component.update();
				}
				this.marked.splice(this.marked.indexOf(thisObject), 1);
			}	
			
			// If the object is not null, process the marking
			if(object != null) {
				if(Array.isArray(object)) {
					if(this.configuration.options.marking.mode == 'multiple') {
						for(var idx = 0; idx < object.length; idx++) {
							var thisObject = object[idx];
							var component = this.componentMap[thisObject._key];
							component.marked = true;
							component.update();
							this.marked.push(thisObject);
						}
					}
					// If we receive a list of objects but can only mark one, mark only the first one
					else if(this.configuration.options.marking.mode == 'single' && object.length > 0) {
						var component = this.componentMap[object[0]._key];
						component.marked = true;
						component.update();
						this.marked.push(object[0]);
					}
				}
				else {
					var component = this.componentMap[object._key];
					component.marked = true;
					component.update();
					this.marked.push(object);
				}
			}
			
			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.marking.passthrough == true) {
				this.sink.dataset.trigger('marked', this.sink, object);
			}			

		}
	};
		
	/**
	 * Marks all components within the specified lat and lon corner positions.
	 * @param {Double[]} latlng1 - first corner of the bounding box
	 * @param {Double[]} latlng2 - second cornder of the bounding box
	 */
	Layer.prototype.markBounds = function(latlng1, latlng2) {
		if(this.configuration.options.marking.enabled == false) return;
		
		// Skip if this layer is currently not visible
		if(this.node._map == null) return;
		
		// Loop over all the components and test the position of the node against the 
		//   bounding latlngs
		var bounds = L.latLngBounds(latlng1, latlng2);
		var selected = [];
		for(var idx = 0; idx < this.componentArr.length; idx++) {
			var latlng = this.componentArr[idx].node.getLatLng();
			if(bounds.contains(latlng) == true)
				selected.push(this.componentArr[idx]);
		}
		
		// If nothing is selected, then do not change the marking for the layer
		if(selected.length == 0) return;
		var mode = this.configuration.options.marking.mode;
			
		// Single mode
		if(mode == 'single') {
			var component = selected[0];
			if(component.marked == true)
				this.mark(null, true);
			else
				this.mark(component.object, true);
		}
		// Multiple mode
		else if(mode == 'multiple') {
			var objects = [];
			for(var idx = 0; idx < selected.length; idx++)
				objects.push(selected[idx].object);
			this.mark(objects, true);
		}
		
	};
	
	/**
	 * Applies a filter and updates the map. Will also optionally notify the
	 * dataset of a filter event, if enabled.
	 * @param {Array} filter - array of filter objects
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Layer.prototype.filter = function(filters, notify) {
		if(this.configuration.options.filter.enabled != false) {
			// This clones the array so we don't get duplicates
			this.filterDef.filter.filters = filters.slice(0);
			this.filterDef.filter.fields = this.configuration.options.filter.fields;
			
			for(var idx = 0; idx < this.componentArr.length; idx++) {
				var include = DataFlow.Utilities.Filtering.filterItem(this.componentArr[idx], 
					this.filterDef.filter.filters, this.filterDef.filter.fields, 'object');
				var filtered = !include;
				this.componentArr[idx].filter(filtered);
			}
			
			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.filter.passthrough == true) {
				this.sink.dataset.trigger('filter', this.sink, this.filterDef.filter.filters);
			}
			
			this.update();
		}
	};
	
	/**
	 * Applies a search value against the layer. This will not be passed through
	 * to the dataset and is local only to the current layer.
	 * @param {String} value - the search string
	 */
	Layer.prototype.search = function(value) {
		if(value == undefined || value.length == 0) {
			delete this.filterDef.search.expression;
			for(var idx = 0; idx < this.componentArr.length; idx++) {
				this.componentArr[idx].search(null);
			}
		}
		else {
			this.filterDef.search.expression = value;
			for(var idx = 0; idx < this.componentArr.length; idx++) {
				var include = DataFlow.Utilities.Filtering.searchItem(this.componentArr[idx], 
					this.filterDef.search.expression, this.filterDef.search.fields, 'object');
				this.componentArr[idx].search(include);
			}
		}
	};
	
	
	/**
	 * Draws the UI components for the layer.
	 */
	Layer.prototype.draw = function() {
		// Create the layer and append to the map
		this.node = L.layerGroup();
		this.node.addTo(this.map.node);
		if(this.map.controlLayers.layers != null) {
			this.map.controlLayers.layers.addOverlay(this.node, this.name);
		}
	};
	
	/**
	 * Updates the contents of the layer. Normally this will happen in a microbatch
	 * to prevent too many updates from occurring. However it is possible to trigger
	 * an immediate update by setting the flag.
	 * @param {Boolean} immediate - trigger an immediate update
	 */
	Layer.prototype.update = function(immediate) {
		var self = this;

		if(immediate == true) {
			run();
			return;
		}

		// Microbatch handler to prevent too many updates
		if(self.microbatch.timeout === false) {
			self.microbatch.timeout = true;
			setTimeout(eventEnd, self.microbatch.delta);
		}					
		function eventEnd() {
			var d = new Date() - self.microbatch.rtime;
			if (new Date() - self.microbatch.rtime < self.microbatch.delta)
				setTimeout(eventEnd, self.microbatch.delta);
			else {
				self.microbatch.rtime = new Date();
				self.microbatch.timeout = false;
				run();
			}               
		};
		
		// Run the update
		function run() {
			if(self.paused == true || self.map.paused == true) return;
			for(var key in self.updateQueue) {
				var component = self.componentMap[key];
				if(component == undefined) continue;
				component.update();
			}
			self.updateQueue = {};
		};
	};
	
	/**
	 * Pauses updates on a layer.
	 */
	Layer.prototype.pause = function() {
		this.paused = true;
	};
	
	/**
	 * Resumes a paused layer.
	 */
	Layer.prototype.resume = function() {
		if(this.paused == false) return;
		this.paused = false;
		for(var idx = 0; idx < this.layersArr.length; idx++)
			this.layersArr[idx].update();
	};
	
	/** 
	 * Returns a color for the given value from the color map for this layer.
	 * @param {Object} value - the value to lookup, can be anything but typically a String
	 * @return {String} the color
	 */
	Layer.prototype.colormap = function(value) {
		var color = this.colorValMap[value];
		if(color == undefined) {
			color = this.colorscale[this.colorIdx++];
			if(this.colorIdx == this.colorscale.length) {
				this.colorIdx = 0;
			}
			this.colorValMap[value] = color; 
		}
		return color;
	};
	
	/**
	 * Creates a UI component for the given object based on configuration for this layer.
	 * @param {Object} object - the object to create component for
	 */
	Layer.prototype.createComponent = function(object) {
		if(this.configuration.marker != undefined) {
			return new Marker(object, this.configuration.marker, this);
		}
		else if(this.configuration.polyline != undefined) {
			return new Polyline(object, this.configuration.polyline, this);
		}
		else if(this.configuration.polygon != undefined) {
			return new Polygon(object, this.configuration.polygon, this);
		}
		else if(this.configuration.rectangle != undefined) {
			return new Rectangle(object, this.configuration.rectangle, this);
		}
		else if(this.configuration.circle != undefined) {
			return new Circle(object, this.configuration.circle, this);
		}
		else if(this.configuration.circleMarker != undefined) {
			return new CircleMarker(object, this.configuration.circleMarker, this);
		}
		else
			throw 'Unknown layer component type';
	}
	
	
	
	/** 
	 * Creates a new component on the specified layer.
	 * @param {Object} object - the object to bind to the component
	 * @param {Object} configuration - component configuration
	 * @param {Object} layer - layer to add the component on
	 */
	function AbstractComponent(object, configuration, layer) {
		this._key			= object._key;
		this.object 		= object;
		this.configuration	= configuration;
		this.layer			= layer;

		this.node			= null;
		this.addedToLayer	= false;
		
		this.tooltip		= null;
		this.popup			= null;
		
		this.hovered		= false;
		this.marked			= false;
		this.filtered		= false;
		this.searched		= null;
		
		// Create the property index
		var propertyIndexResult = DataFlow.Utilities.PropertyMap.index(this.configuration, ['property', 'eval']);
		this.propertyIndex = propertyIndexResult.propertyIndex;
		this.properties = propertyIndexResult.properties;
		
		// Create a property index for tooltip
		if(this.layer.configuration.tooltip != undefined) {
			var tooltipPropertyIndexResult = DataFlow.Utilities.PropertyMap.index(this.layer.configuration.tooltip, ['property', 'eval']);
			this.tooltipPropertyIndex = tooltipPropertyIndexResult.propertyIndex;
			this.tooltipProperties = tooltipPropertyIndexResult.properties;
		}
		
		// Create a property index for popup
		if(this.layer.configuration.popup != undefined) {
			var popupPropertyIndexResult = DataFlow.Utilities.PropertyMap.index(this.layer.configuration.popup, ['property', 'eval']);
			this.popupPropertyIndex = popupPropertyIndexResult.propertyIndex;
			this.popupProperties = popupPropertyIndexResult.properties;

			// If there is a contentUrl load that now synchronously
			var self = this;
			if(this.layer.configuration.popup.contentUrl != undefined) {
				$.ajax({
					type: 'GET',
					url: this.layer.configuration.popup.contentUrl,
					dataType: 'html',
					cache: false,
					aync: false,
					success: function(data, textStatus, xhr) {
						self.popupPropertyIndex.arr.push({'prop': 'content', 'type': 'eval'});
						var thisData = data;
						if(!thisData.startsWith('`')) thisData = '`' + thisData;
						if(!thisData.endsWith('`')) thisData = thisData + '`';
						self.popupPropertyIndex.map.content = thisData;
					},
					error: function(xhr, textStatus, errorThrown) {
						console.error("Layer content url not found or invalid - " + self.configuration.contentUrl);
					}
				});	
			}
		}
	};

	/**
	 * Draws a new component using base configuration. 
	 */
	AbstractComponent.prototype.draw = function() {
		var self = this;
		
		// Add popup if enabled
		if(this.layer.configuration.popup != undefined) {
			this.node.bindPopup('', this.popupProperties);
		}
			
		// Add tooltip if enabled
		if(this.layer.configuration.tooltip != undefined) {
			this.node.bindTooltip('', this.tooltipProperties);
		}
		
		// Event handler for mouseover
		L.DomEvent.on(this.node, 'mouseover', function(e) {
			self.layer.hover(self.object, true);
		});		

		// Event handler for mouseout
		L.DomEvent.on(this.node, 'mouseout', function(e) {
			if(self.layer.configuration.options.hover.enabled == false) return;
			self.layer.hover(null, true);
		});		

		// Event handler for 
		L.DomEvent.on(this.node, 'click', function(e) {
			if(self.layer.configuration.options.marking.enabled == false) return;
			var ctrlKey = e.originalEvent.ctrlKey;					
			var mode = self.layer.configuration.options.marking.mode;
			var isMarked = self.marked == true;
			
			// If unmarking is disabled and this component is marked, then just call mark again
			//   and take no further action
			if(self.configuration.options.marking.unmark == false && isMarked == true) {
				self.layer.mark(self.object, true);
				return;
			}
			
			// Single mode
			if(mode == 'single') {
				if(isMarked == true)
					self.layer.mark(null, true);
				else
					self.layer.mark(self.object, true);
			}
			// Multiple mode
			else if(mode == 'multiple') {
				if(ctrlKey == true) {
					var objects = [];
					for(var idx = 0; idx < self.layer.marked.length; idx++)
						objects.push(self.layer.marked[idx].object);
					if(isMarked == false)
						objects.push(self.object);
					else if(isMarked == true)
						objects.splice(objects.indexOf(self.object), 1);
					self.layer.mark(objects, true);
				}
				else {
					if(isMarked == true)
						self.layer.mark(null, true);
					else
						self.layer.mark(self.object, true);
				}
			}
		});		
	};
	
	/**
	 * Update the component.
	 */
	AbstractComponent.prototype.update = function() {
		// Update the properties for the component
		DataFlow.Utilities.PropertyMap.evaluate(this.propertyIndex, this.getContext(), this.object, this.properties);
		
		// Update popup content
		if(this.layer.configuration.popup != undefined) {
			DataFlow.Utilities.PropertyMap.evaluate(this.popupPropertyIndex, this.getContext(), this.object, this.popupProperties);
			var contentDiv = Utilities.Template.process(this.popupProperties.content);
			this.node.setPopupContent(contentDiv[0]);
		}
		
		// Update tooltip content
		if(this.layer.configuration.tooltip != undefined) {
			DataFlow.Utilities.PropertyMap.evaluate(this.tooltipPropertyIndex, this.getContext(), this.object, this.tooltipProperties);
			this.node.setTooltipContent(this.tooltipProperties.text);
		}
		
		// Add to layer if not already done so
		if(this.addedToLayer == false) {
			this.node.addTo(this.layer.node);
			this.addedToLayer = true;
			
			// Check initial filters and searches
			var filterInclude = DataFlow.Utilities.Filtering.filterItem(this, 
				this.layer.filterDef.filter.filters, this.layer.filterDef.filter.fields, 'object');
			var filtered = !filterInclude;
			this.filter(filtered);
			
			var searchInclude = DataFlow.Utilities.Filtering.searchItem(this, 
				this.layer.filterDef.search.expression, this.layer.filterDef.search.fields, 'object');
			this.search(searchInclude);
		}
	};
	
	/**
	 * Removes the component from the layer and invalidates the
	 * node for garbage collection.
	 */
	AbstractComponent.prototype.remove = function() {
		if(this.node == null) return;
		L.DomEvent.off(this.node);
		this.layer.node.removeLayer(this.node);
		this.node = null;
	};
	
	/**
	 * Returns a context object for this component.
	 * @return {Object} the context object
	 */
	AbstractComponent.prototype.getContext = function() {
		var self = this;
		function colormap(value, name) {
			if(name == undefined)
				return self.layer.colormap.getColor(value);
			else {
				var colormap = self.layer.map.colorMaps[name];
				if(colormap == undefined) return;
				return colormap.getColor(value);
			}
		};

		// Function to format the specified value, which can be timestamp or number
		function format(val, format) {
			return DataFlow.Utilities.Format.format(val, format);
		};
		
		var context = {
			object:		this.object,
			component:	this,
			layer:		this.layer,
			map:		this.layer.map,
			colormap:	colormap, 
			format:		format
		};		
		
		return context;
	}
	
	/**
	 * Sets the filtered flag for the component and adds or removes from the layer.
	 * @param {Boolean} filtered - true if filtered, otherwise false
	 */
	AbstractComponent.prototype.filter = function(filtered) {
		if(this.filtered == true && filtered == false) {
			this.filtered = false;
			this.node.addTo(this.layer.node);
		}
		else if(this.filtered == false && filtered == true) {
			this.filtered = true;
			this.layer.node.removeLayer(this.node);
		}
	}
	
	/**
	 * Sets the search flag for the component and adds or removes from the layer.
	 * @param {Boolean} include - true if search match, otherwise false
	 */
	AbstractComponent.prototype.search = function(include) {
		if(this.searched == null && include == true) {
			this.searched = true;
		}
		if(this.searched == null && include == false) {
			this.searched = false;
			this.layer.node.removeLayer(this.node);
		}
		
		if(this.searched == true && include == null) {
			this.searched = null;
		}
		if(this.searched == true && include == false) {
			this.searched = false;
			this.layer.node.removeLayer(this.node);
		}
		
		if(this.searched == false && include == null) {
			this.searched = null;
			this.node.addTo(this.layer.node);
		}
		if(this.searched == false && include == true) {
			this.searched = true;
			this.node.addTo(this.layer.node);
		}
	}
	
	
	
	
	/** 
	 * Creates a new marker on the specified layer.
	 * @param {Object} object - the object to bind to the marker
	 * @param {Object} configuration - marker configuration
	 * @param {Object} layer - layer to add the marker on
	 */
	function Marker(object, configuration, layer) {
		AbstractComponent.call(this, object, configuration, layer);
		
		// Create a property index for icon or divIcon, prioritize icon first
		if(this.configuration.icon != undefined) {
			var iconPropertyIndexResult = DataFlow.Utilities.PropertyMap.index(this.configuration.icon, ['property', 'eval']);
			this.iconPropertyIndex = iconPropertyIndexResult.propertyIndex;
			this.iconProperties = iconPropertyIndexResult.properties;
		}		
		else if(this.configuration.divIcon != undefined) {
			var divIconPropertyIndexResult = DataFlow.Utilities.PropertyMap.index(this.configuration.divIcon, ['property', 'eval']);
			this.divIconPropertyIndex = divIconPropertyIndexResult.propertyIndex;
			this.divIconProperties = divIconPropertyIndexResult.properties;
			
			// If there is an htmlUrl load that now synchronously
			var self = this;
			if(this.configuration.divIcon.htmlUrl != undefined) {
				$.ajax({
					type: 'GET',
					url: this.configuration.divIcon.htmlUrl,
					dataType: 'html',
					cache: false,
					aync: false,
					success: function(data, textStatus, xhr) {
						self.divIconPropertyIndex.arr.push({'prop': 'html', 'type': 'eval'});
						var thisData = data;
						if(!thisData.startsWith('`')) thisData = '`' + thisData;
						if(!thisData.endsWith('`')) thisData = thisData + '`';
						self.divIconPropertyIndex.map.html = thisData;
					},
					error: function(xhr, textStatus, errorThrown) {
						console.error("Layer html url not found or invalid - " + self.configuration.divIcon.htmlUrl);
					}
				});	
			}
			
		}		
		
		this.draw();
	}
	
	Marker.prototype = Object.create(AbstractComponent.prototype);
	Marker.prototype.constructor = Marker;
	
	/**
	 * Draws a new marker using base configuration. 
	 */
	Marker.prototype.draw = function() {
		this.node = L.marker([0, 0], this.properties);
		AbstractComponent.prototype.draw.call(this);
		// No update here, will be triggered by layer.update()
	};
	
	/**
	 * Update the marker properties if the node exists. First it
	 * will evaluate properties on the configuration based on the component
	 * context. Then the marker position and icon will be updated
	 * based on the configuration. Finally, if the node has not yet been added
	 * to the containing layer, it will be added now.
	 */
	Marker.prototype.update = function() {
		if(this.node == null) return;

		// Update icon or divIcon
		if(this.configuration.icon != undefined) {
			DataFlow.Utilities.PropertyMap.evaluate(this.iconPropertyIndex, this.getContext(), this.object, this.iconProperties);
			var icon = L.icon(this.iconProperties);
			this.node.setIcon(icon);
		}
		else if(this.configuration.divIcon != undefined) {
			DataFlow.Utilities.PropertyMap.evaluate(this.divIconPropertyIndex, this.getContext(), this.object, this.divIconProperties);
			this.divIconProperties.html = $(Utilities.Template.process(this.divIconProperties.html)).html();
			var icon = L.divIcon(this.divIconProperties);
			this.node.setIcon(icon);
		}
		
		// Update lat/lng
		this.node.setLatLng([this.object[this.configuration.property_lat], this.object[this.configuration.property_lon]]);

		AbstractComponent.prototype.update.call(this);
	};
	
	
	
	
	/**
	 * Creates a new polyline.
	 * @param {Object} object - the object to bind to the polyline
	 * @param {Object} configuration - the configuration for the polyline
	 * @param {Object} layer - the layer to add the polyline on
	 */
	function Polyline(object, configuration, layer) {
		AbstractComponent.call(this, object, configuration, layer);		
		this.draw();
	};
	
	Polyline.prototype = Object.create(AbstractComponent.prototype);
	Polyline.prototype.constructor = Polyline;

	/**
	 * Draws a new polyline using base configuration. 
	 * Polylines are always drawn from the context of a single object.
	 * They may be drawn using two modes:
	 *   1. Start and end component based on properties of the object
	 *   2. Mulitple components based on a property containing an array of
	 *        other objects with lat and lon 
	 */
	Polyline.prototype.draw = function() {
		var coordinates = [[0, 0]];
		this.node = L.polyline(coordinates, this.configuration);
		AbstractComponent.prototype.draw.call(this);
		// No update here, will be triggered by layer.update()
	};
	
	/**
	 * Update the polyline properties if the node exists. First it
	 * will evaluate properties on the configuration based on the component
	 * context. Then the polyline position and style will be updated
	 * based on the configuration. Finally, if the node has not yet been added
	 * to the containing layer, it will be added now.
	 */
	Polyline.prototype.update = function() {
		if(this.node == null) return;
		AbstractComponent.prototype.update.call(this);

		// Update lat/lon and style
		var coordinates = null;
		if(this.properties.lat1 != undefined && this.properties.lon1 != undefined && 
				this.properties.lat2 != undefined && this.properties.lon2 != undefined) {
			coordinates = [[this.properties.lat1, this.properties.lon1], 
				[this.properties.lat2, this.properties.lon2]];
		}
		else if(this.properties.points != undefined) {
			coordinates = [];
			var self = this;
			var sorted = this.properties.points.sort(function(a, b) {
				var dir = self.properties.ordering.direction == 'desc' ? -1 : 1;
				return dir * (a[self.properties.ordering.field] > b[self.properties.ordering.field]) ? 1 : 
					((b[self.properties.ordering.field] > a[self.properties.ordering.field]) ? -1 : 0);
			});
			if(sorted != undefined) {
				for(var idx = 0; idx < sorted.length; idx++) {
					coordinates.push([sorted[idx].lat, sorted[idx].lon]);
				}
			}
		}
		this.node.setLatLngs(coordinates);
		this.node.setStyle(this.properties);

		// Call generic component update
		AbstractComponent.prototype.update.call(this);
	};



	
	/**
	 * Creates a new polygon.
	 * @param {Object} object - the object to bind to the polygon
	 * @param {Object} configuration - the configuration for the polygon
	 * @param {Object} layer - the layer to add the polygon on
	 */
	function Polygon(object, configuration, layer) {
		AbstractComponent.call(this, object, configuration, layer);		
		this.draw();
	};
	
	Polygon.prototype = Object.create(AbstractComponent.prototype);
	Polygon.prototype.constructor = Polygon;

	/**
	 * Draws a new polygon using base configuration. 
	 * Polygones are always drawn from the context of a single object.
	 * There must be a property containing multiple points with lat and lon.
	 */
	Polygon.prototype.draw = function() {
		var coordinates = [[0, 0]];
		this.node = L.polygon(coordinates, this.configuration);
		AbstractComponent.prototype.draw.call(this);
		// No update here, will be triggered by layer.update()
	};
	
	/**
	 * Update the polygon properties if the node exists. First it
	 * will evaluate properties on the configuration based on the component
	 * context. Then the polygon position and style will be updated
	 * based on the configuration. Finally, if the node has not yet been added
	 * to the containing layer, it will be added now.
	 */
	Polygon.prototype.update = function() {
		if(this.node == null) return;
		AbstractComponent.prototype.update.call(this);

		// Update lat/lon and style
		var coordinates = [];
		var self = this;
		var sorted = this.properties.points.sort(function(a, b) {
			var dir = self.properties.ordering.direction == 'desc' ? -1 : 1;
			return dir * (a[self.properties.ordering.field] > b[self.properties.ordering.field]) ? 1 : 
				((b[self.properties.ordering.field] > a[self.properties.ordering.field]) ? -1 : 0);
		});
		if(sorted != undefined) {
			for(var idx = 0; idx < sorted.length; idx++) {
				coordinates.push([sorted[idx].lat, sorted[idx].lon]);
			}
		}
		this.node.setLatLngs(coordinates);
		this.node.setStyle(this.properties);

		// Call generic component update
		AbstractComponent.prototype.update.call(this);
	};

	
	
	
	/**
	 * Creates a new rectangle.
	 * @param {Object} object - the object to bind to the rectangle
	 * @param {Object} configuration - the configuration for the rectangle
	 * @param {Object} layer - the layer to add the rectangle on
	 */
	function Rectangle(object, configuration, layer) {
		AbstractComponent.call(this, object, configuration, layer);		
		this.draw();
	};
	
	Rectangle.prototype = Object.create(AbstractComponent.prototype);
	Rectangle.prototype.constructor = Rectangle;

	/**
	 * Draws a new rectangle using base configuration. 
	 */
	Rectangle.prototype.draw = function() {
		var bounds = [[0, 0], [0, 0]];
		this.node = L.rectangle(bounds, this.configuration);
		AbstractComponent.prototype.draw.call(this);
		// No update here, will be triggered by layer.update()
	};
	
	/**
	 * Update the rectangle properties if the node exists. First it
	 * will evaluate properties on the configuration based on the component
	 * context. Then the rectangle position and style will be updated
	 * based on the configuration. Finally, if the node has not yet been added
	 * to the containing layer, it will be added now.
	 */
	Rectangle.prototype.update = function() {
		if(this.node == null) return;
		AbstractComponent.prototype.update.call(this);

		// Update bounds and style
		var bounds = [[this.properties.lat1, this.properties.lon1], 
			[this.properties.lat2, this.properties.lon2]];
		this.node.setBounds(bounds);
		this.node.setStyle(this.properties);

		// Call generic component update
		AbstractComponent.prototype.update.call(this);
	};

	
	
	
	/**
	 * Creates a new circle.
	 * @param {Object} object - the object to bind to the circle
	 * @param {Object} configuration - the configuration for the circle
	 * @param {Object} layer - the layer to add the circle on
	 */
	function Circle(object, configuration, layer) {
		AbstractComponent.call(this, object, configuration, layer);		
		this.draw();
	};
	
	Circle.prototype = Object.create(AbstractComponent.prototype);
	Circle.prototype.constructor = Circle;

	/**
	 * Draws a new circle using base configuration. 
	 */
	Circle.prototype.draw = function() {
		this.node = L.circle([0,0], this.configuration);
		AbstractComponent.prototype.draw.call(this);
		// No update here, will be triggered by layer.update()
	};
	
	/**
	 * Update the circle properties if the node exists. First it
	 * will evaluate properties on the configuration based on the component
	 * context. Then the circle position, radius, and style will be updated
	 * based on the configuration. Finally, if the node has not yet been added
	 * to the containing layer, it will be added now.
	 */
	Circle.prototype.update = function() {
		if(this.node == null) return;
		AbstractComponent.prototype.update.call(this);

		// Update lat/lon, radius, and style
		this.node.setLatLng([this.properties.lat, this.properties.lon]);
		this.node.setRadius(this.properties.radius);
		this.node.setStyle(this.properties);
		
		// Call generic component update
		AbstractComponent.prototype.update.call(this);
	};
	
	

	
	/**
	 * Creates a new circle marker.
	 * @param {Object} object - the object to bind to the marker
	 * @param {Object} configuration - the configuration for the marker
	 * @param {Object} layer - the layer to add the marker on
	 */
	function CircleMarker(object, configuration, layer) {
		AbstractComponent.call(this, object, configuration, layer);		
		this.draw();
	};
	
	CircleMarker.prototype = Object.create(AbstractComponent.prototype);
	CircleMarker.prototype.constructor = CircleMarker;

	/**
	 * Draws a new circle marker using base configuration. 
	 */
	CircleMarker.prototype.draw = function() {
		this.node = L.circleMarker([0, 0], this.configuration);
		AbstractComponent.prototype.draw.call(this);
		// No update here, will be triggered by layer.update()
	};
	
	/**
	 * Update the circle marker properties if the node exists. First it
	 * will evaluate properties on the configuration based on the component
	 * context. Then the marker position, radius, and style will be updated
	 * based on the configuration. Finally, if the node has not yet been added
	 * to the containing layer, it will be added now.
	 */
	CircleMarker.prototype.update = function() {
		if(this.node == null) return;
		AbstractComponent.prototype.update.call(this);

		// Update lat/lon, radius, and style
		this.node.setLatLng([this.properties.lat, this.properties.lon]);
		this.node.setRadius(this.properties.radius);
		this.node.setStyle(this.properties);
		
		// Call generic component update
		AbstractComponent.prototype.update.call(this);
	};
	
	
	MapConfiguration = {
		defaultConfiguration: {
				leaflet: {
					control: {
						
					},
					tileLayer: {
						provider: 			'http://{s}.tile.openstreetmap.se/hydda/full/{z}/{x}/{y}.png',
						options: {
							attribution:	'Tiles courtesy of <a href="http://openstreetmap.se/" target="_blank">OpenStreetMap Sweden</a> &mdash; Map data &copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>',
							minZoom:		0,
							maxZoom:		18,
							opacity:		1.0
						}
						/* 
							Maporama: http://tilescf.maporama.com/georasterservices/v101/standard_map/{z}/{x}/{y}.png
							OSM: http://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png
							MapQuest Open: http://otile1.mqcdn.com/tiles/1.0.0/map/{z}/{x}/{y}.png
							ArcGIS Aerial: http://server.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile/{z}/{y}/{x}
							Other providers: http://leaflet-extras.github.io/leaflet-providers/preview/
						*/
					}
				},
				options: {
					configurable:		false,
					copy: {
						enabled:		true
					},
					dataset: {
						icon:			false
						//modal:			undefined
					},
					search: {
						fields:			undefined,
						delay:			500,
						enabled:		true,
						label:			"Search"
					},
				},
				
			},
		
		defaultLayerConfiguration: {
				options: {
					filter: {	
						accept:			true,
						fields:			undefined,
						delay:			500,
						enabled:		true,
						passthrough:	true
					},
					hover: {
						accept:			true,
						enabled:		true,
						passthrough:	true					
					},
					marking: {
						accept:			true,
						enabled:		true,
						mode:			"single",
						passthrough:	true,
						unmark:			true
					},
					search: {
						fields:			undefined,
						delay:			500,
						enabled:		true,
						label:			"Search"
					}
				}
			},
			
		// Merges the source configuration object into the target configuration object
		//   Source values will overwrite target values.
		//   Should always merge a DEFAULT configuration source into a USER configuration target
		mergeConfiguration:		function(source, target) {
									for(var idx = 0; idx < Object.keys(source).length; idx++) {
										var key = Object.keys(source)[idx];
										var val = source[key];
										if(typeof val === 'object') {
											if(target[key] == undefined) 
												target[key] = val;
											else
												this.mergeConfiguration(val, target[key]);
										}
										else {
											if(target[key] == undefined) 
												target[key] = val;
										}
									}
									return target;
								},

		// Processes the specified configuration object and merges in default configuration
		//   for base object 
		processConfiguration:	function(config) {
									var configuration = this.mergeConfiguration(this.defaultConfiguration, config);
									for(var idx = 0; idx < Object.keys(configuration.leaflet.layers).length; idx++) {
										var layerKey = Object.keys(configuration.leaflet.layers)[idx];
										configuration.leaflet.layers[layerKey] = this.mergeConfiguration($.extend({}, this.defaultLayerConfiguration), 
											configuration.leaflet.layers[layerKey]);
									}
									return configuration;
								},
	};

	
	module.Map 					= Map;
	module.Layer 				= Layer;
	module.MapConfiguration		= MapConfiguration;
	
	return module;
})();
