/**
 * Represents a passthrough data sink.
 * @constructor
 * @param {Object} dataset - dataset that this sink will be attached to
 * @param {Object} configuration - configuration object defining the sink
 */
function PassthroughDatasink(id, configuration, dataset, containerDiv, overlay, onUpdate) {
	this.id = id;
	this.dataset = dataset;
	this.onUpdate = onUpdate;
	this.passthrough = new LivePassthrough.Passthrough(id, configuration, containerDiv, overlay, this);
}

/**
 * Event handler for dataset events. 
 * @param {String} eventType - string describing the event that occurred
 * @param {Object} object - data object for the event, or null
 */
PassthroughDatasink.prototype.trigger = function(eventType, object) {
	this.passthrough.trigger(eventType, object);
};


// Define a module to hold object defintions for the passthrough objects
var LivePassthrough = (function() {
	var module = {};

	/**
	 * Creates a new Passthrough object.
	 * @constructor
	 * @param {String} id - unique identifier
	 * @param {Object} configuration - the configuration for the passthrough
	 * @param {Object} containerDiv - this node object that passthrough will be appended
	 * @param (Function} overlay - function to call on start and end of the overlay
	 * @param {Object} sink - datasink that will feed this passthrough
	 */
	function Passthrough(id, configuration, containerDiv, overlay, sink) {
		this.id				= id;
		this.configuration	= configuration;
		this.containerDiv 	= containerDiv;
		this.overlay		= overlay;
		this.sink			= sink;
		
		// Pause and snapshot flags
		this.paused			= false;
		this.snapshot		= false;
		
		// Schema for the dataset
		this.schema			= null;		

		// Vars to hold the objects being hovered and marked
		this.hovered		= [];		
		this.marked			= [];

		// Filter definition
		this.filterDef		= {
			search: {
				expression:	undefined,
				fields:		this.configuration.options.search.fields
			},
			filter:	{
				filters:	[],
				fields:		this.configuration.options.filter.fields
			}
		};

		// Ordering definition
		this.orderDef		= [];

		// Color maps
		this.colorMaps		= {};

		// Components
		this.componentArr	= [];
		this.componentMap	= {};

		// Ordered, filtered, and final result rows
		this.orderedArr		= [];
		this.filterArr		= [];
		this.resultArr		= [];
		
		// Hierarchy
		this.hierarchy = {
			level:		0,
			members:	{},
			values:		[]
		};
		
		// Check dependencies. This will throw Error if dependencies are not satisfied.
		this.dependencyCheck();
		
		// Microbatch is used to batch up updates to prevent too many chart redraws
		this.microbatch = {
			rtime:		new Date(1, 1, 2000, 12,00,00),
			timeout:	false,
			delta:		500 // msec between updates
		}

		// Color maps
		if(this.configuration.options.colormaps != undefined) {
			for(var idx = 0; idx < Object.keys(this.configuration.options.colormaps).length; idx++) {
				var thisColorMapName = Object.keys(this.configuration.options.colormaps)[idx];
				var thisColorScale = this.configuration.options.colormaps[thisColorMapName];
				this.colorMaps[thisColorMapName] = new DataFlow.ColorMap(thisColorMapName, thisColorScale);
			}
		}
		
		// On document ready, register UI handlers
		var self = this;
		setTimeout(function() {
			self.registerEventHandlers();
		}), 0;
		
		// Draw the initial empty passthrough
		this.draw();
	};

	/**
	 * Performs a dependency check for rendering the passthrough.
	 */
	Passthrough.prototype.dependencyCheck = function() {
		// No dependencies for this component
	};
		
	/**
	 * Registers event handlers on UI components.
	 */
	Passthrough.prototype.registerEventHandlers = function() {
		// Event handlers
	};
	
	/**
	 * Event handler for all objects delegated from the datasink.
	 * @param {String} eventType - the event type
	 * @param {Object} object - optional object
	 */
	Passthrough.prototype.trigger = function(eventType, object) {
		var self = this;

		if(eventType == 'schemaChange') {
			this.schema = object;			
		}
		else if(eventType == 'snapshotStart') {
			this.snapshot = true;
			this.overlay(true);
		}
		else if(eventType == 'snapshotEnd') {
			setTimeout(function() {
				self.snapshot = false;
				self.overlay(false);
				self.update();
			},500);
		}
		else if(eventType == 'clear') {
			this.clear();
		}
		else if(eventType == 'insert') {
			this.addComponent(object);
		}
		else if(eventType == 'update') {
			this.updateComponent(object);
		}
		else if(eventType == 'upsert') {
		}
		else if(eventType == 'delete') {
			this.removeComponent(object);
		}
		else if(eventType == 'hover') {
			if(this.configuration.options.hover.accept != false)
				this.hover(object);
		}
		else if(eventType == 'marked') {
			if(this.configuration.options.marking.accept != false)
				this.mark(object);
		}					
		else if(eventType == 'filter') {
			if(this.configuration.options.filter.accept != false)
				this.filter(object);
		}

		this.update();		
	};

	/**
	 * Add a new component bound to the specified object on the passthrough.
	 * @param {Object} object - the object to add the bound component
	 */
	Passthrough.prototype.addComponent = function(object) {
		var component = new Component(object, this);
		this.componentMap[component._key] = component;
		this.componentArr.push(component);

		// Process hierarchy if configured
		if(this.configuration.options.hierarchy != undefined) {
			var thisHierarchy = Hierarchy.getHierarchy(object, this.hierarchy, this.configuration.options.hierarchy);
			Hierarchy.addComponent(component, thisHierarchy);
		}

		this.update();
	};

	/**
	 * Updates the component bound to the specified object on the passthrough.
	 * @param {Object} object - the object to update the bound component
	 */
	Passthrough.prototype.updateComponent = function(object) {
		var component = this.componentMap[object._key];
		if(component == undefined) return;		
		component.object = object;

		// Process hierarchy if configured
		if(this.configuration.options.hierarchy != undefined) {
			var thisHierarchy = Hierarchy.getHierarchy(object, this.hierarchy, this.configuration.options.hierarchy);
			Hierarchy.updateComponent(component, thisHierarchy);
		}

		component.update();
		this.update();
	};

	/**
	 * Removes the component bound to the specified object from the passthrough.
	 * @param {Object} object - the object to remove the bound component
	 */
	Passthrough.prototype.removeComponent = function(object) {
		var component = this.componentMap[object._key];
		delete this.componentMap[object._key];
		this.componentArr.splice(this.componentArr.indexOf(component), 1);
		
		// Process hierarchy if configured
		if(this.configuration.options.hierarchy != undefined) {
			Hierarchy.removeComponent(component);
		}
		
		this.update();
	};
	
	
	/**
	 * Clears all objects and all components from the passthrough.
	 */
	Passthrough.prototype.clear = function() {	
		this.componentMap			= {};
		this.componentArr			= [];
		if(this.configuration.options.hierarchy != undefined) {
			this.hierarchy = {
				level:		0,
				members:	{},
				values:		[]
			};
		}
		this.update();
	};

	/**
	 * Draws the UI components for the passthrough.
	 */
	Passthrough.prototype.draw = function() {

	};

	/**
	 * Updates the contents of the passthrough. Normally this will happen in a microbatch
	 * to prevent too many updates from occurring. However it is possible to trigger
	 * an immediate update by setting the flag.
	 * @param {Boolean} immediate - trigger an immediate update
	 */
	Passthrough.prototype.update = function(immediate) {
		var self = this;

		if(immediate == true) {
			run();
			return;
		}
		
		// Microbatch handler to prevent too many updates
		if(self.microbatch.timeout === false) {
			self.microbatch.timeout = true;
			setTimeout(eventEnd, self.microbatch.delta);
		}					
		function eventEnd() {
			var d = new Date() - self.microbatch.rtime;
			if (new Date() - self.microbatch.rtime < self.microbatch.delta)
				setTimeout(eventEnd, self.microbatch.delta);
			else {
				self.microbatch.rtime = new Date();
				self.microbatch.timeout = false;
				run();
			}               
		};
		
		// Run the update
		function run() {
			if(self.paused == true) return;
			
			// Apply ordering if set
			self.orderedArr = self.componentArr;
			if(self.configuration.options.ordering.enabled != false) {
				self.orderedArr = DataFlow.Utilities.Ordering.order(self.orderedArr, self.orderDef, 'object');
			}

			// Apply search and filters if set
			self.filterArr = self.orderedArr;
			if(self.configuration.options.search.enabled != false || self.configuration.options.filter.enabled != false) {
				self.filterArr = DataFlow.Utilities.Filtering.filter(self.filterArr, self.filterDef, 'object');
			}
		
			// Set final results
			self.resultArr = self.filterArr;
			
			// Call the sink onUpdate function
			self.sink.onUpdate();			
		};
	};

	/**
	 * Pauses updates on a passthrough.
	 */
	Passthrough.prototype.pause = function() {
		this.paused = true;
	};
	
	/**
	 * Resumes a paused passthrough.
	 */
	Passthrough.prototype.resume = function() {
		if(this.paused == false) return;
		this.paused = false;
		this.update();
	};

	/**
	 * Resizes a passthrough.
	 */
	Passthrough.prototype.resize = function() {
		// No implementation for this component
	};
	
	/**
	 * Applies a search value against the passthrough. This will not be passed through
	 * to the dataset and is local only to the current passthrough.
	 * @param {String} value - the search string
	 */
	Passthrough.prototype.search = function(value) {
		if(value == undefined || value.length == 0) {
			delete this.filterDef.search.expression;
		}
		else {
			this.filterDef.search.expression = value;
		}

		this.update();
	};
	
	/**
	 * Hovers a single component bound to the specified object.
	 * Will also optionally notify the dataset of a hover event, if enabled.
	 * @param {Object} object - object that the associated component will be hovered
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Passthrough.prototype.hover = function(object, notify) {
		if(this.configuration.options.hover.enabled != false) {
			// Unhover all existing rows
			for(var idx = this.hovered.length - 1; idx >= 0; idx--) {
				var thisObject = this.hovered[idx];
				var component = this.componentMap[thisObject._key];
				if(component != undefined) {
					component.hovered = false;
					component.update();
				}
				this.hovered.splice(this.hovered.indexOf(thisObject), 1);
			}	
			
			// If the object is not null, process the hovering
			if(object != null) {
				if(Array.isArray(object)) {
					for(var idx = 0; idx < object.length; idx++) {
						var thisObject = object[idx];
						var component = this.componentMap[thisObject._key];
						if(component != undefined) {
							component.hovered = true;
							component.update();
						}
						this.hovered.push(thisObject);
					}
				}
				else {
					var component = this.componentMap[object._key];
					if(component != undefined) {
						component.hovered = true;
						component.update();
					}
					this.hovered.push(object);
				}
			}
			
			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.hover.passthrough == true) {
				this.sink.dataset.trigger('hover', this.sink, object);
			}
			
			this.update();
		}
	};

	/**
	 * Marks a single component or a set of components bound to the specified object or array of objects.
	 * Will also optionally notify the dataset of a marked event, if enabled.
	 * @param {Object|Array} object - if a single data object then the associated component will be marked
	 *                                if an array of objects then the associated components will be marked
	 *                                  for single marking components that receive a list of objects then only
	 *                                  the first component will be marked
	 *                                if null then all marked components will be unmarked
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Passthrough.prototype.mark = function(object, notify) {
		if(this.configuration.options.marking.enabled != false) {
			var mode = this.configuration.options.marking.mode;

			// Unmark all existing components
			for(var idx = this.marked.length - 1; idx >= 0; idx--) {
				var thisObject = this.marked[idx];				
				var component = this.componentMap[thisObject._key];
				if(component != undefined) {
					component.marked = false;
					component.update();
				}
				this.marked.splice(this.marked.indexOf(thisObject), 1);
			}	

			// If the object is not null, process the marking
			if(object != null) {
				if(Array.isArray(object)) {
					if(this.configuration.options.marking.mode == 'multiple') {
						for(var idx = 0; idx < object.length; idx++) {
							var thisObject = object[idx];
							var component = this.componentMap[thisObject._key];
							component.marked = true;
							component.update();
							this.marked.push(thisObject);
						}
					}
					// If we receive a list of objects but can only mark one, mark only the first one
					else if(this.configuration.options.marking.mode == 'single' && object.length > 0) {
						var component = this.componentMap[object[0]._key];
						component.marked = true;
						component.update();
						this.marked.push(object[0]);
					}
				}
				else {
					var component = this.componentMap[object._key];
					component.marked = true;
					component.update();
					this.marked.push(object);
				}
			}
			
			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.marking.passthrough == true) {
				this.sink.dataset.trigger('marked', this.sink, object);
			}			

			this.update();
		}
	};

	/**
	 * Applies a filter and updates the passthrough. Will also optionally notify the
	 * dataset of a filter event, if enabled.
	 * @param {Array} filters - array of filter objects
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Passthrough.prototype.filter = function(filters, notify) {
		if(this.configuration.options.filter.enabled != false) {
			// This clones the array so we don't get duplicates
			this.filterDef.filter.filters = filters.slice(0);
			this.filterDef.filter.fields = this.configuration.options.filter.fields;

			// Filter the components
			for(var idx = 0; idx < this.componentArr.length; idx++) {
				var include = DataFlow.Utilities.Filtering.filterItem(this.componentArr[idx], 
					this.filterDef.filter.filters, this.filterDef.filter.fields, 'object');
				var filtered = !include;
				this.componentArr[idx].filter(filtered);
			}

			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.filter.passthrough == true) {
				this.sink.dataset.trigger('filter', this.sink, this.filterDef.filter.filters);
			}
			
			this.update();
		}
	};
		
	/**
	 * Display the passthrough configuration modal.
	 */
	Passthrough.prototype.displayConfiguration = function() {
		if(this.configuration.options.configurable != true) return;
		console.log('display config');
		// TODO display configuration modal
	};

	/**
	 * Displays the dataset configuration modal.
	 */
	Passthrough.prototype.displayDataset = function() {
		if(this.configuration.options.dataset.modal == undefined) return;
		console.log('display dataset');
		// TODO display dataset modal
	};
	
	/** 
	 * Returns a color for the given value from the specified color map.
	 * @param {Object} value - the value to lookup, can be anything but typically a String
	 * @param (Object} name - the name of the color map
	 * @return {String} the color
	 */
	Passthrough.prototype.colormap = function(value, name) {
		var colorMap = this.colorMaps[name];
		if(colorMap == undefined) return;
		return colorMap.getColor(value);
	};
	
	/** 
	 * Formats the value using the specified format expression.
	 * @param {Object} val - the value to format, typically a timestamp or a number
	 * @param {Object} format - the formatting to apply, either a date/time format
	 *                   string or a number for number of decimals to round by
	 */
	Passthrough.prototype.format = function(val, format) {
		return DataFlow.Utilities.Format.format(val, format);
	};

	
	Passthrough.prototype.click = function(e, component) {
		if(this.configuration.options.marking.enabled == false) return;
		var ctrlKey = e.ctrlKey;					
		var mode = this.configuration.options.marking.mode;
		var isMarked = component.marked == true;
		
		// If unmarking is disabled and this component is marked, then just call mark again
		//   and take no further action
		if(this.configuration.options.marking.unmark == false && isMarked == true) {
			this.mark(component.object, true);
			return;
		}
		
		// Single mode
		if(mode == 'single') {
			if(isMarked == true)
				this.mark(null, true);
			else
				this.mark(component.object, true);
		}
		// Multiple mode
		else if(mode == 'multiple') {
			if(ctrlKey == true) {
				var objects = [];
				for(var idx = 0; idx < this.marked.length; idx++)
					objects.push(this.marked[idx].object);
				if(isMarked == false)
					objects.push(component.object);
				else if(isMarked == true)
					objects.splice(objects.indexOf(component.object), 1);
				this.mark(objects, true);
			}
			else {
				if(isMarked == true)
					this.mark(null, true);
				else
					this.mark(component.object, true);
			}
		}
		
	}
	

	/** 
	 * Creates a new component on the passthrough.
	 * @param {Object} object - the object to bind to the component
	 */
	function Component(object, passthrough) {
		this._key			= object._key;
		this.object 		= object;
		this.passthrough	= passthrough;
		
		// Flags for hovered, marked, and filtered
		this.hovered		= false;
		this.marked			= false;
		this.filtered		= false;
	};

	/**
	 * Draws a new component using base configuration. 
	 */
	Component.prototype.draw = function() {
		
	};
	
	/**
	 * Update the component.
	 */
	Component.prototype.update = function() {

	};
	
	/**
	 * Removes the component from the layer and invalidates the
	 * node for garbage collection.
	 */
	Component.prototype.remove = function() {

	};
		
	/**
	 * Sets the filtered flag for the component and adds or removes from the layer.
	 * @param {Boolean} filtered - true if filtered, otherwise false
	 */
	Component.prototype.filter = function(filtered) {
		if(this.filtered == true && filtered == false) {
			this.filtered = false;
		}
		else if(this.filtered == false && filtered == true) {
			this.filtered = true;
		}
	}
	
	/**
	 * Sets the search flag for the component and adds or removes from the layer.
	 * @param {Boolean} include - true if search match, otherwise false
	 */
	Component.prototype.search = function(include) {
		if(this.searched == null && include == true) {
			this.searched = true;
		}
		if(this.searched == null && include == false) {
			this.searched = false;
		}
		
		if(this.searched == true && include == null) {
			this.searched = null;
		}
		if(this.searched == true && include == false) {
			this.searched = false;
		}
		
		if(this.searched == false && include == null) {
			this.searched = null;
		}
		if(this.searched == false && include == true) {
			this.searched = true;
		}
	}
	
	


	/** 
	 * Passthrough configuration helper object.
	 */
	PassthroughConfiguration = {
		// Default passthrough configuration object
		defaultConfiguration: {
				options: {
					//classes:			undefined,
					configurable:		false,
					copy: {
						enabled:		true
					},
					dataset: {
						icon:			true
						//modal:			undefined
					},
					filter: {	
						accept:			true,
						fields:			undefined,
						delay:			500,
						enabled:		true,
						passthrough:	true
					},
					//hierarchy:		undefined,
					hover: {
						accept:			true,
						enabled:		true,
						passthrough:	true					
					},
					marking: {
						accept:			true,
						enabled:		true,
						mode:			"single",
						passthrough:	true,
						unmark:			true
					},
					ordering: {
						//classes:		undefined,
						fields:			undefined,
						enabled:		true,
						//initial:		{},
						multiple:		false
					},
					search: {
						fields:			undefined,
						delay:			500,
						enabled:		true,
						label:			"Search"
					},
				},				
			},
			
		// Merges the source configuration object into the target configuration object
		//   Source values will overwrite target values.
		//   Should always merge a DEFAULT configuration source into a USER configuration target
		mergeConfiguration:		function(source, target) {
									for(var idx = 0; idx < Object.keys(source).length; idx++) {
										var key = Object.keys(source)[idx];
										var val = source[key];
										if(typeof val === 'object') {
											if(target[key] == undefined) 
												target[key] = val;
											else
												this.mergeConfiguration(val, target[key]);
										}
										else {
											if(target[key] == undefined) 
												target[key] = val;
										}
									}
									return target;
								},

		// Processes the specified configuration object and merges in default configuration
		//   for base object
		processConfiguration:	function(config) {
									var configuration = this.mergeConfiguration(this.defaultConfiguration, config);
									return configuration;
								},
	};
	
	
	
	
	// Setup module objects
	module.Passthrough 				= Passthrough;
	module.PassthroughConfiguration	= PassthroughConfiguration;
	
	return module;
})();