/**
 * Represents a Plotly chart data sink.
 * @constructor
 * @param {Object} dataset - dataset that this sink will be attached to
 * @param {Object} configuration - configuration object defining the sink
 */
function PlotlyChartDatasink(id, configuration, dataset, containerDiv, overlay) {
	this.id = id;
	this.dataset = dataset;
	this.chart = new LivePlotlyChart.Chart(id, configuration, containerDiv, overlay, this);
}

/**
 * Event handler for dataset events. 
 * @param {String} eventType - string describing the event that occurred
 * @param {Object} object - data object for the event, or null
 */
PlotlyChartDatasink.prototype.trigger = function(eventType, object) {
	this.chart.trigger(eventType, object);
};


// Define a module to hold object defintions for the chart objects
var LivePlotlyChart = (function() {
	var module = {};

	/**
	 * Creates a new Chart object.
	 * @constructor
	 * @param {String} id - unique identifier
	 * @param {Object} configuration - the configuration for the chart
	 * @param {Object} containerDiv - this node object that chart will be appended
	 * @param (Function} overlay - function to call on start and end of the overlay
	 * @param {Object} sink - datasink that will feed this chart
	 */
	function Chart(id, configuration, containerDiv, overlay, sink) {
		this.id				= id;
		this.configuration	= configuration;
		this.containerDiv 	= containerDiv;
		this.overlay		= overlay;
		this.sink			= sink;

		// Pause and snapshot flags
		this.paused			= false;
		this.snapshot		= false;

		// Schema for the dataset
		this.schema			= null;		

		// Vars to hold the objects being hovered and marked
		this.hovered		= [];		
		this.marked			= [];

		// Filter definition
		this.filterDef		= {
			search: {
				expression:	undefined,
				fields:		this.configuration.options.search.fields
			},
			filter:	{
				filters:	[],
				fields:		this.configuration.options.filter.fields
			}
		};

		// Color maps
		this.colorMaps		= {};

		// Array of trace objects from the series
		this.traceMap		= {};
		this.traceArr		= [];
		
		// Create an array to hold the series objects
		//   Note this is different than this.traceArr which is the actual trace object required
		//   by Plotly
		this.seriesArr 		= [];

		// Check dependencies. This will throw Error if dependencies are not satisfied.
		this.dependencyCheck();
		
		// Microbatch is used to batch up updates to prevent too many chart redraws
		this.microbatch = {
			rtime:		new Date(1, 1, 2000, 12,00,00),
			timeout:	false,
			delta:		500 // msec between updates
		}

		// Color maps
		if(this.configuration.options.colormaps != undefined) {
			for(var idx = 0; idx < Object.keys(this.configuration.options.colormaps).length; idx++) {
				var thisColorMapName = Object.keys(this.configuration.options.colormaps)[idx];
				var thisColors = this.configuration.options.colormaps[thisColorMapName];
				this.colorMaps[thisColorMapName] = new DataFlow.ColorMap(thisColorMapName, thisColors);
			}
		}

		// Loop over the traces configuration and create a series object for each
		//   Add the Series object to the seriesArr and the Series object trace
		//   property to the data array.
		for(var idx = 0; idx < configuration.plotly.traces.length; idx++) {
			var thisSeries = new Series(configuration.plotly.traces[idx], this, idx);
			this.seriesArr.push(thisSeries);
			this.traceArr.push(thisSeries.trace);
			this.traceMap[thisSeries.trace.name] = thisSeries;
		}
		
		// Variables to hold equal axes definitions
		this.minAxis = 0;
		this.maxAxis = 0;					
		
		// On document ready, register UI handlers
		var self = this;
		setTimeout(function() {
			self.registerEventHandlers();
		}), 0;
		
		// Draw the initial empty graph
		this.draw();
	};
	
	/**
	 * Performs a dependency check for rendering the chart.
	 */
	Chart.prototype.dependencyCheck = function() {
		var msg = "Required Plotly libraries not loaded or not present";
		try {
			var p = Plotly;
		}
		catch(err) {
			throw Error("Required Plotly libraries not loaded or not present");
		}
	};
		
	/**
	 * Registers event handlers on UI components.
	 */
	Chart.prototype.registerEventHandlers = function() {
		var self = this;
		
		// Event handlers to pause and resume for wheel scroll, this gives the chart a time to update
		$(this.containerDiv).bind('mousewheel', function(e) {
			self.pause();
			setTimeout(function() {
				self.resume();
			},500);
		});
		
		// Event handlers to pause and resume for click and drag, this gives the chart a time to update
		$(this.containerDiv).on('mousedown', function (evtDown) {
			$(self.containerDiv).on('mouseup', function handler(evt) {
				if(self.paused == true) self.resume();
				$(self.containerDiv).off('mouseup');
				$(self.containerDiv).off('mousemove');
			});
			$(self.containerDiv).on('mousemove', function handler(evt) {
				var delta = Math.max(Math.abs(evtDown.pageX - evt.pageX), Math.abs(evtDown.pageY- evt.pageY));
				if(delta > 5 && self.paused == false) {
					self.pause();
				}
			});
		});
			
		// Event handlers for hover over plotly element
		//   THIS DOES NOT WORK VERY WELL IN PLOTLY, recommend disabling hover
		if(this.configuration.options.hover.enabled != false) {
			this.containerDiv.on('plotly_hover', function(data) {
				if(data.points == undefined) return;
				var thisPoint = data.points[0];
				var curveNumber = thisPoint.curveNumber == undefined ? 0 : thisPoint.curveNumber;
				var series = self.seriesArr[curveNumber];
				var pointNumber = thisPoint.pointNumber == undefined ? thisPoint.i : thisPoint.pointNumber;
				var component = series.pointArr[pointNumber];
				self.hover(component.object, true);
			});
			
			this.containerDiv.on('plotly_unhover', function(data) {
				self.hover(null, true);
			});
		}
		
		// Event handlers for click on plotly element
		//   THIS DOES NOT WORK VERY WELL IN PLOTLY, recommend disabling marking
		if(this.configuration.options.marking.enabled != false) {
			this.containerDiv.on('plotly_click', function(data) {
				if(data.points == undefined) return;
				var thisPoint = data.points[0];
				var curveNumber = thisPoint.curveNumber == undefined ? 0 : thisPoint.curveNumber;
				var series = self.seriesArr[curveNumber];
				var pointNumber = thisPoint.pointNumber == undefined ? thisPoint.i : thisPoint.pointNumber;
				var component = series.pointArr[pointNumber];
				self.mark(component.object, true);
			});
		}
	};
	
		
	/**
	 * Event handler for all objects delegated from the datasink.
	 * @param {String} eventType - the event type
	 * @param {Object} object - optional object
	 */
	Chart.prototype.trigger = function(eventType, object) {
		var self = this;
		
		if(eventType == 'schemaChange') {
			this.schema = object;
		}
		else if(eventType == 'snapshotStart') {
			this.snapshot = true;
			this.overlay(true);
		}
		else if(eventType == 'snapshotEnd') {
			setTimeout(function() {
				self.snapshot = false;
				self.overlay(false);
			},500);
		}
		else if(eventType == 'clear') {
			this.clear();
		}
		else if(eventType == 'insert') {
			this.addPoint(object);
		}
		else if(eventType == 'update') {
			this.updatePoint(object);
		}
		else if(eventType == 'upsert') {
		}
		else if(eventType == 'delete') {
			this.removePoint(object);
		}
		else if(eventType == 'hover') {
			if(this.configuration.options.hover.accept != false)
				this.hover(object);
		}
		else if(eventType == 'marked') {
			if(this.configuration.options.marking.accept != false)
				this.mark(object);
		}
		else if(eventType == 'filter') {
			if(this.configuration.options.filter.accept != false)
				this.filter(object);
		}					
		
	};
	
	/**
	 * Add a new point bound to the specified object on the chart.
	 * @param {Object} object - the object to add the bound point
	 */
	Chart.prototype.addPoint = function(object) {
		for(var idx = 0; idx < this.seriesArr.length; idx++)
			this.seriesArr[idx].addPoint(object);
		this.update();
	};

	/**
	 * Updates the point bound to the specified object on the chart.
	 * @param {Object} object - the object to update the bound point
	 */
	Chart.prototype.updatePoint = function(object) {
		for(var idx = 0; idx < this.seriesArr.length; idx++)
			this.seriesArr[idx].updatePoint(object);
		this.update();
	};
	
	/**
	 * Removes the point bound to the specified object from the chart.
	 * @param {Object} object - the object to remove the bound point
	 */
	Chart.prototype.removePoint = function(object) {
		for(var idx = 0; idx < this.seriesArr.length; idx++)
			this.seriesArr[idx].removePoint(object);
		this.update();
	};
					
	/**
	 * Clears all objects and all points from the chart.
	 */
	Chart.prototype.clear = function() {
		for(var idx = 0; idx < this.seriesArr.length; idx++)
			this.seriesArr[idx].removeAll();
		this.update();
	};
	
	/**
	 * Draws the UI components for the chart.
	 */
	Chart.prototype.draw = function() {
		Plotly.newPlot(this.containerDiv, this.traceArr, this.configuration.plotly.layout, 
			this.configuration.plotly.options);	
	};
	
	/**
	 * Updates the contents of the chart. Normally this will happen in a microbatch
	 * to prevent too many updates from occurring. However it is possible to trigger
	 * an immediate update by setting the flag.
	 * @param {Boolean} immediate - trigger an immediate update
	 */
	Chart.prototype.update = function(immediate) {
		var self = this;

		if(immediate == true) {
			run();
			return;
		}

		// Microbatch handler to prevent too many updates
		if(self.microbatch.timeout === false) {
			self.microbatch.timeout = true;
			setTimeout(eventEnd, self.microbatch.delta);
		}					
		function eventEnd() {
			var d = new Date() - self.microbatch.rtime;
			if (new Date() - self.microbatch.rtime < self.microbatch.delta)
				setTimeout(eventEnd, self.microbatch.delta);
			else {
				self.microbatch.rtime = new Date();
				self.microbatch.timeout = false;
				run();
			}               
		};
		
		// Run the update
		function run() {
			if(self.paused == true) return;
			
			Plotly.redraw(self.containerDiv);
		};
	};
	
	/**
	 * Pauses updates on a chart.
	 */
	Chart.prototype.pause = function() {
		this.paused = true;
	};
	
	/**
	 * Resumes a paused chart.
	 */
	Chart.prototype.resume = function() {
		if(this.paused == false) return;
		this.paused = false;
		this.update();
	};
	
	/**
	 * Resizes a chart.
	 */
	Chart.prototype.resize = function() {
		// Get the new size for the chart
		var update = {
			width: $(this.containerDiv).parent().width(), 
			height: $(this.containerDiv).parent().height()
		}

		// Relayout then redraw the chart						
		Plotly.relayout(this.containerDiv, update);	
		Plotly.redraw(this.containerDiv);
	};
			
	/**
	 * Applies a search value against the chart. This will not be passed through
	 * to the dataset and is local only to the current chart.
	 * @param {String} value - the search string
	 */
	Chart.prototype.search = function(value) {
		if(value == undefined || value.length == 0) {
			delete this.filterDef.search.expression;
		}
		else {
			this.filterDef.search.expression = value;
		}
		
		// Pass the filterDef through to the series
		for(var idx = 0; idx < this.seriesArr.length; idx++) {
			this.seriesArr[idx].filter(this.filterDef);
		}
	};
	
	/**
	 * Hovers a single point bound to the specified object.
	 * Will also optionally notify the dataset of a hover event, if enabled.
	 * @param {Object} object - object that the associated point will be hovered
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Chart.prototype.hover = function(object, notify) {
		if(this.configuration.options.hover.enabled != false) {
		
			// THIS DOES NOT WORK VERY WELL IN PLOTLY, recommend disabling hover
			// Unhover all points
			Plotly.Fx.hover(this.containerDiv,[]);

			// If the object is not null, process the hovering
			if(object != null) {
				if(Array.isArray(object)) {
					this.hovered = object.slice(0);
				}
				else {
					this.hovered = [object];
				}

				var hoverPoints = [];
				for(var seriesIdx = 0; seriesIdx < this.seriesArr.length; seriesIdx++) {
					var thisSeries = this.seriesArr[seriesIdx];
					for(var objIdx = 0; objIdx < this.hovered.length; objIdx++) {
						var thisPoint = thisSeries.pointMap[object[objIdx]._key];
						if(thisPoint == undefined) continue;
						var pointIndex = thisSeries.pointArr.indexOf(thisPoint);
						var hoverPoint = {
							curveNumber: seriesIdx,
							pointNumber: pointIndex
						}
						hoverPoints.push(hoverPoint);
					}
				}
				Plotly.Fx.hover(this.containerDiv, hoverPoints);
			}
		
			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.hover.passthrough == true) {
				this.sink.dataset.trigger('hover', this.sink, object);
			}			
		}
	};
		
	/**
	 * Marks a single point or a set of points bound to the specified object or array of objects.
	 * Will also optionally notify the dataset of a marked event, if enabled.
	 * @param {Object|Array} object - if a single data object then the associated point will be marked
	 *                                if an array of objects then the associated points will be marked
	 *                                  for single marking points that receive a list of objects then only
	 *                                  the first point will be marked
	 *                                if null then all marked points will be unmarked
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Chart.prototype.mark = function(object, notify) {
		if(this.configuration.options.marking.enabled != false) {
			var mode = this.configuration.options.marking.mode;

			// Unmark all existing series points
			for(var idx = 0; idx < this.seriesArr.length; idx++) {
				this.seriesArr[idx].unmark();
			}
			this.marked = [];
			
			// If the object is not null, process the marking
			if(object != null) {
				if(Array.isArray(object)) {
					if(this.configuration.options.marking.mode == 'multiple') {
						for(var idx = 0; idx < object.length; idx++) {
							var thisObject = object[idx];
							this.marked.push(thisObject);
						}
					}
					// If we receive a list of objects but can only mark one, mark only the first one
					else if(this.configuration.options.marking.mode == 'single' && object.length > 0) {
						var thisObject = object[0];
						this.marked.push(thisObject);
					}
				}
				else {
					this.marked.push(object);
				}

				// Update the series marked
				for(var idx = 0; idx < this.seriesArr.length; idx++) {
					this.seriesArr[idx].mark(this.marked);
				}
			}
			
			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.marking.passthrough == true) {
				this.sink.dataset.trigger('marked', this.sink, this.marked);
			}			
		}
	};
		
	/**
	 * Applies a filter and updates the chart. Will also optionally notify the
	 * dataset of a filter event, if enabled.
	 * @param {Array} filter - array of filter objects
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Chart.prototype.filter = function(filters, notify) {
		if(this.configuration.options.filter.enabled != false) {
			// This clones the array so we don't get duplicates
			this.filterDef.filter.filters = filters.slice(0);
			this.filterDef.filter.fields = this.configuration.options.filter.fields;
			
			// Pass the filterDef through to the series
			for(var idx = 0; idx < this.seriesArr.length; idx++) {
				this.seriesArr[idx].filter(this.filterDef);
			}

			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.filter.passthrough == true) {
				this.sink.dataset.trigger('filter', this.sink, this.filtered.filter);
			}
			
			this.update();
		}
	};
	
	/**
	 * Display the chart configuration modal.
	 */
	Chart.prototype.displayConfiguration = function() {
		if(this.configuration.options.configurable != true) return;
		console.log('plotly display config');
		// TODO display configuration modal
	};

	/**
	 * Displays the dataset configuration modal.
	 */
	Chart.prototype.displayDataset = function() {
		if(this.configuration.options.dataset.modal == undefined) return;
		console.log('plotly display dataset');
		// TODO display dataset modal
	};
	
	/**
	 * Equalizes all axes to the same scale factor for both 2d and 3d chart types.
	 */
	Chart.prototype.equalizeAxes = function() {
		var layout = this.configuration.plotly.layout;
		if(layout.scene == undefined) {
			if(layout.xaxis == undefined) layout.xaxis = {};
			if(layout.yaxis == undefined) layout.yaxis = {};
			
			layout.xaxis.range = [this.minAxis, this.maxAxis];
			layout.yaxis.range = [this.minAxis, this.maxAxis];
		}
		else {
			if(layout.scene.xaxis == undefined) layout.scene.xaxis = {};
			if(layout.scene.yaxis == undefined) layout.scene.yaxis = {};
			if(layout.scene.zaxis == undefined) layout.scene.zaxis = {};
			
			layout.scene.xaxis.range = [this.minAxis, this.maxAxis];
			layout.scene.yaxis.range = [this.minAxis, this.maxAxis];
			layout.scene.zaxis.range = [this.minAxis, this.maxAxis];
		}
	};
					
	/** 
	 * Returns a color for the given value from the color map.
	 * @param {Object} value - the value to lookup, can be anything but typically a String
	 * @return {String} the color
	 */
	Chart.prototype.colormap = function(value) {
		return this.colorMap.getColor(value);
	};
	
	
	
	
	/**
	 * Creates a new Series object which represents a set of data on a chart. 
	 * A Series will represent different things depending on the underlying chart type.
	 * Typically a set of points or a line on a chart will be one series. A pie chart
	 * consists of a single series.
	 * @constructor
	 * @param {Object} configuration - the configuration for the series
	 * @param {Object} chart - the chart containing this series
	 */
	function Series(configuration, chart, index) {
		// Set configuration
		this.configuration 	= configuration;
		this.chart 			= chart;
		this.id				= 'series-' + index;

		// Vars for hovered, marked and filtered points
		this.hovered			= [];
		this.marked				= [];
		this.filterDef			= {
			search: {
				expression:	undefined,
				fields:		this.chart.configuration.options.search.fields
			},
			filter:	{
				filters:	[],
				fields:		this.chart.configuration.options.filter.fields
			}
		};

		// Trace object for the series, this will be passed to Plotly
		this.trace			= {};
		
		// Create the point container objects
		this.pointArr 		= [];
		this.pointMap 		= {};
		this.filterArr		= [];
		
		// Create a map of trace property (key) to array of values (value)
		this.arrayMap 		= {};
		
		// Create the property index
		var propertyIndexResult = DataFlow.Utilities.PropertyMap.index(this.configuration, ['property', 'eval']);
		this.propertyIndex = propertyIndexResult.propertyIndex;
		this.trace = propertyIndexResult.properties;
		
		// Create the array map
		for(var idx = 0; idx < this.propertyIndex.arr.length; idx++) {
			var thisProp = this.propertyIndex.arr[idx];
			this.arrayMap[thisProp.prop] = [];
		}		
	};

	/**
	 * Add a new object to the series as a new point.
	 * @param {Object} object - the object to add
	 */
	Series.prototype.addPoint = function(object) {
		var filtered = this.checkFilter(object);
		if(filtered == true) return;

		// Create the point
		var point = {
			_key:			object._key,
			object:			object,
			marked:			false
		}
		
		// Add to the point array and map
		this.pointArr.push(point);
		this.pointMap[point._key] = point;

		// Check include if there is a filter definition
		var include = DataFlow.Utilities.Filtering.filterItem(point, 
			this.filterDef.filter.filters, this.filterDef.filter.fields, 'object');
		
		if(include == true) {
			this.filterArr.push(point);
			
			// Build a context and evaluate the point properties, append to the data object
			for(var idx = 0; idx < this.propertyIndex.arr.length; idx++) {
				var prop = this.propertyIndex.arr[idx].prop;
				var type = this.propertyIndex.arr[idx].type;
				if(type == 'property') {
					var field = this.propertyIndex.map[prop];
					this.arrayMap[prop].push(object[field]);
				}
				else if(type == 'eval') {
					var exp = this.propertyIndex.map[prop];
					var context = {
						object:		object,
						series:		this,
						chart:		this.chart
					};
					var val = DataFlow.Utilities.Evaluate.expression(exp, context);
					this.arrayMap[prop].push(val);
				}
			}
		
			this.equalizeAxes(object);
			this.update();
		}
	};

	/**
	 * Update an existing object point on the series.
	 * @param {Object} object - the object to update
	 */
	Series.prototype.updatePoint = function(object) {
		var point = this.pointMap[object._key];
		if(point == undefined) return;
		
		// Check include if there is a filter definition
		var include = DataFlow.Utilities.Filtering.filterItem(point, 
			this.filterDef.filter.filters, this.filterDef.filter.fields, 'object');

		if(include == true) {
			var index = this.filterArr.indexOf(point);
			if(index == -1) this.filterArr.push(point);
			
			for(var idx = 0; idx < this.propertyIndex.arr.length; idx++) {
				var prop = this.propertyIndex.arr[idx].prop;
				var type = this.propertyIndex.arr[idx].type;
				if(type == 'property') {
					var field = this.propertyIndex.map[prop];
					if(index == -1)
						this.arrayMap[prop].push(object[field]);
					else
						this.arrayMap[prop][index] = object[field];
				}
				else if(type == 'eval') {
					var exp = this.propertyIndex.map[prop];
					var context = {
						object:		object,
						series:		this,
						chart:		this.chart
					};
					var val = DataFlow.Utilities.Evaluate.expression(exp, context);
					if(index == -1)
						this.arrayMap[prop].push(val);
					else
						this.arrayMap[prop][index] = val;
				}
			}		
			
			this.equalizeAxes(object);
			this.update();
		}
		else if(include == false) {
			var index = this.filterArr.indexOf(point);
			if(index > -1) {
				for(var idx = 0; idx < this.propertyIndex.arr.length; idx++) {
					var prop = this.propertyIndex.arr[idx].prop;
					this.arrayMap[prop].splice(index, 1);
				}		
				this.filterArr.splice(index, 1);
			}
			this.update();
		}
		
	};
	
	/**
	 * Remove an existing object point from the series.
	 * @param {Object} object - the object to remove
	 */
	Series.prototype.removePoint = function(object) {
		var point = this.pointMap[object._key];
		if(point == undefined) return;
		var index = this.filterArr.indexOf(point);

		if(index > -1) {
			for(var idx = 0; idx < this.propertyIndex.arr.length; idx++) {
				var prop = this.propertyIndex.arr[idx].prop;
				this.arrayMap[prop].splice(index, 1);
			}		
			this.filterArr.splice(index, 1);
		}
		this.pointArr.splice(this.pointArr.indexOf(point), 1);
		this.pointMap[object._key] = undefined;
		this.update();
		
		// Special case for pie charts when a purge happens
		//   Plotly doesn't like zero length arrays in pie charts, 
		//   so if all data has been removed, call draw to regen the chart
		if(this.trace.type == 'pie' && this.pointArr.length == 0) {
			this.chart.draw();
		}
		
	};
	
	/**
	 * Remove all existing object points from the series.
	 */
	Series.prototype.removeAll = function() {
		for(var idx = 0; idx < this.propertyIndex.arr.length; idx++) {
			var prop = this.propertyIndex.arr[idx].prop;
			this.arrayMap[prop] = [];
		}		
		this.pointArr = [];
		this.pointMap = {};		
		this.update();
	};
	
	/** 
	 * Updates the trace array with the current property arrays.
	 */
	Series.prototype.update = function() {
		for(var idx = 0; idx < this.propertyIndex.arr.length; idx++) {
			var prop = this.propertyIndex.arr[idx].prop;
			var arr = this.arrayMap[prop];
			this.trace[prop] = arr;
		}		
		if(this.configuration.equalAxes == true) this.chart.equalizeAxes();
	};

	/**
	 * Checks to see if the given object has been filtered by any configured
	 * filter condition.
	 * @param {Object} object - the object to test
	 * @return false if the object is included, true if the object is filtered
	 */
	Series.prototype.checkFilter = function(object) {
		if(this.configuration.eval_filter == undefined) return false;
		var context = {
			object:		object,
			series:		this,
			chart:		this.chart
		};
		var b = DataFlow.Utilities.Evaluate.expression(this.configuration.eval_filter, context);
		return !b;
	};
	
	/**
	 * Equalizes all axes for the series.
	 */
	Series.prototype.equalizeAxes = function(object) {
		if(this.configuration.equalAxes != true) return;
		for(var idx = 0; idx < this.propertyIndex.arr.length; idx++) {
			var prop = this.propertyIndex.arr[idx].prop;
			if(prop == 'x' || prop == 'y' || prop == 'z') {
				var field = this.propertyIndex.map[prop];
				var val = object[field];
				if(isNaN(val) == false) {
					this.chart.maxAxis = Math.max(this.chart.maxAxis, val);
					this.chart.minAxis = Math.min(this.chart.minAxis, val);
				}
			}
		}		
	};

	/**
	 * Unhovers the hovered point if there is one.
	 */
	Series.prototype.unhover = function() {
		return;
		
		// TODO hover
	};
	
	/**
	 * Hovers the point for the specified object.
	 * @param {Object} object - the object to hover the point for
	 */
	Series.prototype.hover = function(object) {
		return;
		
		// TODO hover
	};
		
	/**
	 * Unmarks all marked points on the series.
	 */
	Series.prototype.unmark = function() {
		return;
		// TODO mark
		
		if(this.marked == null) return;
		for(var idx = 0; idx < this.marked.length; idx++) {
			var thisObject = this.marked[idx];
			var point = this.pointMap[thisObject._key];
			if(point == undefined) continue;
			var data = point.data;
			if(point.highchartsPoint == null)
				point.highchartsPoint = this.chart.highcharts.get(data.id);
			if(point.highchartsPoint != null)
				point.highchartsPoint.select(false);		
			point.marked = false;
			point.data.selected = false;
		}
		this.marked = [];
	};
	
	/**
	 * Marks the point for the specified object or array.
	 * @param {Object[]} object - the array of objects to mark points for
	 */
	Series.prototype.mark = function(object) {
		return;
		// TODO mark

		if(object == null) return;
		for(var idx = 0; idx < object.length; idx++) {
			var thisObject = object[idx];
			var point = this.pointMap[thisObject._key];
			if(point == undefined) continue;
			var data = point.data;
			if(point.highchartsPoint == null)
				point.highchartsPoint = this.chart.highcharts.get(data.id);
			if(point.highchartsPoint != null)
				point.highchartsPoint.select(true, true);
			point.marked = true;
			point.data.selected = true;
			this.marked.push(thisObject);;
		}
	};
	
	/**
	 * Applies a filter definition on a series.
	 * @param {Object} filterDef - filter definition
	 */
	Series.prototype.filter = function(filterDef) {
		this.filterDef = filterDef;
		
		this.filterArr = [];
		this.arrayMap = {};
				
		// Create the array map
		for(var idx = 0; idx < this.propertyIndex.arr.length; idx++) {
			var thisProp = this.propertyIndex.arr[idx];
			this.arrayMap[thisProp.prop] = [];
		}		

		// Process all the points and build a new filterArr from this 
		for(var pointIdx = 0; pointIdx < this.pointArr.length; pointIdx++) {
			var point = this.pointArr[pointIdx];
			var object = point.object;
			var include = DataFlow.Utilities.Filtering.filterItem(point, 
				this.filterDef.filter.filters, this.filterDef.filter.fields, 'object');
			if(include == true) {
				this.filterArr.push(point);
				
				// Build a context and evaluate the point properties, append to the data object
				for(var propIdx = 0; propIdx < this.propertyIndex.arr.length; propIdx++) {
					var prop = this.propertyIndex.arr[propIdx].prop;
					var type = this.propertyIndex.arr[propIdx].type;
					if(type == 'property') {
						var field = this.propertyIndex.map[prop];
						this.arrayMap[prop].push(object[field]);
					}
					else if(type == 'eval') {
						var exp = this.propertyIndex.map[prop];
						var context = {
							object:		object,
							series:		this,
							chart:		this.chart
						};
						var val = DataFlow.Utilities.Evaluate.expression(exp, context);
						this.arrayMap[prop].push(val);
					}
				}
			
				this.equalizeAxes(object);
			}
		}

		this.update();
	
	};
	
	/** 
	 * Plotly configuration helper object.
	 */
	PlotlyConfiguration = {
		defaultConfiguration: {
				options: {
					configurable:		false,
					copy: {
						enabled:		true
					},
					dataset: {
						icon:			true
						//modal:			undefined
					},
					filter: {	
						accept:			true,
						fields:			undefined,
						delay:			500,
						enabled:		true,
						passthrough:	true
					},
					format: {
						timestamp:		"yyyy-MM-dd HH:mm:ss"
					},
					hover: {
						accept:			false,
						enabled:		false,
						passthrough:	false					
					},
					marking: {
						accept:			true,
						enabled:		true,
						mode:			"single",
						passthrough:	true,
						unmark:			true
					},
					search: {
						fields:			undefined,
						delay:			500,
						enabled:		true,
						label:			"Search"
					}
				},
				
			},
		
		// Merges the source configuration object into the target configuration object
		//   Source values will overwrite target values.
		//   Should always merge a DEFAULT configuration source into a USER configuration target
		mergeConfiguration:		function(source, target) {
									for(var idx = 0; idx < Object.keys(source).length; idx++) {
										var key = Object.keys(source)[idx];
										var val = source[key];
										if(typeof val === 'object') {
											if(target[key] == undefined) 
												target[key] = val;
											else
												this.mergeConfiguration(val, target[key]);
										}
										else {
											if(target[key] == undefined) 
												target[key] = val;
										}
									}
									return target;
								},

		// Processes the specified configuration object and merges in default configuration
		//   for both base object and column objects
		processConfiguration:	function(config) {
									var configuration = this.mergeConfiguration(this.defaultConfiguration, config);
									return configuration;
								},
	};


	
	module.Chart 				= Chart;
	module.PlotlyConfiguration	= PlotlyConfiguration;
	
	return module;
})();
