/**
 * Represents the interface definition for a datasource object. This class should not 
 * be used for an implementation but rather as a reference for source definition.
 * @constructor
 * @param {Object} configuration - configuration object defining the source
 * @param {Object} dataset - dataset that this source will feed into
 */
function AbstractDatasource(configuration, dataset) {
	this.id 			= configuration.id;
	this.name			= configuration.name;
	this.description	= configuration.description;
	this.configuration	= configuration;	
	this.dataset		= dataset;
	this.live			= false;	
}

/** 
 * Initializes the datasource after creation. This function should connect to the datasource
 * successfully and then make a successful call for data. The schema must be set and the 
 * initial data must be sent through to the dataset. For one-time datasources the
 * init function will complete the lifecycle of the data loading and no further data will
 * be processed. For live datasources subsequent updates will be produces through the datasource
 * passed on to the dataset.
 * @return {Promise}
 */
AbstractDatasource.prototype.init = function() {
	var self = this;
	var promise = new Promise(function(resolve, reject) {
		// Initialize the datasource here, including async
		setTimeout(function() {
			// Set the schema
			self.schema = new DatasetSchema();
			for(var idx = 0; idx < self.dataset.key.length; idx++) {
				self.schema.addField(self.dataset.key[idx], 'string');
			}

			// On success
			resolve(self);
			
			// On error
			// reject(Error("It broke"));
		}, 1000);
	});	
	
	return promise;
};

/**
 * Clear any existing cached data inside the data source then calls to the source system
 * and reloads all data. This should be the same process that happens when init() is called.
 */
AbstractDatasource.prototype.reload = function() {

};

/**
 * Clears any existing cached data inside the data source then applies the set of parameters
 * against the source system and reloads all data. This should be the same process that happens
 * when init() is called, except with updated parameters.
 * @param {Object} parameters - object defining the parameters to set
 */
AbstractDatasource.prototype.applyParameters = function(parameters) {

};

/**
 * Clears any existing cached data inside the data source then applies the default set of parameters
 * against the source system and reloads all data. This should be the same process that happens
 * when init() is called.
 */
AbstractDatasource.prototype.resetParameters = function() {

};