/** 
 * Factory object for creating data source instances from configuration.
 */
DatasourceFactory = {
	/**
	 * Creates a new data source from the configuration.
	 * @param {Object} configuration - configuration object defining the source
	 * @param {Object} dataset - dataset object this source is a member of
	 */
	createDatasource: function(configuration, dataset) {
		DatasourceClass = this.getClass(configuration);
		return new DatasourceClass(configuration, dataset);		
	},

	/**
	 * Parses the configuration.type property for package names and then returns the 
	 * full datasource class name.
	 * @return {Function}
	 */
	getClass: function(configuration) {
		// Parse the type to create the correct instance
		var arr = configuration.type.split(".");
		var DatasourceClass = (window || this);
		for (var i = 0, len = arr.length; i < len; i++) {
			DatasourceClass = DatasourceClass[arr[i]];
		}
		
		return DatasourceClass
	},

	/**
	 * Validates the configuration for a single data source config.
	 * @param {Object} configuration - the configuration
	 * @param {Integer} index - position of the datasource in the array
	 * @param {Object} duplicates - object containing existing keys to detect duplicates
	 * @param {Object} warnings - array of warnings for the containing object
	 * @param {Boolean}
	 */
	validate: function(configuration, index, duplicates, warnings) {
		var valid = true;
		
		// Check that the source has an id
		if(configuration.id == undefined) {
			warnings.push('Datasource ' + index + ' has no id');
			valid = false;
		}
		
		// Check that the id is not duplicated
		if(duplicates[configuration.id] != undefined) {
			warnings.push('Datasource ' + configuration.id + ' already exists');
			valid = false;
		}
		
		// Check that the source has a type
		if(configuration.type == undefined) {
			warnings.push('Datasource ' + configuration.id + ' has no type');
			valid = false;
		}
		
		// Check that the type is a valid data source class
		if(valid == true && typeof this.getClass(configuration) !== "function") {
			warnings.push('Datasource ' + configuration.id + ' type ' + configuration.type + ' not found');
			valid = false;
		}
		return valid;
	},
};