/**
 * Represents a LiveView global connection. A page may only make a single connection
 * to the LDM server so this object manages that connection for all datasources.
 */
function LiveViewGlobalConnection() {
	// Set the configuration parameters
	this.url 		= '/lv/client';

	// Initialize objects
	this.connecting = false;
	this.conn		= null;
	this.resolves	= [];
};

/**
 * Connect to the LDM server. When the connection is obtained each of the elements 
 * in the resolves list will be resolved with the new connection.
 */
LiveViewGlobalConnection.prototype.connect = function() {
	var self = this;	
	this.connecting = true;
	
	LiveView.connect({
			url: this.url,
			onClose: function(connection, actor, error) {
				console.warn('LiveViewGlobalConnection connection closed... reconnecting');
				self.connect();
			}
		}
	).then(
		function(connection){
			console.log('LiveViewGlobalConnection connected');
			self.conn = connection;
			self.connecting = false;
			for(var idx = 0; idx < self.resolves.length; idx++) {
				self.resolves[idx](self.conn);
			}
		},
		function(err){
			console.error('LiveViewGlobalConnection error');
			self.connecting = false;
			console.error(err);
		}
	);
};

/**
 * Returns the connection to the LDM server. The connection process is asynchronous
 * so it may be that a request is made to get the connection either before the 
 * connection attempt happens, or while it is occurring. If the conneciton object
 * already exists, the promise is resolved immediately with that connection.
 * Otherwise the resolution is stored in a list of resolves that will be resolved
 * in connect() once the connection is obtained.
 * @param {Promise}
 */
LiveViewGlobalConnection.prototype.getConnection = function() {
	var self = this;
	
	var promise = new Promise(function(resolve, reject) {
		if(self.conn == null) {
			self.resolves.push(resolve);
			if(self.connecting == false)
				self.connect();
		}
		else {
			resolve(self.conn);
		}
	});
	
	return promise;
};

/**
 * Closes all connections to the LDM server.
 */
LiveViewGlobalConnection.prototype.close = function() {
	console.log('LiveViewGlobalConnection closing all connections');
	LiveView.closeAllConnections(
	).then(
		function(connection){
			console.log('LiveViewGlobalConnection disconnected');
		},
		function(error){
			console.error('LiveViewGlobalConnection error');
			console.error(error);
		}
	);
};

/**
 * Sends a tuple to the LDM server with the specified parameters.
 * @param {String} key - optional unique identifier key
 * @param {String} serverUri - server uri for the StreamBase server, e.g. sb://localhost:10000
 * @param {String} inputStream - name of the input stream
 * @param {Object} input - field map for the input tuple
 * @param {String} outputStream - optional name of the output stream, if the request will be 
 *   asynchronous, otherwise a response will be obtained from the output stream
 * @return {Promise}
 */
LiveViewGlobalConnection.prototype.sendTuple = function(key, serverUri, inputStream, input, outputStream) {		
	var self = this;
	var promise = new Promise(function(resolve, reject) {
		// Create a tuple
		var inputTuple = new LiveView.Tuple(null, input);
		
		// If the outputStream is specified, append a key
		if(outputStream != undefined)
			inputTuple.fieldMap.__SendTupleResponseKey = null;
		
		// Send the tuple
		self.conn.sendTuple(inputTuple, inputStream, outputStream, {
			serverUri: serverUri
		}).then(
			function(result){
				var output = undefined;				
				
				// If the outputStream is specified, remove the key
				if(outputStream != undefined) {
					output = result.tuple.fieldMap;
					delete output.__SendTupleResponseKey;
				}
				
				// Resolve with the result
				resolve(result);
			},
			function(error){
				reject(error);
			}
		);
	});
	
	return promise;
};

// Create a LV Global Connection object
var globalConn = new LiveViewGlobalConnection();



/**
 * Represents a LiveView JS API data source. This can be either a one-time or live data source.
 */
function LiveViewJSDatasource(configuration, dataset) {
	this.id 			= configuration.id;
	this.name			= configuration.name;
	this.description	= configuration.description;
	this.template		= configuration.template;
	this.parameters		= configuration.parameters;
	this.snap			= configuration.snapshot == undefined ? false : configuration.snapshot;		
	this.configuration	= configuration;
	this.dataset		= dataset;
	this.live			= true;
	
	this.conn			= null;
	this.tupleMap		= {};
};

/** 
 * Initializes the datasource after creation. Obtains the connection from the LiveViewGlobalConnection
 * and then either performs a snapshot() or a subscribe() depending on the query configuration.
 * The schema will be set on receiving the intial snapshot and all data objects will be sent through
 * to the listeners.
 * @return {Promise}
 */
LiveViewJSDatasource.prototype.init = function() {
	var self = this;
	var promise = new Promise(function(resolve, reject) {
		// Get the connection from the GlobalConnection
		globalConn.getConnection().then(
			// GlobalConnection getConnection resolve
			function(connection) {
				// Set the connection on the data source
				self.conn = connection;				
				
				// If this is a snapshot, then call snapshot, once completed
				//   resolve this promise
				if(self.snap == true) {
					self.snapshot().then(
						function(result) {
							resolve(self);
						},
						function(err) {
							console.error(err);
							resolve(self);
						}
					);
				}
				// Otherwise if subscription, then call subscribe, once completed
				//   resolve this promise
				else {
					self.subscribe().then(
						function(result) {
							resolve(self);
						},
						function(err) {
							console.error(err);
							resolve(self);
						}
					);
				}
			},
			
			// GlobalConnection getConnection reject
			function(err) {
				reject(err);
			}
		);
	});	
	
	return promise;	
};

/**
 * Creates a schema from a received data event. 
 * @param {Object} event - the data event
 */
LiveViewJSDatasource.prototype.createSchema = function(event) {
	this.schema = new DatasetSchema();
	for(var idx = 0; idx < event.schema.fields.length; idx++) {
		var f = event.schema.fields[idx].name;
		var t = event.schema.fields[idx].type;
		this.schema.addField(f, t);
	}
};

/**
 * Returns an object containing key fields from the given field map.
 * @param {Object} fieldMap - the field map
 * @return {Object}
 */
LiveViewJSDatasource.prototype.getKeyFields = function(fieldMap) {
	var obj = {};
	for(var idx = 0; idx < this.dataset.key.length; idx++) {
		var pk = this.dataset.key[idx];
		obj[pk] = fieldMap[pk];
	}
	return obj;
};

/**
 * Performs a snapshot query on the LDM server.
 * @return {Promise}
 */
LiveViewJSDatasource.prototype.snapshot = function() {
	var self = this;
	var promise = new Promise(function(resolve, reject) {
		// Execute the query on the server
		self.conn.execute(
			new LiveView.Query(self.template, self.parameters), {
				onSuccess: function(event){
					// Create the schema
					self.createSchema(event);
					
					// Notify the dataset of snapshot start
					self.dataset.trigger('snapshotStart', self);

					// Loop over the event data and notify the dataset of insert for each
					for(var idx = 0; idx < event.data.length; idx++) {
						var tuple = event.data[idx];
						self.dataset.trigger('insert', self, tuple.fieldMap);
					}					

					// Notify the dataset of snapshot end
					self.dataset.trigger('snapshotEnd', self);
					resolve();
				},
				onProgress: function(event){

				},
				onError: function(error){
					reject(error);
				}
			}
		).then(
			function(result){
			},
			function(error){
				reject(error);
			}
		);				
	});
	
	return promise;
};

/**
 * Performs a subscription query on the LDM server and stored the
 * query object on the datasource.
 * @return {Promise}
 */
LiveViewJSDatasource.prototype.subscribe = function() {
	var self = this;	
	
	var promise = new Promise(function(resolve, reject) {
		// Subscribe to the query on the server
		self.conn.subscribe(
			new LiveView.Query(self.template, self.parameters), {
				onSnapshotStart: function(event){		
					self.createSchema(event);
					self.dataset.trigger('snapshotStart', self);
				},
				onInsert: function(event){
					self.tupleMap[event.tuple.id] = event.tuple.fieldMap;
					self.dataset.trigger('insert', self, event.tuple.fieldMap);
				},
				onUpdate: function(event){
					self.dataset.trigger('update', self, event.tuple.fieldMap);
				},
				onDelete: function(event){
					var fieldMap = self.tupleMap[event.tuple.id];
					self.dataset.trigger('delete', self, fieldMap);
					delete self.tupleMap[event.tuple.id];
				},
				onSnapshotEnd: function(event){
					self.subscription = event.subscription;
					self.dataset.trigger('snapshotEnd', self);
					resolve();
				},
				onError: function(error){
					reject(error);
				}
			}
		).then(
			function(result){
				
			},
			function(error){
				reject(error);
			}
		);					
	});
	
	return promise;
};

/**
 * Unsubscribes the current query object on the LDM server.
 * @return {Promise}
 */
LiveViewJSDatasource.prototype.unsubscribe = function() {
	var self = this;	
	var promise = new Promise(function(resolve, reject) {
		self.conn.unsubscribe(
			self.subscription
		).then(
			function(result){
				delete self.schema;
				resolve(result);
			},
			function(error){
				delete self.schema;
				reject(error);
			}
		);					
	});
	
	return promise;
};

/**
 * Clear the existing tuple map and then calls a new snapshot or does an unsubscribe/subscribe
 * to do a full data reload from the LDM server.
 */
LiveViewJSDatasource.prototype.reload = function() {
	this.tupleMap = {};
	var self = this;
	if(this.snap == true)
		self.snapshot();
	else {
		self.unsubscribe().then(
			function(result) {
				self.subscribe();
			}
		);
	}
};

/**
 * Clears the existing tuple map and updates the parameters on this source. Then calls a new
 * snapshot or unsubscribe/subscribe to do a full data reload from the LDM server.
 */
LiveViewJSDatasource.prototype.applyParameters = function(parameters) {
	this.tupleMap = {};
	this.parameters = parameters;
	var self = this;
	
	// API apply parameters has a known issue, so call snap or unsub/sub as appropriate
	if(this.snap == true)
		self.snapshot();
	else {
		self.unsubscribe().then(
			function(result) {
				self.subscribe();
			},
			function(error) {
				//console.log(error);
			}
		);
	}
};

/**
 * Calls applyParameters() with the default configuration parameters.
 */
 LiveViewJSDatasource.prototype.resetParameters = function() {
	this.applyParameters(this.configuration.parameters);
};




/**
 * Represents a LiveView JS API data source. This can be a one-time data source only.
 */
function LiveViewTupleDatasource(configuration, dataset) {
	this.id 			= configuration.id;
	this.name			= configuration.name;
	this.description	= configuration.description;
	this.serverUri		= configuration.serverUri;
	this.inputStream	= configuration.inputStream;
	this.outputStream	= configuration.outputStream;
	this.parameters		= configuration.parameters;	
	this.configuration	= configuration;	
	this.dataset 		= dataset;
	this.live			= false;
};

/** 
 * Initializes the datasource after creation. Performs an invoke() by sending a message on the configured
 * serverUri and inputStream and then receives an asynchronous response on the configured outputStream.
 * The schema will be set from the response if present, otherwise it will rely on the schema being configured
 * in the data source configuration. Obtains the connection from the LiveViewGlobalConnection
 * and then sends the request tuple.
 * @return {Promise}
 */
LiveViewTupleDatasource.prototype.init = function() {
	var self = this;
	var promise = new Promise(function(resolve, reject) {
		self.invoke().then(
			function(result) {
				resolve(self);
			},
			function(err) {
				reject(err);
			}
		);					
	});	
	
	return promise;	
};

/**
 * Creates a schema from a received data event. If the event does not have a schema element then 
 * creates the schema from the data source configuration.
 * @param {Object} event - the data event
 */
LiveViewTupleDatasource.prototype.createSchema = function(event) {
	// Set the schema from the event
	if(event.schema != undefined) {
		this.schema = new DatasetSchema();
		for(var idx = 0; idx < event.schema.fields.length; idx++) {
			var f = event.schema.fields[idx].name;
			var t = event.schema.fields[idx].type;
			this.schema.addField(f, t);
		}
	}
	// Set the schema from configuration
	else {
		this.schema = new DatasetSchema();
		for(var idx = 0; idx < this.configuration.schema.length; idx++) {
			this.schema.addField(this.configuration.schema[idx].name, this.configuration.schema[idx].type);
		}
	}
};

/**
 * Performs a send tuple on the LDM server to an input stream with an output in an expected format.
 * @return {Promise}
 */
LiveViewTupleDatasource.prototype.invoke = function() {
	var self = this;
	var promise = new Promise(function(resolve, reject) {
		globalConn.getConnection().then(
			function(connection) {
				var input = {};
				if(self.parameters != null) {
					for(var idx = 0; idx < Object.keys(self.parameters).length; idx++) {
						var k = Object.keys(self.parameters)[idx];
						var v = self.parameters[k];
						input[k] = v;
					}
				}
				globalConn.sendTuple(null, self.serverUri, self.inputStream, input, self.outputStream).then(
					function(result) {
						// Resulting data will be in the following format:
						//   result.tuple.fieldMap
						//     schema - optional schema definition
						//     object - array of objects
						
						// Create the schema
						self.createSchema(result.tuple.fieldMap);
					
						// Notify the dataset of snapshot start
						self.dataset.trigger('snapshotStart', self);

						// Loop over the event data and notify the dataset of insert for each
						if(result.tuple.fieldMap.object.length > 0 && result.tuple.fieldMap.object[0] != null) {
							for(var idx = 0; idx < result.tuple.fieldMap.object.length; idx++) {
								var tuple = result.tuple.fieldMap.object[idx];
								self.dataset.trigger('insert', self, tuple);
							}					
						}
						
						// Notify the dataset of snapshot end
						self.dataset.trigger('snapshotEnd', self);
						resolve();
					},
					function(err) {
						reject(err);
					}
				);
			},
			function(err) {
				reject(err);
			}
		);
		
	});
	
	return promise;
};

/**
 * Calls a new invoke to do a full data reload from the server output stream.
 */
LiveViewTupleDatasource.prototype.reload = function() {
	this.invoke();
};

/**
 * Updates the parameters on this source. Then calls a new invoke to do a full data
 * reload from the server output stream.
 */
LiveViewTupleDatasource.prototype.applyParameters = function(parameters) {
	this.parameters = parameters;
	this.invoke();
};

/**
 * Calls applyParameters() with the default configuration parameters.
 */
LiveViewTupleDatasource.prototype.resetParameters = function() {
	this.applyParameters(this.configuration.parameters);
};

/**
 * Sends a tuple to the LDM server with the specified parameters.
 * @param {String} key - optional unique identifier key
 * @param {String} serverUri - server uri for the StreamBase server, e.g. sb://localhost:10000
 * @param {String} inputStream - name of the input stream
 * @param {Object} input - field map for the input tuple
 * @param {String} outputStream - optional name of the output stream, if the request will be 
 *   asynchronous, otherwise a response will be obtained from the output stream
 * @return {Promise}
 */
LiveViewTupleDatasource.sendTuple = function(key, serverUri, inputStream, input, outputStream) {
	return globalConn.sendTuple(key, serverUri, inputStream, input, outputStream);
}
