var Utilities = (function() {
	var module = {};

	/**
	 * Helper object for managing clipboard operations.
	 */
	Clipboard = {
		/**
		 * Copies the specified copy text into the clipboard.
		 * @param {String} copyText - the text to copy
		 * @param {Object} container - the DOM container element 
		 */
		copy: 			function(copyText, container) { 
							var input = $(document.createElement('textarea')).val(copyText);
							container.append(input);
							input.select();
							
							try {
								var successful = document.execCommand('copy');
								//var msg = successful ? 'successful' : 'unsuccessful';
								//console.log('Copying text command was ' + msg);
							} catch (err) {
								//console.log('Unable to copy');
							}

							$(input).remove();				
							window.getSelection().removeAllRanges();					
						}
	};

	/**
	 * Helper object for managing context menu operations.
	 */
	ContextMenu = {
		/**
		 * Displays a new context menu at the specified location with the 
		 * specified context.
		 * @param {Object} containerDiv - the DOM container element
		 * @param {Object} e - mouse event where the context menu action occurred
		 * @param {Object} configuration - configuration object
		 * @param {Objext} context - context object for evaluation
		 */
		display:		function(containerDiv, e, configuration, context) {
							var self = this;

							// Hide any context menus 
							$(document).find('.live-context-menu').remove();
							if(configuration == undefined) return;
							
							// Loop over the groups and process
							var prevCount = 0;
							var count = 0;
							var renderedNodes = [];
							for(var idx = 0; idx < Object.keys(configuration).length; idx++) {
								var thisGroup = Object.keys(configuration)[idx];
								var thisGroupConfiguration = configuration[thisGroup];
								var nodes = self.renderGroup(thisGroupConfiguration, context);
								count = count + nodes.length;
								if(prevCount > 0 && nodes.length > 0) 
									renderedNodes.push(self.renderDivider());
								for(var nodeIdx = 0; nodeIdx < nodes.length; nodeIdx++)
									renderedNodes.push(nodes[nodeIdx]);
								if(nodes.length > 0) prevCount = nodes.length;
							}
							
							if(count > 0) {
								// Create the menu and ul nodes and append to the container
								var menuNode = $(document.createElement('div')).attr('class', 'live-context-menu');
								$(containerDiv).append(menuNode);
								var ulNode = $(document.createElement('ul'));
								menuNode.append(ulNode);
								for(var idx = 0; idx < renderedNodes.length; idx++) {
									ulNode.append(renderedNodes[idx]);
								}
								
								// Calculate the position of the menu based on event location
								var offset = $(containerDiv).parent().offset();
								var x = e.pageX - offset.left;
								var y = e.pageY - offset.top;
								var width = menuNode.width();
								var height = menuNode.height();
								
								// Show the menu
								if(x + width + offset.left + 15 > window.innerWidth)
									x = x - width;
								if(y + height + offset.top + 15 > window.innerHeight)
									y = y - height;
								menuNode.css({position: 'absolute', left: x, top: y, display: 'block'});
								
								// Event listener to hide the menuNode once clicked
								menuNode.on('click', function(e) {
									menuNode.remove();
								});							
							}
							
						},
			
		/**
		 * Renders a group on a context menu. Groups are separated by dividers.
		 * @param {Object} configuration - configuration object for the group
		 * @param {Object} context - context object for evaluation
		 * @return {Object[]} an array of DOM nodes to add to the menu
		 */
		renderGroup:	function(configuration, context) {
							var self = this;
							
							var nodes = [];
							if(Array.isArray(configuration)) {
								for(var idx = 0; idx < configuration.length; idx++) {
									var thisItemConfiguration = configuration[idx];
									var itemNode = self.renderItem(thisItemConfiguration, context);
									if(itemNode != null) nodes.push(itemNode);
								}
							}
							else {
								var itemNode = self.renderItem(configuration, context);
								if(itemNode != null) nodes.push(itemNode);
							}
							return nodes;
						},
		
		/**
		 * Renders a divider for a group.
		 * @return {Object} the DOM element for the divider
		 */
		renderDivider:	function() {
							return $(document.createElement('li')).attr('class', 'divider');
						},
		
		/**
		 * Renders a menu item for a context menu.
		 * @param {Object} configuration - configuration object for the item
		 * @param {Object} context - context object for evaluation
		 * @return {Object} a DOM node to add, or null if no item is to be added
		 */
		renderItem:		function(configuration, context) {
							// Evalute the include condition to determine if the element is added
							var include = true;
							if(configuration.condition != undefined) {
								include = function(str){
									for(key in this) eval('var ' + key + ' = this.' + key);
									return eval('(' + str + ')');
								}.call(context, configuration.condition);
							}
							if(include == false) return;

							// Create the item node and the link node, and append
							var itemNode = $(document.createElement('li'));
							var linkNode = $(document.createElement('a')).attr('tabindex', '-1').attr('href', 'javascript:void(0)');
							itemNode.append(linkNode);

							// Setup the label for the menu item
							var label = configuration.label;
							
							// If there is a string template, evaluate the label
							if(label.indexOf('${') > -1) {
								label = function(str){
									for(key in this) eval('var ' + key + ' = this.' + key);
									return eval('`' + str + '`');
								}.call(context, label);
							}
							
							// Set the label text
							linkNode.text(label);
							
							// Append an event handler to execute the action
							itemNode.on('click', function(e) {
								var result = function(str){
									for(key in this) eval('var ' + key + ' = this.' + key);
									return eval('(' + str + ')');
								}.call(context, configuration.action);
							});
							
							return itemNode;
						}
	};
			
	/**
	 * Helper object for managing debounce.
	 */
	Debounce = {
		/**
		 * Debounce will wait for a specified period of time to capture
		 * keyboard events. Once the interval has expired, then the
		 * results will be passed to the callback.
		 * @param {Function} callback - function to execute when the debounce has expired
		 * @param {Integer} wait - time in msec to wait before calling the callback
		 * @param {Boolean} immediate - if true the callback will be called immediately
		 */
		debounce:	function(callback, wait, immediate) {
						var timeout;
						return function() {
							var context = this, args = arguments;
							var later = function() {
								timeout = null;
								if (!immediate) callback.apply(context, args);
							};
							var callNow = immediate && !timeout;
							clearTimeout(timeout);
							timeout = setTimeout(later, wait);
							if (callNow) callback.apply(context, args);
						};
					},
	}
		
	/**
	 * Helper object for managing evaluation of configurable objects.
	 */
	Evaluate = {
		/**
		 * Evaluates a string into an object that lists a series of classes.
		 * The string must start and end with single braces '{' and '}' otherwise 
		 * the input string will just be returned.
		 * The object keys are the class names and the object values should evaluate
		 * to either true or false. If the evaluation is true, then that class
		 * should be added.
		 * @param {String} classes - string definition for an object defining classes
		 * @param {String} context - the context to evaluate within
		 * @return {Object} object definining which classes to include
		 */
		classes:		function(classes, context) {
							if(classes.indexOf('{') == -1) return classes;
							return function(str){
								for(key in this) eval('var ' + key + ' = this.' + key);
								return eval('(' + str + ')');
							}.call(context, classes);
						},
			
		/**
		 * Evaluates a string expression based on the specified context.
		 * @param {String} expression - string definition for the expression
		 * @param {String} context - the context to evaluate within
		 * @param {Boolean} template - if true then the expression will evaluate as a string template
		 * @return {String} resulting value
		 */
		expression: 	function(expression, context, template) {
							return function(str){
								for(key in this) eval('var ' + key + ' = this.' + key);
								if(template == true)
									return eval('`' + str + '`');
								else
									return eval(str);
							}.call(context, expression);
						},
		/**
		 * Evaluates a string into an object that lists a series of styles.
		 * The string must start and end with single braces '{' and '}' otherwise 
		 * the input string will just be returned.
		 * The object keys are the CSS style attributes object values should evaluate
		 * to the value for those attributes.
		 * @param {String} styles - string definition for an object defining styles
		 * @param {String} context - the context to evaluate within
		 * @return {Object} object definining styles to include
		 */
		styles:			function(styles, context) {
							if(styles.indexOf('{') == -1) return styles;
							return function(str){
								for(key in this) eval('var ' + key + ' = this.' + key);
								return eval('(' + str + ')');
							}.call(context, styles);
						}
	};
	
	/**
	 * Helper object for managing array filters.
	 */
	Filtering = {
		/**
		 * Filters a list of items based on the pagination configuration.
		 * @param {Object[]} items - array of items to filter
		 * @param {Object} pagination - pagination definition
		 * @return {Object[]} subset of items matching the pagination config
		 */
		paginate: 	function(items, pagination){
						if(items == undefined || pagination == undefined) return items;
						var begin = pagination.begin == undefined ? 0 : pagination.begin;
						var end = pagination.end == undefined || pagination.end == 0 ? begin + pagination.size : pagination.end;
						return items.slice(begin, end);
					},
			
		/**
		 * Filters a list of items based on the specified filter object.
		 * The filter definition must be an object with the following schema:
		 *    filter = {
		 *	    search = {        {Object} search definition object
		 *        expression:     {String} - search expression to apply against the property
         *        fields:         {String[]} - list of fields of the object to search against		 
		 *      },
		 *	    filter = {
		 *        filters: [
		 *          {
		 *            expression:  {String} - filter expression to apply against the field
		 *            field:       {String} - field to apply the expression against
		 *          }
		 *        ],
		 *        fields:          {String[]} - list of fields of the object to filter against
		 *      }
		 *    }
		 * @param {Object[]} items - array of items to filter
		 * @param {Object} filter - filter definition
		 * @param {String} property - optional, property of the item to extract the object to 
		 *                              apply the filter against, if omitted then the filter is applied
		 *                              against the item itself
		 * @return {Object[]} subset of items matching the filter definition
		 */
		filter: 	function(items, filter, property){
						if(items == undefined || items.length == 0) return items;
						if(filter == undefined) return items;
						if(filter.search == undefined && (filter.filter == undefined || filter.filter.length == 0)) return items;
						var filtered = [];
						var self = this;
						angular.forEach(items, function(item) {
							var include = true;							
							if(filter.search.expression != undefined)
								include = include && self.searchItem(item, filter.search.expression, filter.search.fields, property);							
							if(filter.filter.filters.length > 0)
								include = include && self.filterItem(item, filter.filter.filters, filter.fields, property);
							if(include == true) filtered.push(item);
						});
						return filtered;
					},
					
		/**
		 * Searches an item based on a search specification. If the item matches, then 
		 * true will be returned. The search specification is a partial match and case
		 * insensitive.
		 * @param {Object} item - item to search
		 * @param {String} expression - search expression to match
		 * @param {String[]} fields - list of fields to search against
		 * @param {String} property - optional, property of the item to extract the object to 
		 *                              apply the search against, if omitted then the search is applied
		 *                              against the item itself
		 * @return {Boolean} true if the item matches the search, otherwise false
		 */
		searchItem:	function(item, expression, fields, property) {
						if(expression == undefined) return true;

						// Filter by the search box text
						//   Search all of the fields if specified, or all fields if not
						//   Simple inexact match, case insensitive
						var include = false;
						
						// Extract the object to search if property is specified
						var object = item;
						if(property != undefined) object = item[property];
						if(object == undefined) return false;
						
						// Search only specified fields
						if(fields != undefined && fields.length > 0) {
							for(var idx = 0; idx < fields.length; idx++) {
								var thisField = fields[idx];
								if(object[thisField] != undefined && (object[thisField] + "").toLowerCase().indexOf(expression.toLowerCase()) >= 0) {
									include = true;
									break;
								}
							}
						}
						// Otherwise search all fields
						else {
							for(var idx = 0; idx < Object.keys(object).length; idx++) {
								var thisField = Object.keys(object)[idx];
								if(thisField.startsWith('_')) continue;								
								if(object[thisField] != undefined && (object[thisField] + "").toLowerCase().indexOf(expression.toLowerCase()) >= 0) {
									include = true;
									break;
								}
							}
						}
						return include;
					},
			
		/**
		 * Filters an item based on a filter expression. If the item matches, then 
		 * true will be returned. The filter specification is either a partial match and case
		 * insensitive, or an expression exact match including the following operators:
		 *     ==, !=, >, >=, <, <=
		 * @param {Object} item - item to filter
		 * @param {Object[]} filters - list of filter objects to apply against the data
		 * @param {String[]} fields - list of fields to search against
		 * @param {String} property - optional, property of the item to extract the object to 
		 *                              apply the filter against, if omitted then the filter is applied
		 *                              against the item itself
		 * @return {Boolean} true if the item matches the filter, otherwise false
		 */
		filterItem:	function(item, filters, fields, property) {
						// Filter by list of filter expressions
						//   Supports inequality operators
						//     >, <, >=, <=, ==, !=
						//   Otherwise inexact match, case insensitive
						var include = true;
						
						// Extract the object to search if property is specified
						var object = item;
						if(property != undefined) object = item[property];
						if(object == undefined) return false;
						
						for(var idx = 0; idx < filters.length; idx++) {
							var thisFilter = filters[idx];
							if(fields != undefined && fields.indexOf(thisFilter.field) == -1) return true;
							var thisExpression = thisFilter.expression;
							var data = object[thisFilter.field];

							// Has exact expression
							if(thisExpression.indexOf('>') == 0 || thisExpression.indexOf('<') == 0 || thisExpression.indexOf('!') == 0 || thisExpression.indexOf('==') == 0) {
								var expression = thisExpression;//.replace(/([.*+?^=!:${}()|\[\]\/\\])/g, "\\$1");
								var expression = expression
									.replace(new RegExp('>', 'g'), ' data >')
									.replace(new RegExp('<', 'g'), ' data <')
									.replace(new RegExp('!=', 'g'), ' data !=')
									.replace(new RegExp('==', 'g'), ' data ==');							
								//console.log(data + " " + expression);
								try {
									if(data != undefined && eval(expression) == false) {
										include = false;
										break;
									}
								}
								catch(err) {
									//console.log(err);
									include = false;
									break;
								}
							}
							// Case insensitive partial match
							else if((data + "").toLowerCase().indexOf(thisExpression.toLowerCase()) == -1) {
								include = false;
								break;
							}
						}			
						return include;
					},
	};
				
	/**
	 * Helper object for managing formatting.
	 */
	Format = {
		/**
		 * Applies formatting against the specified data. If the format value 
		 * is not a number, then the data will be treated as a timestamp and 
		 * formatted using formatTimestamp(). Otherwise the data will be 
		 * treated as a numeric and will be formatted using formatNumber().
		 * @param {Object} data - data to format, typically will be a long in 
		 *                   the case of a timestamp, or an int, long, double
		 *                   in the case of a numeric
		 * @param {String|Number} format - formatting to apply
		 *                            in the case of a timestamp will be a string,
		 *                            otherwise will be a number
		 * @return {String} formatted timestamp or number
		 */
		format:				function(data, format) {
								if(data == undefined) return "";
								if(isNaN(format) == true)
									return this.formatTimestamp(data, format);
								else
									return this.formatNumber(data, format);
							},
							
		/**
		 * Applies number formatting against the specified data.
		 * @param {Number} data - data to format
		 * @param {Integer} decimals - number of decimals to format to
		 * @return {String} formatted number as a string
		 */
		formatNumber:		function(data, decimals) {
								if(decimals === undefined || data === null || data === undefined) return data;
								return parseFloat(data).toFixed(decimals);
							},
							
		/**
		 * Applies date/time formatting against the specified data.
		 * @param {Number} data - timestamp to format
		 * @param {String} format - date/time format
		 * @return {String} formatted timestamp as a string
		 */
		formatTimestamp:	function(data, format) {			
								if(data == undefined) return "";				
								if(format == undefined) return data;
								
								var date = data;
								if(!(date instanceof Date)) {
									date = new Date(date);
									// Firefox hack
									//date = new Date(data.replace(' ', 'T'));
								}

								// Array of short month names
								var shortMonth 	= ['Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec'];	
								var longMonth 	= ['January','February','March','April','May','June','July','August','September','October','November','December'];	
								
								// Array of date format tokens
								var o = {
									"yyyy+" : 	date.getFullYear(),
									"yy+" : 	(date.getFullYear() + "").substring((date.getFullYear() + "").length - 2),
									"dd+" : 	(date.getDate() < 10 ? "0" : "") + date.getDate(),
									"HH+" : 	(date.getHours() < 10 ? "0" : "") + date.getHours(),
									"mm+" : 	(date.getMinutes() < 10 ? "0" : "") + date.getMinutes(),
									"ss+" : 	(date.getSeconds() < 10 ? "0" : "") + date.getSeconds(),
									"S" : 		(date.getMilliseconds() < 10 ? "00" : date.getMilliseconds() < 100 ? "0" : "") + date.getMilliseconds(),
									// Month has to come last and has to be in this order due to the string substitution
									"MMMM+" : 	longMonth[date.getMonth()],
									"MMM+" : 	shortMonth[date.getMonth()],
									"MM+":		(date.getMonth() + 1 < 10 ? "0" : "") + (date.getMonth() + 1),
									"Z":		(date.getTimezoneOffset()),
								}

								var formatted = format;
								for(var k in o) {
									if(new RegExp("("+ k +")").test(formatted))
										formatted = formatted.replace(RegExp.$1,o[k]);
								}
									
								return formatted;
							}
	},
	
	
	/**
	 * Helper object for putting objects into a hierarchy.
	 */
	Hierarchy = {
		/**
		 * Inserts the specified component into the root hierarchy and 
		 * returns it. 
		 * @param {Object} object - the object to insert
		 * @param {Object} rootHierarchy - the root hierarchy object
		 * @param {Object} definition - the hierarchy definition
		 * @return hierarchy object
		 */
		getHierarchy:		function(object, rootHierarchy, definition) {
								var parentHierarchy = rootHierarchy;
								for(var idx = 0; idx < definition.length; idx++) {
									var hierarchyName = definition[idx];
									var hierarchyValue = object[hierarchyName];
									var member = parentHierarchy.members[hierarchyValue];
									if(member == null) {
										member = {
											level:		idx + 1,
											members:	{},
											values:		[],
											parent:		parentHierarchy
										}
										parentHierarchy.members[hierarchyValue] = member;
									}
									parentHierarchy = member;
								}
								
								// Parent hierarchy is now the target
								return parentHierarchy;
			
							},
		
		/**
		 * Adds the component to the hierarchy location.
		 * @param {Object} component - the UI component to add
		 * @param {Object} hierarchyLevel - the level to add it to 
		 */
		addComponent:		function(component, hierarchyLevel) {
								component._hierarchy = hierarchyLevel;
								hierarchyLevel.values.push(component);			
							},
							
		/**
		 * Updates the component in the hierarchy location. If the hierarchy has not changed,
		 * then no action is taken. If it has changed the component is removed from the old
		 * hierarchy and added to the new one.
		 * @param {Object} component - the UI component to add
		 * @param {Object} hierarchyLevel - the level to add it to if changed
		 */
		updateComponent:	function(component, hierarchyLevel) {
								if(component._hierarchy != hierarchyLevel) {
									var oldHierarchy = component._hierarchy;
									component._hierarchy.values.splice(component._hierarchy.values.indexOf(component), 1);
									hierarchyLevel.values.push(component);
									component._hierarchy = hierarchyLevel;
									this.trim(oldHierarchy);
								}			
							},
							
		/** 
		 * Removes the component from the hierarchy location.
		 * @param {Object} component - the UI component to remove
		 */
		removeComponent:	function(component) {
								if(component._hierarchy != undefined) {
									var oldHierarchy = component._hierarchy;
									component._hierarchy.values.splice(component._hierarchy.values.indexOf(component), 1);
									this.trim(oldHierarchy);
								}
							},
							
		/**
		 * Trims the specified hierarchy level from the parent if there are no values included.
		 * @param {Object} hierarchyLevel - the level to trim
		 */
		trim:				function(hierarchyLevel) {
								var parentHierarchy = hierarchyLevel.parent;
								if(parentHierarchy != null) {
									if(hierarchyLevel.values.length == 0 && Object.keys(hierarchyLevel.members).length == 0) {
										for(var idx = 0; idx < Object.keys(parentHierarchy.members).length; idx++) {
											var key = Object.keys(parentHierarchy.members)[idx];
											var val = parentHierarchy.members[key];
											if(val == hierarchyLevel) {
												delete parentHierarchy.members[key];
											}
										}
									}
									this.trim(parentHierarchy);
								}
							}
	};
	
	
	/**
	 * Helper object for managing array ordering.
	 */
	Ordering = {
		/**
		 * Sorts a list of items based on the specified ordering object.
		 * The order definition must be an object with the following schema:
		 *    order = [
		 *      {
		 *        field:     {String} - field name to order by
		 *        orderDir:  {String} - ordering direction, 'asc' or 'desc'
		 *      }
		 *    ]
		 * @param {Object[]} items - array of items to filter
		 * @param {Object[]} ordering - order definition
		 * @param {String} property - optional, property of the item to extract the object to 
		 *                              apply the ordering against, if omitted then the ordering is applied
		 *                              against the item itself
		 * @return {Object[]} ordered array of items matching the order definition
		 */
		order:		function(items, ordering, property) {			
						if(items == undefined || items.length == 0) return items;
						if(ordering == undefined || ordering.length == 0) return items;
						
						var ordered = [];
						angular.forEach(items, function(item) {
							ordered.push(item);
						});

						var compare = function(a, b, dir, field) {
							var aObj = a;
							var bObj = b;
							if(property != undefined) {
								aObj = a[property];
								bObj = b[property];
							}
							return dir * (aObj[field] == bObj[field] ? 0 : aObj[field] < bObj[field] ? 1 : -1);
						};
			
						// This is default sort algorithm
						/*ordered.sort(function (a, b) {
							for(var idx = 0; idx < ordering.length; idx++) {
								var thisOrdering = ordering[idx];
								var dir = thisOrdering.orderDir == 'desc' ? -1 : 1;
								var v = compare(a, b, dir, thisOrdering.field);
								if(v != 0) return v;
							}
							return 0;
						});*/
						// return ordered;
						// End default sort
						
						// This is merge sort algorithm, might be more efficient
						var algo = function(a, b) {
							for(var idx = 0; idx < ordering.length; idx++) {
								var thisOrdering = ordering[idx];
								var dir = thisOrdering.orderDir == 'desc' ? -1 : 1;
								var v = compare(a, b, dir, thisOrdering.field);
								if(v != 0) return v;
							}
							return 0;
						};
						
						var merge = function(left, right) {
							var result = [];
							while((left.length > 0) && (right.length > 0)) {		
								if(algo(left[0], right[0]) >= 0)
									result.push(left.shift());
								else
									result.push(right.shift());
							}
							result = result.concat(left, right);
							return result;
						};

						var sort = function(array) {
							var len = array.length;
							if(len < 2) return array;
							var pivot = Math.ceil(len/2);
							return merge(sort(array.slice(0,pivot)), sort(array.slice(pivot)));
						};

						return sort(ordered);
					}
	};
	
	/**
	 * Helper object for building property maps.
	 */
	PropertyMap = {
		index:		function(configuration, types) {
						if(configuration == null) return null;
						if(types.length == 0) return null;
						
						// Create a property index object
						var propertyIndex = {
							arr:	[],
							map:	{}
						};
						
						// Create a default properties object
						var properties = {};
						
						// Iterate over all the configuration keys and classify
						for(var idx = 0; idx < Object.keys(configuration).length; idx++) {
							var key = Object.keys(configuration)[idx];
							var type = key.substring(0, key.indexOf("_"));
							var prop = key.substring(key.indexOf("_")+1);
							if(types.indexOf(type) > -1) {
								propertyIndex.arr.push({prop: prop, type: type});
								propertyIndex.map[prop] = configuration[key];
							}
							else
								properties[key] = configuration[key];
						}

						return {
							propertyIndex: 	propertyIndex,
							properties:		properties
						};						
					},
					
		evaluate:	function(propertiesIndex, context, object, target) {
						if(propertiesIndex == null) return null;
						
						for(var idx = 0; idx < propertiesIndex.arr.length; idx++) {
							var prop = propertiesIndex.arr[idx].prop;
							var type = propertiesIndex.arr[idx].type;
							if(type == 'property') {
								var field = propertiesIndex.map[prop];
								target[prop] = object[field];
							}
							else if(type == 'eval') {
								var exp = propertiesIndex.map[prop];
								var val = DataFlow.Utilities.Evaluate.expression(exp, context);
								target[prop] = val;
							}
						}
						return target;
					}
	}
	
	/**
	 * Helper object for HTML template processing.
	 */
	Template = {
		process:	function(html) {
						var node = html;
						if($.type(node) === "string")
							node = $(document.createElement('div')).html(html);
						
						// Remove nodes that have if=false
						$(node).find("[if='false']").remove();
						
						return node;
					}
	}
	
	// Other objects
	module.Clipboard	= Clipboard;
	module.ContextMenu	= ContextMenu;
	module.Debounce		= Debounce;
	module.Evaluate		= Evaluate;
	module.Filtering	= Filtering;
	module.Format		= Format;
	module.Hierarchy	= Hierarchy;
	module.Ordering		= Ordering;
	module.PropertyMap	= PropertyMap;
	module.Template		= Template;
	
	return module;
})();

	