package com.streambase.mqtt.broker;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.streambase.sb.*;
import com.streambase.sb.operator.*;

import io.moquette.BrokerConstants;
import io.moquette.server.Server;

import java.io.IOException;
import java.util.Properties;
import java.util.concurrent.atomic.AtomicBoolean;

public class MoquetteMQTTBroker extends Operator implements Parameterizable {
	public static final long serialVersionUID = 1530125058624L;
	
    public static enum PayloadType {
    	STRING("String"),
        BLOB("Blob");

        private final String rep;

        private PayloadType(String rep) {
            this.rep = rep;
        }

        @Override
        public String toString() {
            return rep;
        }
    }
	
	public final static int PORT_PUBLISH = 0;
	public final static int PORT_CONNECT = 1;
	public final static int PORT_DISCONNECT = 2;
	public final static int PORT_CONNECTION_LOST = 3;
	public final static int PORT_SUBSCRIBE = 4;
	public final static int PORT_UNSUBSCRIBE = 5;
	public final static int PORT_MESSAGE_ACKNOWLEDGED = 6;
	
	public final static String DISPLAY_NAME = "Moquette MQTT Broker";
	public final static String SHORT_DISPLAY_NAME = "MQTT Broker";
	
	public final static String FIELD_COMMAND = "Command";
	
	public final static String COMMAND_START = "Start";
	public final static String COMMAND_STOP = "Stop";
	
	Logger logger;	
	private Server mqttBroker; 
	private String host = BrokerConstants.HOST;
	private int port = BrokerConstants.PORT;
	private boolean startOnStartup = true;
	private PayloadType payloadType = PayloadType.STRING;
	
	public MoquetteMQTTBroker() {
		super();
		logger = getLogger();
		setPortHints(1, 1);
		setDisplayName(DISPLAY_NAME);
		setShortDisplayName(SHORT_DISPLAY_NAME);
	}

	public void typecheck() throws TypecheckException {
		if (port <= 0) {
			throw new PropertyTypecheckException(MoquetteMQTTBrokerBeanInfo.PORT_PROPERTY, String.format("The %s must greater than 0", MoquetteMQTTBrokerBeanInfo.PORT_DISPLAY_NAME));
		}
		try {
			if (!getInputSchema(0).hasField(FIELD_COMMAND) || !getInputSchema(0).getField(FIELD_COMMAND).getDataType().equals(DataType.STRING)) {
				throw new TypecheckException(String.format("The input schema must contain a field named '%s' of type '%s'", FIELD_COMMAND, DataType.STRING.toString()));
			}
		} catch (TupleException e) {
			throw new TypecheckException(e);
		}
		
		setOutputSchema(PORT_PUBLISH, PublisherListener.createPublishSchema(getPayloadType()));
		setOutputSchema(PORT_CONNECT, PublisherListener.createConnectSchema());
		setOutputSchema(PORT_DISCONNECT, PublisherListener.createDisconnectSchema());
		setOutputSchema(PORT_CONNECTION_LOST, PublisherListener.createConnectionLostSchema());
		setOutputSchema(PORT_SUBSCRIBE, PublisherListener.createSubscribeSchema());
		setOutputSchema(PORT_UNSUBSCRIBE, PublisherListener.createUnsubscribeSchema());
		setOutputSchema(PORT_MESSAGE_ACKNOWLEDGED, PublisherListener.createMessageAcknowledgedSchema(getPayloadType()));

	}
	
	public void processTuple(int inputPort, Tuple tuple) throws StreamBaseException {
		String command = tuple.isNull(FIELD_COMMAND) ? "" : tuple.getString(FIELD_COMMAND);
		if (command.equals(COMMAND_START)) {
			startServer();
		} else if (command.equals(COMMAND_STOP)) {
			stopServer();
		} else {
			logger.warn(String.format("Invalid command, valid values are %s and %s", COMMAND_START, COMMAND_STOP));
		}

	}

	AtomicBoolean initialResume = new AtomicBoolean(true);
	@Override
	public void resume() throws Exception {
		if (initialResume.get()) {
			initialResume.set(false);
			if (startOnStartup) {
				startServer();
			}
		}
	}
	
	private void startServer() {
        Properties properties = new Properties();
        properties.put(BrokerConstants.PORT_PROPERTY_NAME, "" + port);
        properties.put(BrokerConstants.HOST_PROPERTY_NAME, host);
        properties.put(BrokerConstants.METRICS_ENABLE_PROPERTY_NAME, "false");
		((ch.qos.logback.classic.Logger)LoggerFactory.getLogger("io.moquette")).setLevel(ch.qos.logback.classic.Level.ERROR);
		((ch.qos.logback.classic.Logger)LoggerFactory.getLogger("messageLogger")).setLevel(ch.qos.logback.classic.Level.ERROR);
        
        mqttBroker = new Server();
        
        try {
			mqttBroker.startServer(properties);
			mqttBroker.addInterceptHandler(new PublisherListener(this));
			logger.info("MQTT Broker started on port '{}'", port);
		} catch (IOException e) {
			logger.error(String.format("Error starting MQTT Broker: %s", e.getMessage()), e);
		}
	}
	
	private void stopServer() {		
		mqttBroker.stopServer();
		mqttBroker = null;
		logger.info("MQTT Broker stopped on port '{}'", port);
	}
	
	public void shutdown() {
		stopServer();
	}
	
	public void setStartOnStartup(boolean startOnStartup) {
		this.startOnStartup = startOnStartup;
	}
	
	public boolean getStartOnStartup() {
		return startOnStartup;
	}
	
	public void setHost(String host) {
		this.host = host;
	}
	
	public String getHost() {
		return host;
	}
	
	public void setPort(int port) {
		this.port = port;
	}
	
	public int getPort() {
		return port;
	}
	
	public void setPayloadType(PayloadType payloadType) {
		this.payloadType = payloadType;
	}
	
	public PayloadType getPayloadType() {
		return payloadType;
	}	
}
