package com.streambase.accelerator.datasource.witsml;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.streambase.sb.*;
import com.streambase.sb.Schema.Field;
import com.streambase.sb.adapter.common.AdapterUtil;
import com.streambase.sb.operator.*;

public class MnemonicMapper extends Operator implements Parameterizable {

    public static final long serialVersionUID = 1542054325027L;
    private Schema dataSchema;
    private Schema mnemonicSchema;
    private String displayName = "Mnemonic Mapper";
    private String dataFieldName = "LogData";
    private Map<String, String> fieldMap = new HashMap<>();
    private Pattern dataPointPattern = Pattern.compile("(?i).*DataPoint\\.(.*?)(\\s|$).*");

    public MnemonicMapper() {
        super();
        setPortHints(1, 1);
        setDisplayName(displayName);
        setShortDisplayName(this.getClass().getSimpleName());
    }

    public void typecheck() throws TypecheckException {
        requireInputPortCount(1);
        if (dataSchema == null || dataSchema.getFieldCount() == 0) {
            throw new PropertyTypecheckException("dataSchema", "A data schema is required");
        }

        try {
            if (dataFieldName == null || dataFieldName.length() == 0 || !getInputSchema(0).hasField(dataFieldName) || !getInputSchema(0).getField(dataFieldName).getDataType().equals(DataType.TUPLE)) {
                throw new PropertyTypecheckException("dataFieldName", "The data field name must have a value and the field must exist in the input tuple and be of type tuple");
            }
            AdapterUtil.validateSchema(this, mnemonicSchema, getInputSchema(0).getField(dataFieldName).getSchema(), true, 0, "Data Port");
        } catch (TupleException e) {
            throw new TypecheckException("Error checking data field name type");
        }
        List<Field> outputFields = new ArrayList<>();
        for (int x = 0; x < getInputSchema(0).getFieldCount(); x++) {
            Field inputField = getInputSchema(0).getField(x);
            if (inputField.getName().equals(dataFieldName)) {
                List<String> matches = new ArrayList<>();
                for (Field mnemonicField : mnemonicSchema.fields()) {
                	Matcher matcher = dataPointPattern.matcher(mnemonicField.getDescription());
                    matches.add(mnemonicField.getDescription() + " Matches: " + matcher.matches() + "Count: " + matcher.groupCount());
                    if (matcher.matches() && matcher.groupCount() == 2 && dataSchema.hasField(matcher.group(1))) {
                        try {
                            Field outputDataField = dataSchema.getField(matcher.group(1));
                            matches.add(outputDataField.getName());
                            fieldMap.put(mnemonicField.getName(), outputDataField.getName());
                            if (!mnemonicField.getCompleteDataType().equals(outputDataField.getCompleteDataType())) {
                                throw new PropertyTypecheckException("dataSchema", String.format("Input field '%s' and output field '%s' must have the same datatype.", mnemonicField.getName(), outputDataField.getName()));
                            }
                        } catch (TupleException e) {
                            throw new TypecheckException(String.format("Error getting data schema field '%s'", matcher.group()));
                        }

                    } else {
                    	throw new TypecheckException(String.format("Mnemonic field '%s' has not corrisponding field in the data schema using the Mnemonic field description with pattern 'DataPoint.mnemonic'", mnemonicField.getName(), mnemonicField.getDescription()));
                    }
                }                
                outputFields.add(new Field(dataFieldName, matches.toString(), CompleteDataType.forTuple(dataSchema)));
            } else {
                outputFields.add(inputField);
            }
        }
        setOutputSchema(0, new Schema("", outputFields));
    }

    public void processTuple(int inputPort, Tuple tuple) throws StreamBaseException {
        Tuple outputTuple = getOutputSchema(0).createTuple();
        for (int x = 0; x < outputTuple.getSchema().getFieldCount(); ++x) {
            if (outputTuple.getSchema().getField(x).getName().equals(dataFieldName)) {
            	Tuple inputDataTuple = tuple.getTuple(x);
            	Tuple outputDataTuple = dataSchema.createTuple();
            	for (Field field : inputDataTuple.getSchema().fields()) {
            		if (fieldMap.containsKey(field.getName())) {
            			outputDataTuple.setField(fieldMap.get(field.getName()), inputDataTuple.getField(field));
            		}
            	}
                outputTuple.setTuple(x, outputDataTuple);
            } else {
                outputTuple.setField(x, tuple.getField(x));
            }
        }
        sendOutput(0, outputTuple);
    }

    public void setMnemonicSchema(Schema mnemonicSchema) {
        this.mnemonicSchema = mnemonicSchema;
    }

    public Schema getMnemonicSchema() {
        return this.mnemonicSchema;
    }

    public void setDataSchema(Schema dataSchema) {
        this.dataSchema = dataSchema;
    }

    public Schema getDataSchema() {
        return this.dataSchema;
    }

    public void setDataFieldName(String dataFieldName) {
        this.dataFieldName = dataFieldName;
    }

    public String getDataFieldName() {
        return this.dataFieldName;
    }
}
