var DataFlow = (function() {
	var module = {};

	
	
/**
 * Represents a dataset.
 * @constructor
 * @param {Object} container - container DOM object this dataset will be added to
 * @param {Object} configuration - configuration object defining the options
 *   for this dataset
 */
function Dataset(container, configuration) {
	// Parent container object containing this dataset
	this.container		= container;
	
	// Dataset properties from configuration
	this.id				= configuration.id;
	this.name			= configuration.name;
	this.description	= configuration.description;
	this.key			= configuration.key;
	this.grouped		= configuration.grouped;
	this.links			= configuration.links == undefined ? [] : configuration.links;
	
	// Full configuration object
	this.configuration	= configuration;
	
	// Schema for the dataset
	this.schema			= null;		
	
	// Vars to hold the objects being hovered and marked
	this.hovered		= null;		
	this.marked			= null;

	// List of sinks registered on this dataset
	this.sinks 			= [];
	
	// List of sources registered on this dataset
	this.sources 		= [];
	
	// List of data objects and a map of key to object for this dataset
	this.data			= [];
	this.dataMap		= {};

	// Map of grouped key to object for this dataset
	this.groupedDataMap	= {};
	
	// List of warnings for this dataset
	this.warnings		= [];
	
	// Live flat
	this.live			= false;
}

/**
 * Initializes the dataset by reading the configuration and then creating
 * and initializing all contained datasources.
 * @return {Promise} 
 */
Dataset.prototype.init = function() {	
	var self = this;
	var promise = new Promise(function(resolve, reject) {
		// Setup an object for datasource id to detect duplicates
		var duplicates = {};
		
		// Count the configured datasources. If there are none, resolve the promise
		var count = self.configuration.source == undefined ? 0 : self.configuration.source.length;
		if(count == 0) resolve(self);
		
		// Function to complete processing by decrementing the count
		//   Once the count reaches zero, rebuild the dataset schema and
		//   resolve the promise
		var complete = function() {
			count = count - 1;
			if(count == 0) {
				self.rebuildSchema();			
				resolve(self);
			}
		};
		
		// Loop over each source in the configuration
		for(var idx = 0; idx < self.configuration.source.length; idx++) {			
			// Get the configuration and validate
			var thisConfiguration = self.configuration.source[idx];
			var valid = DatasourceFactory.validate(thisConfiguration, idx, duplicates, self.warnings);

			// If valid, create and initialize the source, which will return it's own promise
			if(valid == true) {
				// Add the ID to the duplicate container
				duplicates[thisConfiguration.id] = thisConfiguration.id;
				
				// Create and initialize the datasource
				DatasourceFactory.createDatasource(thisConfiguration, self).init().then(
					function(source) {
						// Add to the source collection, then call complete
						if(source.live == true) 
							self.live = true;
						self.sources.push(source);
						complete();
					},
					function(err) {
						// Log the error, then call complete
						console.error(err);
						complete();
					}
				);
			}
			// Otherwise skip this source and just call complete
			else {
				complete();
			}
		}
	});
	
	return promise;
};

/**
 * Trigger an event on the dataset. Depending on the trigger type, all contained sinks
 * and/or sources will be notified with the appropriate information.
 * @param {String} eventType - string describing the event to trigger
 * @param {Object} eventSource - object event source for the event, can be either a 
 *   datasink or a datasource depending on the event type
 * @param {Object} object - data object associated with the event, may also be 
 *   null for certain event types
 */
Dataset.prototype.trigger = function(eventType, eventSource, object) {
	// Choose what to do based on trigger type

	// Snapshot start, notify all sinks of the event
	// eventSource - datasource that called trigger
	// object - null or undefined
	if(eventType == 'snapshotStart') {
		for(var idx = 0; idx < this.sinks.length; idx++) {
			this.sinks[idx].trigger(eventType);
		}
	}
	// Snapshot end, notify all sinks of the event
	// eventSource - datasource that called trigger
	// object - null or undefined
	else if(eventType == 'snapshotEnd') {
		for(var idx = 0; idx < this.sinks.length; idx++) {
			this.sinks[idx].trigger(eventType);
		}
	}
	// Insert, Update, Upsert, get the object key for the object and any existing
	//   instance matching the object key. If not found, create a new dataset object,
	//   otherwise merge the fields into the existing dataset object. 
	//   Note that insert/update/upsert are all handled the same here. An 'insert' 
	//   eventType might result in a merge if there are multiple datasources.
	// eventSource - datasource that called trigger
	// object - field map defining all fields on the insert or changed fields on the update
	else if(eventType == 'insert' || eventType == 'update' || eventType == 'upsert') {
		var objectKey = this.getObjectKey(object);
		var existing = this.dataMap[objectKey];

		if(existing == undefined)
			this.createNewDSObject(eventType, objectKey, object, eventSource);
		else
			this.mergeDSObject(eventType, objectKey, object, existing, eventSource);
	}
	// Delete, get the object key for the object and any existing instance matching the
	//   object key. If not found, do nothing, otherwise perform a dataset delete, 
	//   which may result in an internal update or a delete of the object depending
	//   on circumstances. See deleteDSObject for details.
	// eventSource - datasource that called trigger
	// object - field map defining at a minimum only key fields for the deleted object
	else if(eventType == 'delete') {
		var objectKey = this.getObjectKey(object);
		var existing = this.dataMap[objectKey];
		if(existing != undefined) {
			this.deleteDSObject(eventType, objectKey, object, existing, eventSource);
		}
	}
	// Clear, purge all data from the dataset, then notify all sinks to clear all data
	// eventSource - 
	// object - null or undefined
	else if(eventType == 'clear') {
		this.data = [];
		this.dataMap = {};
		for(var idx = 0; idx < this.sinks.length; idx++) {
			this.sinks[idx].trigger(eventType);
		}
	}
	// Hover, a data object is being hovered over in a UI element. If the dataset has
	//   any links defined, loop over the links and find any with eventType of hover.
	//   For each of these links, call apply link take any linking action on the
	//   target dataset.
	//   Also loop over all configured data sinks and notify all sinks of the event
	//     except for the sink that caused the trigger
	// eventSource - datasink that produced the hover event
	// object - object being hovered over, or null when the hover has stopped
	else if(eventType == 'hover') {
		if(eventSource != null && eventSource.dataset == this) {
			for(var idx = 0; idx < this.links.length; idx++) {
				var link = this.links[idx];
				if(link.trigger == 'hover') {
					this.applyLink(link, object);
				}
			}
		}

		for(var idx = 0; idx < this.sinks.length; idx++) {
			if(this.sinks[idx] != eventSource)
				this.sinks[idx].trigger(eventType, object);
		}
		
		this.hovered = object;
		if(this.container.onUpdate != undefined) 
			this.container.onUpdate();
	}
	// Marked, a data object is being marked in a UI element. If the dataset has
	//   any links defined, loop over the links and find any with eventType of marked.
	//   For each of these links, call apply link take any linking action on the
	//   target dataset.
	//   Also loop over all configured data sinks and notify all sinks of the event
	//     except for the sink that caused the trigger
	// eventSource - datasink that produced the marking event
	// object - object being marked, or null when the marking has stopped
	else if(eventType == 'marked') {
		if(eventSource != null && eventSource.dataset == this) {
			for(var idx = 0; idx < this.links.length; idx++) {
				var link = this.links[idx];
				if(link.trigger == 'marked') {
					this.applyLink(link, object);
				}
			}
		}

		for(var idx = 0; idx < this.sinks.length; idx++) {
			if(this.sinks[idx] != eventSource)
				this.sinks[idx].trigger(eventType, object);
		}
		
		this.marked = object;
		if(this.container.onUpdate != undefined) 
			this.container.onUpdate();
	}
	// Filter, a data set is being filtered in a UI element. If the dataset has
	//   any links defined, loop over the links and find any with eventType of filter.
	//   For each of these links, call apply link take any linking action on the
	//   target dataset.
	//   Also loop over all configured data sinks and notify all sinks of the event
	//     except for the sink that caused the trigger
	// eventSource - datasink that produced the marking event
	// filter - filter definition object with predicate that defines when a row
	//   is included in the result
	else if(eventType == 'filter') {
		for(var idx = 0; idx < this.sinks.length; idx++) {
			if(this.sinks[idx] != eventSource)
				this.sinks[idx].trigger(eventType, object);
		}
	}
	// Otherwise log a warning about an unknown event type
	else {
		console.warn('Dataset ' + this.id + ' unknown eventType ' + eventType);
	}
};

/** 
 * Registers a new sink on the dataset. This is called whenever a sink is added,
 * either dynamically or during application initialization.
 * @param {Object} sink - datasink to register on the dataset
 */
Dataset.prototype.registerSink = function(sink) {
	// Add the sink to the array
	this.sinks.push(sink);
	
	// Trigger a schema change, then a snapshot start on the sink
	sink.trigger('schemaChange', this.schema);
	sink.trigger('snapshotStart');
	
	// If there is any data in the dataset, iterate over and trigger
	//   an insert for each object on the sink
	if(this.data.length > 0) {
		for(var idx = 0; idx < this.data.length; idx++) {
			object = this.data[idx];
			sink.trigger('insert', object);
		}
	}
	
	// Trigger a snapshot end on the sink
	sink.trigger('snapshotEnd');
};

/** 
 * Unregisters an existing sink on the dataset. This will remove the sink
 * and it will stop receiving events from this dataset.
 * @param {Object} sink - datasink to unregister from the dataset
 */
Dataset.prototype.unregisterSink = function(sink) {
	// Get the index of the sink and remove it from the array, if it exists
	var idx = this.sinks.indexOf(sink);
	if(idx > -1) this.sinks.splice(idx, 1);
};

/**
 * Registers a new source on the dataset. This is called only when a source
 * is added dynamically and NOT during intialization. The sources will
 * automatically be added during initialization based on initial configuration.
 * @param {Object} source - the datasource to register on the dataset
 */
Dataset.prototype.registerSource = function(source) {
	// Add the source to the array, rebuild the schema, and reload all
	//   of the datasources. This will cause a full rebuild of the 
	//   dataset and all associated datasinks.
	source.init().then(
		function(source) {
			// Add to the source collection, then call complete
			if(source.live == true) 
				self.live = true;
			self.sources.push(source);
			self.rebuildSchema();	
			self.reload();
		},
		function(err) {
			// Log the error, then call complete
			console.error(err);
		}
	);
};

/**
 * Unregisters an existing sink on the dataset. This will remove the sink
 * and it will stop producing events for this dataset.
 * @param {Object} source - the datasource to unregister from the dataset
 */
Dataset.prototype.unregisterSource = function(source) {
	var idx = this.sources.indexOf(source);
	if(idx > -1) this.sources.splice(idx, 1);
	this.rebuildSchema();	
	this.reload();
};

/**
 * Rebuild the dataset schema from all the configured datasource schemas.
 * This will automatically be called on initialization once all datasources
 * have been added to the dataset. It will also automatically be called when
 * a new datasource is registered with the registerSource() function.
 */
Dataset.prototype.rebuildSchema = function() {
	// Create a new schema
	this.schema = new DatasetSchema();
	
	// Loop over each of the sources
	for(var sourceIdx = 0; sourceIdx < this.sources.length; sourceIdx++) {
		var source = this.sources[sourceIdx];
		
		// Get the source schema and throw an error if it is undefined
		var sourceSchema = source.schema;
		if(sourceSchema == undefined) 
			throw Error('Datasource ' + source.id + ' has no schema');
		
		// Loop over the source fields and add to the dataset schema
		//   Duplicate field names will overwrite existing configured fields
		//   therefore it's important that field configurations are consistent
		//   in particular for key fields which will always be duplicated
		for(var fieldIdx = 0; fieldIdx < sourceSchema.fields.length; fieldIdx++) {
			var f = sourceSchema.fields[fieldIdx];
			this.schema.addField(f.name, f.type);
		}
	}		
	
	// Trigger a schema change on all sinks
	for(var idx = 0; idx < this.sinks.length; idx++) {
		this.sinks[idx].trigger('schemaChange', this.schema);
	}
};

/**
 * Clear the existing set of data objects and reload all from each of the datasources.
 * This will automatically be called when a new datasource is registered with the 
 * registerSource() function.
 */
Dataset.prototype.reload = function() {
	// Clear all data from the local data
	this.data = [];
	this.dataMap = {};
	
	// Trigger a clear on all sinks, this will purge all existing rendered data
	for(var idx = 0; idx < this.sinks.length; idx++) {
		this.sinks[idx].trigger('clear');
	}

	// Run reload on all sources. This will cause a snapshotStart, insert set, snapshotEnd
	//   sequence to repeat which will rebuild the full set of data in the dataset
	//   and notify all sinks of these events.  
	for(var idx = 0; idx < this.sources.length; idx++) {
		this.sources[idx].reload();
	}	
};

/**
 * Builds the object key for the given object. The dataset must have a key configured
 * otherwise an error will be thrown during dataset creation.
 * @param {Object} object - the data object
 * @return {String} the object key
 */
Dataset.prototype.getObjectKey = function(object) {
	// The dataset will be validated at create time to ensure
	//   there is actually a key defined
	
	// Get the field value for the first key element and set this
	//   as the initial object key
	var objectKey = object[this.key[0]];
	if(objectKey == undefined) {
		console.warn('Dataset ' + this.id + ' object key not found for ' + this.key[0]);
		objectKey = "";
	}
	
	// For each remaining key elements, get the field value and concatenate
	//   to the object key with pipes between elements
	for(var idx = 1; idx < this.key.length; idx++) {
		var thisKey = object[this.key[idx]];
		if(thisKey == undefined) {
			console.warn('Dataset ' + this.id + ' object key not found for ' + this.key[idx]);
			thisKey = "";
		}
		objectKey = objectKey + '|' + thisKey;
	}
	
	return objectKey;
};

/**
 * Builds the grouped object key for the given object. The dataset group must have a key configured
 * otherwise an error will be thrown during dataset creation.
 * @param {Object} object - the data object
 * @return {String} the group object key
 */
Dataset.prototype.getGroupedObjectKey = function(object) {
	// The dataset will be validated at create time to ensure
	//   there is actually a group key defined
	
	// Get the field value for the first key element and set this
	//   as the initial object key
	var objectKey = object[this.grouped.key[0]];
	if(objectKey == undefined) {
		console.warn('Dataset ' + this.id + ' grouped object key not found for ' + this.grouped.key[0]);
		objectKey = "";
	}
	
	// For each remaining key elements, get the field value and concatenate
	//   to the group object key with pipes between elements
	for(var idx = 1; idx < this.grouped.key.length; idx++) {
		var thisKey = object[this.grouped.key[idx]];
		if(thisKey == undefined) {
			console.warn('Dataset ' + this.id + ' grouped object key not found for ' + this.grouped.key[idx]);
			thisKey = "";
		}
		objectKey = objectKey + '|' + thisKey;
	}
	
	return objectKey;
};

/** 
 * Apply a configured link against a target dataset in response to a UI action.
 * @param {Object} link - the link configuration to apply
 * @param {Object} object - the data object that caused the link, or null 
 */
Dataset.prototype.applyLink = function(link, object) {
	// Decision based on the action attribute of the link

	// Navigation
	if(link.action == 'navigation') {
		// Check to see if there is a navigate function on the container navigator object
		if(this.container.navigator == undefined) return;
		if(this.container.navigator.navigate == undefined) return;
		this.container.navigator.navigate(link.navigate);
	}
	// All others are dataset related
	else {
		// Get the linked dataset from the target property
		var linkedDataset = this.container[link.target];
		if(linkedDataset == undefined) return;
		
		// Parameters link action
		if(link.action == 'parameters') {
			// Create the target parameters
			var targetParameters = {};
			
			// If object is null, reset the parameters
			if(object == null)
				linkedDataset.resetParameters();
			// Otherwise evaluate the parameters
			else {
				for(var parameterIdx = 0; parameterIdx < Object.keys(link.parameters).length; parameterIdx++) {
					var parameterName = Object.keys(link.parameters)[parameterIdx];
					var parameterExpression = link.parameters[parameterName];
					var thisObject = object;
					if(Array.isArray(thisObject) == true) 
						thisObject = object[0];
					var source = thisObject;
					var parameterValue = eval(parameterExpression);
					targetParameters[parameterName] = parameterValue;
				}

				// Call apply on the linked dataset
				linkedDataset.applyParameters(targetParameters);
			}
		}
		// Other link actions, hovered and marked
		else {
			// Loop over the objects in the target dataset and apply conditions
			var targetObjects = [];
			if(object == null) {
				linkedDataset.trigger(link.action, this, null);
			}
			else {
				for(var datasetIdx = 0; datasetIdx < linkedDataset.data.length; datasetIdx++) {
					var target = linkedDataset.data[datasetIdx];
					if(Array.isArray(object)) {
						for(var objectIdx = 0; objectIdx < object.length; objectIdx++) {
							var source = object[objectIdx];
							var result = eval(link.condition);
							if(result == true)
								targetObjects.push(target);
						}
					}
					else {
						var source = object;
						var result = eval(link.condition);
						if(result == true)
							targetObjects.push(target);
					}
				}

				if(targetObjects.length == 0)
					targetObjects = null;
				else if(targetObjects.length == 1)
					targetObjects = targetObjects[0];
				linkedDataset.trigger(link.action, this, targetObjects);
			}
		}
	}
};

/**
 * Creates a new data object in the dataset from a field map. 
 * @param {String} eventType - string describing the event that triggered
 * @param {String} objectKey - internal unique identifier object key for the dataset object
 * @param {Object} fieldMap - field map defining all fields on the new object
 * @param {Object} source - datasource that produced the event causing this create
 */
Dataset.prototype.createNewDSObject = function(eventType, objectKey, fieldMap, source) {
	// Create the object and set the _key to the object key and set the _source to the
	//   id of the original source. Source will be used as part of the delete step.
	var object = {
		_key:		objectKey,
		_source:	[source.id],
	};
		
	// If grouped, add a group field
	var groupedObject = null;
	if(this.grouped != undefined) {
		object[this.grouped.property] = [];
	}

	// Set all the fields in the field map on the object
	for(var idx = 0; idx < Object.keys(fieldMap).length; idx++) {
		var fieldName = Object.keys(fieldMap)[idx];
		if(this.grouped != undefined && this.grouped.fields.indexOf(fieldName) > -1) {
			if(groupedObject == null) {
				groupedObject = {
					_key:		this.getGroupedObjectKey(fieldMap),
					_source:	[source.id]
				}				
			}
			groupedObject[fieldName] = fieldMap[fieldName];
		}
		else {
			object[fieldName] = fieldMap[fieldName];
		}
	}

	// Put the object in the data array and in the data map keyed on the object key
	this.data.push(object);
	this.dataMap[object._key] = object;
		
	// If there is a grouped object, add it to the appropriate arrays and maps
	if(groupedObject != null) {
		object[this.grouped.property].push(groupedObject);
		this.groupedDataMap[groupedObject._key] = groupedObject;
	}

	// Run callbacks on all sinks and notify of the inserted object
	for(var idx = 0; idx < this.sinks.length; idx++)
		this.sinks[idx].trigger('insert', object);	
};

/**
 * Merges a field map into an existing dataset object.
 * @param {String} eventType - string describing the event that triggered
 * @param {String} objectKey - internal unique identifier object key for the dataset object
 * @param {Object} fieldMap - field map defining all changed fields on the object
 * @param {Object} existing - existing dataset object
 * @param {Object} source - datasource that produced the event causing this create
 */
Dataset.prototype.mergeDSObject = function(eventType, objectKey, fieldMap, existing, source) {
	// Check the if the source is set on the object, if not set it now
	if(existing._source.indexOf(source.id) == -1)
		existing._source.push(source.id);
	
	// If grouped, retrieve the existing grouped field
	var groupedObject = null;
	var newGroupedObject = false;
	if(this.grouped != undefined) {
		groupedObject = this.groupedDataMap[this.getGroupedObjectKey(fieldMap)];
	}
	
	// Set all the fields in the field map on the object
	for(var idx = 0; idx < Object.keys(fieldMap).length; idx++) {
		var fieldName = Object.keys(fieldMap)[idx];
		if(this.grouped != undefined && this.grouped.fields.indexOf(fieldName) > -1) {
			if(groupedObject == null) {
				groupedObject = {
					_key:		this.getGroupedObjectKey(fieldMap),
					_source:	[source.id]
				}	
				newGroupedObject = true;
			}
			else {
				if(groupedObject._source.indexOf(source.id) == -1)
					groupedObject._source.push(source.id);
			}
			groupedObject[fieldName] = fieldMap[fieldName];
		}
		else {
			existing[fieldName] = fieldMap[fieldName];
		}
	}

	// If there is a new grouped object, add it to the appropriate arrays and maps
	if(groupedObject != null && newGroupedObject == true) {
		existing[this.grouped.property].push(groupedObject);
		this.groupedDataMap[groupedObject._key] = groupedObject;
	}

	// Run callbacks on all sinks and notify of the merged object
	for(var idx = 0; idx < this.sinks.length; idx++)
		this.sinks[idx].trigger('update', existing);		
};

/**
 * Deletes or updates an existing dataset object.
 * @param {String} eventType - string describing the event that triggered
 * @param {String} objectKey - internal unique identifier object key for the dataset object
 * @param {Object} fieldMap - field map defining all changed fields on the object
 * @param {Object} existing - existing dataset object
 * @param {Object} source - datasource that produced the event causing this create
 */
Dataset.prototype.deleteDSObject = function(eventType, objectKey, fieldMap, existing, source) {
	// If grouped, retrieve the existing grouped field
	var groupedObject = null;
	if(this.grouped != undefined) {
		groupedObject = this.groupedDataMap[this.getGroupedObjectKey(fieldMap)];
	}

	// If there is a grouped object, remove it now
	if(groupedObject != null) {
		delete this.groupedDataMap[groupedObject._key];
		existing[this.grouped.property].splice(existing[this.grouped.property].indexOf(groupedObject), 1);
	}
	
	// If not grouped, or all group elements are removed, then 
	//   remove all the other fields from the object based on the source 
	//   schema, unless the field is part of the key in which case leave it
	if(this.grouped == undefined || existing[this.grouped.property].length == 0) {
		for(var idx = 0; idx < source.schema.fields.length; idx++) {
			var field = source.schema.fields[idx];
			if(this.key.indexOf(field.name) == -1) {
				delete existing[field.name];
			}
			
			// Remove the source from the list of contributing sources
			var sourceIdx = existing._source.indexOf(source.id);
			if(sourceIdx > -1) existing._source.splice(sourceIdx, 1);
		}
		if(this.grouped != undefined) {
			delete existing[this.grouped.property];
			groupedObject = null;
		}
	}

	// If the contributing source array is length zero, then delete the whole object, 
	//   otherwise just update the sinks
	if(existing._source.length == 0) {
		delete this.dataMap[existing._key];
		var sourceIdx = this.data.indexOf(existing);
		if(sourceIdx > -1) this.data.splice(sourceIdx, 1);
		for(var idx = 0; idx < this.sinks.length; idx++)
			this.sinks[idx].trigger('delete', existing);		
	}
	else {
		for(var idx = 0; idx < this.sinks.length; idx++)
			this.sinks[idx].trigger('update', existing);		
	}
};

/**
 * Clears all data in the dataset and notifies all data sinks to clear data. Then applies
 * passed parameters on all data sources in the dataset. 
 * data which will be rebuilt from the data sources once the parameters are applied.
 * @param {Object} parameters - object defining the parameters to set
 */
Dataset.prototype.applyParameters = function(parameters) {
	// Clear all data from the local data
	this.data = [];
	this.dataMap = {};
	this.parameters = parameters;
	
	// Run sink.trigger clear data, hover, and marking
	for(var idx = 0; idx < this.sinks.length; idx++) {
		this.sinks[idx].trigger('clear');
		this.sinks[idx].trigger('hover', null);
		this.sinks[idx].trigger('marked', null);
	}
	
	this.hovered = null;
	this.marked = null;	
	if(this.container.onUpdate != undefined) 
		this.container.onUpdate();
	
	// Run source.applyParameters on each of the sources
	for(var idx = 0; idx < this.sources.length; idx++) {
		this.sources[idx].applyParameters(parameters);
	}	
};

/**
 * Clears all data in the dataset and notifies all data sinks to clear data. Then applies
 * default parameters on all data sources in the dataset.
 */
Dataset.prototype.resetParameters = function() {
	// Clear all data from the local data
	this.data = [];
	this.dataMap = {};
	this.parameters = {};
	
	// Run sink.trigger clear data, hover, and marking
	for(var idx = 0; idx < this.sinks.length; idx++) {
		this.sinks[idx].trigger('clear');
		this.sinks[idx].trigger('hover', null);
		this.sinks[idx].trigger('marked', null);
	}

	this.hovered = null;
	this.marked = null;	
	if(this.container.onUpdate != undefined) 
		this.container.onUpdate();

	// Run source.resetParameters on each of the sources
	for(var idx = 0; idx < this.sources.length; idx++) {
		this.sources[idx].resetParameters();
	}	
};
/** 
 * Factory object for creating dataset instances from file or configuration.
 */
DatasetFactory = {
	/** 
	 * Initializes the factory. Datasets are created from the uri or configuration 
	 * passed through in the options object.
	 * @param {Object} options - options object containing dataset creation options
	 * @return {Promise} 
	 */
	init: function(options) {
		var self = this;
		
		var promise = new Promise(function(resolve, reject) {
			// Validate the options
			if(options == undefined)
				reject(Error('DatasetFactory options are undefined'));
			if(options.container == undefined)
				reject(Error('DatasetFactory container is undefined'));		

			// Setup the container objects
			container 			= options.container;
			container.datasets 	= [];
			container.warnings	= [];
			
			// Options can be either uri or configuration
			// If uri, load configuration from the file which will return it's own promise
			// If both uri and configuration are specified then uri will take precedence
			if(options.uri != null) {
				self.loadFile(options.uri).then(
					function(configuration) {
						// Process the configuration loaded from the file
						self.processConfiguration(configuration, resolve, reject);
					},
					function(err) {
						// Error happened loading the file, reject with the passed error
						reject(Error('URI ' + err.message));
					}
				);
			}
			// If configuration object then process that configuration
			else if(options.configuration != null) {
				self.processConfiguration(options.configuration, resolve, reject);
			}
			// Otherwise reject with an error for unknown configuration method
			else {
				reject(Error('DatasetFactory uri and configuration are both undefined'));		
			}
		});	
		
		return promise;
	},
	
	/** 
	 * Processes the configuration object. This can be passed directly or loaded
	 * from URI.
	 * @param {Object} configuration - the configuration object
	 */
	processConfiguration: function(configuration, resolve, reject) {
		var self = this;
		
		// Setup an object to hold dataset id to detect duplicates
		var duplicates = {};
		
		// Count the configured datasets. If there are none, resolve the promise
		var count = configuration.dataset == undefined ? 0 : configuration.dataset.length;
		if(count == 0) resolve();
		
		// Function to complete processing
		//   Once the count reaches zero, resolve the promise
		var complete = function() {
			count = count - 1;
			if(count == 0) resolve();
		};
		
		// Loop over each dataset in the configuration
		for(var idx = 0; idx < configuration.dataset.length; idx++) {
			// Get the configuration and validate
			var thisConfiguration = configuration.dataset[idx];
			var valid = self.validate(thisConfiguration, idx, duplicates, container.warnings);
			
			// If valid, create and initialize the dataset which will return it's own promise
			if(valid == true) {
				// Add the ID to the duplicate container
				duplicates[thisConfiguration.id] = thisConfiguration.id;
				
				// Create and initialize the dataset			
				new Dataset(container, thisConfiguration).init().then(
					function(dataset) {
						// Add the dataset to the container, then call complete
						container[dataset.id] = dataset;
						container.datasets.push(dataset);
						complete();
					},
					function(err) {
						// Log the error, then call complete
						console.error(err);
						complete();
					}
				);
			}
			// Otherwise skip this dataset and just call complete
			else {
				complete();
			}
		}
	},
	
	/** 
	 * Loads dataset configuration from a file.
	 * @param {String} uri - the uri for the file
	 * @return {Promise}
	 */
	loadFile: function(uri) {
		var self = this;
		var promise = new Promise(function(resolve, reject) {
			// Load the file asynchronously
			$.ajax({
				type: 'GET',
				url: uri,
				dataType: 'json',
				cache: false,
				success: function(data, textStatus, xhr) {
					// Resolve with the loaded data
					resolve(data);
				},
				error: function(xhr, textStatus, errorThrown) {
					// Reject with an error
					reject(Error(xhr.statusText));
				}
			});	
		});
		
		return promise;
	},
	
	/**
	 * Validates the configuration for a single dataset config.
	 * @param {Object} configuration - the configuration for the individual dataset
	 * @param {Integer} index - the index of the dataset in the overall configuration
	 * @param {Object} duplicates - object containing dataset keys to detect duplicates
	 * @param {String[]} warnings - list of warnings
	 * @param {Boolean} true if the configuration is valid, otherwise false
	 */
	validate: function(configuration, index, duplicates, warnings) {
		var valid = true;
		
		// Check the dataset has an id
		if(configuration.id == undefined) {
			warnings.push('Dataset ' + index + ' has no id');
			valid = false;
		}
		
		// Check the dataset id is not duplicated
		if(duplicates[configuration.id] != undefined) {
			warnings.push('Dataset ' + configuration.id + ' already exists');
			valid = false;
		}
		
		// Check the dataset has a key defined and it is an array
		if(configuration.key == undefined || Array.isArray(configuration.key) == false || configuration.key.length == 0) {
			warnings.push('Dataset ' + configuration.id + ' has no key or invalid key');
			valid = false;
		}
		
		// Check that if the dataset has a grouped config
		if(configuration.grouped != undefined) {
			// Check the grouped has a property 
			if(configuration.grouped.property == undefined || configuration.grouped.property.length == 0) {
				warnings.push('Dataset ' + configuration.id + ' grouped has no property');
				valid = false;
			}
			
			// Check the grouped has a key and it is an array
			if(configuration.grouped.key == undefined || Array.isArray(configuration.grouped.key) == false || configuration.grouped.key.length == 0) {
				warnings.push('Dataset ' + configuration.id + ' grouped has no key or invalid key');
				valid = false;
			}

			// Check the grouped has fields and it is an array
			if(configuration.grouped.fields == undefined || Array.isArray(configuration.grouped.fields) == false || configuration.grouped.fields.length == 0) {
				warnings.push('Dataset ' + configuration.id + ' grouped has no fields or invalid fields');
				valid = false;
			}
		}
		return valid;
	},
	
};
/**
 * Represents a schema for a dataset or a datasource.
 * @constructor
 */
function DatasetSchema() {
	this.fields 	= [];
	this.fieldMap 	= {};
};

/**
 * Adds a new field to the schema. If the field with the given name
 * already exists then the type will be updated. Otherwise a new field
 * will be added.
 * @param {String} name - field name
 * @param {String} type - field data type
 */
DatasetSchema.prototype.addField = function(name, type) {
	// Retrieve the existing field.
	var field = this.fieldMap[name];
	
	// If field is not found, create it and add to the schema
	if(field == undefined) {
		field = {
			name:	name,
			type:	type
		};
		this.fieldMap[name] = field;
		this.fields.push(field);
	}
	// Otherwise update the type
	else {
		field.type = type;
	}
};
/**
 * Represents a grid data sink.
 * @constructor
 * @param {Object} dataset - dataset that this sink will be attached to
 * @param {Object} configuration - configuration object defining the sink
 */
function GridDatasink(id, configuration, dataset, containerDiv, overlay) {
	this.id = id;
	this.dataset = dataset;
	this.grid = new LiveGrid.Grid(id, configuration, containerDiv, overlay, this);
}

/**
 * Event handler for dataset events. 
 * @param {String} eventType - string describing the event that occurred
 * @param {Object} object - data object for the event, or null
 */
GridDatasink.prototype.trigger = function(eventType, object) {
	this.grid.trigger(eventType, object);
};


// Define a module to hold object defintions for the grid objects
var LiveGrid = (function() {
	var module = {};

	/**
	 * Creates a new Grid object.
	 * @constructor
	 * @param {String} id - unique identifier
	 * @param {Object} configuration - the configuration for the grid
	 * @param {Object} containerDiv - this node object that grid will be appended
	 * @param (Function} overlay - function to call on start and end of the overlay
	 * @param {Object} sink - datasink that will feed this grid
	 */
	function Grid(id, configuration, containerDiv, overlay, sink) {
		this.id				= id;
		this.configuration	= configuration;
		this.containerDiv 	= containerDiv;
		this.overlay		= overlay;
		this.sink			= sink;
		
		// Pause and snapshot flags
		this.paused			= false;
		this.snapshot		= false;
		
		// Schema for the dataset
		this.schema			= null;		
		
		// Vars to hold the objects being hovered and marked
		this.hovered		= [];		
		this.marked			= [];
		
		// Filter definition
		this.filterDef		= {
			search: {
				expression:	undefined,
				fields:		this.configuration.options.search.fields
			},
			filter:	{
				filters:	[],
				fields:		this.configuration.options.filter.fields
			}
		};
		
		// Flag for filters being visible
		this.filtersVisible	= false;
		
		// Ordering definition
		this.orderDef		= [];
		
		// Pagination defintion
		this.pagination		= null;
		
		// Color maps
		this.colorMaps		= {};

		// Columns
		this.columnMap		= {};
		this.columnArr		= [];
		
		// Rows
		this.rowMap			= {};
		this.rowArr			= [];
		
		// Ordered, filtered, and final result rows
		this.orderedArr		= [];
		this.filterArr		= [];
		this.resultArr		= [];
		
		// Check dependencies. This will throw Error if dependencies are not satisfied.
		this.dependencyCheck();
		
		// Microbatch is used to batch up updates to prevent too many chart redraws
		this.microbatch = {
			rtime:		new Date(1, 1, 2000, 12,00,00),
			timeout:	false,
			delta:		500 // msec between updates
		}

		// Color maps
		if(this.configuration.options.colormaps != undefined) {
			for(var idx = 0; idx < Object.keys(this.configuration.options.colormaps).length; idx++) {
				var thisColorMapName = Object.keys(this.configuration.options.colormaps)[idx];
				var thisColors = this.configuration.options.colormaps[thisColorMapName];
				this.colorMaps[thisColorMapName] = new DataFlow.ColorMap(thisColorMapName, thisColors);
			}
		}
		
		// Pagination
		if(this.configuration.options.pagination.enabled == true) {
			this.pagination = {
				size:	configuration.options.pagination.size,
				begin: 	0,
				end:	configuration.options.pagination.size
			};
		}
		
		// On document ready, register UI handlers
		var self = this;
		setTimeout(function() {
			self.registerEventHandlers();
		}), 0;
		
		// Draw the initial empty grid
		this.draw();
	};

	/**
	 * Performs a dependency check for rendering the grid.
	 */
	Grid.prototype.dependencyCheck = function() {
		// No dependencies for this component
	};
		
	/**
	 * Registers event handlers on UI components.
	 */
	Grid.prototype.registerEventHandlers = function() {
		var self = this;
		
		// Event handler to hide any open context menu on click
		$(document).on('click', function() {
			$(document).find('.live-context-menu').remove();
		});
		
		// Event handler for horizontal scrolling to bind the position of header row with
		//   body of the grid
		this.lastLeftPos = $(this.containerDiv).find('.grid-rows').scrollLeft();
		$(this.containerDiv).find('.grid-rows').scroll(function () {
			var currLeftPos = $(this).scrollLeft();
			if(self.lastLeftPos != currLeftPos) {
				$(self.containerDiv).find('.header-table').offset({ left: -1 * this.scrollLeft + 11});
				self.lastLeftPos = currLeftPos;
			}
			$(document).find('.live-context-menu').remove();
		});
		
		// Event handler to show/hide the header filter row
		$(this.containerDiv).find('.grid-filtered').on('click', function() {
			self.toggleFilters();
		});
		
		// Event handler to display the configuration modal
		$(this.containerDiv).find('.grid-config').on('click', function(e) {
			self.displayConfiguration();
		});

		// Event handler to display the dataset modal
		$(this.containerDiv).find('.grid-icon').on('click', function(e) {
			self.displayDataset();
		});
		
		// Event handler to apply search
		$(this.containerDiv).find('.grid-search').find('input').on('input', DataFlow.Utilities.Debounce.debounce(function(e) {
			var val = $(e.target).val();
			self.search(val);
		}, this.configuration.options.search.delay));

		// Event handler to clear search
		$(this.containerDiv).find('.grid-search').find('.grid-search-clear').on('click', function(e) {
			$(self.containerDiv).find('.grid-search').find('input').val('');
			self.search('');
		});

		// Event handler to clear hover when mouse moves out of boundary for the grid
		if(this.configuration.options.hover.enabled != false) {
			$(this.containerDiv).on('mouseout', function() {
				self.hover(null, true);
			});
		}	
		
		// Event handler for adjusting pagination on scroll
		if(this.configuration.options.pagination.enabled == true) {
			var dataTableDiv = $(this.containerDiv).find('.data-table');
			$(this.containerDiv).find('.grid-rows').scroll(function(e) {
				var scrollPercentage = $(this).scrollTop() / (dataTableDiv.height() - $(this).height());
				if(scrollPercentage >= 0.75 && self.resultArr.length < self.rowArr.length) {
					self.pagination.end = self.pagination.end + self.pagination.size;
					self.update(true);
				}
			});			
		}

		// Event handler to draw selection area for multiple marking
		if(this.configuration.options.marking.enabled != false && this.configuration.options.marking.mode == 'multiple') {
			$(self.containerDiv).on('dragstart', function(e) { return false; });
			$(self.containerDiv).find('.data-table').on('mousedown', function(e) {				
				// Function to update the size and position of the selection area
				var positionSelectionArea = function(startY, endY, dataTableDiv, gridRowsDiv, selectionAreaDiv, startScrollTop) {
					if(endY + (gridRowsDiv.scrollTop() - startScrollTop) - startY > 0) {
						//console.log("DN " + startY + " -> " + endY + " --- " + gridRowsDiv.scrollTop() + " " + gridRowsDiv.height() + " " + dataTableDiv.height());
						if(endY >= dataTableDiv.height())
							endY = dataTableDiv.height();
						selectionAreaDiv.css({height: endY + (gridRowsDiv.scrollTop() - startScrollTop) - startY});
					}
					else { 
						//console.log("UP " + startY + " -> " + endY + " --- " + gridRowsDiv.scrollTop() + " " + gridRowsDiv.height() + " " + dataTableDiv.height());
						selectionAreaDiv.css({top: endY + (gridRowsDiv.scrollTop() - startScrollTop), height: startY - (endY + (gridRowsDiv.scrollTop() - startScrollTop))});
					}
				};
				
				// Create a selection area div
				var appended = false;
				var selectionAreaDiv = $(document.createElement('div')).attr('class', 'selection-area');
				
				// Data table, containing all the rows
				var dataTableDiv = $(self.containerDiv).find('.data-table');
				var offset = dataTableDiv.offset();
				
				// Grid rows, the scrolling container
				var gridRowsDiv = $(self.containerDiv).find('.grid-rows');				
				var startScrollTop = gridRowsDiv.scrollTop();

				// Get initial start and end position
				var startY = e.pageY - offset.top;	
				var endY = null;
				var ctrlKey = e.ctrlKey;
				
				// Set intial position for the selection area
				selectionAreaDiv.css({'left': 0, 'top': startY});
				
				// Mouse exits the scroll area, create an auto-scroll
				var interval = null;
				gridRowsDiv.on('mouseleave', function(e) {
					// Compute the direction
					var direction = e.pageY - offset.top < startY ? 'moveup' : 'movedown';
					
					// If already auto-scrolling don't start another one
					if(interval == null) {
						var thisStartScroll = gridRowsDiv.scrollTop(); 
						interval = setInterval(function() {
							var scrollDelta = gridRowsDiv.scrollTop() - thisStartScroll;
							if(direction == 'movedown' && endY + scrollDelta < dataTableDiv.height() - 1) {
								gridRowsDiv.scrollTop(gridRowsDiv.scrollTop() + 10);
							}
							else if(direction == 'moveup') {
								gridRowsDiv.scrollTop(gridRowsDiv.scrollTop() - 10);
							}
							else {
								clearInterval(interval);
								interval = null;
							}
							
							// Update the selection area
							positionSelectionArea(startY, endY, dataTableDiv, gridRowsDiv, selectionAreaDiv, startScrollTop);
						},50);
					}
				});

				// Mouse enters the scroll area, cancel any auto-scroll
				gridRowsDiv.on('mouseenter', function(e) {
					if(interval != null) {
						clearInterval(interval);
						interval = null;
					}
				});
				
				// Move the mouse in the scroll area
				gridRowsDiv.on('mousemove', function(e) {
					endY = e.pageY - offset.top;
					if(appended == false && Math.abs(startY - endY) > 10) {
						dataTableDiv.append(selectionAreaDiv);				
						appended = true;
					}
					
					// Update the selection area
					positionSelectionArea(startY, endY, dataTableDiv, gridRowsDiv, selectionAreaDiv, startScrollTop);
				});
				
				// Complete the select
				$(document).on('mouseup', function() {
					// If the selection area exists, execute marking
					if(selectionAreaDiv.height() > 0) {

						// Create array of objects to mark
						var objects = [];

						// If CTRL key is down, then pre-populate the array with existing
						if(ctrlKey == true) {
							for(var idx = 0; idx < self.marked.length; idx++) 
								objects.push(self.marked[idx]);
						}
						
						// Loop over all rows and mark those inside the selection area
						var selectionStart = selectionAreaDiv.offset().top - offset.top + gridRowsDiv.scrollTop();
						var selectionEnd = selectionStart + selectionAreaDiv.height();
						for(var idx = 0; idx < Object.keys(self.rowMap).length; idx++) {
							var id = Object.keys(self.rowMap)[idx];
							var row = self.rowMap[id];
							if(row.node == null) continue;
							var rowStart = row.node.offset().top - offset.top + gridRowsDiv.scrollTop();
							var rowEnd = rowStart + row.node.height();
							if(rowEnd >= selectionStart && rowStart <= selectionEnd && objects.indexOf(row.object) == -1) {
								objects.push(row.object);
							}
						}
						self.mark(objects, true);
					}
					
					// Remove the selection area
					selectionAreaDiv.remove();
					
					// If there is an auto-scroll, then clear it
					if(interval != null) {
						clearInterval(interval);
						interval = null;
					}					
					
					// Turn off other listeners
					$(document).off('mouseup');
					gridRowsDiv.off('mousemove');
					gridRowsDiv.off('mouseenter');
					gridRowsDiv.off('mouseleave');
				});
			});
		}			
		
	};
	
	/**
	 * Event handler for all objects delegated from the datasink.
	 * @param {String} eventType - the event type
	 * @param {Object} object - optional object
	 */
	Grid.prototype.trigger = function(eventType, object) {
		var self = this;

		if(eventType == 'schemaChange') {
			this.schema = object;			
			this.buildColumnModel();
		}
		else if(eventType == 'snapshotStart') {
			this.snapshot = true;
			this.overlay(true);
		}
		else if(eventType == 'snapshotEnd') {
			setTimeout(function() {
				self.snapshot = false;
				self.overlay(false);
			},500);
		}
		else if(eventType == 'clear') {
			this.clear();
		}
		else if(eventType == 'insert') {
			this.addRow(object);
		}
		else if(eventType == 'update') {
			this.updateRow(object);
		}
		else if(eventType == 'upsert') {
		}
		else if(eventType == 'delete') {
			this.removeRow(object);
		}
		else if(eventType == 'hover') {
			if(this.configuration.options.hover.accept != false)
				this.hover(object);
		}
		else if(eventType == 'marked') {
			if(this.configuration.options.marking.accept != false)
				this.mark(object);
		}					
		else if(eventType == 'filter') {
			if(this.configuration.options.filter.accept != false)
				this.filter(object);
		}					
	};

	/**
	 * Builds the column model based on the column configuration.
	 */
	Grid.prototype.buildColumnModel = function() {
		// Clear the header row
		var headerRowDiv = $(this.containerDiv).find('.header-row');
		headerRowDiv.empty();

		// Clear the header filter row
		var headerFilterRowDiv = $(this.containerDiv).find('.header-filter-row');
		headerFilterRowDiv.empty();

		// Reset the column list and then build new column objects
		this.columnMap = {};
		this.columnArr = [];
		for(var idx = 0; idx < this.configuration.columns.length; idx++) {
			var column = new Column(this.configuration.columns[idx], this, headerRowDiv, headerFilterRowDiv);
			this.columnMap[column.field] = column;
			this.columnArr.push(column);
			column.render();
		}
	};
		
	/**
	 * Add a new row bound to the specified object on the grid.
	 * @param {Object} object - the object to add the bound row
	 */
	Grid.prototype.addRow = function(object) {
		var row = new Row(object, this.columnArr, this);
		this.rowMap[row._key] = row;
		this.rowArr.push(row);
		this.update();
	};

	/**
	 * Updates the row bound to the specified object on the grid.
	 * @param {Object} object - the object to update the bound row
	 */
	Grid.prototype.updateRow = function(object) {
		var row = this.rowMap[object._key];
		if(row == undefined) return;		
		row.object = object;
		row.update();
		this.update();
	};

	/**
	 * Removes the row bound to the specified object from the grid.
	 * @param {Object} object - the object to remove the bound row
	 */
	Grid.prototype.removeRow = function(object) {
		var row = this.rowMap[object._key];
		delete this.rowMap[object._key];
		this.rowArr.splice(this.rowArr.indexOf(row), 1);
		row.node = null;
		this.update();
	};

	/**
	 * Clears all objects and all rows from the grid.
	 */
	Grid.prototype.clear = function() {	
		this.rowMap			= {};
		this.rowArr			= [];
		this.update();
	};

	/**
	 * Draws the UI components for the grid.
	 */
	Grid.prototype.draw = function() {
		var self = this;
		var context = {};
		
		// Set classes on the container div
		if(this.configuration.options.classes != undefined) {
			$(this.containerDiv).addClass(DataFlow.Utilities.Evaluate.classes(this.configuration.options.classes, context));
		}
		
		// Set styles on the container div, always must be an object
		if(this.configuration.options.styles != undefined) {
			var styles = DataFlow.Utilities.Evaluate.styles(this.configuration.options.styles, context);
			for(style in styles) $(this.containerDiv).css(style, styles[style]);
		}	
	};

	/**
	 * Updates the contents of the grid. Normally this will happen in a microbatch
	 * to prevent too many updates from occurring. However it is possible to trigger
	 * an immediate update by setting the flag.
	 * @param {Boolean} immediate - trigger an immediate update
	 */
	Grid.prototype.update = function(immediate) {
		var self = this;

		if(immediate == true) {
			run();
			return;
		}
		
		// Microbatch handler to prevent too many updates
		if(self.microbatch.timeout === false) {
			self.microbatch.timeout = true;
			setTimeout(eventEnd, self.microbatch.delta);
		}					
		function eventEnd() {
			var d = new Date() - self.microbatch.rtime;
			if (new Date() - self.microbatch.rtime < self.microbatch.delta)
				setTimeout(eventEnd, self.microbatch.delta);
			else {
				self.microbatch.rtime = new Date();
				self.microbatch.timeout = false;
				run();
			}               
		};
		
		// Run the update
		function run() {
			if(self.paused == true) return;
			
			// Detach all child data rows, and reattach later in the correct order
			var dataTableDiv = $(self.containerDiv).find('.data-table');
			dataTableDiv.children('.data-row').detach();

			// Apply ordering if set
			self.orderedArr = self.rowArr;
			if(self.configuration.options.ordering.enabled != false) {
				self.orderedArr = DataFlow.Utilities.Ordering.order(self.orderedArr, self.orderDef, 'object');
			}

			// Apply search and filters if set
			self.filterArr = self.orderedArr;
			if(self.configuration.options.search.enabled != false || self.configuration.options.filter.enabled != false) {
				self.filterArr = DataFlow.Utilities.Filtering.filter(self.filterArr, self.filterDef, 'object');
			}
		
			// Apply pagination if enabled
			self.resultArr = self.filterArr;
			if(self.pagination != null) {
				self.resultArr = DataFlow.Utilities.Filtering.paginate(self.resultArr, self.pagination);
			}
			
			// Show or hide the no data message depending on the number of rows in the result
			var noDataDiv = $(self.containerDiv).find('.no-data');
			if(self.resultArr.length > 0)
				noDataDiv.hide();
			else
				noDataDiv.show();
				
			// Loop over all the result rows
			for(var idx = 0; idx < self.resultArr.length; idx++) {
				// If the node isn't rendered for this row, render it now
				if(self.resultArr[idx].node == null) 
					self.resultArr[idx].render();
				dataTableDiv.append(self.resultArr[idx].node);
			}

			// Update the info panel
			self.updateInfo();
		};
	};

	/**
	 * Pauses updates on a grid.
	 */
	Grid.prototype.pause = function() {
		this.paused = true;
	};
	
	/**
	 * Resumes a paused grid.
	 */
	Grid.prototype.resume = function() {
		if(this.paused == false) return;
		this.paused = false;
		this.update();
	};

	/**
	 * Resizes a grid.
	 */
	Grid.prototype.resize = function() {
		// No implementation for this component
	};
	
	/**
	 * Applies a search value against the grid. This will not be passed through
	 * to the dataset and is local only to the current grid.
	 * @param {String} value - the search string
	 */
	Grid.prototype.search = function(value) {
		if(value == undefined || value.length == 0) {
			delete this.filterDef.search.expression;
		}
		else {
			this.filterDef.search.expression = value;
		}
		
		this.update();
	};
	
	/**
	 * Hovers a single row bound to the specified object.
	 * Will also optionally notify the dataset of a hover event, if enabled.
	 * @param {Object} object - object that the associated row will be hovered
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Grid.prototype.hover = function(object, notify) {
		if(this.configuration.options.hover.enabled != false) {
			// Unhover all existing rows
			for(var idx = this.hovered.length - 1; idx >= 0; idx--) {
				var thisObject = this.hovered[idx];
				var row = this.rowMap[thisObject._key];
				if(row != undefined) row.hover(false);
				this.hovered.splice(this.hovered.indexOf(thisObject), 1);
			}	
					
			// If the object is not null, process the hovering
			if(object != null) {
				if(Array.isArray(object)) {
					for(var idx = 0; idx < object.length; idx++) {
						var thisObject = object[idx];
						var row = this.rowMap[thisObject._key];
						if(row != undefined) row.hover(true);
						this.hovered.push(thisObject);
					}
				}
				else {
					var row = this.rowMap[object._key];
					if(row != undefined) row.hover(true);
					this.hovered.push(object);
				}
			}
					
			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.hover.passthrough == true) {
				this.sink.dataset.trigger('hover', this.sink, object);
			}			
		}
	};

	/**
	 * Marks a single row or a set of rows bound to the specified object or array of objects.
	 * Will also optionally notify the dataset of a marked event, if enabled.
	 * @param {Object|Array} object - if a single data object then the associated row will be marked
	 *                                if an array of objects then the associated rows will be marked
	 *                                  for single marking rows that receive a list of objects then only
	 *                                  the first row will be marked
	 *                                if null then all marked rows will be unmarked
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Grid.prototype.mark = function(object, notify) {
		if(this.configuration.options.marking.enabled != false) {
			var mode = this.configuration.options.marking.mode;

			// Unmark all existing rows
			for(var idx = this.marked.length - 1; idx >= 0; idx--) {
				var thisObject = this.marked[idx];
				var row = this.rowMap[thisObject._key];
				if(row != undefined) row.mark(false);
				this.marked.splice(this.marked.indexOf(thisObject), 1);
			}	
			
			// If the object is not null, process the marking
			if(object != null) {
				if(Array.isArray(object)) {
					if(this.configuration.options.marking.mode == 'multiple') {
						for(var idx = 0; idx < object.length; idx++) {
							var thisObject = object[idx];
							var row = this.rowMap[thisObject._key];
							row.mark(true);
							this.marked.push(thisObject);
						}
					}
					// If we receive a list of objects but can only mark one, mark only the first one
					else if(this.configuration.options.marking.mode == 'single' && object.length > 0) {
						var thisObject = object[0];
						var row = this.rowMap[thisObject._key];
						row.mark(true);
						this.marked.push(thisObject);
					}
				}
				else {
					var row = this.rowMap[object._key];
					row.mark(true);
					this.marked.push(object);
				}
			}
			
			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.marking.passthrough == true) {
				this.sink.dataset.trigger('marked', this.sink, object);
			}			

			this.updateInfo();
		}
	};

	/**
	 * Applies a filter and updates the grid. Will also optionally notify the
	 * dataset of a filter event, if enabled.
	 * @param {Array} filters - array of filter objects
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Grid.prototype.filter = function(filters, notify) {
		if(this.configuration.options.filter.enabled != false) {			
			// This clones the array so we don't get duplicates
			this.filterDef.filter.filters = filters.slice(0);
			this.filterDef.filter.fields = this.configuration.options.filter.fields;
			
			// Clear all existing filters
			for(var idx = 0; idx < this.columnArr.length; idx++) {
				this.columnArr[idx].applyFilter(null, false);
			}
			
			// Apply the new filters on each column
			for(var idx = 0; idx < filters.length; idx++) {
				var column = this.columnMap[filters[idx].field];
				column.applyFilter(filters[idx].expression, false);
			}

			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.filter.passthrough == true) {
				this.sink.dataset.trigger('filter', this.sink, this.filterDef.filter.filters);
			}
			
			this.update();
		}
	};
	
	/**
	 * Updates the contents of the information bar, if present.
	 */
	Grid.prototype.updateInfo = function() {
		$(this.containerDiv).find('.grid-showing').text(this.filterArr.length + " of " + this.rowArr.length + ' rows ' + this.marked.length + ' marked');			
	}

	/**
	 * Clears all ordering and updates the grid.
	 */
	Grid.prototype.clearOrdering = function() {
		if(this.orderDef.length > 0) {
			for(var idx = this.orderDef.length - 1; idx >= 0; idx--) {
				var prev = this.orderDef[idx];
				prev.applyOrder(null, false);
				this.orderDef.splice(idx, 1);
			}
		}
		this.update();
	}
	
	/** 
	 * Clears all field-level filters and updates the grid. Does NOT
	 * clear any Search value.
	 */
	Grid.prototype.clearFilters = function() {
		for(var idx = 0; idx < this.columnArr.length; idx++) {
			this.columnArr[idx].applyFilter(null, false);
		}
		this.filter([], true);
	};
	
	/**
	 * Toggles the visibility of the field-level filters.
	 */
	Grid.prototype.toggleFilters = function() {
		if(this.filtersVisible == true)
			$(this.containerDiv).find('.grid-header').removeClass('display-filters');
		else
			$(this.containerDiv).find('.grid-header').addClass('display-filters');
		this.filtersVisible = !this.filtersVisible;
	};
	
	/**
	 * Display the grid configuration modal.
	 */
	Grid.prototype.displayConfiguration = function() {
		if(this.configuration.options.configurable != true) return;
		console.log('display config');
		// TODO display configuration modal
	};

	/**
	 * Displays the dataset configuration modal.
	 */
	Grid.prototype.displayDataset = function() {
		if(this.configuration.options.dataset.modal == undefined) return;
		console.log('display dataset');
		// TODO display dataset modal
	};
	
	/** 
	 * Returns a color for the given value from the specified color map.
	 * @param {Object} value - the value to lookup, can be anything but typically a String
	 * @param (Object} name - the name of the color map
	 * @return {String} the color
	 */
	Grid.prototype.colormap = function(value, name) {
		var colorMap = this.colorMaps[name];
		if(colorValMap == undefined) return;
		return colorMap.getColor(value);
	};
	


	
	/**
	 * Creates a new Column object which represents a column in the grid.
	 * @constructor
	 * @param {Object} configuration - the configuration for the column
	 * @param {Object} grid - the grid containing this column
	 * @param {Object} containerDiv - this node object that header columns will be appended
	 * @param (Object} filterContainerDiv - the node object that column filters will be appended
	 */
	function Column(configuration, grid, containerDiv, filterContainerDiv) {
		this.field				= configuration.field;
		this.configuration 		= configuration;
		this.grid				= grid;
		this.containerDiv		= containerDiv;
		this.filterContainerDiv	= filterContainerDiv;

		// Set the title and datatype based on configuration
		this.title				= this.configuration.title == undefined ? this.configuration.field : this.configuration.title;
		this.type				= this.grid.schema.fieldMap[this.field] == undefined ? "string" : this.grid.schema.fieldMap[this.field].type;
		
		// The UI node object which is the DIV in the HTML for the column header
		this.node				= null;

		// The UI node object which is the DIV in the HTML for the column title
		this.titleNode			= null;
		
		// The UI node object which is the DIV in the HTML for the column filter
		this.filterNode			= null;
		
		// The UI node object which is the INPUT in the HTML for the column filter input box
		this.filterInputNode	= null;

		// Content to render for each row cell for this column. Ultimately this will end up being one of the following:
		//   1. undefined - default option, only the formatted data field will be appended to the cell 
		//   2. string - content field in column configuration in the form of a JS string template
		//   3. string - contentUrl loaded from the server in the form of a JS string template, used for complex, 
		//        multi-line type contents		
		this.content			= null;
		
		// Set the content value based on configuration, prioritize contentUrl over content
		var self = this;
		if(this.configuration.contentUrl != undefined) {
			$.ajax({
				type: 'GET',
				url: this.configuration.contentUrl,
				dataType: 'html',
				cache: false,
				aync: false,
				success: function(data, textStatus, xhr) {
					self.content = data;
				},
				error: function(xhr, textStatus, errorThrown) {
					console.error("Grid content url not found or invalid - " + self.configuration.contentUrl);
				}
			});	
		}
		else if(this.configuration.content != undefined) {
			this.content = this.configuration.content;
		}
		
		// Color map for this column
		this.colormap 			= new DataFlow.ColorMap(null, this.configuration.colormap);
		
		// Filter definition applied on the column
		this.filterDef			= null;
		
		// Flag to indicate if the column can be filtered
		this.filterable			= this.grid.configuration.options.filter.fields == undefined || 
									this.grid.configuration.options.filter.fields.length == 0 || 
									this.grid.configuration.options.filter.fields.indexOf(this.field) > -1;
									
		// Order direction, asc, desc, or null
		this.orderDir			= null;
		
		// Flag to indicate if the column can be ordered
		this.orderable			= this.grid.configuration.options.ordering.fields == undefined || 
									this.grid.configuration.options.ordering.fields.length == 0 ||
									this.grid.configuration.options.ordering.fields.indexOf(this.field) > -1;
		
	}
	
	/**
	 * Renders the UI components for a column header. This is a one-time render,
	 * if already rendered it will not be re-rendered.
	 */
	Column.prototype.render = function() {
		var self = this;
		
		// If node doesn't exist, then create it now
		if(this.node == null) {
			var context = {
				column:		this,
				grid:		this.grid
			}

			// Create the node and append to the header row
			this.node = $(document.createElement('div')).attr('class', 'header');
			this.containerDiv.append(this.node);
			
			// Create a filter node
			var filterIconNode = $(document.createElement('div')).attr('class', 'header-filter');
			this.node.append(filterIconNode);
			
			// Create a title node
			this.titleNode = $(document.createElement('div')).attr('class', 'header-title');
			this.node.append(this.titleNode);
						
			// Append default classes
			if(this.grid.configuration.options.columnHeaders.classes != undefined) {
				this.node.addClass(DataFlow.Utilities.Evaluate.classes(this.grid.configuration.options.columnHeaders.classes, context));
			}
			
			// Append default styles
			if(this.grid.configuration.options.columnHeaders.styles != undefined) {
				var styles = DataFlow.Utilities.Evaluate.styles(this.grid.configuration.options.columnHeaders.styles, context);
				for(style in styles) this.node.css(style, styles[style]);
			}

			// If filtering is enabled, append filter row
			if(this.grid.configuration.options.filter.enabled != false) {
				// Create the filter node and append to the header row
				this.filterNode = $(document.createElement('div')).attr('class', 'header-filter');
				this.filterContainerDiv.append(this.filterNode);
				
				// Append an input box to this filter node if there are no columns specified, or this column
				//   is one of the specified columns
				if(this.filterable == true) {
					var filterContainerNode = $(document.createElement('div'));
					this.filterNode.append(filterContainerNode);
					
					// Create the filter input node and append to the filter node
					this.filterInputNode = $(document.createElement('input')).attr('type', 'text');
					filterContainerNode.append(this.filterInputNode);
					
					// Create the filter input node clear and append
					var filterInputClearNode = $(document.createElement('div')).attr('class', 'header-filter-clear');
					filterContainerNode.append(filterInputClearNode);
					
					// Create event handler for change in input box value
					this.filterInputNode.on('input', DataFlow.Utilities.Debounce.debounce(function(e) {
						var val = $(e.target).val();
						self.applyFilter(val, true);
					}, this.grid.configuration.options.filter.delay));
					
					// Create event handler for clearing input box value
					filterInputClearNode.on('click', function(e) {
						self.filterInputNode.val('');
						var val = $(e.target).val();
						self.applyFilter(null, true);
					});
				}
			}
		
			// If ordering is enabled, create listeners
			if(this.grid.configuration.options.ordering.enabled == true) {
				// Apply initial ordering to this column if no columns are specified, or this 
				//   column is one of the specified columns
				if(this.orderable == true) {					
					// Add class
					this.node.addClass('orderable');
					
					// If there is initial ordering for this column, apply that sort
					if(this.grid.configuration.options.ordering.initial != undefined) {
						var initOrder = this.grid.configuration.options.ordering.initial[this.field];
						if(initOrder != undefined) {
							this.applyOrder(initOrder, false); // no update because that will happen below
						}
					}
					
					// Create a listener for click of the column header for sorting
					this.node.on('click', function(e) {
						if(self.orderDir == null)
							self.applyOrder('asc', true)
						else if(self.orderDir == 'asc')
							self.applyOrder('desc', true)
						else if(self.orderDir == 'desc')
							self.applyOrder(null, true)
					});
				}
			}
		
			// Add listener for the context menu
			this.node.on('contextmenu', function(e) {
				var context = {
					column:		self,
					grid:		self.grid
				}
				var container = self.grid.containerDiv;
				if(self.grid.configuration.options.menuContainer != undefined)
					container = $(self.grid.containerDiv).closest('.' + self.grid.configuration.options.menuContainer)[0];
				DataFlow.Utilities.ContextMenu.display(container, e, self.grid.configuration.headerMenu, context);
				return false;
			});
		}
		this.update();
	};
	
	/**
	 * Updates the column header.
	 */
	Column.prototype.update = function() {
		// Column header content will be the title if configured, otherwise
		//   set to the field name
		var content = this.title;
		this.titleNode.html(content);
		
		// Set the visibility
		if(this.configuration.visible == false) 
			this.node.css('display', 'none');
		else
			this.node.css('display', '');
		
		// Set the width
		if(this.configuration.width != undefined && this.node.width() != this.configuration.width) {
			this.node.css({'width': this.configuration.width});
		}
	};
	
	/**
	 * Applies a column ordering on the grid.
	 * @param {String} dir - the direction, 'asc', 'desc', or null to remove ordering
	 * @param {Boolean} update - flag indicating if the grid should be updated
	 */
	Column.prototype.applyOrder = function(dir, update) {
		// If multiple ordering is disabled, remove previous ordering
		if(this.grid.configuration.options.ordering.multiple == false) {
			if(this.grid.orderDef.length > 0) {
				var prev = this.grid.orderDef[0];
				this.grid.orderDef.splice(0, 1);
				prev.applyOrder(null, false); // no update because that will be happen below
			}
		}
		
		// Remove previous ordering class
		this.node.removeClass('order-asc order-desc');
		
		// Set the orderDir to the specified direction
		//   May be one of asc, desc, or null for no ordering
		this.orderDir = dir;
		
		// Add appropriate classes to the header
		if(dir == 'asc')
			this.node.addClass('order-asc');
		else if(dir == 'desc')
			this.node.addClass('order-desc');
		else if(dir == null) {
			this.grid.orderDef.splice(this.grid.orderDef.indexOf(this), 1);
		}

		if(dir != null && this.grid.orderDef.indexOf(this) == -1)
			this.grid.orderDef.push(this);
		
		if(update == true)
			this.grid.update();
	};
	
	/** 
	 * Applies a column-level filter on the grid.
	 * @param {String} val - the filter string to apply
	 */
	Column.prototype.applyFilter = function(val, notify) {
		// Create the filter definition
		if(val == null) {
			this.filterDef = null;
			if(this.filterInputNode != null)
				this.filterInputNode.val('');
		}
		else {
			this.filterDef = {
				field:		this.field,
				expression:	val
			}
			if(this.filterInputNode != null)
				this.filterInputNode.val(val);
		}
		
		// Call grid filter
		if(notify == true) {
			// Create an array of filters and populate with the current filters from the grid
			//   except for any filter for the current column
			var filters = [];
			for(var idx = 0; idx < this.grid.filterDef.filter.filters.length; idx++) {
				var thisFilter = this.grid.filterDef.filter.filters[idx];
				if(thisFilter.field != this.field)
					filters.push(thisFilter);
			}
			
			// Add this filter definition if it isn't null		
			if(this.filterDef != null)
				filters.push(this.filterDef);
		
			this.grid.filter(filters, true);
		}
		
	};
	
	/** 
	 * Returns a color for the given value from the color map for this column.
	 * @param {Object} value - the value to lookup, can be anything but typically a String
	 * @return {String} the color
	 */
	Column.prototype.colormap = function(value) {
		var color = this.colorValMap[value];
		if(color == undefined) {
			color = this.colorscale[this.colorIdx++];
			if(this.colorIdx == this.colorscale.length) {
				this.colorIdx = 0;
			}
			this.colorValMap[value] = color; 
		}
		return color;
	};
	
	
	
	/**
	 * Creates a new Row object which represents a row in the grid.
	 * @constructor
	 * @param {Object} object - the data object for the row
	 * @param {Object} columns - this list of columns for the row
	 * @param {Object} grid - the grid containing this row
	 */
	function Row(object, columns, grid) {
		this._key			= object._key;
		this.object 		= object;
		this.columns 		= columns;
		this.grid			= grid;
		this.configuration	= grid.configuration;
		
		// The UI node object which is the DIV in the HTML
		this.node			= null;
		
		// List of cell objects inside this row
		this.cells			= [];
		
		// Flags for hovered, marked, and filtered
		this.hovered		= false;
		this.marked			= false;
		this.filtered		= false;
		
		// Create a cell for each column in the configuration
		for(var idx = 0; idx < this.columns.length; idx++) {
			var thisColumn = this.columns[idx];
			this.cells.push(new Cell(this, thisColumn));
		}
	}
	
	/**
	 * Renders the UI components for a row. This is a one-time render,
	 * if already rendered it will not be re-rendered.
	 */
	Row.prototype.render = function() {
		var self = this;
		
		var context = {
			row:		this,
			grid:		this.grid
		}

		// Create the node element
		this.node = $(document.createElement('div')).attr('class', 'data-row');
		
		// Append default classes
		if(this.grid.configuration.options.rows.classes != undefined) {
			this.node.addClass(DataFlow.Utilities.Evaluate.classes(this.grid.configuration.options.rows.classes, context));
		}
		
		// If stripe classes are enabled, add class
		if(this.grid.configuration.options.rows.stripeClasses == true) {
			this.node.addClass('striped');
		}
		
		// Append default styles
		if(this.grid.configuration.options.rows.styles != undefined) {
			var styles = DataFlow.Utilities.Evaluate.styles(this.grid.configuration.options.rows.styles, context);
			for(style in styles) this.node.css(style, styles[style]);
		}
		
		// If marking is enabled, add a markable class and add the marking classes
		if(this.configuration.options.marking.enabled != false) {
			this.node.addClass('markable');
			this.node.addClass(this.configuration.options.marking.classes);
		}
	
		// If hover is enabled, add the hovering classes
		if(this.configuration.options.hover.enabled != false) {
			this.node.addClass(this.configuration.options.hover.classes);
		}

		// If row is already marked, add a marked class
		if(this.marked == true) {
			this.node.addClass('marked');
		}
		
		// Render each of the cells
		for(var idx = 0; idx < this.cells.length; idx++) {
			this.cells[idx].render();
		}
		
		// Create listener for mouseover to trigger hover
		this.node.on('mouseover', function() {
			self.grid.hover(self.object, true);
		});
		
		// Create listener for click to trigger marking
		this.node.on('click', function(e) {
			if(self.configuration.options.marking.enabled != false) {
				var ctrlKey = e.ctrlKey;					
				var mode = self.configuration.options.marking.mode;
				var isMarked = self.marked == true;
				
				// If unmarking is disabled and this row is marked, then just call mark again
				//   and take no further action
				if(self.configuration.options.marking.unmark == false && isMarked == true) {
					self.grid.mark(self.object, true);
					return;
				}
				
				// Single mode
				if(mode == 'single') {
					if(isMarked == true)
						self.grid.mark(null, true);
					else
						self.grid.mark(self.object, true);
				}
				// Multiple mode
				else if(mode == 'multiple') {
					if(ctrlKey == true) {
						var objects = [];
						for(var idx = 0; idx < self.grid.marked.length; idx++)
							objects.push(self.grid.marked[idx]);
						if(isMarked == false)
							objects.push(self.object);
						else if(isMarked == true)
							objects.splice(objects.indexOf(self.object), 1);
						self.grid.mark(objects, true);
					}
					else {
						if(isMarked == true)
							self.grid.mark(null, true);
						else
							self.grid.mark(self.object, true);
					}
				}
			}
		});			
		
		// Update the row
		this.update();
	};

	/** 
	 * Updates the contents of the row with the given object data.
	 * This will trigger cell.update() on each of the cells in the row.
	 */
	Row.prototype.update = function() {
		if(this.node == null) return;
		for(var idx = 0; idx < this.cells.length; idx++)
			this.cells[idx].update(this.object);
	};
	
	/**
	 * Copies the data for all cells to the clipboard in CSV format. 
	 * This will not include any formatting, just the raw data from the object.
	 */
	Row.prototype.copyToClipboard = function() {
		var text = "";
		for(key in this.object) text = text + "," + this.object[key];
		text = text.substring(1, text.length);
		DataFlow.Utilities.Clipboard.copy(text, this.node);
	};
	
	/**
	 * Sets the row hovered state based on the flag and updates the row classes.
	 * @param {Boolean} flag - true if the row is hovered, otherwise false
	 */
	Row.prototype.hover = function(flag) {
		if(flag == true && this.hovered == false) {
			this.hovered = true;
			if(this.node == null) return;		
			this.node.addClass('hover');
		}
		else if(flag == false && this.hovered == true) {
			this.hovered = false;
			if(this.node == null) return;		
			this.node.removeClass('hover');
		}
	};
	
	/**
	 * Sets the row marked state based on the flag and updates the row classes.
	 * @param {Boolean} flag - true if the row is marked, otherwise false
	 */
	Row.prototype.mark = function(flag) {
		if(flag == true && this.marked == false) {
			this.marked = true;
			if(this.node == null) return;
			this.node.addClass('marked');			
		}
		else if(flag == false && this.marked == true) {
			this.marked = false;
			if(this.node == null) return;
			this.node.removeClass('marked');			
		}
	}
	
	

	
	/**
	 * Creates a new Cell object which represents a cell in the grid.
	 * @constructor
	 * @param {Object} row - the row this cell belongs to
	 * @param {Object} column - this column this cell belongs to
	 */
	function Cell(row, column) {
		this.row			= row;
		this.column			= column;

		// The UI node object which is the DIV in the HTML
		this.node			= null;

		// Data object for the row, all fields
		this.object			= null;
		
		// Field data from the data object for this column
		this.data			= null;
		
		// Formatted field data
		this.formatted		= null;
	};
	
	/**
	 * Renders the UI components for a cell. This is a one-time render,
	 * if already rendered it will not be re-rendered.
	 */
	Cell.prototype.render = function() {
		var self = this;
		
		// Create a context object.
		var context = {
			data:		this.data,
			formatted:	this.formatted,
			object:		this.object,
			cell:		this,
			column:		this.column,
			row:		this.row,
			grid:		this.row.grid,
		}

		// Create the element
		this.node = $(document.createElement('div')).attr('class', 'cell');
		
		// Add default classes 
		if(this.row.grid.configuration.options.cells.classes != undefined) {
			this.node.addClass(DataFlow.Utilities.Evaluate.classes(this.row.grid.configuration.options.cells.classes, context));
		}
		
		if(this.column.content != undefined)
			this.node.addClass('custom');
		
		// Add default styles
		if(this.row.grid.configuration.options.cells.styles != undefined) {
			var styles = DataFlow.Utilities.Evaluate.styles(this.row.grid.configuration.options.cells.styles, context);
			for(style in styles) this.node.css(style, styles[style]);
		}

		// Add listener for the context menu
		this.node.on('contextmenu', function(e) {
			var context = {
				data:		self.data,
				formatted:	self.formatted,
				object:		self.object,
				cell:		self,
				column:		self.column,
				row:		self.row,
				grid:		self.row.grid,
			}
			var container = self.row.grid.containerDiv;
			if(self.row.grid.configuration.options.menuContainer != undefined)
				container = $(self.row.grid.containerDiv).closest('.' + self.row.grid.configuration.options.menuContainer)[0];
			DataFlow.Utilities.ContextMenu.display(container, e, self.row.grid.configuration.bodyMenu, context);
			return false;
		});
		
		// Append the cell to the row
		this.row.node.append(this.node);
		
		// No update here, will be triggered by row.update()
	};

	/**
	 * Updates the contents of the cell with the given object data.
	 * @param {Object} object - the data object to update
	 */
	Cell.prototype.update = function(object) {
		var self 		= this;
		this.data 		= object[this.column.field] != undefined ? object[this.column.field] : null;
		this.object 	= object;
		
		// Function to retrieve a color for the given value
		//   If name is undefined then the column colormap is used, 
		//   otherwise the grid colormap is used
		function colormap(value, name) {
			if(name == undefined)
				return self.column.colormap.getColor(value);
			else {
				var colormap = self.row.grid.colorMaps[name];
				if(colormap == undefined) return;
				return colormap.getColor(value);
			}
		}

		// Function to format the specified value, which can be timestamp or number
		function format(val, format) {
			return DataFlow.Utilities.Format.format(val, format);
		};
				
		// Apply any default formatting by data type
		this.formatted	= this.data;
		if(this.row.grid.configuration.options.format[self.column.type] != undefined) {
			var thisFormat = this.row.grid.configuration.options.format[self.column.type];
			if(self.column.type == 'timestamp')
				this.formatted = DataFlow.Utilities.Format.formatTimestamp(this.data, thisFormat);
			else if(self.column.type == 'double' || self.column.type == 'int' || self.column.type == 'long')
				this.formatted = DataFlow.Utilities.Format.formatNumber(this.data, thisFormat);
		}
		
		// Create the context
		var context = {
			data:		this.data,
			formatted:	this.formatted,
			object:		this.object,
			cell:		this,
			column:		this.column,
			row:		this.row,
			grid:		this.row.grid,
			colormap:	colormap,
			format:		format
		}
			
		// Set the node content
		var content = this.formatted;
		if(this.column.content != undefined) {
			content = function(str){
				for(key in this) eval('var ' + key + ' = this.' + key);
				return eval('`' + str + '`');
			}.call(context, this.column.content);
		}
		this.node.empty().html(content);
		if(this.node.children().length > 0)
			$(Utilities.Template.process(this.node.children()[0]));
		
		// Add or remove classes
		if(this.column.configuration.classes != undefined) {
			var classes = DataFlow.Utilities.Evaluate.classes(this.column.configuration.classes, context);
			if(typeof classes === 'object') {
				for(className in classes) {
					var value = classes[className];
					if(value == true)
						this.node.addClass(className)
					else
						this.node.removeClass(className);
				}
			}
			else {
				this.node.addClass(classes);
			}			
		}

		// Set styles
		if(this.column.configuration.styles != undefined) {
			var styles = DataFlow.Utilities.Evaluate.styles(this.column.configuration.styles, context);
			for(style in styles) this.node.css(style, styles[style]);
		}
	
		// Set the visibility
		if(this.column.configuration.visible == false) 
			this.node.css('display', 'none');
		else
			this.node.css('display', '');
		
		// Set the width
		if(this.column.configuration.width != undefined && this.node.width() != this.column.configuration.width) {
			this.node.css({'width': this.column.configuration.width});
		}
	};
	
	/**
	 * Copies the cell data to the clipboard. This will not include
	 * any formatting, just the raw data from the object.
	 */
	Cell.prototype.copyToClipboard = function() {
		DataFlow.Utilities.Clipboard.copy(this.data, this.node);
	};
	
	
	

	/** 
	 * Grid configuration helper object.
	 */
	GridConfiguration = {
		// Default grid configuration object
		defaultConfiguration: {
				columns:	[],
				options: {
					//classes:			undefined,
					configurable:		false,
					cells: {
						//classes:		undefined,
						//styles:		undefined
					},
					columnHeaders: {
						//classes:		undefined,
						//styles:		undefined
					},
					copy: {
						enabled:		true
					},
					dataset: {
						icon:			true
						//modal:			undefined
					},
					filter: {	
						accept:			true,
						fields:			undefined,
						delay:			500,
						enabled:		true,
						passthrough:	true
					},
					format: {
						timestamp:		"yyyy-MM-dd HH:mm:ss"
					},
					hover: {
						accept:			true,
						classes:		"hover-background",
						enabled:		true,
						passthrough:	true
					},
					// info:			false,
					marking: {
						accept:			true,
						classes:		"mark-background",
						enabled:		true,
						mode:			"single",
						passthrough:	true,
						unmark:			true
					},
					ordering: {
						//classes:		undefined,
						fields:			undefined,
						enabled:		true,
						//initial:		{},
						multiple:		false
					},
					pagination: {
						enabled:		true,
						size:			25
					},
					rows: {
						//classes:		undefined,
						stripes:		false
						//styles:		undefined
					},
					search: {
						fields:			undefined,
						delay:			500,
						enabled:		true,
						label:			"Search"
					},
					//styles:			undefined,
					//title:			undefined,
				},
				
				headerMenu: {
					ordering: [
						{
							label:		"Order ${column.title} Ascending",
							condition:	"column.orderable == true && (column.orderDir == null || column.orderDir == 'desc')",
							action:		"column.applyOrder('asc', true)"
						},
						{
							label:		"Order ${column.title} Descending",
							condition:	"column.orderable == true && (column.orderDir == null || column.orderDir == 'asc')",
							action:		"column.applyOrder('desc', true)"
						},
						{
							label:		"Clear ${column.title} Ordering",
							condition:	"column.orderable == true && (column.orderDir != null)",
							action:		"column.applyOrder(null, true)"
						},
						{
							label:		"Clear All Ordering",
							condition:	"grid.configuration.options.ordering.enabled != false && grid.orderDef.length > 0",
							action:		"grid.clearOrdering()"
						}
					],
					filtering: [
						{
							label:		"Clear ${column.title} Filter",
							condition:	"column.filterable == true && column.filterDef != null",
							action:		"column.applyFilter(null, true, true)"
						},
						{
							label:		"Clear All Filters",
							condition:	"grid.configuration.options.filter.enabled != false && grid.filterDef.filter.filters != undefined && grid.filterDef.filter.filters.length > 0",
							action:		"grid.clearFilters()"
						},
						{
							label:		"Show Field Filters",
							condition:	"grid.configuration.options.filter.enabled != false && grid.filtersVisible == false",
							action:		"grid.toggleFilters()"
						},
						{
							label:		"Hide Field Filters",
							condition:	"grid.configuration.options.filter.enabled != false && grid.filtersVisible == true",
							action:		"grid.toggleFilters()"
						}
					],
					panels: [
						{
							label:		"Display Configuration",
							condition:	"grid.configuration.options.configurable == true",
							action:		"grid.displayConfiguration()"
						},
						{
							label:		"Display Dataset",
							condition:	"grid.configuration.options.dataset.modal != undefined",
							action:		"grid.displayDataset()"
						}
					]
				},
				
				bodyMenu: {
					clipboard: [
						{
							label:		"Copy Cell",
							condition:	"document.queryCommandSupported('copy') == true && grid.configuration.options.copy.enabled != false",
							action:		"cell.copyToClipboard()"
						},
						{
							label:		"Copy Row",
							condition:	"document.queryCommandSupported('copy') == true && grid.configuration.options.copy.enabled != false",
							action:		"row.copyToClipboard()"
						}
					],
					filtering: [
						{
							label:		"Filter ${column.title} to value ${data}",
							condition:	"column.filterable == true",
							action:		"column.applyFilter(\"== '\" + data + \"'\", true, true)"
						},
						{
							label:		"Clear ${column.title} Filter",
							condition:	"column.filterable == true && column.filterDef != null",
							action:		"column.applyFilter(null, true, true)"
						},
						{
							label:		"Clear All Filters",
							condition:	"grid.configuration.options.filter.enabled != false && grid.filterDef.filter.filters != undefined && grid.filterDef.filter.filters.length > 0",
							action:		"grid.clearFilters()"
						}
					],
					panels: [
						{
							label:		"Display Configuration",
							condition:	"grid.configuration.options.configurable == true",
							action:		"grid.displayConfiguration()"
						},
						{
							label:		"Display Dataset",
							condition:	"grid.configuration.options.dataset.modal != undefined",
							action:		"grid.displayDataset()"
						}
					]
				}
			},
			
		// Default grid column configuration object
		defaultColumnConfiguration: {
				//field:				undefined,
				//classes:				undefined, 
				//styles:				undefined,
				//title:				undefined,
				visible:				true,
				//width:				undefined
			},
		
		// Merges the source configuration object into the target configuration object
		//   Source values will overwrite target values.
		//   Should always merge a DEFAULT configuration source into a USER configuration target
		mergeConfiguration:		function(source, target) {
									for(var idx = 0; idx < Object.keys(source).length; idx++) {
										var key = Object.keys(source)[idx];
										var val = source[key];
										if(typeof val === 'object') {
											if(target[key] == undefined) 
												target[key] = val;
											else
												this.mergeConfiguration(val, target[key]);
										}
										else {
											if(target[key] == undefined) 
												target[key] = val;
										}
									}
									return target;
								},

		// Processes the specified configuration object and merges in default configuration
		//   for both base object and column objects
		processConfiguration:	function(config) {
									var configuration = this.mergeConfiguration(this.defaultConfiguration, config);
									for(var idx = 0; idx < configuration.columns.length; idx++) {
										configuration.columns[idx] = this.mergeConfiguration($.extend({}, this.defaultColumnConfiguration), 
											configuration.columns[idx]);
									}
									return configuration;
								},
	};
	
	
	
	
	// Setup module objects
	module.Grid 				= Grid;
	module.GridConfiguration	= GridConfiguration;
	
	return module;
})();
/**
 * Represents a Leaflet map data sink.
 * @constructor
 * @param {Object} dataset - dataset that this sink will be attached to
 * @param {Object} configuration - configuration object defining the sink
 */
function LeafletLayerDatasink(id, configuration, dataset, map, overlay) {
	this.id = id;
	this.dataset = dataset;
	this.layer = new LiveLeafletMap.Layer(id, configuration, map, overlay, this);
}

/**
 * Event handler for dataset events. 
 * @param {String} eventType - string describing the event that occurred
 * @param {Object} object - data object for the event, or null
 */
LeafletLayerDatasink.prototype.trigger = function(eventType, object) {
	this.layer.trigger(eventType, object);
};


// Define a module to hold object defintions for the map objects
var LiveLeafletMap = (function() {
	var module = {};

	/**
	 * Creates a new Map object.
	 * @constructor
	 * @param {String} id - unique identifier
	 * @param {Object} configuration - the configuration for the map
	 * @param {Object} containerDiv - this node object that map will be appended
	 */
	function Map(id, configuration, containerDiv) {
		this.id					= id;
		this.configuration		= configuration;
		this.containerDiv 		= containerDiv;

		// Pause and snapshot flags
		this.paused				= false;
		this.snapshot			= false;

		// Leaflet node object, this is the leaflet map
		this.node				= null;

		// Color maps
		this.colorMaps			= {};

		// Layers
		this.layersArr			= [];

		// Control layers
		this.controlLayers		= {};
		
		// Check dependencies. This will throw Error if dependencies are not satisfied.
		this.dependencyCheck();
		
		// Color maps
		if(this.configuration.options.colormaps != undefined) {
			for(var idx = 0; idx < Object.keys(this.configuration.options.colormaps).length; idx++) {
				var thisColorMapName = Object.keys(this.configuration.options.colormaps)[idx];
				var thisColors = this.configuration.options.colormaps[thisColorMapName];
				this.colorMaps[thisColorMapName] = new DataFlow.ColorMap(thisColorMapName, thisColors);
			}
		}

		// On document ready, register UI handlers
		var self = this;
		setTimeout(function() {
			self.registerEventHandlers();
		}), 0;
		
		// Draw the initial empty map
		this.draw();
	};
	
	/**
	 * Performs a dependency check for rendering the map.
	 */
	Map.prototype.dependencyCheck = function() {
		var msg = "Required Leaflet libraries not loaded or not present";
		try {
			var p = L;
		}
		catch(err) {
			console.log(err);
			throw Error("Required Leaflet libraries not loaded or not present");
		}
	};
		
	/**
	 * Registers event handlers on UI components.
	 */
	Map.prototype.registerEventHandlers = function() {
		var self = this;
		
		this.node.on('mousedown', function(e) {
			if(e.originalEvent.ctrlKey == true) {
				self.node.dragging.disable();
				
				// Create a selection area div
				var appended = false;
				var selectionAreaDiv = $(document.createElement('div')).attr('class', 'selection-area');
				
				// Get initial start event
				var startEvent = e;
				var startX = e.containerPoint.x;
				var startY = e.containerPoint.y;

				// Function to update the size and position of the selection area
				var positionSelectionArea = function(endX, endY) {
					selectionAreaDiv.css({
						top: Math.min(startY, endY), 
						left: Math.min(startX, endX),
						height: Math.abs(startY - endY),
						width: Math.abs(startX - endX)
					});
				};
				
				self.node.on('mousemove', function(e) {
					var endX = e.containerPoint.x;
					var endY = e.containerPoint.y;
					if(appended == false && (Math.abs(startY - endY) > 10 || Math.abs(startX - endX) > 10)) {
						$(self.containerDiv).append(selectionAreaDiv);
						appended = true;
					}
					
					// Update the selection area
					positionSelectionArea(endX, endY);
				});
				
				self.node.on('mouseup', function(e) {
					var endEvent = e;
					
					// Instruct each layer to mark bounds
					for(var idx = 0; idx < self.layersArr.length; idx++)
						self.layersArr[idx].markBounds(startEvent.latlng, endEvent.latlng);
					
					// Remove the selection area
					selectionAreaDiv.remove();
					self.node.off('mouseup');
					self.node.off('mousemove');
					self.node.dragging.enable();
				});
			}
		});
	};
			
	/**
	 * Draws the UI components for the map.
	 */
	Map.prototype.draw = function() {
		var self = this;
		
		// Create the map
		this.node = L.map(this.containerDiv, this.configuration.leaflet.map);
		
		// Setup controls
		if(this.configuration.leaflet.control.attribution != undefined) {
			this.controlLayers.attributionControl = L.control.attribution(this.configuration.leaflet.control.attribution);
			this.controlLayers.attributionControl.addTo(this.node);
		}

		// Setup multiple marking control
		var multipleMarkingEnabled = false;
		for(var layerId in this.configuration.leaflet.layers) {
			var thisConfiguration = this.configuration.leaflet.layers[layerId];
			if(thisConfiguration.options.marking.enabled == true && thisConfiguration.options.marking.mode == 'multiple') {
				multipleMarkingEnabled = true;
				break;
			}
		}
		
		if(this.configuration.options.search.enabled == true) {
			L.Control.Search = L.Control.extend({
				onAdd: function(map) {
					var searchDiv = $(document.createElement('div')).attr('class', 'leaflet-control-search');
					
					var searchInputDiv = $(document.createElement('div')).attr('class', 'leaflet-control-search-input');
					searchDiv.append(searchInputDiv);
					searchInputDiv.append($(document.createElement('span')).attr('class', 'leaflet-control-search-layers'));
					var searchSelect = $(document.createElement('select'));
					searchInputDiv.append(searchSelect);
					var searchInput = $(document.createElement('input'));
					searchInputDiv.append(searchInput);
					var searchClear = $(document.createElement('div')).attr('class', 'leaflet-control-search-clear');
					searchInputDiv.append(searchClear);					
					var searchLink = $(document.createElement('a')).attr('class', 'leaflet-control-search-toggle').attr('href', '#');
					searchDiv.append(searchLink);
					
					var opened = false;
					searchDiv.on('dblclick', function(e) {
						return false;
					});
					searchDiv.on('mouseover', function(e) {
						self.node.dragging.disable();
					});
					searchDiv.on('mouseout', function(e) {
						self.node.dragging.enable();
					});
					searchLink.on('click', function(e) {
						if(opened == false) {
							searchSelect.empty();
							searchSelect.append($(document.createElement('option')).attr('value', null).html('All'));
							for(var idx = 0; idx < self.layersArr.length; idx++) {
								var layer = self.layersArr[idx];
								searchSelect.append($(document.createElement('option')).attr('value', self.layersArr[idx].name).html(self.layersArr[idx].name));
							}
							opened = true;
							searchInputDiv.addClass('visible');
							searchInput.focus();
						}
						else if(searchInput.val().length == 0) {
							opened = false;
							searchInputDiv.removeClass('visible');						
						}
						return false;
					});
					searchClear.on('click', function(e) {
						searchInput.val('');
						applySearch(searchSelect.val(), searchInput.val());
						return false;
					});
					
					var applySearch = function(layer, val) {
						for(var idx = 0; idx < self.layersArr.length; idx++) {
							if(val == '')
								self.layersArr[idx].search('');
							else if(layer == 'All' || layer == self.layersArr[idx].name) 
								self.layersArr[idx].search(val);
							else
								self.layersArr[idx].search('');
						}
					};
					
					searchSelect.on('change', function(e) {
						applySearch(searchSelect.val(), searchInput.val());
					});
					
					searchInput.on('input', DataFlow.Utilities.Debounce.debounce(function(e) {
						applySearch(searchSelect.val(), searchInput.val());
					}, self.configuration.options.search.delay));

					return searchDiv[0];
				},

				onRemove: function(map) {
					
				}
			});

			L.control.search = function(opts) {
				return new L.Control.Search(opts);
			}
			var thisConfiguration = this.configuration.leaflet.control.search == undefined ? {position: 'topright'} : this.configuration.leaflet.control.search;
			this.controlLayers.searchControl = L.control.search(thisConfiguration);
			this.controlLayers.searchControl.addTo(this.node);
		}
		
		if(this.configuration.leaflet.control.layers != undefined) {
			this.controlLayers.layers = L.control.layers(this.baseLayers, this.overlayLayers, this.configuration.leaflet.control.layers);
			this.controlLayers.layers.addTo(this.node);
		}
		
		if(this.configuration.leaflet.control.scale != undefined) {
			this.controlLayers.scale = L.control.scale(this.configuration.leaflet.control.scale);
			this.controlLayers.scale.addTo(this.node);
		}
		
		if(this.configuration.leaflet.control.zoom != undefined) {
			this.controlLayers.zoomControl = L.control.zoom(this.configuration.leaflet.control.zoom);
			this.controlLayers.zoomControl.addTo(this.node);
		}

		// Setup configuration control
		if(this.configuration.options.configurable == true) {
			L.Control.Configuration = L.Control.extend({
				onAdd: function(map) {
					var configurationDiv = L.DomUtil.create('div', 'leaflet-control-configuration');
					var configurationLink = L.DomUtil.create('a', 'leaflet-control-configuration-toggle', configurationDiv);
					configurationLink.href = '#';
					L.DomEvent.on(configurationLink, 'click', function(e) {
						self.displayConfiguration();
					});
					return configurationDiv;
				},

				onRemove: function(map) {
					
				}
			});

			L.control.configuration = function(opts) {
				return new L.Control.Configuration(opts);
			}
			var thisConfiguration = this.configuration.leaflet.control.configuration == undefined ? {position: 'topright'} : this.configuration.leaflet.control.configuration;
			this.controlLayers.configurationControl = L.control.configuration(thisConfiguration);
			this.controlLayers.configurationControl.addTo(this.node);
		}
		
		// Setup dataset control
		if(this.configuration.options.dataset.icon == true) {
			L.Control.Dataset = L.Control.extend({
				onAdd: function(map) {
					var live = true;
					var datasetDiv = L.DomUtil.create('div', 'leaflet-control-dataset');
					var datasetLink = L.DomUtil.create('a', 'leaflet-control-dataset-toggle ' + (live == true ? 'live' : 'notlive'), datasetDiv);
					datasetLink.href = '#';
					L.DomEvent.on(datasetLink, 'click', function(e) {
						self.displayDataset();
					});
					return datasetDiv;
				},

				onRemove: function(map) {
					
				}
			});

			L.control.dataset = function(opts) {
				return new L.Control.Dataset(opts);
			}
			var thisConfiguration = this.configuration.leaflet.control.dataset == undefined ? {position: 'topright'} : this.configuration.leaflet.control.dataset;
			this.controlLayers.datasetControl = L.control.dataset(thisConfiguration);
			this.controlLayers.datasetControl.addTo(this.node);
		}
		
		// Create and append tile layer
		var tileLayer = L.tileLayer(this.configuration.leaflet.tileLayer.provider, this.configuration.leaflet.tileLayer.options);
		tileLayer.addTo(this.node);
		if(this.controlLayers.layers != null) {
			this.controlLayers.layers.addBaseLayer(tileLayer, this.configuration.leaflet.tileLayer.name);
		}
	};
	
	/**
	 * Pauses updates on a map.
	 */
	Map.prototype.pause = function() {
		this.paused = true;
	};
	
	/**
	 * Resumes a paused map.
	 */
	Map.prototype.resume = function() {
		if(this.paused == false) return;
		this.paused = false;
		for(var idx = 0; idx < this.layersArr.length; idx++)
			this.layersArr[idx].update();
	};
	
	/**
	 * Resizes a map.
	 */
	Map.prototype.resize = function() {
		this.node.invalidateSize();		
	};
			
	/**
	 * Display the map configuration modal.
	 */
	Map.prototype.displayConfiguration = function() {
		if(this.configuration.options.configurable != true) return;
		console.log('leaflet display config');
		// TODO display configuration modal
	};

	/**
	 * Displays the dataset configuration modal.
	 */
	Map.prototype.displayDataset = function() {
		if(this.configuration.options.dataset.modal == undefined) return;
		console.log('leaflet display dataset');
		// TODO display dataset modal
	};
	
	/** 
	 * Returns a color for the given value from the specified color map.
	 * @param {Object} value - the value to lookup, can be anything but typically a String
	 * @param (Object} name - the name of the color map
	 * @return {String} the color
	 */
	Map.prototype.colormap = function(value, name) {
		var colorMap = this.colorMaps[name];
		if(colorMap == undefined) return;
		return colorMap.getColor(value);
	};
	
	
	
	
	/**
	 * Creates a new Layer object.
	 * @constructor
	 * @param {String} id - unique identifier
	 * @param {Object} configuration - the configuration for the layer
	 * @param {Object} map - the map that this layer will be added to
	 * @param (Function} overlay - function to call on start and end of the overlay
	 * @param {Object} sink - datasink that will feed this layer
	 */
	function Layer(id, configuration, map, overlay, sink) {
		this.id				= id;
		this.name			= configuration.name;
		this.configuration	= configuration;
		this.map			= map;
		this.overlay		= overlay;
		this.sink			= sink;
		
		// Pause and snapshot flags
		this.paused			= false;
		this.snapshot		= false;

		// Schema for the dataset
		this.schema			= null;		
		
		// Leaflet layer object
		this.node			= null;
		
		// Vars to hold the objects being hovered and marked
		this.hovered		= [];		
		this.marked			= [];

		// Filter definition
		this.filterDef		= {
			search: {
				expression:	undefined,
				fields:		this.configuration.options.search.fields
			},
			filter:	{
				filters:	[],
				fields:		this.configuration.options.filter.fields
			}
		};

		// Color map
		this.colormap 		= new DataFlow.ColorMap(null, this.configuration.colormap);

		// Component array and map, AbstractComponent objects
		this.componentArr	= [];
		this.componentMap	= {};
		
		// Add this layer to the map
		this.map.layersArr.push(this);
				
		// Microbatch is used to batch up updates to prevent too many map redraws
		this.updateQueue = {};
		this.microbatch = {
			rtime:		new Date(1, 1, 2000, 12,00,00),
			timeout:	false,
			delta:		500 // msec between updates
		}

		// Draw initial empty layer
		this.draw();
	};
	
	/**
	 * Event handler for all objects delegated from the datasink.
	 * @param {String} eventType - the event type
	 * @param {Object} object - optional object
	 */
	Layer.prototype.trigger = function(eventType, object) {
		var self = this;
		
		if(eventType == 'schemaChange') {
			this.schema = object;
		}
		else if(eventType == 'snapshotStart') {
			this.snapshot = true;
			this.overlay(true);
		}
		else if(eventType == 'snapshotEnd') {
			setTimeout(function() {
				self.snapshot = false;
				self.overlay(false);
			},500);
		}
		else if(eventType == 'clear') {
			this.clear();
		}
		else if(eventType == 'insert') {
			this.addObject(object);
		}
		else if(eventType == 'update') {
			this.updateObject(object);
		}
		else if(eventType == 'upsert') {
		}
		else if(eventType == 'delete') {
			this.removeObject(object);
		}
		else if(eventType == 'hover') {
			if(this.configuration.options.hover.accept != false)
				this.hover(object);
		}
		else if(eventType == 'marked') {
			if(this.configuration.options.marking.accept != false)
				this.mark(object);
		}
		else if(eventType == 'filter') {
			if(this.configuration.options.filter.accept != false)
				this.filter(object);
		}
	};
	
	/**
	 * Add a new object bound to the specified object on the layer group.
	 * @param {Object} object - the object to add the bound object
	 */
	Layer.prototype.addObject = function(object) {
		var component = this.createComponent(object);
		this.componentMap[component._key] = component;
		this.componentArr.push(component);
		this.updateQueue[component._key] = component;
		this.update();
	};

	/**
	 * Updates the component bound to the specified object on the layer group.
	 * @param {Object} object - the object to update the bound object
	 */
	Layer.prototype.updateObject = function(object) {
		var component = this.componentMap[object._key];
		if(component == undefined) return;		
		component.object = object;
		this.updateQueue[component._key] = component;
		this.update();
	};
	
	/**
	 * Removes the component bound to the specified object from the layer group.
	 * @param {Object} object - the object to remove the bound object
	 */
	Layer.prototype.removeObject = function(object) {
		var component = this.componentMap[object._key];
		component.remove();
		delete this.componentMap[object._key];
		this.componentArr.splice(this.componentArr.indexOf(component), 1);
		this.update();
	};
					
	/**
	 * Clears all objects and all components from the layer group.
	 */
	Layer.prototype.clear = function() {
		for(var idx = 0; idx < this.componentArr.length; idx++)
			this.componentArr[idx].remove();
		this.componentMap	= {};
		this.componentArr	= [];
		this.update();
	};
	
	/**
	 * Hovers a single component bound to the specified object.
	 * Will also optionally notify the dataset of a hover event, if enabled.
	 * @param {Object} object - object that the associated component will be hovered
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Layer.prototype.hover = function(object, notify) {
		if(this.configuration.options.hover.enabled != false) {

			// Unhover all existing rows
			for(var idx = this.hovered.length - 1; idx >= 0; idx--) {
				var thisObject = this.hovered[idx];
				var component = this.componentMap[thisObject._key];
				if(component != undefined) {
					component.hovered = false;
					component.update();
				};
				this.hovered.splice(this.hovered.indexOf(thisObject), 1);
			}	

			// If the object is not null, process the hovering
			if(object != null) {
				if(Array.isArray(object)) {
					for(var idx = 0; idx < object.length; idx++) {
						var thisObject = object[idx];
						var component = this.componentMap[thisObject._key];
						if(component != undefined) {
							component.hovered = true;
							component.update();
						}
						this.hovered.push(thisObject);
					}
				}
				else {
					var component = this.componentMap[object._key];
					if(component != undefined) {
						component.hovered = true;
						component.update();
					}
					this.hovered.push(object);
				}
			}
			
			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.hover.passthrough == true) {
				this.sink.dataset.trigger('hover', this.sink, object);
			}
		}
	};
		
	/**
	 * Marks a single component or a set of components bound to the specified object or array of objects.
	 * Will also optionally notify the dataset of a marked event, if enabled.
	 * @param {Object|Array} object - if a single data object then the associated component will be marked
	 *                                if an array of objects then the associated components will be marked
	 *                                  for single marking components that receive a list of objects then only
	 *                                  the first component will be marked
	 *                                if null then all marked components will be unmarked
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Layer.prototype.mark = function(object, notify) {
		if(this.configuration.options.marking.enabled != false) {
			var mode = this.configuration.options.marking.mode;

			// Unmark all existing components
			for(var idx = this.marked.length - 1; idx >= 0; idx--) {
				var thisObject = this.marked[idx];				
				var component = this.componentMap[thisObject._key];
				if(component != undefined) {
					component.marked = false;
					component.update();
				}
				this.marked.splice(this.marked.indexOf(thisObject), 1);
			}	
			
			// If the object is not null, process the marking
			if(object != null) {
				if(Array.isArray(object)) {
					if(this.configuration.options.marking.mode == 'multiple') {
						for(var idx = 0; idx < object.length; idx++) {
							var thisObject = object[idx];
							var component = this.componentMap[thisObject._key];
							component.marked = true;
							component.update();
							this.marked.push(thisObject);
						}
					}
					// If we receive a list of objects but can only mark one, mark only the first one
					else if(this.configuration.options.marking.mode == 'single' && object.length > 0) {
						var component = this.componentMap[object[0]._key];
						component.marked = true;
						component.update();
						this.marked.push(object[0]);
					}
				}
				else {
					var component = this.componentMap[object._key];
					component.marked = true;
					component.update();
					this.marked.push(object);
				}
			}
			
			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.marking.passthrough == true) {
				this.sink.dataset.trigger('marked', this.sink, object);
			}			

		}
	};
		
	/**
	 * Marks all components within the specified lat and lon corner positions.
	 * @param {Double[]} latlng1 - first corner of the bounding box
	 * @param {Double[]} latlng2 - second cornder of the bounding box
	 */
	Layer.prototype.markBounds = function(latlng1, latlng2) {
		if(this.configuration.options.marking.enabled == false) return;
		
		// Skip if this layer is currently not visible
		if(this.node._map == null) return;
		
		// Loop over all the components and test the position of the node against the 
		//   bounding latlngs
		var bounds = L.latLngBounds(latlng1, latlng2);
		var selected = [];
		for(var idx = 0; idx < this.componentArr.length; idx++) {
			var latlng = this.componentArr[idx].node.getLatLng();
			if(bounds.contains(latlng) == true)
				selected.push(this.componentArr[idx]);
		}
		
		// If nothing is selected, then do not change the marking for the layer
		if(selected.length == 0) return;
		var mode = this.configuration.options.marking.mode;
			
		// Single mode
		if(mode == 'single') {
			var component = selected[0];
			if(component.marked == true)
				this.mark(null, true);
			else
				this.mark(component.object, true);
		}
		// Multiple mode
		else if(mode == 'multiple') {
			var objects = [];
			for(var idx = 0; idx < selected.length; idx++)
				objects.push(selected[idx].object);
			this.mark(objects, true);
		}
		
	};
	
	/**
	 * Applies a filter and updates the map. Will also optionally notify the
	 * dataset of a filter event, if enabled.
	 * @param {Array} filter - array of filter objects
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Layer.prototype.filter = function(filters, notify) {
		if(this.configuration.options.filter.enabled != false) {
			// This clones the array so we don't get duplicates
			this.filterDef.filter.filters = filters.slice(0);
			this.filterDef.filter.fields = this.configuration.options.filter.fields;
			
			for(var idx = 0; idx < this.componentArr.length; idx++) {
				var include = DataFlow.Utilities.Filtering.filterItem(this.componentArr[idx], 
					this.filterDef.filter.filters, this.filterDef.filter.fields, 'object');
				var filtered = !include;
				this.componentArr[idx].filter(filtered);
			}
			
			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.filter.passthrough == true) {
				this.sink.dataset.trigger('filter', this.sink, this.filterDef.filter.filters);
			}
			
			this.update();
		}
	};
	
	/**
	 * Applies a search value against the layer. This will not be passed through
	 * to the dataset and is local only to the current layer.
	 * @param {String} value - the search string
	 */
	Layer.prototype.search = function(value) {
		if(value == undefined || value.length == 0) {
			delete this.filterDef.search.expression;
			for(var idx = 0; idx < this.componentArr.length; idx++) {
				this.componentArr[idx].search(null);
			}
		}
		else {
			this.filterDef.search.expression = value;
			for(var idx = 0; idx < this.componentArr.length; idx++) {
				var include = DataFlow.Utilities.Filtering.searchItem(this.componentArr[idx], 
					this.filterDef.search.expression, this.filterDef.search.fields, 'object');
				this.componentArr[idx].search(include);
			}
		}
	};
	
	
	/**
	 * Draws the UI components for the layer.
	 */
	Layer.prototype.draw = function() {
		// Create the layer and append to the map
		this.node = L.layerGroup();
		this.node.addTo(this.map.node);
		if(this.map.controlLayers.layers != null) {
			this.map.controlLayers.layers.addOverlay(this.node, this.name);
		}
	};
	
	/**
	 * Updates the contents of the layer. Normally this will happen in a microbatch
	 * to prevent too many updates from occurring. However it is possible to trigger
	 * an immediate update by setting the flag.
	 * @param {Boolean} immediate - trigger an immediate update
	 */
	Layer.prototype.update = function(immediate) {
		var self = this;

		if(immediate == true) {
			run();
			return;
		}

		// Microbatch handler to prevent too many updates
		if(self.microbatch.timeout === false) {
			self.microbatch.timeout = true;
			setTimeout(eventEnd, self.microbatch.delta);
		}					
		function eventEnd() {
			var d = new Date() - self.microbatch.rtime;
			if (new Date() - self.microbatch.rtime < self.microbatch.delta)
				setTimeout(eventEnd, self.microbatch.delta);
			else {
				self.microbatch.rtime = new Date();
				self.microbatch.timeout = false;
				run();
			}               
		};
		
		// Run the update
		function run() {
			if(self.paused == true || self.map.paused == true) return;
			for(var key in self.updateQueue) {
				var component = self.componentMap[key];
				if(component == undefined) continue;
				component.update();
			}
			self.updateQueue = {};
		};
	};
	
	/**
	 * Pauses updates on a layer.
	 */
	Layer.prototype.pause = function() {
		this.paused = true;
	};
	
	/**
	 * Resumes a paused layer.
	 */
	Layer.prototype.resume = function() {
		if(this.paused == false) return;
		this.paused = false;
		for(var idx = 0; idx < this.layersArr.length; idx++)
			this.layersArr[idx].update();
	};
	
	/** 
	 * Returns a color for the given value from the color map for this layer.
	 * @param {Object} value - the value to lookup, can be anything but typically a String
	 * @return {String} the color
	 */
	Layer.prototype.colormap = function(value) {
		var color = this.colorValMap[value];
		if(color == undefined) {
			color = this.colorscale[this.colorIdx++];
			if(this.colorIdx == this.colorscale.length) {
				this.colorIdx = 0;
			}
			this.colorValMap[value] = color; 
		}
		return color;
	};
	
	/**
	 * Creates a UI component for the given object based on configuration for this layer.
	 * @param {Object} object - the object to create component for
	 */
	Layer.prototype.createComponent = function(object) {
		if(this.configuration.marker != undefined) {
			return new Marker(object, this.configuration.marker, this);
		}
		else if(this.configuration.polyline != undefined) {
			return new Polyline(object, this.configuration.polyline, this);
		}
		else if(this.configuration.polygon != undefined) {
			return new Polygon(object, this.configuration.polygon, this);
		}
		else if(this.configuration.rectangle != undefined) {
			return new Rectangle(object, this.configuration.rectangle, this);
		}
		else if(this.configuration.circle != undefined) {
			return new Circle(object, this.configuration.circle, this);
		}
		else if(this.configuration.circleMarker != undefined) {
			return new CircleMarker(object, this.configuration.circleMarker, this);
		}
		else
			throw 'Unknown layer component type';
	}
	
	
	
	/** 
	 * Creates a new component on the specified layer.
	 * @param {Object} object - the object to bind to the component
	 * @param {Object} configuration - component configuration
	 * @param {Object} layer - layer to add the component on
	 */
	function AbstractComponent(object, configuration, layer) {
		this._key			= object._key;
		this.object 		= object;
		this.configuration	= configuration;
		this.layer			= layer;

		this.node			= null;
		this.addedToLayer	= false;
		
		this.tooltip		= null;
		this.popup			= null;
		
		this.hovered		= false;
		this.marked			= false;
		this.filtered		= false;
		this.searched		= null;
		
		// Create the property index
		var propertyIndexResult = DataFlow.Utilities.PropertyMap.index(this.configuration, ['property', 'eval']);
		this.propertyIndex = propertyIndexResult.propertyIndex;
		this.properties = propertyIndexResult.properties;
		
		// Create a property index for tooltip
		if(this.layer.configuration.tooltip != undefined) {
			var tooltipPropertyIndexResult = DataFlow.Utilities.PropertyMap.index(this.layer.configuration.tooltip, ['property', 'eval']);
			this.tooltipPropertyIndex = tooltipPropertyIndexResult.propertyIndex;
			this.tooltipProperties = tooltipPropertyIndexResult.properties;
		}
		
		// Create a property index for popup
		if(this.layer.configuration.popup != undefined) {
			var popupPropertyIndexResult = DataFlow.Utilities.PropertyMap.index(this.layer.configuration.popup, ['property', 'eval']);
			this.popupPropertyIndex = popupPropertyIndexResult.propertyIndex;
			this.popupProperties = popupPropertyIndexResult.properties;

			// If there is a contentUrl load that now synchronously
			var self = this;
			if(this.layer.configuration.popup.contentUrl != undefined) {
				$.ajax({
					type: 'GET',
					url: this.layer.configuration.popup.contentUrl,
					dataType: 'html',
					cache: false,
					aync: false,
					success: function(data, textStatus, xhr) {
						self.popupPropertyIndex.arr.push({'prop': 'content', 'type': 'eval'});
						var thisData = data;
						if(!thisData.startsWith('`')) thisData = '`' + thisData;
						if(!thisData.endsWith('`')) thisData = thisData + '`';
						self.popupPropertyIndex.map.content = thisData;
					},
					error: function(xhr, textStatus, errorThrown) {
						console.error("Layer content url not found or invalid - " + self.configuration.contentUrl);
					}
				});	
			}
		}
	};

	/**
	 * Draws a new component using base configuration. 
	 */
	AbstractComponent.prototype.draw = function() {
		var self = this;
		
		// Add popup if enabled
		if(this.layer.configuration.popup != undefined) {
			this.node.bindPopup('', this.popupProperties);
		}
			
		// Add tooltip if enabled
		if(this.layer.configuration.tooltip != undefined) {
			this.node.bindTooltip('', this.tooltipProperties);
		}
		
		// Event handler for mouseover
		L.DomEvent.on(this.node, 'mouseover', function(e) {
			self.layer.hover(self.object, true);
		});		

		// Event handler for mouseout
		L.DomEvent.on(this.node, 'mouseout', function(e) {
			if(self.layer.configuration.options.hover.enabled == false) return;
			self.layer.hover(null, true);
		});		

		// Event handler for 
		L.DomEvent.on(this.node, 'click', function(e) {
			if(self.layer.configuration.options.marking.enabled == false) return;
			var ctrlKey = e.originalEvent.ctrlKey;					
			var mode = self.layer.configuration.options.marking.mode;
			var isMarked = self.marked == true;
			
			// If unmarking is disabled and this component is marked, then just call mark again
			//   and take no further action
			if(self.configuration.options.marking.unmark == false && isMarked == true) {
				self.layer.mark(self.object, true);
				return;
			}
			
			// Single mode
			if(mode == 'single') {
				if(isMarked == true)
					self.layer.mark(null, true);
				else
					self.layer.mark(self.object, true);
			}
			// Multiple mode
			else if(mode == 'multiple') {
				if(ctrlKey == true) {
					var objects = [];
					for(var idx = 0; idx < self.layer.marked.length; idx++)
						objects.push(self.layer.marked[idx].object);
					if(isMarked == false)
						objects.push(self.object);
					else if(isMarked == true)
						objects.splice(objects.indexOf(self.object), 1);
					self.layer.mark(objects, true);
				}
				else {
					if(isMarked == true)
						self.layer.mark(null, true);
					else
						self.layer.mark(self.object, true);
				}
			}
		});		
	};
	
	/**
	 * Update the component.
	 */
	AbstractComponent.prototype.update = function() {
		// Update the properties for the component
		DataFlow.Utilities.PropertyMap.evaluate(this.propertyIndex, this.getContext(), this.object, this.properties);
		
		// Update popup content
		if(this.layer.configuration.popup != undefined) {
			DataFlow.Utilities.PropertyMap.evaluate(this.popupPropertyIndex, this.getContext(), this.object, this.popupProperties);
			var contentDiv = Utilities.Template.process(this.popupProperties.content);
			this.node.setPopupContent(contentDiv[0]);
		}
		
		// Update tooltip content
		if(this.layer.configuration.tooltip != undefined) {
			DataFlow.Utilities.PropertyMap.evaluate(this.tooltipPropertyIndex, this.getContext(), this.object, this.tooltipProperties);
			this.node.setTooltipContent(this.tooltipProperties.text);
		}
		
		// Add to layer if not already done so
		if(this.addedToLayer == false) {
			this.node.addTo(this.layer.node);
			this.addedToLayer = true;
			
			// Check initial filters and searches
			var filterInclude = DataFlow.Utilities.Filtering.filterItem(this, 
				this.layer.filterDef.filter.filters, this.layer.filterDef.filter.fields, 'object');
			var filtered = !filterInclude;
			this.filter(filtered);
			
			var searchInclude = DataFlow.Utilities.Filtering.searchItem(this, 
				this.layer.filterDef.search.expression, this.layer.filterDef.search.fields, 'object');
			this.search(searchInclude);
		}
	};
	
	/**
	 * Removes the component from the layer and invalidates the
	 * node for garbage collection.
	 */
	AbstractComponent.prototype.remove = function() {
		if(this.node == null) return;
		L.DomEvent.off(this.node);
		this.layer.node.removeLayer(this.node);
		this.node = null;
	};
	
	/**
	 * Returns a context object for this component.
	 * @return {Object} the context object
	 */
	AbstractComponent.prototype.getContext = function() {
		var self = this;
		function colormap(value, name) {
			if(name == undefined)
				return self.layer.colormap.getColor(value);
			else {
				var colormap = self.layer.map.colorMaps[name];
				if(colormap == undefined) return;
				return colormap.getColor(value);
			}
		};

		// Function to format the specified value, which can be timestamp or number
		function format(val, format) {
			return DataFlow.Utilities.Format.format(val, format);
		};
		
		var context = {
			object:		this.object,
			component:	this,
			layer:		this.layer,
			map:		this.layer.map,
			colormap:	colormap, 
			format:		format
		};		
		
		return context;
	}
	
	/**
	 * Sets the filtered flag for the component and adds or removes from the layer.
	 * @param {Boolean} filtered - true if filtered, otherwise false
	 */
	AbstractComponent.prototype.filter = function(filtered) {
		if(this.filtered == true && filtered == false) {
			this.filtered = false;
			this.node.addTo(this.layer.node);
		}
		else if(this.filtered == false && filtered == true) {
			this.filtered = true;
			this.layer.node.removeLayer(this.node);
		}
	}
	
	/**
	 * Sets the search flag for the component and adds or removes from the layer.
	 * @param {Boolean} include - true if search match, otherwise false
	 */
	AbstractComponent.prototype.search = function(include) {
		if(this.searched == null && include == true) {
			this.searched = true;
		}
		if(this.searched == null && include == false) {
			this.searched = false;
			this.layer.node.removeLayer(this.node);
		}
		
		if(this.searched == true && include == null) {
			this.searched = null;
		}
		if(this.searched == true && include == false) {
			this.searched = false;
			this.layer.node.removeLayer(this.node);
		}
		
		if(this.searched == false && include == null) {
			this.searched = null;
			this.node.addTo(this.layer.node);
		}
		if(this.searched == false && include == true) {
			this.searched = true;
			this.node.addTo(this.layer.node);
		}
	}
	
	
	
	
	/** 
	 * Creates a new marker on the specified layer.
	 * @param {Object} object - the object to bind to the marker
	 * @param {Object} configuration - marker configuration
	 * @param {Object} layer - layer to add the marker on
	 */
	function Marker(object, configuration, layer) {
		AbstractComponent.call(this, object, configuration, layer);
		
		// Create a property index for icon or divIcon, prioritize icon first
		if(this.configuration.icon != undefined) {
			var iconPropertyIndexResult = DataFlow.Utilities.PropertyMap.index(this.configuration.icon, ['property', 'eval']);
			this.iconPropertyIndex = iconPropertyIndexResult.propertyIndex;
			this.iconProperties = iconPropertyIndexResult.properties;
		}		
		else if(this.configuration.divIcon != undefined) {
			var divIconPropertyIndexResult = DataFlow.Utilities.PropertyMap.index(this.configuration.divIcon, ['property', 'eval']);
			this.divIconPropertyIndex = divIconPropertyIndexResult.propertyIndex;
			this.divIconProperties = divIconPropertyIndexResult.properties;
			
			// If there is an htmlUrl load that now synchronously
			var self = this;
			if(this.configuration.divIcon.htmlUrl != undefined) {
				$.ajax({
					type: 'GET',
					url: this.configuration.divIcon.htmlUrl,
					dataType: 'html',
					cache: false,
					aync: false,
					success: function(data, textStatus, xhr) {
						self.divIconPropertyIndex.arr.push({'prop': 'html', 'type': 'eval'});
						var thisData = data;
						if(!thisData.startsWith('`')) thisData = '`' + thisData;
						if(!thisData.endsWith('`')) thisData = thisData + '`';
						self.divIconPropertyIndex.map.html = thisData;
					},
					error: function(xhr, textStatus, errorThrown) {
						console.error("Layer html url not found or invalid - " + self.configuration.divIcon.htmlUrl);
					}
				});	
			}
			
		}		
		
		this.draw();
	}
	
	Marker.prototype = Object.create(AbstractComponent.prototype);
	Marker.prototype.constructor = Marker;
	
	/**
	 * Draws a new marker using base configuration. 
	 */
	Marker.prototype.draw = function() {
		this.node = L.marker([0, 0], this.properties);
		AbstractComponent.prototype.draw.call(this);
		// No update here, will be triggered by layer.update()
	};
	
	/**
	 * Update the marker properties if the node exists. First it
	 * will evaluate properties on the configuration based on the component
	 * context. Then the marker position and icon will be updated
	 * based on the configuration. Finally, if the node has not yet been added
	 * to the containing layer, it will be added now.
	 */
	Marker.prototype.update = function() {
		if(this.node == null) return;

		// Update icon or divIcon
		if(this.configuration.icon != undefined) {
			DataFlow.Utilities.PropertyMap.evaluate(this.iconPropertyIndex, this.getContext(), this.object, this.iconProperties);
			var icon = L.icon(this.iconProperties);
			this.node.setIcon(icon);
		}
		else if(this.configuration.divIcon != undefined) {
			DataFlow.Utilities.PropertyMap.evaluate(this.divIconPropertyIndex, this.getContext(), this.object, this.divIconProperties);
			this.divIconProperties.html = $(Utilities.Template.process(this.divIconProperties.html)).html();
			var icon = L.divIcon(this.divIconProperties);
			this.node.setIcon(icon);
		}
		
		// Update lat/lng
		this.node.setLatLng([this.object[this.configuration.property_lat], this.object[this.configuration.property_lon]]);

		AbstractComponent.prototype.update.call(this);
	};
	
	
	
	
	/**
	 * Creates a new polyline.
	 * @param {Object} object - the object to bind to the polyline
	 * @param {Object} configuration - the configuration for the polyline
	 * @param {Object} layer - the layer to add the polyline on
	 */
	function Polyline(object, configuration, layer) {
		AbstractComponent.call(this, object, configuration, layer);		
		this.draw();
	};
	
	Polyline.prototype = Object.create(AbstractComponent.prototype);
	Polyline.prototype.constructor = Polyline;

	/**
	 * Draws a new polyline using base configuration. 
	 * Polylines are always drawn from the context of a single object.
	 * They may be drawn using two modes:
	 *   1. Start and end component based on properties of the object
	 *   2. Mulitple components based on a property containing an array of
	 *        other objects with lat and lon 
	 */
	Polyline.prototype.draw = function() {
		var coordinates = [[0, 0]];
		this.node = L.polyline(coordinates, this.configuration);
		AbstractComponent.prototype.draw.call(this);
		// No update here, will be triggered by layer.update()
	};
	
	/**
	 * Update the polyline properties if the node exists. First it
	 * will evaluate properties on the configuration based on the component
	 * context. Then the polyline position and style will be updated
	 * based on the configuration. Finally, if the node has not yet been added
	 * to the containing layer, it will be added now.
	 */
	Polyline.prototype.update = function() {
		if(this.node == null) return;
		AbstractComponent.prototype.update.call(this);

		// Update lat/lon and style
		var coordinates = null;
		if(this.properties.lat1 != undefined && this.properties.lon1 != undefined && 
				this.properties.lat2 != undefined && this.properties.lon2 != undefined) {
			coordinates = [[this.properties.lat1, this.properties.lon1], 
				[this.properties.lat2, this.properties.lon2]];
		}
		else if(this.properties.points != undefined) {
			coordinates = [];
			var self = this;
			var sorted = this.properties.points.sort(function(a, b) {
				var dir = self.properties.ordering.direction == 'desc' ? -1 : 1;
				return dir * (a[self.properties.ordering.field] > b[self.properties.ordering.field]) ? 1 : 
					((b[self.properties.ordering.field] > a[self.properties.ordering.field]) ? -1 : 0);
			});
			if(sorted != undefined) {
				for(var idx = 0; idx < sorted.length; idx++) {
					coordinates.push([sorted[idx].lat, sorted[idx].lon]);
				}
			}
		}
		this.node.setLatLngs(coordinates);
		this.node.setStyle(this.properties);

		// Call generic component update
		AbstractComponent.prototype.update.call(this);
	};



	
	/**
	 * Creates a new polygon.
	 * @param {Object} object - the object to bind to the polygon
	 * @param {Object} configuration - the configuration for the polygon
	 * @param {Object} layer - the layer to add the polygon on
	 */
	function Polygon(object, configuration, layer) {
		AbstractComponent.call(this, object, configuration, layer);		
		this.draw();
	};
	
	Polygon.prototype = Object.create(AbstractComponent.prototype);
	Polygon.prototype.constructor = Polygon;

	/**
	 * Draws a new polygon using base configuration. 
	 * Polygones are always drawn from the context of a single object.
	 * There must be a property containing multiple points with lat and lon.
	 */
	Polygon.prototype.draw = function() {
		var coordinates = [[0, 0]];
		this.node = L.polygon(coordinates, this.configuration);
		AbstractComponent.prototype.draw.call(this);
		// No update here, will be triggered by layer.update()
	};
	
	/**
	 * Update the polygon properties if the node exists. First it
	 * will evaluate properties on the configuration based on the component
	 * context. Then the polygon position and style will be updated
	 * based on the configuration. Finally, if the node has not yet been added
	 * to the containing layer, it will be added now.
	 */
	Polygon.prototype.update = function() {
		if(this.node == null) return;
		AbstractComponent.prototype.update.call(this);

		// Update lat/lon and style
		var coordinates = [];
		var self = this;
		var sorted = this.properties.points.sort(function(a, b) {
			var dir = self.properties.ordering.direction == 'desc' ? -1 : 1;
			return dir * (a[self.properties.ordering.field] > b[self.properties.ordering.field]) ? 1 : 
				((b[self.properties.ordering.field] > a[self.properties.ordering.field]) ? -1 : 0);
		});
		if(sorted != undefined) {
			for(var idx = 0; idx < sorted.length; idx++) {
				coordinates.push([sorted[idx].lat, sorted[idx].lon]);
			}
		}
		this.node.setLatLngs(coordinates);
		this.node.setStyle(this.properties);

		// Call generic component update
		AbstractComponent.prototype.update.call(this);
	};

	
	
	
	/**
	 * Creates a new rectangle.
	 * @param {Object} object - the object to bind to the rectangle
	 * @param {Object} configuration - the configuration for the rectangle
	 * @param {Object} layer - the layer to add the rectangle on
	 */
	function Rectangle(object, configuration, layer) {
		AbstractComponent.call(this, object, configuration, layer);		
		this.draw();
	};
	
	Rectangle.prototype = Object.create(AbstractComponent.prototype);
	Rectangle.prototype.constructor = Rectangle;

	/**
	 * Draws a new rectangle using base configuration. 
	 */
	Rectangle.prototype.draw = function() {
		var bounds = [[0, 0], [0, 0]];
		this.node = L.rectangle(bounds, this.configuration);
		AbstractComponent.prototype.draw.call(this);
		// No update here, will be triggered by layer.update()
	};
	
	/**
	 * Update the rectangle properties if the node exists. First it
	 * will evaluate properties on the configuration based on the component
	 * context. Then the rectangle position and style will be updated
	 * based on the configuration. Finally, if the node has not yet been added
	 * to the containing layer, it will be added now.
	 */
	Rectangle.prototype.update = function() {
		if(this.node == null) return;
		AbstractComponent.prototype.update.call(this);

		// Update bounds and style
		var bounds = [[this.properties.lat1, this.properties.lon1], 
			[this.properties.lat2, this.properties.lon2]];
		this.node.setBounds(bounds);
		this.node.setStyle(this.properties);

		// Call generic component update
		AbstractComponent.prototype.update.call(this);
	};

	
	
	
	/**
	 * Creates a new circle.
	 * @param {Object} object - the object to bind to the circle
	 * @param {Object} configuration - the configuration for the circle
	 * @param {Object} layer - the layer to add the circle on
	 */
	function Circle(object, configuration, layer) {
		AbstractComponent.call(this, object, configuration, layer);		
		this.draw();
	};
	
	Circle.prototype = Object.create(AbstractComponent.prototype);
	Circle.prototype.constructor = Circle;

	/**
	 * Draws a new circle using base configuration. 
	 */
	Circle.prototype.draw = function() {
		this.node = L.circle([0,0], this.configuration);
		AbstractComponent.prototype.draw.call(this);
		// No update here, will be triggered by layer.update()
	};
	
	/**
	 * Update the circle properties if the node exists. First it
	 * will evaluate properties on the configuration based on the component
	 * context. Then the circle position, radius, and style will be updated
	 * based on the configuration. Finally, if the node has not yet been added
	 * to the containing layer, it will be added now.
	 */
	Circle.prototype.update = function() {
		if(this.node == null) return;
		AbstractComponent.prototype.update.call(this);

		// Update lat/lon, radius, and style
		this.node.setLatLng([this.properties.lat, this.properties.lon]);
		this.node.setRadius(this.properties.radius);
		this.node.setStyle(this.properties);
		
		// Call generic component update
		AbstractComponent.prototype.update.call(this);
	};
	
	

	
	/**
	 * Creates a new circle marker.
	 * @param {Object} object - the object to bind to the marker
	 * @param {Object} configuration - the configuration for the marker
	 * @param {Object} layer - the layer to add the marker on
	 */
	function CircleMarker(object, configuration, layer) {
		AbstractComponent.call(this, object, configuration, layer);		
		this.draw();
	};
	
	CircleMarker.prototype = Object.create(AbstractComponent.prototype);
	CircleMarker.prototype.constructor = CircleMarker;

	/**
	 * Draws a new circle marker using base configuration. 
	 */
	CircleMarker.prototype.draw = function() {
		this.node = L.circleMarker([0, 0], this.configuration);
		AbstractComponent.prototype.draw.call(this);
		// No update here, will be triggered by layer.update()
	};
	
	/**
	 * Update the circle marker properties if the node exists. First it
	 * will evaluate properties on the configuration based on the component
	 * context. Then the marker position, radius, and style will be updated
	 * based on the configuration. Finally, if the node has not yet been added
	 * to the containing layer, it will be added now.
	 */
	CircleMarker.prototype.update = function() {
		if(this.node == null) return;
		AbstractComponent.prototype.update.call(this);

		// Update lat/lon, radius, and style
		this.node.setLatLng([this.properties.lat, this.properties.lon]);
		this.node.setRadius(this.properties.radius);
		this.node.setStyle(this.properties);
		
		// Call generic component update
		AbstractComponent.prototype.update.call(this);
	};
	
	
	MapConfiguration = {
		defaultConfiguration: {
				leaflet: {
					control: {
						
					},
					tileLayer: {
						provider: 			'http://{s}.tile.openstreetmap.se/hydda/full/{z}/{x}/{y}.png',
						options: {
							attribution:	'Tiles courtesy of <a href="http://openstreetmap.se/" target="_blank">OpenStreetMap Sweden</a> &mdash; Map data &copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>',
							minZoom:		0,
							maxZoom:		18,
							opacity:		1.0
						}
						/* 
							Maporama: http://tilescf.maporama.com/georasterservices/v101/standard_map/{z}/{x}/{y}.png
							OSM: http://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png
							MapQuest Open: http://otile1.mqcdn.com/tiles/1.0.0/map/{z}/{x}/{y}.png
							ArcGIS Aerial: http://server.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile/{z}/{y}/{x}
							Other providers: http://leaflet-extras.github.io/leaflet-providers/preview/
						*/
					}
				},
				options: {
					configurable:		false,
					copy: {
						enabled:		true
					},
					dataset: {
						icon:			false
						//modal:			undefined
					},
					search: {
						fields:			undefined,
						delay:			500,
						enabled:		true,
						label:			"Search"
					},
				},
				
			},
		
		defaultLayerConfiguration: {
				options: {
					filter: {	
						accept:			true,
						fields:			undefined,
						delay:			500,
						enabled:		true,
						passthrough:	true
					},
					hover: {
						accept:			true,
						enabled:		true,
						passthrough:	true					
					},
					marking: {
						accept:			true,
						enabled:		true,
						mode:			"single",
						passthrough:	true,
						unmark:			true
					},
					search: {
						fields:			undefined,
						delay:			500,
						enabled:		true,
						label:			"Search"
					}
				}
			},
			
		// Merges the source configuration object into the target configuration object
		//   Source values will overwrite target values.
		//   Should always merge a DEFAULT configuration source into a USER configuration target
		mergeConfiguration:		function(source, target) {
									for(var idx = 0; idx < Object.keys(source).length; idx++) {
										var key = Object.keys(source)[idx];
										var val = source[key];
										if(typeof val === 'object') {
											if(target[key] == undefined) 
												target[key] = val;
											else
												this.mergeConfiguration(val, target[key]);
										}
										else {
											if(target[key] == undefined) 
												target[key] = val;
										}
									}
									return target;
								},

		// Processes the specified configuration object and merges in default configuration
		//   for base object 
		processConfiguration:	function(config) {
									var configuration = this.mergeConfiguration(this.defaultConfiguration, config);
									for(var idx = 0; idx < Object.keys(configuration.leaflet.layers).length; idx++) {
										var layerKey = Object.keys(configuration.leaflet.layers)[idx];
										configuration.leaflet.layers[layerKey] = this.mergeConfiguration($.extend({}, this.defaultLayerConfiguration), 
											configuration.leaflet.layers[layerKey]);
									}
									return configuration;
								},
	};

	
	module.Map 					= Map;
	module.Layer 				= Layer;
	module.MapConfiguration		= MapConfiguration;
	
	return module;
})();

/**
 * Represents a Highcharts chart data sink.
 * @constructor
 * @param {Object} dataset - dataset that this sink will be attached to
 * @param {Object} configuration - configuration object defining the sink
 */
function HighchartsChartDatasink(id, configuration, dataset, containerDiv, overlay) {
	this.id = id;
	this.dataset = dataset;
	this.chart = new LiveHighchartsChart.Chart(id, configuration, containerDiv, overlay, this);
}

/**
 * Event handler for dataset events. 
 * @param {String} eventType - string describing the event that occurred
 * @param {Object} object - data object for the event, or null
 */
HighchartsChartDatasink.prototype.trigger = function(eventType, object) {
	this.chart.trigger(eventType, object);
};


// Define a module to hold object defintions for the chart objects
var LiveHighchartsChart = (function() {
	var module = {};

	/**
	 * Creates a new Chart object.
	 * @constructor
	 * @param {String} id - unique identifier
	 * @param {Object} configuration - the configuration for the chart
	 * @param {Object} containerDiv - this node object that chart will be appended
	 * @param (Function} overlay - function to call on start and end of the overlay
	 * @param {Object} sink - datasink that will feed this chart
	 */
	function Chart(id, configuration, containerDiv, overlay, sink) {
		this.id				= id;
		this.configuration	= configuration;
		this.containerDiv 	= containerDiv;
		this.overlay		= overlay;
		this.sink			= sink;

		// Pause and snapshot flags
		this.paused			= false;
		this.snapshot		= false;

		// Schema for the dataset
		this.schema			= null;		
		
		// Highcharts charting object
		this.highcharts		= null;
		
		// Vars to hold the objects being hovered and marked
		this.hovered		= [];		
		this.marked			= [];

		// Filter definition
		this.filterDef		= {
			search: {
				expression:	undefined,
				fields:		this.configuration.options.search.fields
			},
			filter:	{
				filters:	[],
				fields:		this.configuration.options.filter.fields
			}
		};

		// Color maps
		this.colorMaps		= {};

		// Series array, but not Highcharts series objects, LiveHighchartsChart.Series objects
		this.seriesArr 		= [];
		this.seriesMap		= {};
		
		// Check dependencies. This will throw Error if dependencies are not satisfied.
		this.dependencyCheck();
		
		// Microbatch is used to batch up updates to prevent too many chart redraws
		this.microbatch = {
			rtime:		new Date(1, 1, 2000, 12,00,00),
			timeout:	false,
			delta:		500 // msec between updates
		}

		// Color maps
		if(this.configuration.options.colormaps != undefined) {
			for(var idx = 0; idx < Object.keys(this.configuration.options.colormaps).length; idx++) {
				var thisColorMapName = Object.keys(this.configuration.options.colormaps)[idx];
				var thisColors = this.configuration.options.colormaps[thisColorMapName];
				this.colorMaps[thisColorMapName] = new DataFlow.ColorMap(thisColorMapName, thisColors);
			}
		}
		
		// Loop over the series configuration and create a series object for each
		for(var idx = 0; idx < configuration.highcharts.series.length; idx++) {
			var thisSeries = new Series(configuration.highcharts.series[idx], this, idx);
			this.seriesArr.push(thisSeries);
			this.seriesMap[thisSeries.id] = thisSeries;
		}
		
		// Register UI handlers
		this.registerEventHandlers();

		// Draw the initial empty chart
		this.draw();
	};
	
	/**
	 * Performs a dependency check for rendering the chart.
	 */
	Chart.prototype.dependencyCheck = function() {
		var msg = "Required Highcharts libraries not loaded or not present";
		try {
			var h = Highcharts;
		}
		catch(err) {
			throw Error("Required Highcharts libraries not loaded or not present");
		}
	};
		
	/**
	 * Registers event handlers on UI components.
	 */
	Chart.prototype.registerEventHandlers = function() {
		var self = this;
		
		// Event handler to hide any open context menu on click
		$(document).on('click', function() {
			$(document).find('.live-context-menu').remove();
		});
		
		// Event handler to display the configuration modal
		$(this.containerDiv).find('.chart-config').on('click', function(e) {
			self.displayConfiguration();
		});

		// Event handler to display the dataset modal
		$(this.containerDiv).find('.chart-icon').on('click', function(e) {
			self.displayDataset();
		});
		
		// Event handler to detect mouse over and mouse out of points on the chart to set hover
		if(this.configuration.options.hover.enabled == true) {
			this.configuration.highcharts.plotOptions.series.point.events.mouseOver = function() {
				var object = self.seriesArr[0].pointMap[this._key].object;
				self.hover(object, true);
			};
			this.configuration.highcharts.plotOptions.series.point.events.mouseOut = function() {
				self.hover(null, true);
			};
		}
		
		// Event handler to detect clicks on points to set marking
		if(this.configuration.options.marking.enabled == true) {
			// Adjust default configuration to allow point select
			this.configuration.highcharts.plotOptions.series.allowPointSelect = true;
			
			// Register event handler
			this.configuration.highcharts.plotOptions.series.point.events.click = function(e) {
				var ctrlKey = e.ctrlKey;					
				var mode = self.configuration.options.marking.mode;
				var point = self.seriesArr[0].pointMap[this._key];
				var object = point.object;
				var isMarked = point.marked == true;
				
				// If unmarking is disabled and this component is marked, then just call mark again
				//   and take no further action
				if(self.configuration.options.marking.unmark == false && isMarked == true) {
					self.mark(object, true);
					return;
				}
				
				// Single mode
				if(mode == 'single') {
					if(isMarked == true)
						self.mark(null, true);
					else
						self.mark(object, true);
				}
				// Multiple mode
				else if(mode == 'multiple') {
					if(ctrlKey == true) {
						var objects = [];
						for(var idx = 0; idx < self.marked.length; idx++)
							objects.push(self.marked[idx]);
						if(isMarked == false)
							objects.push(object);
						else if(isMarked == true)
							objects.splice(objects.indexOf(self.object), 1);
						self.mark(objects, true);
					}
					else {
						if(isMarked == true)
							self.mark(null, true);
						else
							self.mark(object, true);
					}
				}
						
				return;
			};
		}
		
		// Event handler for 3D rotation, if enabled
		if(this.configuration.options.rotate3d.enabled == true) {
			$(this.containerDiv).on('mousedown.hc touchstart.hc', function (eStart) {
				var eStart = self.highcharts.pointer.normalize(eStart);

				var posX = eStart.pageX;
				var posY = eStart.pageY;
				var alpha = self.highcharts.options.chart.options3d.alpha;
				var beta = self.highcharts.options.chart.options3d.beta;
				var newAlpha = null;
				var newBeta = null;
				var sensitivity = 5; // lower is more sensitive

				$(document).on({
					'mousemove.hc touchdrag.hc': function (e) {
						// Run beta
						newBeta = beta + (posX - e.pageX) / sensitivity;
						self.highcharts.options.chart.options3d.beta = newBeta;

						// Run alpha
						newAlpha = alpha + (e.pageY - posY) / sensitivity;
						self.highcharts.options.chart.options3d.alpha = newAlpha;

						self.highcharts.redraw(false);
					},
					'mouseup touchend': function () {
						$(document).off('.hc');
					}
				});
			});
		}
	};
	
		
	/**
	 * Event handler for all objects delegated from the datasink.
	 * @param {String} eventType - the event type
	 * @param {Object} object - optional object
	 */
	Chart.prototype.trigger = function(eventType, object) {
		var self = this;
		
		if(eventType == 'schemaChange') {
			this.schema = object;
		}
		else if(eventType == 'snapshotStart') {
			this.snapshot = true;
			this.overlay(true);
		}
		else if(eventType == 'snapshotEnd') {
			setTimeout(function() {
				self.snapshot = false;
				self.overlay(false);
			},500);
		}
		else if(eventType == 'clear') {
			this.clear();
		}
		else if(eventType == 'insert') {
			this.addPoint(object);
		}
		else if(eventType == 'update') {
			this.updatePoint(object);
		}
		else if(eventType == 'upsert') {
		}
		else if(eventType == 'delete') {
			this.removePoint(object);
		}
		else if(eventType == 'hover') {
			if(this.configuration.options.hover.accept != false)
				this.hover(object);
		}
		else if(eventType == 'marked') {
			if(this.configuration.options.marking.accept != false)
				this.mark(object);
		}
		else if(eventType == 'filter') {
			if(this.configuration.options.filter.accept != false)
				this.filter(object);
		}					
		
	};
	
	/**
	 * Add a new point bound to the specified object on the chart.
	 * @param {Object} object - the object to add the bound point
	 */
	Chart.prototype.addPoint = function(object) {
		for(var idx = 0; idx < this.seriesArr.length; idx++)
			this.seriesArr[idx].addPoint(object);
		this.update();
	};

	/**
	 * Updates the point bound to the specified object on the chart.
	 * @param {Object} object - the object to update the bound point
	 */
	Chart.prototype.updatePoint = function(object) {
		for(var idx = 0; idx < this.seriesArr.length; idx++)
			this.seriesArr[idx].updatePoint(object);
		this.update();
	};
	
	/**
	 * Removes the point bound to the specified object from the chart.
	 * @param {Object} object - the object to remove the bound point
	 */
	Chart.prototype.removePoint = function(object) {
		for(var idx = 0; idx < this.seriesArr.length; idx++)
			this.seriesArr[idx].removePoint(object);
		this.update();
	};
					
	/**
	 * Clears all objects and all points from the chart.
	 */
	Chart.prototype.clear = function() {
		for(var idx = 0; idx < this.seriesArr.length; idx++)
			this.seriesArr[idx].removeAll();
		this.update();
	};
	
	/**
	 * Draws the UI components for the chart.
	 */
	Chart.prototype.draw = function() {
		this.highcharts = Highcharts.chart(this.containerDiv, this.configuration.highcharts);	;
	};
	
	/**
	 * Updates the contents of the chart. Normally this will happen in a microbatch
	 * to prevent too many updates from occurring. However it is possible to trigger
	 * an immediate update by setting the flag.
	 * @param {Boolean} immediate - trigger an immediate update
	 */
	Chart.prototype.update = function(immediate) {
		var self = this;

		if(immediate == true) {
			run();
			return;
		}

		// Microbatch handler to prevent too many updates
		if(self.microbatch.timeout === false) {
			self.microbatch.timeout = true;
			setTimeout(eventEnd, self.microbatch.delta);
		}					
		function eventEnd() {
			var d = new Date() - self.microbatch.rtime;
			if (new Date() - self.microbatch.rtime < self.microbatch.delta)
				setTimeout(eventEnd, self.microbatch.delta);
			else {
				self.microbatch.rtime = new Date();
				self.microbatch.timeout = false;
				run();
			}               
		};
		
		// Run the update
		function run() {
			if(self.paused == true) return;
			self.highcharts.redraw();
		};
	};
	
	/**
	 * Pauses updates on a chart.
	 */
	Chart.prototype.pause = function() {
		this.paused = true;
	};
	
	/**
	 * Resumes a paused chart.
	 */
	Chart.prototype.resume = function() {
		if(this.paused == false) return;
		this.paused = false;
		this.update();
	};
	
	/**
	 * Resizes a chart.
	 */
	Chart.prototype.resize = function() {
		this.highcharts.setSize($(this.containerDiv).parent().width(), 
			$(this.containerDiv).parent().height());
	};
			
	/**
	 * Applies a search value against the chart. This will not be passed through
	 * to the dataset and is local only to the current chart.
	 * @param {String} value - the search string
	 */
	Chart.prototype.search = function(value) {
		if(value == undefined || value.length == 0) {
			delete this.filterDef.search.expression;
		}
		else {
			this.filterDef.search.expression = value;
		}
		
		// Pass the filterDef through to the series
		for(var idx = 0; idx < this.seriesArr.length; idx++) {
			this.seriesArr[idx].filter(this.filterDef);
		}
	};
	
	/**
	 * Hovers a single point bound to the specified object.
	 * Will also optionally notify the dataset of a hover event, if enabled.
	 * @param {Object} object - object that the associated point will be hovered
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Chart.prototype.hover = function(object, notify) {
		if(this.configuration.options.hover.enabled != false) {
			// Unhover all existing series
			for(var idx = 0; idx < this.seriesArr.length; idx++) {
				this.seriesArr[idx].unhover();
			}
			
			// If the object is not null, process the hovering
			if(object != null) {
				if(Array.isArray(object)) {
					this.hovered = object.slice(0);
				}
				else {
					this.hovered = [object];
				}

				for(var idx = 0; idx < this.seriesArr.length; idx++) {
					this.seriesArr[idx].hover(this.hovered);
				}
			}

			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.hover.passthrough == true) {
				this.sink.dataset.trigger('hover', this.sink, object);
			}			
		}
	};
		
	/**
	 * Marks a single point or a set of points bound to the specified object or array of objects.
	 * Will also optionally notify the dataset of a marked event, if enabled.
	 * @param {Object|Array} object - if a single data object then the associated point will be marked
	 *                                if an array of objects then the associated points will be marked
	 *                                  for single marking points that receive a list of objects then only
	 *                                  the first point will be marked
	 *                                if null then all marked points will be unmarked
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Chart.prototype.mark = function(object, notify) {
		if(this.configuration.options.marking.enabled != false) {
			var mode = this.configuration.options.marking.mode;

			// Unmark all existing series
			for(var idx = 0; idx < this.seriesArr.length; idx++) {
				this.seriesArr[idx].unmark();
			}
			this.marked = [];
			
			// If the object is not null, process the marking
			if(object != null) {
				if(Array.isArray(object)) {
					if(this.configuration.options.marking.mode == 'multiple') {
						for(var idx = 0; idx < object.length; idx++) {
							var thisObject = object[idx];
							this.marked.push(thisObject);
						}
					}
					// If we receive a list of objects but can only mark one, mark only the first one
					else if(this.configuration.options.marking.mode == 'single' && object.length > 0) {
						var thisObject = object[0];
						this.marked.push(thisObject);
					}
				}
				else {
					this.marked.push(object);
				}

				// Update the series marked
				for(var idx = 0; idx < this.seriesArr.length; idx++) {
					this.seriesArr[idx].mark(this.marked);
				}
			}
			
			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.marking.passthrough == true) {
				this.sink.dataset.trigger('marked', this.sink, this.marked);
			}			
		}
	};
		
	/**
	 * Applies a filter and updates the chart. Will also optionally notify the
	 * dataset of a filter event, if enabled.
	 * @param {Array} filters - array of filter objects
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Chart.prototype.filter = function(filters, notify) {
		if(this.configuration.options.filter.enabled != false) {
			// This clones the array so we don't get duplicates
			this.filterDef.filter.filters = filters.slice(0);
			this.filterDef.filter.fields = this.configuration.options.filter.fields;

			// Pass the filterDef through to the series
			for(var idx = 0; idx < this.seriesArr.length; idx++) {
				this.seriesArr[idx].filter(this.filterDef);
			}

			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.filter.passthrough == true) {
				this.sink.dataset.trigger('filter', this.sink, this.filtered.filter);
			}
			
			this.update();
		}
	};
	
	/**
	 * Display the chart configuration modal.
	 */
	Chart.prototype.displayConfiguration = function() {
		if(this.configuration.options.configurable != true) return;
		console.log('highcharts display config');
		// TODO display configuration modal
	};

	/**
	 * Displays the dataset configuration modal.
	 */
	Chart.prototype.displayDataset = function() {
		if(this.configuration.options.dataset.modal == undefined) return;
		console.log('highcharts display dataset');
		// TODO display dataset modal
	};
						
	/** 
	 * Returns a color for the given value from the color map.
	 * @param {Object} value - the value to lookup, can be anything but typically a String
	 * @return {String} the color
	 */
	Chart.prototype.colormap = function(value) {
		return this.colorMap.getColor(value);
	};
	
	
	
	
	/**
	 * Creates a new Series object which represents a set of data on a chart. 
	 * A Series will represent different things depending on the underlying chart type.
	 * Typically a set of points or a line on a chart will be one series. A pie chart
	 * consists of a single series.
	 * @constructor
	 * @param {Object} configuration - the configuration for the series
	 * @param {Object} chart - the chart containing this series
	 */
	function Series(configuration, chart, index) {
		// Set configuration
		this.configuration 		= configuration;
		this.chart 				= chart;
		this.id					= 'series-' + index;
		
		this.highchartsSeries	= null;
		
		this.hovered			= [];
		this.marked				= [];
		this.filterDef			= {
			search: {
				expression:	undefined,
				fields:		this.chart.configuration.options.search.fields
			},
			filter:	{
				filters:	[],
				fields:		this.chart.configuration.options.filter.fields
			}
		};

		// Initialize map for point objects
		this.pointMap			= {};
		
		// Create the property index
		var pointPropertiesIndexResult = DataFlow.Utilities.PropertyMap.index(this.configuration.pointProperties, ['property', 'eval']);
		if(pointPropertiesIndexResult != null) {
			this.pointPropertiesIndex = pointPropertiesIndexResult.propertyIndex;
			this.pointProperties = pointPropertiesIndexResult.properties;
		}

		this.configuration.id 	= this.id;
		if(this.configuration.data == undefined)
			this.configuration.data = [];
	};

	/**
	 * Add a new object to the series as a new point.
	 * @param {Object} object - the object to add
	 */
	Series.prototype.addPoint = function(object) {
		if(this.pointPropertiesIndex == null) return;
		
		var filtered = this.checkFilter(object);
		if(filtered == true) return;

		var data = {
			_key:				object._key,
			id:					this.id + "-" + object._key,
			object:				object,
			selected:			false
		}
		var point = {
			data:				data,
			object:				object,
			highchartsPoint: 	null,
			marked:				false,
			added:				false
		}

		// Build a context and evaluate the point properties, append to the data object
		var context = {
			object:		object,
			series:		this,
			chart:		this.chart
		};
		DataFlow.Utilities.PropertyMap.evaluate(this.pointPropertiesIndex, context, object, data);
		
		// Add the point to the point map
		this.pointMap[object._key] = point;

		// Check the filter
		var include = DataFlow.Utilities.Filtering.filterItem(data, 
			this.filterDef.filter.filters, this.filterDef.filter.fields, 'object');
		
		if(include == true) {
			if(this.highchartsSeries == null) 
				this.highchartsSeries = this.chart.highcharts.get(this.id);
			this.highchartsSeries.addPoint(data, false);
			point.added = true;
		}
	};

	/**
	 * Update an existing object point on the series.
	 * @param {Object} object - the object to update
	 */
	Series.prototype.updatePoint = function(object) {
		if(this.pointPropertiesIndex == null) return;

		var point = this.pointMap[object._key];
		if(point == undefined) return;
		var data = point.data;
		
		// Build a context and evaluate the point properties, append to the data object
		var context = {
			object:		object,
			series:		this,
			chart:		this.chart
		};
		DataFlow.Utilities.PropertyMap.evaluate(this.pointPropertiesIndex, context, object, data);
		
		// Check the filter
		var include = DataFlow.Utilities.Filtering.filterItem(data, 
			this.filterDef.filter.filters, this.filterDef.filter.fields, 'object');
		
		var self = this;
		var doUpdate = function(thisData) {
			if(include == true) {
				if(point.added == false) {
					if(self.highchartsSeries == null) 
						self.highchartsSeries = self.chart.highcharts.get(self.id);
					self.highchartsSeries.addPoint(data, false);
					point.added = true;
				}
				else {
					point.highchartsPoint.update(thisData, false);
				}
			}
			else if(include == false && point.added == true) {
				if(point.highchartsPoint != null) {
					data.color = point.highchartsPoint.color;
					point.highchartsPoint.select(false);		
					point.highchartsPoint.remove(data);
					point.highchartsPoint = null;
				}
				point.added = false;
			}				
		};

		if(point.highchartsPoint == null)
			point.highchartsPoint = this.chart.highcharts.get(data.id);
		if(point.highchartsPoint == null) {
			var self = this;
			setTimeout(function() {
				point.highchartsPoint = self.chart.highcharts.get(data.id);		
				if(point.highchartsPoint == null && point.added == true) return;
				doUpdate(data);
			},1000); // this is highcharts nonsense because sometimes we can't find the point too soon after add
		}
		else {
			doUpdate(data);
		}
	};
	
	/**
	 * Remove an existing object point from the series.
	 * @param {Object} object - the object to remove
	 */
	Series.prototype.removePoint = function(object) {
		if(this.pointPropertiesIndex == null) return;

		var point = this.pointMap[object._key];
		if(point == undefined) return;
		var data = point.data;
		
		var doRemove = function(thisData) {
			point.highchartsPoint.remove(thisData);
			point.added = false;
		};
		
		if(point.highchartsPoint == null)
			point.highchartsPoint = this.chart.highcharts.get(data.id);
		if(point.highchartsPoint == null) {
			var self = this;
			setTimeout(function() {
				point.highchartsPoint = self.chart.highcharts.get(data.id);		
				if(point.highchartsPoint == null) return;
				doRemove(data);
			},1000); // this is highcharts nonsense because sometimes we can't find the point too soon after add
		}
		else {
			doRemove(data);
		}
		
		delete this.pointMap[object._key];		
	};
	
	/**
	 * Remove all existing object points from the series.
	 */
	Series.prototype.removeAll = function() {
		if(this.pointPropertiesIndex == null) return;

		if(this.highchartsSeries == null) 
			this.highchartsSeries = this.chart.highcharts.get(this.id);
		this.highchartsSeries.setData([]);
		this.pointMap = {};		
	};
	
	/** 
	 * Updates the trace array with the current property arrays.
	 */
	Series.prototype.update = function() {

	};

	/**
	 * Checks to see if the given object has been filtered by any configured
	 * filter condition.
	 * @param {Object} object - the object to test
	 * @return false if the object is included, true if the object is filtered
	 */
	Series.prototype.checkFilter = function(object) {
		if(this.configuration.pointProperties == null) return false;
		if(this.configuration.pointProperties.eval_filter == undefined) return false;
		var context = {
			object:		object,
			series:		this,
			chart:		this.chart
		};
		var b = DataFlow.Utilities.Evaluate.expression(this.configuration.pointProperties.eval_filter, context);
		return !b;
	};
	
	/**
	 * Unhovers the hovered point if there is one.
	 */
	Series.prototype.unhover = function() {
		for(var idx = 0; idx < this.hovered.length; idx++) {
			var thisObject = this.hovered[idx];
			var point = this.pointMap[thisObject._key];
			if(point == undefined) return;
			var data = point.data;
			if(point.highchartsPoint == null)
				point.highchartsPoint = this.chart.highcharts.get(data.id);
			point.highchartsPoint.setState();
		}
		this.hovered = [];
	};
	
	/**
	 * Hovers the point for the specified object.
	 * @param {Object} object - the object to hover the point for
	 */
	Series.prototype.hover = function(object) {
		for(var idx = 0; idx < object.length; idx++) {
			var thisObject = object[idx];
			var point = this.pointMap[thisObject._key];
			if(point == undefined) return;
			var data = point.data;
			if(point.highchartsPoint == null)
				point.highchartsPoint = this.chart.highcharts.get(data.id);
			point.highchartsPoint.setState('hover');
		}
		this.hovered = object;
	};
		
	/**
	 * Unmarks all marked points on the series.
	 */
	Series.prototype.unmark = function() {
		if(this.marked == null) return;
		for(var idx = 0; idx < this.marked.length; idx++) {
			var thisObject = this.marked[idx];
			var point = this.pointMap[thisObject._key];
			if(point == undefined) continue;
			var data = point.data;
			if(point.highchartsPoint == null)
				point.highchartsPoint = this.chart.highcharts.get(data.id);
			if(point.highchartsPoint != null)
				point.highchartsPoint.select(false);		
			point.marked = false;
			point.data.selected = false;
		}
		this.marked = [];
	};
	
	/**
	 * Marks the point for the specified object or array.
	 * @param {Object[]} object - the array of objects to mark points for
	 */
	Series.prototype.mark = function(object) {
		if(object == null) return;
		for(var idx = 0; idx < object.length; idx++) {
			var thisObject = object[idx];
			var point = this.pointMap[thisObject._key];
			if(point == undefined) continue;
			var data = point.data;
			if(point.highchartsPoint == null)
				point.highchartsPoint = this.chart.highcharts.get(data.id);
			if(point.highchartsPoint != null)
				point.highchartsPoint.select(true, true);
			point.marked = true;
			point.data.selected = true;
			this.marked.push(thisObject);;
		}
	};
	
	/**
	 * Applies a filter definition on a series.
	 * @param {Object} filterDef - filter definition
	 */
	Series.prototype.filter = function(filterDef) {
		var self = this;
		this.filterDef = filterDef;
		
		// Remove all points from the series, but don't discard them from the component
		for(var idx = 0; idx < Object.keys(this.pointMap).length; idx++) {
			var key = Object.keys(this.pointMap)[idx];
			var point = this.pointMap[key];
			var data = point.data;
			if(point.highchartsPoint == null)
				point.highchartsPoint = this.chart.highcharts.get(data.id);
			if(point.highchartsPoint != undefined) {
				data.color = point.highchartsPoint.color;
				point.highchartsPoint.select(false);		
				point.highchartsPoint.remove(data);
				point.highchartsPoint = null;
				point.added = false;
			}
		}

		// Add the points back in to the series
		for(var idx = 0; idx < Object.keys(this.pointMap).length; idx++) {
			var key = Object.keys(this.pointMap)[idx];
			var point = this.pointMap[key];
			var data = point.data;
			var include = DataFlow.Utilities.Filtering.filterItem(data, 
				this.filterDef.filter.filters, this.filterDef.filter.fields, 'object');
			if(include == true) {
				this.highchartsSeries.addPoint(data, false);
				point.added = true;
			}
		}

		setTimeout(function() {
			var marked = self.marked;
			self.unmark();
			self.mark(marked);
		}, 750);
	};
	
	
	
	
	/** 
	 * Highcharts configuration helper object.
	 */
	HighchartsConfiguration = {
		defaultConfiguration: {
				highcharts: {
					chart: {
						animation:		false
					},
					credits: {
						enabled:		false
					},
					exporting: {
						enabled: 		false
					},
					plotOptions: {
						series: {
							point: {
								events: {
								}
							}
						}
					}
				},
				options: {
					configurable:		false,
					copy: {
						enabled:		true
					},
					dataset: {
						icon:			true
						//modal:			undefined
					},
					filter: {	
						accept:			true,
						fields:			undefined,
						delay:			500,
						enabled:		true,
						passthrough:	true
					},
					format: {
						timestamp:		"yyyy-MM-dd HH:mm:ss"
					},
					hover: {
						accept:			true,
						enabled:		true,
						passthrough:	true					
					},
					marking: {
						accept:			true,
						enabled:		true,
						mode:			"single",
						passthrough:	true,
						unmark:			true
					},
					rotate3d: {
					},
					search: {
						fields:			undefined,
						delay:			500,
						enabled:		true,
						label:			"Search"
					}
				},
				
			},
		
		// Merges the source configuration object into the target configuration object
		//   Source values will overwrite target values.
		//   Should always merge a DEFAULT configuration source into a USER configuration target
		mergeConfiguration:		function(source, target) {
									for(var idx = 0; idx < Object.keys(source).length; idx++) {
										var key = Object.keys(source)[idx];
										var val = source[key];
										if(typeof val === 'object') {
											if(target[key] == undefined) 
												target[key] = val;
											else
												this.mergeConfiguration(val, target[key]);
										}
										else {
											if(target[key] == undefined) 
												target[key] = val;
										}
									}
									return target;
								},

		// Processes the specified configuration object and merges in default configuration
		//   for both base object and column objects
		processConfiguration:	function(config) {
									var configuration = this.mergeConfiguration(this.defaultConfiguration, config);
									return configuration;
								},
	};


	
	module.Chart 					= Chart;
	module.HighchartsConfiguration	= HighchartsConfiguration;
	
	return module;
})();

/**
 * Represents a passthrough data sink.
 * @constructor
 * @param {Object} dataset - dataset that this sink will be attached to
 * @param {Object} configuration - configuration object defining the sink
 */
function PassthroughDatasink(id, configuration, dataset, containerDiv, overlay, onUpdate) {
	this.id = id;
	this.dataset = dataset;
	this.onUpdate = onUpdate;
	this.passthrough = new LivePassthrough.Passthrough(id, configuration, containerDiv, overlay, this);
}

/**
 * Event handler for dataset events. 
 * @param {String} eventType - string describing the event that occurred
 * @param {Object} object - data object for the event, or null
 */
PassthroughDatasink.prototype.trigger = function(eventType, object) {
	this.passthrough.trigger(eventType, object);
};


// Define a module to hold object defintions for the passthrough objects
var LivePassthrough = (function() {
	var module = {};

	/**
	 * Creates a new Passthrough object.
	 * @constructor
	 * @param {String} id - unique identifier
	 * @param {Object} configuration - the configuration for the passthrough
	 * @param {Object} containerDiv - this node object that passthrough will be appended
	 * @param (Function} overlay - function to call on start and end of the overlay
	 * @param {Object} sink - datasink that will feed this passthrough
	 */
	function Passthrough(id, configuration, containerDiv, overlay, sink) {
		this.id				= id;
		this.configuration	= configuration;
		this.containerDiv 	= containerDiv;
		this.overlay		= overlay;
		this.sink			= sink;
		
		// Pause and snapshot flags
		this.paused			= false;
		this.snapshot		= false;
		
		// Schema for the dataset
		this.schema			= null;		

		// Vars to hold the objects being hovered and marked
		this.hovered		= [];		
		this.marked			= [];

		// Filter definition
		this.filterDef		= {
			search: {
				expression:	undefined,
				fields:		this.configuration.options.search.fields
			},
			filter:	{
				filters:	[],
				fields:		this.configuration.options.filter.fields
			}
		};

		// Ordering definition
		this.orderDef		= [];

		// Color maps
		this.colorMaps		= {};

		// Components
		this.componentArr	= [];
		this.componentMap	= {};

		// Ordered, filtered, and final result rows
		this.orderedArr		= [];
		this.filterArr		= [];
		this.resultArr		= [];
		
		// Hierarchy
		this.hierarchy = {
			level:		0,
			members:	{},
			values:		[]
		};
		
		// Check dependencies. This will throw Error if dependencies are not satisfied.
		this.dependencyCheck();
		
		// Microbatch is used to batch up updates to prevent too many chart redraws
		this.microbatch = {
			rtime:		new Date(1, 1, 2000, 12,00,00),
			timeout:	false,
			delta:		500 // msec between updates
		}

		// Color maps
		if(this.configuration.options.colormaps != undefined) {
			for(var idx = 0; idx < Object.keys(this.configuration.options.colormaps).length; idx++) {
				var thisColorMapName = Object.keys(this.configuration.options.colormaps)[idx];
				var thisColorScale = this.configuration.options.colormaps[thisColorMapName];
				this.colorMaps[thisColorMapName] = new DataFlow.ColorMap(thisColorMapName, thisColorScale);
			}
		}
		
		// On document ready, register UI handlers
		var self = this;
		setTimeout(function() {
			self.registerEventHandlers();
		}), 0;
		
		// Draw the initial empty passthrough
		this.draw();
	};

	/**
	 * Performs a dependency check for rendering the passthrough.
	 */
	Passthrough.prototype.dependencyCheck = function() {
		// No dependencies for this component
	};
		
	/**
	 * Registers event handlers on UI components.
	 */
	Passthrough.prototype.registerEventHandlers = function() {
		// Event handlers
	};
	
	/**
	 * Event handler for all objects delegated from the datasink.
	 * @param {String} eventType - the event type
	 * @param {Object} object - optional object
	 */
	Passthrough.prototype.trigger = function(eventType, object) {
		var self = this;

		if(eventType == 'schemaChange') {
			this.schema = object;			
		}
		else if(eventType == 'snapshotStart') {
			this.snapshot = true;
			this.overlay(true);
		}
		else if(eventType == 'snapshotEnd') {
			setTimeout(function() {
				self.snapshot = false;
				self.overlay(false);
				self.update();
			},500);
		}
		else if(eventType == 'clear') {
			this.clear();
		}
		else if(eventType == 'insert') {
			this.addComponent(object);
		}
		else if(eventType == 'update') {
			this.updateComponent(object);
		}
		else if(eventType == 'upsert') {
		}
		else if(eventType == 'delete') {
			this.removeComponent(object);
		}
		else if(eventType == 'hover') {
			if(this.configuration.options.hover.accept != false)
				this.hover(object);
		}
		else if(eventType == 'marked') {
			if(this.configuration.options.marking.accept != false)
				this.mark(object);
		}					
		else if(eventType == 'filter') {
			if(this.configuration.options.filter.accept != false)
				this.filter(object);
		}

		this.update();		
	};

	/**
	 * Add a new component bound to the specified object on the passthrough.
	 * @param {Object} object - the object to add the bound component
	 */
	Passthrough.prototype.addComponent = function(object) {
		var component = new Component(object, this);
		this.componentMap[component._key] = component;
		this.componentArr.push(component);

		// Process hierarchy if configured
		if(this.configuration.options.hierarchy != undefined) {
			var thisHierarchy = Hierarchy.getHierarchy(object, this.hierarchy, this.configuration.options.hierarchy);
			Hierarchy.addComponent(component, thisHierarchy);
		}

		this.update();
	};

	/**
	 * Updates the component bound to the specified object on the passthrough.
	 * @param {Object} object - the object to update the bound component
	 */
	Passthrough.prototype.updateComponent = function(object) {
		var component = this.componentMap[object._key];
		if(component == undefined) return;		
		component.object = object;

		// Process hierarchy if configured
		if(this.configuration.options.hierarchy != undefined) {
			var thisHierarchy = Hierarchy.getHierarchy(object, this.hierarchy, this.configuration.options.hierarchy);
			Hierarchy.updateComponent(component, thisHierarchy);
		}

		component.update();
		this.update();
	};

	/**
	 * Removes the component bound to the specified object from the passthrough.
	 * @param {Object} object - the object to remove the bound component
	 */
	Passthrough.prototype.removeComponent = function(object) {
		var component = this.componentMap[object._key];
		delete this.componentMap[object._key];
		this.componentArr.splice(this.componentArr.indexOf(component), 1);
		
		// Process hierarchy if configured
		if(this.configuration.options.hierarchy != undefined) {
			Hierarchy.removeComponent(component);
		}
		
		this.update();
	};
	
	
	/**
	 * Clears all objects and all components from the passthrough.
	 */
	Passthrough.prototype.clear = function() {	
		this.componentMap			= {};
		this.componentArr			= [];
		if(this.configuration.options.hierarchy != undefined) {
			this.hierarchy = {
				level:		0,
				members:	{},
				values:		[]
			};
		}
		this.update();
	};

	/**
	 * Draws the UI components for the passthrough.
	 */
	Passthrough.prototype.draw = function() {

	};

	/**
	 * Updates the contents of the passthrough. Normally this will happen in a microbatch
	 * to prevent too many updates from occurring. However it is possible to trigger
	 * an immediate update by setting the flag.
	 * @param {Boolean} immediate - trigger an immediate update
	 */
	Passthrough.prototype.update = function(immediate) {
		var self = this;

		if(immediate == true) {
			run();
			return;
		}
		
		// Microbatch handler to prevent too many updates
		if(self.microbatch.timeout === false) {
			self.microbatch.timeout = true;
			setTimeout(eventEnd, self.microbatch.delta);
		}					
		function eventEnd() {
			var d = new Date() - self.microbatch.rtime;
			if (new Date() - self.microbatch.rtime < self.microbatch.delta)
				setTimeout(eventEnd, self.microbatch.delta);
			else {
				self.microbatch.rtime = new Date();
				self.microbatch.timeout = false;
				run();
			}               
		};
		
		// Run the update
		function run() {
			if(self.paused == true) return;
			
			// Apply ordering if set
			self.orderedArr = self.componentArr;
			if(self.configuration.options.ordering.enabled != false) {
				self.orderedArr = DataFlow.Utilities.Ordering.order(self.orderedArr, self.orderDef, 'object');
			}

			// Apply search and filters if set
			self.filterArr = self.orderedArr;
			if(self.configuration.options.search.enabled != false || self.configuration.options.filter.enabled != false) {
				self.filterArr = DataFlow.Utilities.Filtering.filter(self.filterArr, self.filterDef, 'object');
			}
		
			// Set final results
			self.resultArr = self.filterArr;
			
			// Call the sink onUpdate function
			self.sink.onUpdate();			
		};
	};

	/**
	 * Pauses updates on a passthrough.
	 */
	Passthrough.prototype.pause = function() {
		this.paused = true;
	};
	
	/**
	 * Resumes a paused passthrough.
	 */
	Passthrough.prototype.resume = function() {
		if(this.paused == false) return;
		this.paused = false;
		this.update();
	};

	/**
	 * Resizes a passthrough.
	 */
	Passthrough.prototype.resize = function() {
		// No implementation for this component
	};
	
	/**
	 * Applies a search value against the passthrough. This will not be passed through
	 * to the dataset and is local only to the current passthrough.
	 * @param {String} value - the search string
	 */
	Passthrough.prototype.search = function(value) {
		if(value == undefined || value.length == 0) {
			delete this.filterDef.search.expression;
		}
		else {
			this.filterDef.search.expression = value;
		}

		this.update();
	};
	
	/**
	 * Hovers a single component bound to the specified object.
	 * Will also optionally notify the dataset of a hover event, if enabled.
	 * @param {Object} object - object that the associated component will be hovered
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Passthrough.prototype.hover = function(object, notify) {
		if(this.configuration.options.hover.enabled != false) {
			// Unhover all existing rows
			for(var idx = this.hovered.length - 1; idx >= 0; idx--) {
				var thisObject = this.hovered[idx];
				var component = this.componentMap[thisObject._key];
				if(component != undefined) {
					component.hovered = false;
					component.update();
				}
				this.hovered.splice(this.hovered.indexOf(thisObject), 1);
			}	
			
			// If the object is not null, process the hovering
			if(object != null) {
				if(Array.isArray(object)) {
					for(var idx = 0; idx < object.length; idx++) {
						var thisObject = object[idx];
						var component = this.componentMap[thisObject._key];
						if(component != undefined) {
							component.hovered = true;
							component.update();
						}
						this.hovered.push(thisObject);
					}
				}
				else {
					var component = this.componentMap[object._key];
					if(component != undefined) {
						component.hovered = true;
						component.update();
					}
					this.hovered.push(object);
				}
			}
			
			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.hover.passthrough == true) {
				this.sink.dataset.trigger('hover', this.sink, object);
			}
			
			this.update();
		}
	};

	/**
	 * Marks a single component or a set of components bound to the specified object or array of objects.
	 * Will also optionally notify the dataset of a marked event, if enabled.
	 * @param {Object|Array} object - if a single data object then the associated component will be marked
	 *                                if an array of objects then the associated components will be marked
	 *                                  for single marking components that receive a list of objects then only
	 *                                  the first component will be marked
	 *                                if null then all marked components will be unmarked
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Passthrough.prototype.mark = function(object, notify) {
		if(this.configuration.options.marking.enabled != false) {
			var mode = this.configuration.options.marking.mode;

			// Unmark all existing components
			for(var idx = this.marked.length - 1; idx >= 0; idx--) {
				var thisObject = this.marked[idx];				
				var component = this.componentMap[thisObject._key];
				if(component != undefined) {
					component.marked = false;
					component.update();
				}
				this.marked.splice(this.marked.indexOf(thisObject), 1);
			}	

			// If the object is not null, process the marking
			if(object != null) {
				if(Array.isArray(object)) {
					if(this.configuration.options.marking.mode == 'multiple') {
						for(var idx = 0; idx < object.length; idx++) {
							var thisObject = object[idx];
							var component = this.componentMap[thisObject._key];
							component.marked = true;
							component.update();
							this.marked.push(thisObject);
						}
					}
					// If we receive a list of objects but can only mark one, mark only the first one
					else if(this.configuration.options.marking.mode == 'single' && object.length > 0) {
						var component = this.componentMap[object[0]._key];
						component.marked = true;
						component.update();
						this.marked.push(object[0]);
					}
				}
				else {
					var component = this.componentMap[object._key];
					component.marked = true;
					component.update();
					this.marked.push(object);
				}
			}
			
			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.marking.passthrough == true) {
				this.sink.dataset.trigger('marked', this.sink, object);
			}			

			this.update();
		}
	};

	/**
	 * Applies a filter and updates the passthrough. Will also optionally notify the
	 * dataset of a filter event, if enabled.
	 * @param {Array} filters - array of filter objects
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Passthrough.prototype.filter = function(filters, notify) {
		if(this.configuration.options.filter.enabled != false) {
			// This clones the array so we don't get duplicates
			this.filterDef.filter.filters = filters.slice(0);
			this.filterDef.filter.fields = this.configuration.options.filter.fields;

			// Filter the components
			for(var idx = 0; idx < this.componentArr.length; idx++) {
				var include = DataFlow.Utilities.Filtering.filterItem(this.componentArr[idx], 
					this.filterDef.filter.filters, this.filterDef.filter.fields, 'object');
				var filtered = !include;
				this.componentArr[idx].filter(filtered);
			}

			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.filter.passthrough == true) {
				this.sink.dataset.trigger('filter', this.sink, this.filterDef.filter.filters);
			}
			
			this.update();
		}
	};
		
	/**
	 * Display the passthrough configuration modal.
	 */
	Passthrough.prototype.displayConfiguration = function() {
		if(this.configuration.options.configurable != true) return;
		console.log('display config');
		// TODO display configuration modal
	};

	/**
	 * Displays the dataset configuration modal.
	 */
	Passthrough.prototype.displayDataset = function() {
		if(this.configuration.options.dataset.modal == undefined) return;
		console.log('display dataset');
		// TODO display dataset modal
	};
	
	/** 
	 * Returns a color for the given value from the specified color map.
	 * @param {Object} value - the value to lookup, can be anything but typically a String
	 * @param (Object} name - the name of the color map
	 * @return {String} the color
	 */
	Passthrough.prototype.colormap = function(value, name) {
		var colorMap = this.colorMaps[name];
		if(colorMap == undefined) return;
		return colorMap.getColor(value);
	};
	
	/** 
	 * Formats the value using the specified format expression.
	 * @param {Object} val - the value to format, typically a timestamp or a number
	 * @param {Object} format - the formatting to apply, either a date/time format
	 *                   string or a number for number of decimals to round by
	 */
	Passthrough.prototype.format = function(val, format) {
		return DataFlow.Utilities.Format.format(val, format);
	};

	
	Passthrough.prototype.click = function(e, component) {
		if(this.configuration.options.marking.enabled == false) return;
		var ctrlKey = e.ctrlKey;					
		var mode = this.configuration.options.marking.mode;
		var isMarked = component.marked == true;
		
		// If unmarking is disabled and this component is marked, then just call mark again
		//   and take no further action
		if(this.configuration.options.marking.unmark == false && isMarked == true) {
			this.mark(component.object, true);
			return;
		}
		
		// Single mode
		if(mode == 'single') {
			if(isMarked == true)
				this.mark(null, true);
			else
				this.mark(component.object, true);
		}
		// Multiple mode
		else if(mode == 'multiple') {
			if(ctrlKey == true) {
				var objects = [];
				for(var idx = 0; idx < this.marked.length; idx++)
					objects.push(this.marked[idx].object);
				if(isMarked == false)
					objects.push(component.object);
				else if(isMarked == true)
					objects.splice(objects.indexOf(component.object), 1);
				this.mark(objects, true);
			}
			else {
				if(isMarked == true)
					this.mark(null, true);
				else
					this.mark(component.object, true);
			}
		}
		
	}
	

	/** 
	 * Creates a new component on the passthrough.
	 * @param {Object} object - the object to bind to the component
	 */
	function Component(object, passthrough) {
		this._key			= object._key;
		this.object 		= object;
		this.passthrough	= passthrough;
		
		// Flags for hovered, marked, and filtered
		this.hovered		= false;
		this.marked			= false;
		this.filtered		= false;
	};

	/**
	 * Draws a new component using base configuration. 
	 */
	Component.prototype.draw = function() {
		
	};
	
	/**
	 * Update the component.
	 */
	Component.prototype.update = function() {

	};
	
	/**
	 * Removes the component from the layer and invalidates the
	 * node for garbage collection.
	 */
	Component.prototype.remove = function() {

	};
		
	/**
	 * Sets the filtered flag for the component and adds or removes from the layer.
	 * @param {Boolean} filtered - true if filtered, otherwise false
	 */
	Component.prototype.filter = function(filtered) {
		if(this.filtered == true && filtered == false) {
			this.filtered = false;
		}
		else if(this.filtered == false && filtered == true) {
			this.filtered = true;
		}
	}
	
	/**
	 * Sets the search flag for the component and adds or removes from the layer.
	 * @param {Boolean} include - true if search match, otherwise false
	 */
	Component.prototype.search = function(include) {
		if(this.searched == null && include == true) {
			this.searched = true;
		}
		if(this.searched == null && include == false) {
			this.searched = false;
		}
		
		if(this.searched == true && include == null) {
			this.searched = null;
		}
		if(this.searched == true && include == false) {
			this.searched = false;
		}
		
		if(this.searched == false && include == null) {
			this.searched = null;
		}
		if(this.searched == false && include == true) {
			this.searched = true;
		}
	}
	
	


	/** 
	 * Passthrough configuration helper object.
	 */
	PassthroughConfiguration = {
		// Default passthrough configuration object
		defaultConfiguration: {
				options: {
					//classes:			undefined,
					configurable:		false,
					copy: {
						enabled:		true
					},
					dataset: {
						icon:			true
						//modal:			undefined
					},
					filter: {	
						accept:			true,
						fields:			undefined,
						delay:			500,
						enabled:		true,
						passthrough:	true
					},
					//hierarchy:		undefined,
					hover: {
						accept:			true,
						enabled:		true,
						passthrough:	true					
					},
					marking: {
						accept:			true,
						enabled:		true,
						mode:			"single",
						passthrough:	true,
						unmark:			true
					},
					ordering: {
						//classes:		undefined,
						fields:			undefined,
						enabled:		true,
						//initial:		{},
						multiple:		false
					},
					search: {
						fields:			undefined,
						delay:			500,
						enabled:		true,
						label:			"Search"
					},
				},				
			},
			
		// Merges the source configuration object into the target configuration object
		//   Source values will overwrite target values.
		//   Should always merge a DEFAULT configuration source into a USER configuration target
		mergeConfiguration:		function(source, target) {
									for(var idx = 0; idx < Object.keys(source).length; idx++) {
										var key = Object.keys(source)[idx];
										var val = source[key];
										if(typeof val === 'object') {
											if(target[key] == undefined) 
												target[key] = val;
											else
												this.mergeConfiguration(val, target[key]);
										}
										else {
											if(target[key] == undefined) 
												target[key] = val;
										}
									}
									return target;
								},

		// Processes the specified configuration object and merges in default configuration
		//   for base object
		processConfiguration:	function(config) {
									var configuration = this.mergeConfiguration(this.defaultConfiguration, config);
									return configuration;
								},
	};
	
	
	
	
	// Setup module objects
	module.Passthrough 				= Passthrough;
	module.PassthroughConfiguration	= PassthroughConfiguration;
	
	return module;
})();
/**
 * Represents a Plotly chart data sink.
 * @constructor
 * @param {Object} dataset - dataset that this sink will be attached to
 * @param {Object} configuration - configuration object defining the sink
 */
function PlotlyChartDatasink(id, configuration, dataset, containerDiv, overlay) {
	this.id = id;
	this.dataset = dataset;
	this.chart = new LivePlotlyChart.Chart(id, configuration, containerDiv, overlay, this);
}

/**
 * Event handler for dataset events. 
 * @param {String} eventType - string describing the event that occurred
 * @param {Object} object - data object for the event, or null
 */
PlotlyChartDatasink.prototype.trigger = function(eventType, object) {
	this.chart.trigger(eventType, object);
};


// Define a module to hold object defintions for the chart objects
var LivePlotlyChart = (function() {
	var module = {};

	/**
	 * Creates a new Chart object.
	 * @constructor
	 * @param {String} id - unique identifier
	 * @param {Object} configuration - the configuration for the chart
	 * @param {Object} containerDiv - this node object that chart will be appended
	 * @param (Function} overlay - function to call on start and end of the overlay
	 * @param {Object} sink - datasink that will feed this chart
	 */
	function Chart(id, configuration, containerDiv, overlay, sink) {
		this.id				= id;
		this.configuration	= configuration;
		this.containerDiv 	= containerDiv;
		this.overlay		= overlay;
		this.sink			= sink;

		// Pause and snapshot flags
		this.paused			= false;
		this.snapshot		= false;

		// Schema for the dataset
		this.schema			= null;		

		// Vars to hold the objects being hovered and marked
		this.hovered		= [];		
		this.marked			= [];

		// Filter definition
		this.filterDef		= {
			search: {
				expression:	undefined,
				fields:		this.configuration.options.search.fields
			},
			filter:	{
				filters:	[],
				fields:		this.configuration.options.filter.fields
			}
		};

		// Color maps
		this.colorMaps		= {};

		// Array of trace objects from the series
		this.traceMap		= {};
		this.traceArr		= [];
		
		// Create an array to hold the series objects
		//   Note this is different than this.traceArr which is the actual trace object required
		//   by Plotly
		this.seriesArr 		= [];

		// Check dependencies. This will throw Error if dependencies are not satisfied.
		this.dependencyCheck();
		
		// Microbatch is used to batch up updates to prevent too many chart redraws
		this.microbatch = {
			rtime:		new Date(1, 1, 2000, 12,00,00),
			timeout:	false,
			delta:		500 // msec between updates
		}

		// Color maps
		if(this.configuration.options.colormaps != undefined) {
			for(var idx = 0; idx < Object.keys(this.configuration.options.colormaps).length; idx++) {
				var thisColorMapName = Object.keys(this.configuration.options.colormaps)[idx];
				var thisColors = this.configuration.options.colormaps[thisColorMapName];
				this.colorMaps[thisColorMapName] = new DataFlow.ColorMap(thisColorMapName, thisColors);
			}
		}

		// Loop over the traces configuration and create a series object for each
		//   Add the Series object to the seriesArr and the Series object trace
		//   property to the data array.
		for(var idx = 0; idx < configuration.plotly.traces.length; idx++) {
			var thisSeries = new Series(configuration.plotly.traces[idx], this, idx);
			this.seriesArr.push(thisSeries);
			this.traceArr.push(thisSeries.trace);
			this.traceMap[thisSeries.trace.name] = thisSeries;
		}
		
		// Variables to hold equal axes definitions
		this.minAxis = 0;
		this.maxAxis = 0;					
		
		// On document ready, register UI handlers
		var self = this;
		setTimeout(function() {
			self.registerEventHandlers();
		}), 0;
		
		// Draw the initial empty graph
		this.draw();
	};
	
	/**
	 * Performs a dependency check for rendering the chart.
	 */
	Chart.prototype.dependencyCheck = function() {
		var msg = "Required Plotly libraries not loaded or not present";
		try {
			var p = Plotly;
		}
		catch(err) {
			throw Error("Required Plotly libraries not loaded or not present");
		}
	};
		
	/**
	 * Registers event handlers on UI components.
	 */
	Chart.prototype.registerEventHandlers = function() {
		var self = this;
		
		// Event handlers to pause and resume for wheel scroll, this gives the chart a time to update
		$(this.containerDiv).bind('mousewheel', function(e) {
			self.pause();
			setTimeout(function() {
				self.resume();
			},500);
		});
		
		// Event handlers to pause and resume for click and drag, this gives the chart a time to update
		$(this.containerDiv).on('mousedown', function (evtDown) {
			$(self.containerDiv).on('mouseup', function handler(evt) {
				if(self.paused == true) self.resume();
				$(self.containerDiv).off('mouseup');
				$(self.containerDiv).off('mousemove');
			});
			$(self.containerDiv).on('mousemove', function handler(evt) {
				var delta = Math.max(Math.abs(evtDown.pageX - evt.pageX), Math.abs(evtDown.pageY- evt.pageY));
				if(delta > 5 && self.paused == false) {
					self.pause();
				}
			});
		});
			
		// Event handlers for hover over plotly element
		//   THIS DOES NOT WORK VERY WELL IN PLOTLY, recommend disabling hover
		if(this.configuration.options.hover.enabled != false) {
			this.containerDiv.on('plotly_hover', function(data) {
				if(data.points == undefined) return;
				var thisPoint = data.points[0];
				var curveNumber = thisPoint.curveNumber == undefined ? 0 : thisPoint.curveNumber;
				var series = self.seriesArr[curveNumber];
				var pointNumber = thisPoint.pointNumber == undefined ? thisPoint.i : thisPoint.pointNumber;
				var component = series.pointArr[pointNumber];
				self.hover(component.object, true);
			});
			
			this.containerDiv.on('plotly_unhover', function(data) {
				self.hover(null, true);
			});
		}
		
		// Event handlers for click on plotly element
		//   THIS DOES NOT WORK VERY WELL IN PLOTLY, recommend disabling marking
		if(this.configuration.options.marking.enabled != false) {
			this.containerDiv.on('plotly_click', function(data) {
				if(data.points == undefined) return;
				var thisPoint = data.points[0];
				var curveNumber = thisPoint.curveNumber == undefined ? 0 : thisPoint.curveNumber;
				var series = self.seriesArr[curveNumber];
				var pointNumber = thisPoint.pointNumber == undefined ? thisPoint.i : thisPoint.pointNumber;
				var component = series.pointArr[pointNumber];
				self.mark(component.object, true);
			});
		}
	};
	
		
	/**
	 * Event handler for all objects delegated from the datasink.
	 * @param {String} eventType - the event type
	 * @param {Object} object - optional object
	 */
	Chart.prototype.trigger = function(eventType, object) {
		var self = this;
		
		if(eventType == 'schemaChange') {
			this.schema = object;
		}
		else if(eventType == 'snapshotStart') {
			this.snapshot = true;
			this.overlay(true);
		}
		else if(eventType == 'snapshotEnd') {
			setTimeout(function() {
				self.snapshot = false;
				self.overlay(false);
			},500);
		}
		else if(eventType == 'clear') {
			this.clear();
		}
		else if(eventType == 'insert') {
			this.addPoint(object);
		}
		else if(eventType == 'update') {
			this.updatePoint(object);
		}
		else if(eventType == 'upsert') {
		}
		else if(eventType == 'delete') {
			this.removePoint(object);
		}
		else if(eventType == 'hover') {
			if(this.configuration.options.hover.accept != false)
				this.hover(object);
		}
		else if(eventType == 'marked') {
			if(this.configuration.options.marking.accept != false)
				this.mark(object);
		}
		else if(eventType == 'filter') {
			if(this.configuration.options.filter.accept != false)
				this.filter(object);
		}					
		
	};
	
	/**
	 * Add a new point bound to the specified object on the chart.
	 * @param {Object} object - the object to add the bound point
	 */
	Chart.prototype.addPoint = function(object) {
		for(var idx = 0; idx < this.seriesArr.length; idx++)
			this.seriesArr[idx].addPoint(object);
		this.update();
	};

	/**
	 * Updates the point bound to the specified object on the chart.
	 * @param {Object} object - the object to update the bound point
	 */
	Chart.prototype.updatePoint = function(object) {
		for(var idx = 0; idx < this.seriesArr.length; idx++)
			this.seriesArr[idx].updatePoint(object);
		this.update();
	};
	
	/**
	 * Removes the point bound to the specified object from the chart.
	 * @param {Object} object - the object to remove the bound point
	 */
	Chart.prototype.removePoint = function(object) {
		for(var idx = 0; idx < this.seriesArr.length; idx++)
			this.seriesArr[idx].removePoint(object);
		this.update();
	};
					
	/**
	 * Clears all objects and all points from the chart.
	 */
	Chart.prototype.clear = function() {
		for(var idx = 0; idx < this.seriesArr.length; idx++)
			this.seriesArr[idx].removeAll();
		this.update();
	};
	
	/**
	 * Draws the UI components for the chart.
	 */
	Chart.prototype.draw = function() {
		Plotly.newPlot(this.containerDiv, this.traceArr, this.configuration.plotly.layout, 
			this.configuration.plotly.options);	
	};
	
	/**
	 * Updates the contents of the chart. Normally this will happen in a microbatch
	 * to prevent too many updates from occurring. However it is possible to trigger
	 * an immediate update by setting the flag.
	 * @param {Boolean} immediate - trigger an immediate update
	 */
	Chart.prototype.update = function(immediate) {
		var self = this;

		if(immediate == true) {
			run();
			return;
		}

		// Microbatch handler to prevent too many updates
		if(self.microbatch.timeout === false) {
			self.microbatch.timeout = true;
			setTimeout(eventEnd, self.microbatch.delta);
		}					
		function eventEnd() {
			var d = new Date() - self.microbatch.rtime;
			if (new Date() - self.microbatch.rtime < self.microbatch.delta)
				setTimeout(eventEnd, self.microbatch.delta);
			else {
				self.microbatch.rtime = new Date();
				self.microbatch.timeout = false;
				run();
			}               
		};
		
		// Run the update
		function run() {
			if(self.paused == true) return;
			
			Plotly.redraw(self.containerDiv);
		};
	};
	
	/**
	 * Pauses updates on a chart.
	 */
	Chart.prototype.pause = function() {
		this.paused = true;
	};
	
	/**
	 * Resumes a paused chart.
	 */
	Chart.prototype.resume = function() {
		if(this.paused == false) return;
		this.paused = false;
		this.update();
	};
	
	/**
	 * Resizes a chart.
	 */
	Chart.prototype.resize = function() {
		// Get the new size for the chart
		var update = {
			width: $(this.containerDiv).parent().width(), 
			height: $(this.containerDiv).parent().height()
		}

		// Relayout then redraw the chart						
		Plotly.relayout(this.containerDiv, update);	
		Plotly.redraw(this.containerDiv);
	};
			
	/**
	 * Applies a search value against the chart. This will not be passed through
	 * to the dataset and is local only to the current chart.
	 * @param {String} value - the search string
	 */
	Chart.prototype.search = function(value) {
		if(value == undefined || value.length == 0) {
			delete this.filterDef.search.expression;
		}
		else {
			this.filterDef.search.expression = value;
		}
		
		// Pass the filterDef through to the series
		for(var idx = 0; idx < this.seriesArr.length; idx++) {
			this.seriesArr[idx].filter(this.filterDef);
		}
	};
	
	/**
	 * Hovers a single point bound to the specified object.
	 * Will also optionally notify the dataset of a hover event, if enabled.
	 * @param {Object} object - object that the associated point will be hovered
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Chart.prototype.hover = function(object, notify) {
		if(this.configuration.options.hover.enabled != false) {
		
			// THIS DOES NOT WORK VERY WELL IN PLOTLY, recommend disabling hover
			// Unhover all points
			Plotly.Fx.hover(this.containerDiv,[]);

			// If the object is not null, process the hovering
			if(object != null) {
				if(Array.isArray(object)) {
					this.hovered = object.slice(0);
				}
				else {
					this.hovered = [object];
				}

				var hoverPoints = [];
				for(var seriesIdx = 0; seriesIdx < this.seriesArr.length; seriesIdx++) {
					var thisSeries = this.seriesArr[seriesIdx];
					for(var objIdx = 0; objIdx < this.hovered.length; objIdx++) {
						var thisPoint = thisSeries.pointMap[object[objIdx]._key];
						if(thisPoint == undefined) continue;
						var pointIndex = thisSeries.pointArr.indexOf(thisPoint);
						var hoverPoint = {
							curveNumber: seriesIdx,
							pointNumber: pointIndex
						}
						hoverPoints.push(hoverPoint);
					}
				}
				Plotly.Fx.hover(this.containerDiv, hoverPoints);
			}
		
			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.hover.passthrough == true) {
				this.sink.dataset.trigger('hover', this.sink, object);
			}			
		}
	};
		
	/**
	 * Marks a single point or a set of points bound to the specified object or array of objects.
	 * Will also optionally notify the dataset of a marked event, if enabled.
	 * @param {Object|Array} object - if a single data object then the associated point will be marked
	 *                                if an array of objects then the associated points will be marked
	 *                                  for single marking points that receive a list of objects then only
	 *                                  the first point will be marked
	 *                                if null then all marked points will be unmarked
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Chart.prototype.mark = function(object, notify) {
		if(this.configuration.options.marking.enabled != false) {
			var mode = this.configuration.options.marking.mode;

			// Unmark all existing series points
			for(var idx = 0; idx < this.seriesArr.length; idx++) {
				this.seriesArr[idx].unmark();
			}
			this.marked = [];
			
			// If the object is not null, process the marking
			if(object != null) {
				if(Array.isArray(object)) {
					if(this.configuration.options.marking.mode == 'multiple') {
						for(var idx = 0; idx < object.length; idx++) {
							var thisObject = object[idx];
							this.marked.push(thisObject);
						}
					}
					// If we receive a list of objects but can only mark one, mark only the first one
					else if(this.configuration.options.marking.mode == 'single' && object.length > 0) {
						var thisObject = object[0];
						this.marked.push(thisObject);
					}
				}
				else {
					this.marked.push(object);
				}

				// Update the series marked
				for(var idx = 0; idx < this.seriesArr.length; idx++) {
					this.seriesArr[idx].mark(this.marked);
				}
			}
			
			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.marking.passthrough == true) {
				this.sink.dataset.trigger('marked', this.sink, this.marked);
			}			
		}
	};
		
	/**
	 * Applies a filter and updates the chart. Will also optionally notify the
	 * dataset of a filter event, if enabled.
	 * @param {Array} filter - array of filter objects
	 * @param {Boolean} notify - if true and filter passthrough is enabled, 
	 *                    then the dataset will be notified
	 */
	Chart.prototype.filter = function(filters, notify) {
		if(this.configuration.options.filter.enabled != false) {
			// This clones the array so we don't get duplicates
			this.filterDef.filter.filters = filters.slice(0);
			this.filterDef.filter.fields = this.configuration.options.filter.fields;
			
			// Pass the filterDef through to the series
			for(var idx = 0; idx < this.seriesArr.length; idx++) {
				this.seriesArr[idx].filter(this.filterDef);
			}

			// If notify is true and passthrough is true, then notify the dataset
			if(notify == true && this.configuration.options.filter.passthrough == true) {
				this.sink.dataset.trigger('filter', this.sink, this.filtered.filter);
			}
			
			this.update();
		}
	};
	
	/**
	 * Display the chart configuration modal.
	 */
	Chart.prototype.displayConfiguration = function() {
		if(this.configuration.options.configurable != true) return;
		console.log('plotly display config');
		// TODO display configuration modal
	};

	/**
	 * Displays the dataset configuration modal.
	 */
	Chart.prototype.displayDataset = function() {
		if(this.configuration.options.dataset.modal == undefined) return;
		console.log('plotly display dataset');
		// TODO display dataset modal
	};
	
	/**
	 * Equalizes all axes to the same scale factor for both 2d and 3d chart types.
	 */
	Chart.prototype.equalizeAxes = function() {
		var layout = this.configuration.plotly.layout;
		if(layout.scene == undefined) {
			if(layout.xaxis == undefined) layout.xaxis = {};
			if(layout.yaxis == undefined) layout.yaxis = {};
			
			layout.xaxis.range = [this.minAxis, this.maxAxis];
			layout.yaxis.range = [this.minAxis, this.maxAxis];
		}
		else {
			if(layout.scene.xaxis == undefined) layout.scene.xaxis = {};
			if(layout.scene.yaxis == undefined) layout.scene.yaxis = {};
			if(layout.scene.zaxis == undefined) layout.scene.zaxis = {};
			
			layout.scene.xaxis.range = [this.minAxis, this.maxAxis];
			layout.scene.yaxis.range = [this.minAxis, this.maxAxis];
			layout.scene.zaxis.range = [this.minAxis, this.maxAxis];
		}
	};
					
	/** 
	 * Returns a color for the given value from the color map.
	 * @param {Object} value - the value to lookup, can be anything but typically a String
	 * @return {String} the color
	 */
	Chart.prototype.colormap = function(value) {
		return this.colorMap.getColor(value);
	};
	
	
	
	
	/**
	 * Creates a new Series object which represents a set of data on a chart. 
	 * A Series will represent different things depending on the underlying chart type.
	 * Typically a set of points or a line on a chart will be one series. A pie chart
	 * consists of a single series.
	 * @constructor
	 * @param {Object} configuration - the configuration for the series
	 * @param {Object} chart - the chart containing this series
	 */
	function Series(configuration, chart, index) {
		// Set configuration
		this.configuration 	= configuration;
		this.chart 			= chart;
		this.id				= 'series-' + index;

		// Vars for hovered, marked and filtered points
		this.hovered			= [];
		this.marked				= [];
		this.filterDef			= {
			search: {
				expression:	undefined,
				fields:		this.chart.configuration.options.search.fields
			},
			filter:	{
				filters:	[],
				fields:		this.chart.configuration.options.filter.fields
			}
		};

		// Trace object for the series, this will be passed to Plotly
		this.trace			= {};
		
		// Create the point container objects
		this.pointArr 		= [];
		this.pointMap 		= {};
		this.filterArr		= [];
		
		// Create a map of trace property (key) to array of values (value)
		this.arrayMap 		= {};
		
		// Create the property index
		var propertyIndexResult = DataFlow.Utilities.PropertyMap.index(this.configuration, ['property', 'eval']);
		this.propertyIndex = propertyIndexResult.propertyIndex;
		this.trace = propertyIndexResult.properties;
		
		// Create the array map
		for(var idx = 0; idx < this.propertyIndex.arr.length; idx++) {
			var thisProp = this.propertyIndex.arr[idx];
			this.arrayMap[thisProp.prop] = [];
		}		
	};

	/**
	 * Add a new object to the series as a new point.
	 * @param {Object} object - the object to add
	 */
	Series.prototype.addPoint = function(object) {
		var filtered = this.checkFilter(object);
		if(filtered == true) return;

		// Create the point
		var point = {
			_key:			object._key,
			object:			object,
			marked:			false
		}
		
		// Add to the point array and map
		this.pointArr.push(point);
		this.pointMap[point._key] = point;

		// Check include if there is a filter definition
		var include = DataFlow.Utilities.Filtering.filterItem(point, 
			this.filterDef.filter.filters, this.filterDef.filter.fields, 'object');
		
		if(include == true) {
			this.filterArr.push(point);
			
			// Build a context and evaluate the point properties, append to the data object
			for(var idx = 0; idx < this.propertyIndex.arr.length; idx++) {
				var prop = this.propertyIndex.arr[idx].prop;
				var type = this.propertyIndex.arr[idx].type;
				if(type == 'property') {
					var field = this.propertyIndex.map[prop];
					this.arrayMap[prop].push(object[field]);
				}
				else if(type == 'eval') {
					var exp = this.propertyIndex.map[prop];
					var context = {
						object:		object,
						series:		this,
						chart:		this.chart
					};
					var val = DataFlow.Utilities.Evaluate.expression(exp, context);
					this.arrayMap[prop].push(val);
				}
			}
		
			this.equalizeAxes(object);
			this.update();
		}
	};

	/**
	 * Update an existing object point on the series.
	 * @param {Object} object - the object to update
	 */
	Series.prototype.updatePoint = function(object) {
		var point = this.pointMap[object._key];
		if(point == undefined) return;
		
		// Check include if there is a filter definition
		var include = DataFlow.Utilities.Filtering.filterItem(point, 
			this.filterDef.filter.filters, this.filterDef.filter.fields, 'object');

		if(include == true) {
			var index = this.filterArr.indexOf(point);
			if(index == -1) this.filterArr.push(point);
			
			for(var idx = 0; idx < this.propertyIndex.arr.length; idx++) {
				var prop = this.propertyIndex.arr[idx].prop;
				var type = this.propertyIndex.arr[idx].type;
				if(type == 'property') {
					var field = this.propertyIndex.map[prop];
					if(index == -1)
						this.arrayMap[prop].push(object[field]);
					else
						this.arrayMap[prop][index] = object[field];
				}
				else if(type == 'eval') {
					var exp = this.propertyIndex.map[prop];
					var context = {
						object:		object,
						series:		this,
						chart:		this.chart
					};
					var val = DataFlow.Utilities.Evaluate.expression(exp, context);
					if(index == -1)
						this.arrayMap[prop].push(val);
					else
						this.arrayMap[prop][index] = val;
				}
			}		
			
			this.equalizeAxes(object);
			this.update();
		}
		else if(include == false) {
			var index = this.filterArr.indexOf(point);
			if(index > -1) {
				for(var idx = 0; idx < this.propertyIndex.arr.length; idx++) {
					var prop = this.propertyIndex.arr[idx].prop;
					this.arrayMap[prop].splice(index, 1);
				}		
				this.filterArr.splice(index, 1);
			}
			this.update();
		}
		
	};
	
	/**
	 * Remove an existing object point from the series.
	 * @param {Object} object - the object to remove
	 */
	Series.prototype.removePoint = function(object) {
		var point = this.pointMap[object._key];
		if(point == undefined) return;
		var index = this.filterArr.indexOf(point);

		if(index > -1) {
			for(var idx = 0; idx < this.propertyIndex.arr.length; idx++) {
				var prop = this.propertyIndex.arr[idx].prop;
				this.arrayMap[prop].splice(index, 1);
			}		
			this.filterArr.splice(index, 1);
		}
		this.pointArr.splice(this.pointArr.indexOf(point), 1);
		this.pointMap[object._key] = undefined;
		this.update();
		
		// Special case for pie charts when a purge happens
		//   Plotly doesn't like zero length arrays in pie charts, 
		//   so if all data has been removed, call draw to regen the chart
		if(this.trace.type == 'pie' && this.pointArr.length == 0) {
			this.chart.draw();
		}
		
	};
	
	/**
	 * Remove all existing object points from the series.
	 */
	Series.prototype.removeAll = function() {
		for(var idx = 0; idx < this.propertyIndex.arr.length; idx++) {
			var prop = this.propertyIndex.arr[idx].prop;
			this.arrayMap[prop] = [];
		}		
		this.pointArr = [];
		this.pointMap = {};		
		this.update();
	};
	
	/** 
	 * Updates the trace array with the current property arrays.
	 */
	Series.prototype.update = function() {
		for(var idx = 0; idx < this.propertyIndex.arr.length; idx++) {
			var prop = this.propertyIndex.arr[idx].prop;
			var arr = this.arrayMap[prop];
			this.trace[prop] = arr;
		}		
		if(this.configuration.equalAxes == true) this.chart.equalizeAxes();
	};

	/**
	 * Checks to see if the given object has been filtered by any configured
	 * filter condition.
	 * @param {Object} object - the object to test
	 * @return false if the object is included, true if the object is filtered
	 */
	Series.prototype.checkFilter = function(object) {
		if(this.configuration.eval_filter == undefined) return false;
		var context = {
			object:		object,
			series:		this,
			chart:		this.chart
		};
		var b = DataFlow.Utilities.Evaluate.expression(this.configuration.eval_filter, context);
		return !b;
	};
	
	/**
	 * Equalizes all axes for the series.
	 */
	Series.prototype.equalizeAxes = function(object) {
		if(this.configuration.equalAxes != true) return;
		for(var idx = 0; idx < this.propertyIndex.arr.length; idx++) {
			var prop = this.propertyIndex.arr[idx].prop;
			if(prop == 'x' || prop == 'y' || prop == 'z') {
				var field = this.propertyIndex.map[prop];
				var val = object[field];
				if(isNaN(val) == false) {
					this.chart.maxAxis = Math.max(this.chart.maxAxis, val);
					this.chart.minAxis = Math.min(this.chart.minAxis, val);
				}
			}
		}		
	};

	/**
	 * Unhovers the hovered point if there is one.
	 */
	Series.prototype.unhover = function() {
		return;
		
		// TODO hover
	};
	
	/**
	 * Hovers the point for the specified object.
	 * @param {Object} object - the object to hover the point for
	 */
	Series.prototype.hover = function(object) {
		return;
		
		// TODO hover
	};
		
	/**
	 * Unmarks all marked points on the series.
	 */
	Series.prototype.unmark = function() {
		return;
		// TODO mark
		
		if(this.marked == null) return;
		for(var idx = 0; idx < this.marked.length; idx++) {
			var thisObject = this.marked[idx];
			var point = this.pointMap[thisObject._key];
			if(point == undefined) continue;
			var data = point.data;
			if(point.highchartsPoint == null)
				point.highchartsPoint = this.chart.highcharts.get(data.id);
			if(point.highchartsPoint != null)
				point.highchartsPoint.select(false);		
			point.marked = false;
			point.data.selected = false;
		}
		this.marked = [];
	};
	
	/**
	 * Marks the point for the specified object or array.
	 * @param {Object[]} object - the array of objects to mark points for
	 */
	Series.prototype.mark = function(object) {
		return;
		// TODO mark

		if(object == null) return;
		for(var idx = 0; idx < object.length; idx++) {
			var thisObject = object[idx];
			var point = this.pointMap[thisObject._key];
			if(point == undefined) continue;
			var data = point.data;
			if(point.highchartsPoint == null)
				point.highchartsPoint = this.chart.highcharts.get(data.id);
			if(point.highchartsPoint != null)
				point.highchartsPoint.select(true, true);
			point.marked = true;
			point.data.selected = true;
			this.marked.push(thisObject);;
		}
	};
	
	/**
	 * Applies a filter definition on a series.
	 * @param {Object} filterDef - filter definition
	 */
	Series.prototype.filter = function(filterDef) {
		this.filterDef = filterDef;
		
		this.filterArr = [];
		this.arrayMap = {};
				
		// Create the array map
		for(var idx = 0; idx < this.propertyIndex.arr.length; idx++) {
			var thisProp = this.propertyIndex.arr[idx];
			this.arrayMap[thisProp.prop] = [];
		}		

		// Process all the points and build a new filterArr from this 
		for(var pointIdx = 0; pointIdx < this.pointArr.length; pointIdx++) {
			var point = this.pointArr[pointIdx];
			var object = point.object;
			var include = DataFlow.Utilities.Filtering.filterItem(point, 
				this.filterDef.filter.filters, this.filterDef.filter.fields, 'object');
			if(include == true) {
				this.filterArr.push(point);
				
				// Build a context and evaluate the point properties, append to the data object
				for(var propIdx = 0; propIdx < this.propertyIndex.arr.length; propIdx++) {
					var prop = this.propertyIndex.arr[propIdx].prop;
					var type = this.propertyIndex.arr[propIdx].type;
					if(type == 'property') {
						var field = this.propertyIndex.map[prop];
						this.arrayMap[prop].push(object[field]);
					}
					else if(type == 'eval') {
						var exp = this.propertyIndex.map[prop];
						var context = {
							object:		object,
							series:		this,
							chart:		this.chart
						};
						var val = DataFlow.Utilities.Evaluate.expression(exp, context);
						this.arrayMap[prop].push(val);
					}
				}
			
				this.equalizeAxes(object);
			}
		}

		this.update();
	
	};
	
	/** 
	 * Plotly configuration helper object.
	 */
	PlotlyConfiguration = {
		defaultConfiguration: {
				options: {
					configurable:		false,
					copy: {
						enabled:		true
					},
					dataset: {
						icon:			true
						//modal:			undefined
					},
					filter: {	
						accept:			true,
						fields:			undefined,
						delay:			500,
						enabled:		true,
						passthrough:	true
					},
					format: {
						timestamp:		"yyyy-MM-dd HH:mm:ss"
					},
					hover: {
						accept:			false,
						enabled:		false,
						passthrough:	false					
					},
					marking: {
						accept:			true,
						enabled:		true,
						mode:			"single",
						passthrough:	true,
						unmark:			true
					},
					search: {
						fields:			undefined,
						delay:			500,
						enabled:		true,
						label:			"Search"
					}
				},
				
			},
		
		// Merges the source configuration object into the target configuration object
		//   Source values will overwrite target values.
		//   Should always merge a DEFAULT configuration source into a USER configuration target
		mergeConfiguration:		function(source, target) {
									for(var idx = 0; idx < Object.keys(source).length; idx++) {
										var key = Object.keys(source)[idx];
										var val = source[key];
										if(typeof val === 'object') {
											if(target[key] == undefined) 
												target[key] = val;
											else
												this.mergeConfiguration(val, target[key]);
										}
										else {
											if(target[key] == undefined) 
												target[key] = val;
										}
									}
									return target;
								},

		// Processes the specified configuration object and merges in default configuration
		//   for both base object and column objects
		processConfiguration:	function(config) {
									var configuration = this.mergeConfiguration(this.defaultConfiguration, config);
									return configuration;
								},
	};


	
	module.Chart 				= Chart;
	module.PlotlyConfiguration	= PlotlyConfiguration;
	
	return module;
})();

/** 
 * Factory object for creating data source instances from configuration.
 */
DatasourceFactory = {
	/**
	 * Creates a new data source from the configuration.
	 * @param {Object} configuration - configuration object defining the source
	 * @param {Object} dataset - dataset object this source is a member of
	 */
	createDatasource: function(configuration, dataset) {
		DatasourceClass = this.getClass(configuration);
		return new DatasourceClass(configuration, dataset);		
	},

	/**
	 * Parses the configuration.type property for package names and then returns the 
	 * full datasource class name.
	 * @return {Function}
	 */
	getClass: function(configuration) {
		// Parse the type to create the correct instance
		var arr = configuration.type.split(".");
		var DatasourceClass = (window || this);
		for (var i = 0, len = arr.length; i < len; i++) {
			DatasourceClass = DatasourceClass[arr[i]];
		}
		
		return DatasourceClass
	},

	/**
	 * Validates the configuration for a single data source config.
	 * @param {Object} configuration - the configuration
	 * @param {Integer} index - position of the datasource in the array
	 * @param {Object} duplicates - object containing existing keys to detect duplicates
	 * @param {Object} warnings - array of warnings for the containing object
	 * @param {Boolean}
	 */
	validate: function(configuration, index, duplicates, warnings) {
		var valid = true;
		
		// Check that the source has an id
		if(configuration.id == undefined) {
			warnings.push('Datasource ' + index + ' has no id');
			valid = false;
		}
		
		// Check that the id is not duplicated
		if(duplicates[configuration.id] != undefined) {
			warnings.push('Datasource ' + configuration.id + ' already exists');
			valid = false;
		}
		
		// Check that the source has a type
		if(configuration.type == undefined) {
			warnings.push('Datasource ' + configuration.id + ' has no type');
			valid = false;
		}
		
		// Check that the type is a valid data source class
		if(valid == true && typeof this.getClass(configuration) !== "function") {
			warnings.push('Datasource ' + configuration.id + ' type ' + configuration.type + ' not found');
			valid = false;
		}
		return valid;
	},
};
/**
 * Represents a LiveView global connection. A page may only make a single connection
 * to the LDM server so this object manages that connection for all datasources.
 */
function LiveViewGlobalConnection() {
	// Set the configuration parameters
	this.url 		= '/lv/client';

	// Initialize objects
	this.connecting = false;
	this.conn		= null;
	this.resolves	= [];
};

/**
 * Connect to the LDM server. When the connection is obtained each of the elements 
 * in the resolves list will be resolved with the new connection.
 */
LiveViewGlobalConnection.prototype.connect = function() {
	var self = this;	
	this.connecting = true;
	
	LiveView.connect({
			url: this.url,
			onClose: function(connection, actor, error) {
				console.warn('LiveViewGlobalConnection connection closed... reconnecting');
				self.connect();
			}
		}
	).then(
		function(connection){
			console.log('LiveViewGlobalConnection connected');
			self.conn = connection;
			self.connecting = false;
			for(var idx = 0; idx < self.resolves.length; idx++) {
				self.resolves[idx](self.conn);
			}
		},
		function(err){
			console.error('LiveViewGlobalConnection error');
			self.connecting = false;
			console.error(err);
		}
	);
};

/**
 * Returns the connection to the LDM server. The connection process is asynchronous
 * so it may be that a request is made to get the connection either before the 
 * connection attempt happens, or while it is occurring. If the conneciton object
 * already exists, the promise is resolved immediately with that connection.
 * Otherwise the resolution is stored in a list of resolves that will be resolved
 * in connect() once the connection is obtained.
 * @param {Promise}
 */
LiveViewGlobalConnection.prototype.getConnection = function() {
	var self = this;
	
	var promise = new Promise(function(resolve, reject) {
		if(self.conn == null) {
			self.resolves.push(resolve);
			if(self.connecting == false)
				self.connect();
		}
		else {
			resolve(self.conn);
		}
	});
	
	return promise;
};

/**
 * Closes all connections to the LDM server.
 */
LiveViewGlobalConnection.prototype.close = function() {
	console.log('LiveViewGlobalConnection closing all connections');
	LiveView.closeAllConnections(
	).then(
		function(connection){
			console.log('LiveViewGlobalConnection disconnected');
		},
		function(error){
			console.error('LiveViewGlobalConnection error');
			console.error(error);
		}
	);
};

/**
 * Sends a tuple to the LDM server with the specified parameters.
 * @param {String} key - optional unique identifier key
 * @param {String} serverUri - server uri for the StreamBase server, e.g. sb://localhost:10000
 * @param {String} inputStream - name of the input stream
 * @param {Object} input - field map for the input tuple
 * @param {String} outputStream - optional name of the output stream, if the request will be 
 *   asynchronous, otherwise a response will be obtained from the output stream
 * @return {Promise}
 */
LiveViewGlobalConnection.prototype.sendTuple = function(key, serverUri, inputStream, input, outputStream) {		
	var self = this;
	var promise = new Promise(function(resolve, reject) {
		// Create a tuple
		var inputTuple = new LiveView.Tuple(null, input);
		
		// If the outputStream is specified, append a key
		if(outputStream != undefined)
			inputTuple.fieldMap.__SendTupleResponseKey = null;
		
		// Send the tuple
		self.conn.sendTuple(inputTuple, inputStream, outputStream, {
			serverUri: serverUri
		}).then(
			function(result){
				var output = undefined;				
				
				// If the outputStream is specified, remove the key
				if(outputStream != undefined) {
					output = result.tuple.fieldMap;
					delete output.__SendTupleResponseKey;
				}
				
				// Resolve with the result
				resolve(result);
			},
			function(error){
				reject(error);
			}
		);
	});
	
	return promise;
};

// Create a LV Global Connection object
var globalConn = new LiveViewGlobalConnection();



/**
 * Represents a LiveView JS API data source. This can be either a one-time or live data source.
 */
function LiveViewJSDatasource(configuration, dataset) {
	this.id 			= configuration.id;
	this.name			= configuration.name;
	this.description	= configuration.description;
	this.template		= configuration.template;
	this.parameters		= configuration.parameters;
	this.snap			= configuration.snapshot == undefined ? false : configuration.snapshot;		
	this.configuration	= configuration;
	this.dataset		= dataset;
	this.live			= true;
	
	this.conn			= null;
	this.tupleMap		= {};
};

/** 
 * Initializes the datasource after creation. Obtains the connection from the LiveViewGlobalConnection
 * and then either performs a snapshot() or a subscribe() depending on the query configuration.
 * The schema will be set on receiving the intial snapshot and all data objects will be sent through
 * to the listeners.
 * @return {Promise}
 */
LiveViewJSDatasource.prototype.init = function() {
	var self = this;
	var promise = new Promise(function(resolve, reject) {
		// Get the connection from the GlobalConnection
		globalConn.getConnection().then(
			// GlobalConnection getConnection resolve
			function(connection) {
				// Set the connection on the data source
				self.conn = connection;				
				
				// If this is a snapshot, then call snapshot, once completed
				//   resolve this promise
				if(self.snap == true) {
					self.snapshot().then(
						function(result) {
							resolve(self);
						},
						function(err) {
							console.error(err);
							resolve(self);
						}
					);
				}
				// Otherwise if subscription, then call subscribe, once completed
				//   resolve this promise
				else {
					self.subscribe().then(
						function(result) {
							resolve(self);
						},
						function(err) {
							console.error(err);
							resolve(self);
						}
					);
				}
			},
			
			// GlobalConnection getConnection reject
			function(err) {
				reject(err);
			}
		);
	});	
	
	return promise;	
};

/**
 * Creates a schema from a received data event. 
 * @param {Object} event - the data event
 */
LiveViewJSDatasource.prototype.createSchema = function(event) {
	this.schema = new DatasetSchema();
	for(var idx = 0; idx < event.schema.fields.length; idx++) {
		var f = event.schema.fields[idx].name;
		var t = event.schema.fields[idx].type;
		this.schema.addField(f, t);
	}
};

/**
 * Returns an object containing key fields from the given field map.
 * @param {Object} fieldMap - the field map
 * @return {Object}
 */
LiveViewJSDatasource.prototype.getKeyFields = function(fieldMap) {
	var obj = {};
	for(var idx = 0; idx < this.dataset.key.length; idx++) {
		var pk = this.dataset.key[idx];
		obj[pk] = fieldMap[pk];
	}
	return obj;
};

/**
 * Performs a snapshot query on the LDM server.
 * @return {Promise}
 */
LiveViewJSDatasource.prototype.snapshot = function() {
	var self = this;
	var promise = new Promise(function(resolve, reject) {
		// Execute the query on the server
		self.conn.execute(
			new LiveView.Query(self.template, self.parameters), {
				onSuccess: function(event){
					// Create the schema
					self.createSchema(event);
					
					// Notify the dataset of snapshot start
					self.dataset.trigger('snapshotStart', self);

					// Loop over the event data and notify the dataset of insert for each
					for(var idx = 0; idx < event.data.length; idx++) {
						var tuple = event.data[idx];
						self.dataset.trigger('insert', self, tuple.fieldMap);
					}					

					// Notify the dataset of snapshot end
					self.dataset.trigger('snapshotEnd', self);
					resolve();
				},
				onProgress: function(event){

				},
				onError: function(error){
					reject(error);
				}
			}
		).then(
			function(result){
			},
			function(error){
				reject(error);
			}
		);				
	});
	
	return promise;
};

/**
 * Performs a subscription query on the LDM server and stored the
 * query object on the datasource.
 * @return {Promise}
 */
LiveViewJSDatasource.prototype.subscribe = function() {
	var self = this;	
	
	var promise = new Promise(function(resolve, reject) {
		// Subscribe to the query on the server
		self.conn.subscribe(
			new LiveView.Query(self.template, self.parameters), {
				onSnapshotStart: function(event){		
					self.createSchema(event);
					self.dataset.trigger('snapshotStart', self);
				},
				onInsert: function(event){
					self.tupleMap[event.tuple.id] = event.tuple.fieldMap;
					self.dataset.trigger('insert', self, event.tuple.fieldMap);
				},
				onUpdate: function(event){
					self.dataset.trigger('update', self, event.tuple.fieldMap);
				},
				onDelete: function(event){
					var fieldMap = self.tupleMap[event.tuple.id];
					self.dataset.trigger('delete', self, fieldMap);
					delete self.tupleMap[event.tuple.id];
				},
				onSnapshotEnd: function(event){
					self.subscription = event.subscription;
					self.dataset.trigger('snapshotEnd', self);
					resolve();
				},
				onError: function(error){
					reject(error);
				}
			}
		).then(
			function(result){
				
			},
			function(error){
				reject(error);
			}
		);					
	});
	
	return promise;
};

/**
 * Unsubscribes the current query object on the LDM server.
 * @return {Promise}
 */
LiveViewJSDatasource.prototype.unsubscribe = function() {
	var self = this;	
	var promise = new Promise(function(resolve, reject) {
		self.conn.unsubscribe(
			self.subscription
		).then(
			function(result){
				delete self.schema;
				resolve(result);
			},
			function(error){
				delete self.schema;
				reject(error);
			}
		);					
	});
	
	return promise;
};

/**
 * Clear the existing tuple map and then calls a new snapshot or does an unsubscribe/subscribe
 * to do a full data reload from the LDM server.
 */
LiveViewJSDatasource.prototype.reload = function() {
	this.tupleMap = {};
	var self = this;
	if(this.snap == true)
		self.snapshot();
	else {
		self.unsubscribe().then(
			function(result) {
				self.subscribe();
			}
		);
	}
};

/**
 * Clears the existing tuple map and updates the parameters on this source. Then calls a new
 * snapshot or unsubscribe/subscribe to do a full data reload from the LDM server.
 */
LiveViewJSDatasource.prototype.applyParameters = function(parameters) {
	this.tupleMap = {};
	this.parameters = parameters;
	var self = this;
	
	// API apply parameters has a known issue, so call snap or unsub/sub as appropriate
	if(this.snap == true)
		self.snapshot();
	else {
		self.unsubscribe().then(
			function(result) {
				self.subscribe();
			},
			function(error) {
				//console.log(error);
			}
		);
	}
};

/**
 * Calls applyParameters() with the default configuration parameters.
 */
 LiveViewJSDatasource.prototype.resetParameters = function() {
	this.applyParameters(this.configuration.parameters);
};




/**
 * Represents a LiveView JS API data source. This can be a one-time data source only.
 */
function LiveViewTupleDatasource(configuration, dataset) {
	this.id 			= configuration.id;
	this.name			= configuration.name;
	this.description	= configuration.description;
	this.serverUri		= configuration.serverUri;
	this.inputStream	= configuration.inputStream;
	this.outputStream	= configuration.outputStream;
	this.parameters		= configuration.parameters;	
	this.configuration	= configuration;	
	this.dataset 		= dataset;
	this.live			= false;
};

/** 
 * Initializes the datasource after creation. Performs an invoke() by sending a message on the configured
 * serverUri and inputStream and then receives an asynchronous response on the configured outputStream.
 * The schema will be set from the response if present, otherwise it will rely on the schema being configured
 * in the data source configuration. Obtains the connection from the LiveViewGlobalConnection
 * and then sends the request tuple.
 * @return {Promise}
 */
LiveViewTupleDatasource.prototype.init = function() {
	var self = this;
	var promise = new Promise(function(resolve, reject) {
		self.invoke().then(
			function(result) {
				resolve(self);
			},
			function(err) {
				reject(err);
			}
		);					
	});	
	
	return promise;	
};

/**
 * Creates a schema from a received data event. If the event does not have a schema element then 
 * creates the schema from the data source configuration.
 * @param {Object} event - the data event
 */
LiveViewTupleDatasource.prototype.createSchema = function(event) {
	// Set the schema from the event
	if(event.schema != undefined) {
		this.schema = new DatasetSchema();
		for(var idx = 0; idx < event.schema.fields.length; idx++) {
			var f = event.schema.fields[idx].name;
			var t = event.schema.fields[idx].type;
			this.schema.addField(f, t);
		}
	}
	// Set the schema from configuration
	else {
		this.schema = new DatasetSchema();
		for(var idx = 0; idx < this.configuration.schema.length; idx++) {
			this.schema.addField(this.configuration.schema[idx].name, this.configuration.schema[idx].type);
		}
	}
};

/**
 * Performs a send tuple on the LDM server to an input stream with an output in an expected format.
 * @return {Promise}
 */
LiveViewTupleDatasource.prototype.invoke = function() {
	var self = this;
	var promise = new Promise(function(resolve, reject) {
		globalConn.getConnection().then(
			function(connection) {
				var input = {};
				if(self.parameters != null) {
					for(var idx = 0; idx < Object.keys(self.parameters).length; idx++) {
						var k = Object.keys(self.parameters)[idx];
						var v = self.parameters[k];
						input[k] = v;
					}
				}
				globalConn.sendTuple(null, self.serverUri, self.inputStream, input, self.outputStream).then(
					function(result) {
						// Resulting data will be in the following format:
						//   result.tuple.fieldMap
						//     schema - optional schema definition
						//     object - array of objects
						
						// Create the schema
						self.createSchema(result.tuple.fieldMap);
					
						// Notify the dataset of snapshot start
						self.dataset.trigger('snapshotStart', self);

						// Loop over the event data and notify the dataset of insert for each
						if(result.tuple.fieldMap.object.length > 0 && result.tuple.fieldMap.object[0] != null) {
							for(var idx = 0; idx < result.tuple.fieldMap.object.length; idx++) {
								var tuple = result.tuple.fieldMap.object[idx];
								self.dataset.trigger('insert', self, tuple);
							}					
						}
						
						// Notify the dataset of snapshot end
						self.dataset.trigger('snapshotEnd', self);
						resolve();
					},
					function(err) {
						reject(err);
					}
				);
			},
			function(err) {
				reject(err);
			}
		);
		
	});
	
	return promise;
};

/**
 * Calls a new invoke to do a full data reload from the server output stream.
 */
LiveViewTupleDatasource.prototype.reload = function() {
	this.invoke();
};

/**
 * Updates the parameters on this source. Then calls a new invoke to do a full data
 * reload from the server output stream.
 */
LiveViewTupleDatasource.prototype.applyParameters = function(parameters) {
	this.parameters = parameters;
	this.invoke();
};

/**
 * Calls applyParameters() with the default configuration parameters.
 */
LiveViewTupleDatasource.prototype.resetParameters = function() {
	this.applyParameters(this.configuration.parameters);
};

/**
 * Sends a tuple to the LDM server with the specified parameters.
 * @param {String} key - optional unique identifier key
 * @param {String} serverUri - server uri for the StreamBase server, e.g. sb://localhost:10000
 * @param {String} inputStream - name of the input stream
 * @param {Object} input - field map for the input tuple
 * @param {String} outputStream - optional name of the output stream, if the request will be 
 *   asynchronous, otherwise a response will be obtained from the output stream
 * @return {Promise}
 */
LiveViewTupleDatasource.sendTuple = function(key, serverUri, inputStream, input, outputStream) {
	return globalConn.sendTuple(key, serverUri, inputStream, input, outputStream);
}

/**
 * Color map returns a consistent color for a given value from
 * an array of colors. 
 * The colorscale can be specified as one of three ways:
 *   null - then a default color scale will be used
 *   object - defines mappings where the key is the field value and
 *     the value is the color to return for that value
 *   array - defines an array of color string
 * @constructor 
 * @param {String} name - name of the color map
 * @param {Object|String[]} colors - optional, object defining a mapping, or
 *                            and array of color strings, or null
 */
function ColorMap(name, colors) {
	var defaultColorScale = ['#1f77b4','#aec7e8','#ff7f0e','#ffbb78','#2ca02c','#98df8a',
		'#d62728','#ff9896','#9467bd','#c5b0d5','#8c564b','#c49c94','#e377c2','#f7b6d2',
		'#7f7f7f','#c7c7c7','#bcbd22','#dbdb8d','#17becf','#9edae5'];
	
	this.name 		= name;
	this.mode		= colors == null || Array.isArray(colors) ? 'scale' : 'mapping';

	this.colorscale	= null;
	this.colorIdx 	= 0;
	this.map 		= {};

	if(this.mode == 'scale')
		this.colorscale = colors == null || colors.length == 0 ? defaultColorScale : colors;
	else if(this.mode == 'mapping') {
		for(var key in colors) {
			this.map[key] = colors[key];
		}
	}
};

/**
 * Returns a color for the specified value. If the value was never requested
 * previously then the color map will return the next color in the scale.
 * Once the color scale has been exhausted, it will reset back to the first
 * element and duplicate colors will be returned. If the value was previously
 * requested then that color will be returned again.
 * @param {Object} val - the value to retrieve, typically a string but can be 
 *                   any data type
 * @return {String} a color
 */
ColorMap.prototype.getColor = function(val) {
	var color = this.map[val];
	if(color == undefined && this.mode == 'scale') {
		color = this.colorscale[this.colorIdx++];
		if(this.colorIdx == this.colorscale.length) {
			this.colorIdx = 0;
		}
		this.map[val] = color; 
	}
	return color;
};


var Utilities = (function() {
	var module = {};

	/**
	 * Helper object for managing clipboard operations.
	 */
	Clipboard = {
		/**
		 * Copies the specified copy text into the clipboard.
		 * @param {String} copyText - the text to copy
		 * @param {Object} container - the DOM container element 
		 */
		copy: 			function(copyText, container) { 
							var input = $(document.createElement('textarea')).val(copyText);
							container.append(input);
							input.select();
							
							try {
								var successful = document.execCommand('copy');
								//var msg = successful ? 'successful' : 'unsuccessful';
								//console.log('Copying text command was ' + msg);
							} catch (err) {
								//console.log('Unable to copy');
							}

							$(input).remove();				
							window.getSelection().removeAllRanges();					
						}
	};

	/**
	 * Helper object for managing context menu operations.
	 */
	ContextMenu = {
		/**
		 * Displays a new context menu at the specified location with the 
		 * specified context.
		 * @param {Object} containerDiv - the DOM container element
		 * @param {Object} e - mouse event where the context menu action occurred
		 * @param {Object} configuration - configuration object
		 * @param {Objext} context - context object for evaluation
		 */
		display:		function(containerDiv, e, configuration, context) {
							var self = this;

							// Hide any context menus 
							$(document).find('.live-context-menu').remove();
							if(configuration == undefined) return;
							
							// Loop over the groups and process
							var prevCount = 0;
							var count = 0;
							var renderedNodes = [];
							for(var idx = 0; idx < Object.keys(configuration).length; idx++) {
								var thisGroup = Object.keys(configuration)[idx];
								var thisGroupConfiguration = configuration[thisGroup];
								var nodes = self.renderGroup(thisGroupConfiguration, context);
								count = count + nodes.length;
								if(prevCount > 0 && nodes.length > 0) 
									renderedNodes.push(self.renderDivider());
								for(var nodeIdx = 0; nodeIdx < nodes.length; nodeIdx++)
									renderedNodes.push(nodes[nodeIdx]);
								if(nodes.length > 0) prevCount = nodes.length;
							}
							
							if(count > 0) {
								// Create the menu and ul nodes and append to the container
								var menuNode = $(document.createElement('div')).attr('class', 'live-context-menu');
								$(containerDiv).append(menuNode);
								var ulNode = $(document.createElement('ul'));
								menuNode.append(ulNode);
								for(var idx = 0; idx < renderedNodes.length; idx++) {
									ulNode.append(renderedNodes[idx]);
								}
								
								// Calculate the position of the menu based on event location
								var offset = $(containerDiv).parent().offset();
								var x = e.pageX - offset.left;
								var y = e.pageY - offset.top;
								var width = menuNode.width();
								var height = menuNode.height();
								
								// Show the menu
								if(x + width + offset.left + 15 > window.innerWidth)
									x = x - width;
								if(y + height + offset.top + 15 > window.innerHeight)
									y = y - height;
								menuNode.css({position: 'absolute', left: x, top: y, display: 'block'});
								
								// Event listener to hide the menuNode once clicked
								menuNode.on('click', function(e) {
									menuNode.remove();
								});							
							}
							
						},
			
		/**
		 * Renders a group on a context menu. Groups are separated by dividers.
		 * @param {Object} configuration - configuration object for the group
		 * @param {Object} context - context object for evaluation
		 * @return {Object[]} an array of DOM nodes to add to the menu
		 */
		renderGroup:	function(configuration, context) {
							var self = this;
							
							var nodes = [];
							if(Array.isArray(configuration)) {
								for(var idx = 0; idx < configuration.length; idx++) {
									var thisItemConfiguration = configuration[idx];
									var itemNode = self.renderItem(thisItemConfiguration, context);
									if(itemNode != null) nodes.push(itemNode);
								}
							}
							else {
								var itemNode = self.renderItem(configuration, context);
								if(itemNode != null) nodes.push(itemNode);
							}
							return nodes;
						},
		
		/**
		 * Renders a divider for a group.
		 * @return {Object} the DOM element for the divider
		 */
		renderDivider:	function() {
							return $(document.createElement('li')).attr('class', 'divider');
						},
		
		/**
		 * Renders a menu item for a context menu.
		 * @param {Object} configuration - configuration object for the item
		 * @param {Object} context - context object for evaluation
		 * @return {Object} a DOM node to add, or null if no item is to be added
		 */
		renderItem:		function(configuration, context) {
							// Evalute the include condition to determine if the element is added
							var include = true;
							if(configuration.condition != undefined) {
								include = function(str){
									for(key in this) eval('var ' + key + ' = this.' + key);
									return eval('(' + str + ')');
								}.call(context, configuration.condition);
							}
							if(include == false) return;

							// Create the item node and the link node, and append
							var itemNode = $(document.createElement('li'));
							var linkNode = $(document.createElement('a')).attr('tabindex', '-1').attr('href', 'javascript:void(0)');
							itemNode.append(linkNode);

							// Setup the label for the menu item
							var label = configuration.label;
							
							// If there is a string template, evaluate the label
							if(label.indexOf('${') > -1) {
								label = function(str){
									for(key in this) eval('var ' + key + ' = this.' + key);
									return eval('`' + str + '`');
								}.call(context, label);
							}
							
							// Set the label text
							linkNode.text(label);
							
							// Append an event handler to execute the action
							itemNode.on('click', function(e) {
								var result = function(str){
									for(key in this) eval('var ' + key + ' = this.' + key);
									return eval('(' + str + ')');
								}.call(context, configuration.action);
							});
							
							return itemNode;
						}
	};
			
	/**
	 * Helper object for managing debounce.
	 */
	Debounce = {
		/**
		 * Debounce will wait for a specified period of time to capture
		 * keyboard events. Once the interval has expired, then the
		 * results will be passed to the callback.
		 * @param {Function} callback - function to execute when the debounce has expired
		 * @param {Integer} wait - time in msec to wait before calling the callback
		 * @param {Boolean} immediate - if true the callback will be called immediately
		 */
		debounce:	function(callback, wait, immediate) {
						var timeout;
						return function() {
							var context = this, args = arguments;
							var later = function() {
								timeout = null;
								if (!immediate) callback.apply(context, args);
							};
							var callNow = immediate && !timeout;
							clearTimeout(timeout);
							timeout = setTimeout(later, wait);
							if (callNow) callback.apply(context, args);
						};
					},
	}
		
	/**
	 * Helper object for managing evaluation of configurable objects.
	 */
	Evaluate = {
		/**
		 * Evaluates a string into an object that lists a series of classes.
		 * The string must start and end with single braces '{' and '}' otherwise 
		 * the input string will just be returned.
		 * The object keys are the class names and the object values should evaluate
		 * to either true or false. If the evaluation is true, then that class
		 * should be added.
		 * @param {String} classes - string definition for an object defining classes
		 * @param {String} context - the context to evaluate within
		 * @return {Object} object definining which classes to include
		 */
		classes:		function(classes, context) {
							if(classes.indexOf('{') == -1) return classes;
							return function(str){
								for(key in this) eval('var ' + key + ' = this.' + key);
								return eval('(' + str + ')');
							}.call(context, classes);
						},
			
		/**
		 * Evaluates a string expression based on the specified context.
		 * @param {String} expression - string definition for the expression
		 * @param {String} context - the context to evaluate within
		 * @param {Boolean} template - if true then the expression will evaluate as a string template
		 * @return {String} resulting value
		 */
		expression: 	function(expression, context, template) {
							return function(str){
								for(key in this) eval('var ' + key + ' = this.' + key);
								if(template == true)
									return eval('`' + str + '`');
								else
									return eval(str);
							}.call(context, expression);
						},
		/**
		 * Evaluates a string into an object that lists a series of styles.
		 * The string must start and end with single braces '{' and '}' otherwise 
		 * the input string will just be returned.
		 * The object keys are the CSS style attributes object values should evaluate
		 * to the value for those attributes.
		 * @param {String} styles - string definition for an object defining styles
		 * @param {String} context - the context to evaluate within
		 * @return {Object} object definining styles to include
		 */
		styles:			function(styles, context) {
							if(styles.indexOf('{') == -1) return styles;
							return function(str){
								for(key in this) eval('var ' + key + ' = this.' + key);
								return eval('(' + str + ')');
							}.call(context, styles);
						}
	};
	
	/**
	 * Helper object for managing array filters.
	 */
	Filtering = {
		/**
		 * Filters a list of items based on the pagination configuration.
		 * @param {Object[]} items - array of items to filter
		 * @param {Object} pagination - pagination definition
		 * @return {Object[]} subset of items matching the pagination config
		 */
		paginate: 	function(items, pagination){
						if(items == undefined || pagination == undefined) return items;
						var begin = pagination.begin == undefined ? 0 : pagination.begin;
						var end = pagination.end == undefined || pagination.end == 0 ? begin + pagination.size : pagination.end;
						return items.slice(begin, end);
					},
			
		/**
		 * Filters a list of items based on the specified filter object.
		 * The filter definition must be an object with the following schema:
		 *    filter = {
		 *	    search = {        {Object} search definition object
		 *        expression:     {String} - search expression to apply against the property
         *        fields:         {String[]} - list of fields of the object to search against		 
		 *      },
		 *	    filter = {
		 *        filters: [
		 *          {
		 *            expression:  {String} - filter expression to apply against the field
		 *            field:       {String} - field to apply the expression against
		 *          }
		 *        ],
		 *        fields:          {String[]} - list of fields of the object to filter against
		 *      }
		 *    }
		 * @param {Object[]} items - array of items to filter
		 * @param {Object} filter - filter definition
		 * @param {String} property - optional, property of the item to extract the object to 
		 *                              apply the filter against, if omitted then the filter is applied
		 *                              against the item itself
		 * @return {Object[]} subset of items matching the filter definition
		 */
		filter: 	function(items, filter, property){
						if(items == undefined || items.length == 0) return items;
						if(filter == undefined) return items;
						if(filter.search == undefined && (filter.filter == undefined || filter.filter.length == 0)) return items;
						var filtered = [];
						var self = this;
						angular.forEach(items, function(item) {
							var include = true;							
							if(filter.search.expression != undefined)
								include = include && self.searchItem(item, filter.search.expression, filter.search.fields, property);							
							if(filter.filter.filters.length > 0)
								include = include && self.filterItem(item, filter.filter.filters, filter.fields, property);
							if(include == true) filtered.push(item);
						});
						return filtered;
					},
					
		/**
		 * Searches an item based on a search specification. If the item matches, then 
		 * true will be returned. The search specification is a partial match and case
		 * insensitive.
		 * @param {Object} item - item to search
		 * @param {String} expression - search expression to match
		 * @param {String[]} fields - list of fields to search against
		 * @param {String} property - optional, property of the item to extract the object to 
		 *                              apply the search against, if omitted then the search is applied
		 *                              against the item itself
		 * @return {Boolean} true if the item matches the search, otherwise false
		 */
		searchItem:	function(item, expression, fields, property) {
						if(expression == undefined) return true;

						// Filter by the search box text
						//   Search all of the fields if specified, or all fields if not
						//   Simple inexact match, case insensitive
						var include = false;
						
						// Extract the object to search if property is specified
						var object = item;
						if(property != undefined) object = item[property];
						if(object == undefined) return false;
						
						// Search only specified fields
						if(fields != undefined && fields.length > 0) {
							for(var idx = 0; idx < fields.length; idx++) {
								var thisField = fields[idx];
								if(object[thisField] != undefined && (object[thisField] + "").toLowerCase().indexOf(expression.toLowerCase()) >= 0) {
									include = true;
									break;
								}
							}
						}
						// Otherwise search all fields
						else {
							for(var idx = 0; idx < Object.keys(object).length; idx++) {
								var thisField = Object.keys(object)[idx];
								if(thisField.startsWith('_')) continue;								
								if(object[thisField] != undefined && (object[thisField] + "").toLowerCase().indexOf(expression.toLowerCase()) >= 0) {
									include = true;
									break;
								}
							}
						}
						return include;
					},
			
		/**
		 * Filters an item based on a filter expression. If the item matches, then 
		 * true will be returned. The filter specification is either a partial match and case
		 * insensitive, or an expression exact match including the following operators:
		 *     ==, !=, >, >=, <, <=
		 * @param {Object} item - item to filter
		 * @param {Object[]} filters - list of filter objects to apply against the data
		 * @param {String[]} fields - list of fields to search against
		 * @param {String} property - optional, property of the item to extract the object to 
		 *                              apply the filter against, if omitted then the filter is applied
		 *                              against the item itself
		 * @return {Boolean} true if the item matches the filter, otherwise false
		 */
		filterItem:	function(item, filters, fields, property) {
						// Filter by list of filter expressions
						//   Supports inequality operators
						//     >, <, >=, <=, ==, !=
						//   Otherwise inexact match, case insensitive
						var include = true;
						
						// Extract the object to search if property is specified
						var object = item;
						if(property != undefined) object = item[property];
						if(object == undefined) return false;
						
						for(var idx = 0; idx < filters.length; idx++) {
							var thisFilter = filters[idx];
							if(fields != undefined && fields.indexOf(thisFilter.field) == -1) return true;
							var thisExpression = thisFilter.expression;
							var data = object[thisFilter.field];

							// Has exact expression
							if(thisExpression.indexOf('>') == 0 || thisExpression.indexOf('<') == 0 || thisExpression.indexOf('!') == 0 || thisExpression.indexOf('==') == 0) {
								var expression = thisExpression;//.replace(/([.*+?^=!:${}()|\[\]\/\\])/g, "\\$1");
								var expression = expression
									.replace(new RegExp('>', 'g'), ' data >')
									.replace(new RegExp('<', 'g'), ' data <')
									.replace(new RegExp('!=', 'g'), ' data !=')
									.replace(new RegExp('==', 'g'), ' data ==');							
								//console.log(data + " " + expression);
								try {
									if(data != undefined && eval(expression) == false) {
										include = false;
										break;
									}
								}
								catch(err) {
									//console.log(err);
									include = false;
									break;
								}
							}
							// Case insensitive partial match
							else if((data + "").toLowerCase().indexOf(thisExpression.toLowerCase()) == -1) {
								include = false;
								break;
							}
						}			
						return include;
					},
	};
				
	/**
	 * Helper object for managing formatting.
	 */
	Format = {
		/**
		 * Applies formatting against the specified data. If the format value 
		 * is not a number, then the data will be treated as a timestamp and 
		 * formatted using formatTimestamp(). Otherwise the data will be 
		 * treated as a numeric and will be formatted using formatNumber().
		 * @param {Object} data - data to format, typically will be a long in 
		 *                   the case of a timestamp, or an int, long, double
		 *                   in the case of a numeric
		 * @param {String|Number} format - formatting to apply
		 *                            in the case of a timestamp will be a string,
		 *                            otherwise will be a number
		 * @return {String} formatted timestamp or number
		 */
		format:				function(data, format) {
								if(data == undefined) return "";
								if(isNaN(format) == true)
									return this.formatTimestamp(data, format);
								else
									return this.formatNumber(data, format);
							},
							
		/**
		 * Applies number formatting against the specified data.
		 * @param {Number} data - data to format
		 * @param {Integer} decimals - number of decimals to format to
		 * @return {String} formatted number as a string
		 */
		formatNumber:		function(data, decimals) {
								if(decimals === undefined || data === null || data === undefined) return data;
								return parseFloat(data).toFixed(decimals);
							},
							
		/**
		 * Applies date/time formatting against the specified data.
		 * @param {Number} data - timestamp to format
		 * @param {String} format - date/time format
		 * @return {String} formatted timestamp as a string
		 */
		formatTimestamp:	function(data, format) {			
								if(data == undefined) return "";				
								if(format == undefined) return data;
								
								var date = data;
								if(!(date instanceof Date)) {
									date = new Date(date);
									// Firefox hack
									//date = new Date(data.replace(' ', 'T'));
								}

								// Array of short month names
								var shortMonth 	= ['Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec'];	
								var longMonth 	= ['January','February','March','April','May','June','July','August','September','October','November','December'];	
								
								// Array of date format tokens
								var o = {
									"yyyy+" : 	date.getFullYear(),
									"yy+" : 	(date.getFullYear() + "").substring((date.getFullYear() + "").length - 2),
									"dd+" : 	(date.getDate() < 10 ? "0" : "") + date.getDate(),
									"HH+" : 	(date.getHours() < 10 ? "0" : "") + date.getHours(),
									"mm+" : 	(date.getMinutes() < 10 ? "0" : "") + date.getMinutes(),
									"ss+" : 	(date.getSeconds() < 10 ? "0" : "") + date.getSeconds(),
									"S" : 		(date.getMilliseconds() < 10 ? "00" : date.getMilliseconds() < 100 ? "0" : "") + date.getMilliseconds(),
									// Month has to come last and has to be in this order due to the string substitution
									"MMMM+" : 	longMonth[date.getMonth()],
									"MMM+" : 	shortMonth[date.getMonth()],
									"MM+":		(date.getMonth() + 1 < 10 ? "0" : "") + (date.getMonth() + 1),
									"Z":		(date.getTimezoneOffset()),
								}

								var formatted = format;
								for(var k in o) {
									if(new RegExp("("+ k +")").test(formatted))
										formatted = formatted.replace(RegExp.$1,o[k]);
								}
									
								return formatted;
							}
	},
	
	
	/**
	 * Helper object for putting objects into a hierarchy.
	 */
	Hierarchy = {
		/**
		 * Inserts the specified component into the root hierarchy and 
		 * returns it. 
		 * @param {Object} object - the object to insert
		 * @param {Object} rootHierarchy - the root hierarchy object
		 * @param {Object} definition - the hierarchy definition
		 * @return hierarchy object
		 */
		getHierarchy:		function(object, rootHierarchy, definition) {
								var parentHierarchy = rootHierarchy;
								for(var idx = 0; idx < definition.length; idx++) {
									var hierarchyName = definition[idx];
									var hierarchyValue = object[hierarchyName];
									var member = parentHierarchy.members[hierarchyValue];
									if(member == null) {
										member = {
											level:		idx + 1,
											members:	{},
											values:		[],
											parent:		parentHierarchy
										}
										parentHierarchy.members[hierarchyValue] = member;
									}
									parentHierarchy = member;
								}
								
								// Parent hierarchy is now the target
								return parentHierarchy;
			
							},
		
		/**
		 * Adds the component to the hierarchy location.
		 * @param {Object} component - the UI component to add
		 * @param {Object} hierarchyLevel - the level to add it to 
		 */
		addComponent:		function(component, hierarchyLevel) {
								component._hierarchy = hierarchyLevel;
								hierarchyLevel.values.push(component);			
							},
							
		/**
		 * Updates the component in the hierarchy location. If the hierarchy has not changed,
		 * then no action is taken. If it has changed the component is removed from the old
		 * hierarchy and added to the new one.
		 * @param {Object} component - the UI component to add
		 * @param {Object} hierarchyLevel - the level to add it to if changed
		 */
		updateComponent:	function(component, hierarchyLevel) {
								if(component._hierarchy != hierarchyLevel) {
									var oldHierarchy = component._hierarchy;
									component._hierarchy.values.splice(component._hierarchy.values.indexOf(component), 1);
									hierarchyLevel.values.push(component);
									component._hierarchy = hierarchyLevel;
									this.trim(oldHierarchy);
								}			
							},
							
		/** 
		 * Removes the component from the hierarchy location.
		 * @param {Object} component - the UI component to remove
		 */
		removeComponent:	function(component) {
								if(component._hierarchy != undefined) {
									var oldHierarchy = component._hierarchy;
									component._hierarchy.values.splice(component._hierarchy.values.indexOf(component), 1);
									this.trim(oldHierarchy);
								}
							},
							
		/**
		 * Trims the specified hierarchy level from the parent if there are no values included.
		 * @param {Object} hierarchyLevel - the level to trim
		 */
		trim:				function(hierarchyLevel) {
								var parentHierarchy = hierarchyLevel.parent;
								if(parentHierarchy != null) {
									if(hierarchyLevel.values.length == 0 && Object.keys(hierarchyLevel.members).length == 0) {
										for(var idx = 0; idx < Object.keys(parentHierarchy.members).length; idx++) {
											var key = Object.keys(parentHierarchy.members)[idx];
											var val = parentHierarchy.members[key];
											if(val == hierarchyLevel) {
												delete parentHierarchy.members[key];
											}
										}
									}
									this.trim(parentHierarchy);
								}
							}
	};
	
	
	/**
	 * Helper object for managing array ordering.
	 */
	Ordering = {
		/**
		 * Sorts a list of items based on the specified ordering object.
		 * The order definition must be an object with the following schema:
		 *    order = [
		 *      {
		 *        field:     {String} - field name to order by
		 *        orderDir:  {String} - ordering direction, 'asc' or 'desc'
		 *      }
		 *    ]
		 * @param {Object[]} items - array of items to filter
		 * @param {Object[]} ordering - order definition
		 * @param {String} property - optional, property of the item to extract the object to 
		 *                              apply the ordering against, if omitted then the ordering is applied
		 *                              against the item itself
		 * @return {Object[]} ordered array of items matching the order definition
		 */
		order:		function(items, ordering, property) {			
						if(items == undefined || items.length == 0) return items;
						if(ordering == undefined || ordering.length == 0) return items;
						
						var ordered = [];
						angular.forEach(items, function(item) {
							ordered.push(item);
						});

						var compare = function(a, b, dir, field) {
							var aObj = a;
							var bObj = b;
							if(property != undefined) {
								aObj = a[property];
								bObj = b[property];
							}
							return dir * (aObj[field] == bObj[field] ? 0 : aObj[field] < bObj[field] ? 1 : -1);
						};
			
						// This is default sort algorithm
						/*ordered.sort(function (a, b) {
							for(var idx = 0; idx < ordering.length; idx++) {
								var thisOrdering = ordering[idx];
								var dir = thisOrdering.orderDir == 'desc' ? -1 : 1;
								var v = compare(a, b, dir, thisOrdering.field);
								if(v != 0) return v;
							}
							return 0;
						});*/
						// return ordered;
						// End default sort
						
						// This is merge sort algorithm, might be more efficient
						var algo = function(a, b) {
							for(var idx = 0; idx < ordering.length; idx++) {
								var thisOrdering = ordering[idx];
								var dir = thisOrdering.orderDir == 'desc' ? -1 : 1;
								var v = compare(a, b, dir, thisOrdering.field);
								if(v != 0) return v;
							}
							return 0;
						};
						
						var merge = function(left, right) {
							var result = [];
							while((left.length > 0) && (right.length > 0)) {		
								if(algo(left[0], right[0]) >= 0)
									result.push(left.shift());
								else
									result.push(right.shift());
							}
							result = result.concat(left, right);
							return result;
						};

						var sort = function(array) {
							var len = array.length;
							if(len < 2) return array;
							var pivot = Math.ceil(len/2);
							return merge(sort(array.slice(0,pivot)), sort(array.slice(pivot)));
						};

						return sort(ordered);
					}
	};
	
	/**
	 * Helper object for building property maps.
	 */
	PropertyMap = {
		index:		function(configuration, types) {
						if(configuration == null) return null;
						if(types.length == 0) return null;
						
						// Create a property index object
						var propertyIndex = {
							arr:	[],
							map:	{}
						};
						
						// Create a default properties object
						var properties = {};
						
						// Iterate over all the configuration keys and classify
						for(var idx = 0; idx < Object.keys(configuration).length; idx++) {
							var key = Object.keys(configuration)[idx];
							var type = key.substring(0, key.indexOf("_"));
							var prop = key.substring(key.indexOf("_")+1);
							if(types.indexOf(type) > -1) {
								propertyIndex.arr.push({prop: prop, type: type});
								propertyIndex.map[prop] = configuration[key];
							}
							else
								properties[key] = configuration[key];
						}

						return {
							propertyIndex: 	propertyIndex,
							properties:		properties
						};						
					},
					
		evaluate:	function(propertiesIndex, context, object, target) {
						if(propertiesIndex == null) return null;
						
						for(var idx = 0; idx < propertiesIndex.arr.length; idx++) {
							var prop = propertiesIndex.arr[idx].prop;
							var type = propertiesIndex.arr[idx].type;
							if(type == 'property') {
								var field = propertiesIndex.map[prop];
								target[prop] = object[field];
							}
							else if(type == 'eval') {
								var exp = propertiesIndex.map[prop];
								var val = DataFlow.Utilities.Evaluate.expression(exp, context);
								target[prop] = val;
							}
						}
						return target;
					}
	}
	
	/**
	 * Helper object for HTML template processing.
	 */
	Template = {
		process:	function(html) {
						var node = html;
						if($.type(node) === "string")
							node = $(document.createElement('div')).html(html);
						
						// Remove nodes that have if=false
						$(node).find("[if='false']").remove();
						
						return node;
					}
	}
	
	// Other objects
	module.Clipboard	= Clipboard;
	module.ContextMenu	= ContextMenu;
	module.Debounce		= Debounce;
	module.Evaluate		= Evaluate;
	module.Filtering	= Filtering;
	module.Format		= Format;
	module.Hierarchy	= Hierarchy;
	module.Ordering		= Ordering;
	module.PropertyMap	= PropertyMap;
	module.Template		= Template;
	
	return module;
})();

	

	// Dataset objects
	module.Dataset 					= Dataset;
	module.DatasetFactory 			= DatasetFactory;
	module.DatasetSchema 			= DatasetSchema;

	// Datasink objects
	module.GridDatasink				= GridDatasink;
	module.HighchartsChartDatasink	= HighchartsChartDatasink;
	module.LeafletLayerDatasink		= LeafletLayerDatasink;
	module.PassthroughDatasink		= PassthroughDatasink;
	module.PlotlyChartDatasink		= PlotlyChartDatasink;
	
	// Datasource objects
	module.LiveViewJSDatasource 	= LiveViewJSDatasource;
	module.LiveViewTupleDatasource 	= LiveViewTupleDatasource;
	
	// Other objects
	module.ColorMap					= ColorMap;
	module.LiveGrid					= LiveGrid;
	module.LiveLeafletMap			= LiveLeafletMap;
	module.LiveHighchartsChart		= LiveHighchartsChart;
	module.LivePassthrough			= LivePassthrough;
	module.LivePlotlyChart			= LivePlotlyChart;
	module.Utilities				= Utilities;
	
	return module;
})();


